//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#include <sstream>

#include <Foundation/const.h>
#include <Foundation/datatype.h>
#include <Foundation/error.h>
#include <Foundation/util.h>

#include "AreaCommLink.h"
#include "SklSolverFoundation.h"

AreaCommLink::AreaCommLink(
		const CfgInfo& cfg,
		const string& lbl, const int nid, const int gid,
		const CfgInfo::VoxelDivInfo& slfdiv,
		const CfgInfo::VoxelDivInfo& tgtdiv,
		const std::vector<char> Es)
: label(lbl), nID(nid), gID(gid), aryN(Es)
{
	for (int i = 0; i < 3; i++) {
		SlfVxRect.head[i] = slfdiv.head[i];
		SlfVxRect.tail[i] = slfdiv.tail[i];
		SlfVxRect.size[i] = slfdiv.size[i];
	}
	for (int i = 0; i < 3; i++) {
		TgtVxRect.head[i] = tgtdiv.head[i];
		TgtVxRect.tail[i] = tgtdiv.tail[i];
		TgtVxRect.size[i] = tgtdiv.size[i];
	}

#ifdef SPHERE175
	//ParaComponent
	SklParaComponent *para_cmp = cfg.solver->SklGetParaComponent();	//2009.12.07

	//ParaManager
	SklParaManager *para_mng = para_cmp->GetParaManager();	//2009.12.07
#endif

	// exchange VoxeclRect and Assigned equation array
	const char* m = "Commnunication Error - Broadcast";

#ifdef SPHERE175
	TRY(para_mng->Broadcast(TgtVxRect.head,3,SKL_ARRAY_DTYPE_INT,nID,gID), m);	//2009.12.07
	TRY(para_mng->Broadcast(TgtVxRect.tail,3,SKL_ARRAY_DTYPE_INT,nID,gID), m);	//2009.12.07
#else
	TRY(ParaMngr.Broadcast(TgtVxRect.head,3,SKL_ARRAY_DTYPE_INT,nID,gID), m);
	TRY(ParaMngr.Broadcast(TgtVxRect.tail,3,SKL_ARRAY_DTYPE_INT,nID,gID), m);
#endif

	for (int i = 0; i < 3; i++) {
		TgtVxRect.size[i] = TgtVxRect.tail[i] - TgtVxRect.size[i] + 1;
	}

	// Calc Wrapped Rect
	for (int i = 0; i < 3; i++) {
		WrappedRect.head[i] = grater(TgtVxRect.head[i], SlfVxRect.head[i]);
		WrappedRect.tail[i] = lesser(TgtVxRect.tail[i], SlfVxRect.tail[i]);
		WrappedRect.size[i] = WrappedRect.tail[i] - WrappedRect.head[i] + 1;
		if (WrappedRect.size[i] < 0) { WrappedRect.size[i] = 0; }
	}

	// allocate S4D_REAL buffer
	const int ix = WrappedRect.size[0];
	const int jx = WrappedRect.size[1];
	const int kx = WrappedRect.size[2];

	int N = 0; for (int i = 0; i < aryN.size(); i++) { N += aryN[i]? 1: 0; }
	buffer = (ix*jx*kx*N > 0)?
				new S4D_REAL(SKL_ARRAY_DTYPE_REAL, ix, jx, kx, N, VC):
				NULL;
}

AreaCommLink::~AreaCommLink()
{
	delete buffer;
	buffer = NULL;
}

bool AreaCommLink::enable() const
{
	return buffer;
}

int AreaCommLink::Send(const S4D_REAL* src, const bool comm_vc) const
{
	if (!enable()) return -1;

	SKL_REAL* buf = buffer->GetData();
	int       sz  = buffer->GetArrayLength();
	int       vc  = comm_vc?
						lesser(buffer->GetVCellSize(), src->GetVCellSize()):
						0;

	// pack data
	int n  = 0;
	for (int e = 0; e < aryN.size(); e++) {
		if (!aryN[e]) continue;

		const int I  = static_cast<int>(src->GetSize()[0]);
		const int J  = static_cast<int>(src->GetSize()[1]);
		const int K  = static_cast<int>(src->GetSize()[2]);

		const int dvc = buffer->GetVCellSize();
		const int svc = src->GetVCellSize();

		SKL_REAL       *pdst = UtilGetS4DN(buffer,n);
		const SKL_REAL *psrc = UtilGetS4DN(src,   e);
		ricscpys3dareacomm_(pdst,
			&WrappedRect.size[0],&WrappedRect.size[1],&WrappedRect.size[2],
			&dvc,
			psrc,&I,&J,&K,&svc,
			&WrappedRect.head[0],&WrappedRect.head[1],&WrappedRect.head[2],
			&SlfVxRect.head[0],&SlfVxRect.head[1],&SlfVxRect.head[2],&vc);


		//for (int bi = 0 - vc; bi < WrappedRect.size[0] + vc; bi++) {
		//for (int bj = 0 - vc; bj < WrappedRect.size[1] + vc; bj++) {
		//for (int bk = 0 - vc; bk < WrappedRect.size[2] + vc; bk++) {
			//const int si = bi + WrappedRect.head[0] - SlfVxRect.head[0];
			//const int sj = bj + WrappedRect.head[1] - SlfVxRect.head[1];
			//const int sk = bk + WrappedRect.head[2] - SlfVxRect.head[2];
			//
			//buffer->val(bi,bj,bk,n) = src->val(si,sj,sk,e);
		//}}}

		n++;
	}

#ifdef SPHERE175
	const CfgInfo *cfg = UtilGetCfg();					//2009.12.07
	//ParaComponent
	SklParaComponent *para_cmp = cfg->solver->SklGetParaComponent();	//2009.12.07

	//ParaManager
	SklParaManager *para_mng = para_cmp->GetParaManager();			//2009.12.07
#endif

#ifdef SPHERE175
	int cid = para_cmp->ImmediateSend(buf, sz, SKL_ARRAY_DTYPE_REAL, nID, gID);	//2009.12.07
#else
	int cid = ParaMngr.ImmediateSend(buf, sz, SKL_ARRAY_DTYPE_REAL, nID, gID);
#endif

	TRY((cid != -1), "Communication Error - ImmediateSend");

	return cid;
}


int AreaCommLink::Recv(S4D_REAL* dest, const bool comm_vc) const
{
	if (!enable()) return -1;

	SKL_REAL* buf = buffer->GetData();
	int       sz  = buffer->GetArrayLength();
	int       vc  = comm_vc?
						lesser(buffer->GetVCellSize(), dest->GetVCellSize()):
						0;

#ifdef SPHERE175
	const CfgInfo *cfg = UtilGetCfg();					//2009.12.07
	//ParaComponent
	SklParaComponent *para_cmp = cfg->solver->SklGetParaComponent();	//2009.12.07

	//ParaManager
	SklParaManager *para_mng = para_cmp->GetParaManager();			//2009.12.07
#endif

#ifdef SPHERE175
	int cid = para_mng->Recv(buf, sz, SKL_ARRAY_DTYPE_REAL, nID, gID);	//2009.12.07
#else
	int cid = ParaMngr.Recv(buf, sz, SKL_ARRAY_DTYPE_REAL, nID, gID);
#endif

	TRY((cid != -1), "Communication Error - Recv");

	// recv data unpacking
	int n = 0;
	for (int e = 0; e < aryN.size(); e++) {
		if (!aryN[e]) continue;

		const int I = static_cast<int>(dest->GetSize()[0]);
		const int J = static_cast<int>(dest->GetSize()[1]);
		const int K = static_cast<int>(dest->GetSize()[2]);

		const int dvc = dest->GetVCellSize();
		const int svc = buffer->GetVCellSize();

		SKL_REAL *pdst = UtilGetS4DN(dest,  e);
		SKL_REAL *psrc = UtilGetS4DN(buffer,n);

		ricscpys3dareacomm_(pdst,
			&I,&J,&K,&dvc,
			psrc,
			&WrappedRect.size[0],&WrappedRect.size[1],&WrappedRect.size[2],
			&svc,
			&WrappedRect.head[0],&WrappedRect.head[1],&WrappedRect.head[2],
			&SlfVxRect.head[0],&SlfVxRect.head[1],&SlfVxRect.head[2],&vc);


		//for (int bi = 0 - vc; bi < WrappedRect.size[0] + vc; bi++) {
		//for (int bj = 0 - vc; bj < WrappedRect.size[1] + vc; bj++) {
		//for (int bk = 0 - vc; bk < WrappedRect.size[2] + vc; bk++) {
			//const int si = bi + WrappedRect.head[0] - SlfVxRect.head[0];
			//const int sj = bj + WrappedRect.head[1] - SlfVxRect.head[1];
			//const int sk = bk + WrappedRect.head[2] - SlfVxRect.head[2];
			//
			//dest->val(si,sj,sk,e) = buffer->val(bi,bj,bk,n);
		//}}}

		n++;
	}

	return 0;
}

string AreaCommLink::toString() const
{
#ifdef SPHERE175
	const CfgInfo *cfg = UtilGetCfg();					//2009.12.07
	//ParaComponent
	SklParaComponent *para_cmp = cfg->solver->SklGetParaComponent();	//2009.12.07

	//ParaManager
	SklParaManager *para_mng = para_cmp->GetParaManager();			//2009.12.07
#endif

	stringstream ss;

	ss << label << " @"  << gID;

#ifdef SPHERE175
	ss << " #" << para_cmp->GetMyID(gID) << ':';	//2009.12.08
#else
	ss << " #" << ParaMngr.GetMyID(gID) << ':';
#endif

	ss << "("  << SlfVxRect.head[0]
		<< ',' << SlfVxRect.head[1]
		<< ',' << SlfVxRect.head[2]
		<< ')';                      
	ss << "-(" << SlfVxRect.tail[0]
		<< ',' << SlfVxRect.tail[1]
		<< ',' << SlfVxRect.tail[2]
		<< ')';
	ss << " =>";
	ss << " #" << nID << ":";
	ss << "("  << TgtVxRect.head[0]
		<< ',' << TgtVxRect.head[1]
		<< ',' << TgtVxRect.head[2]
		<< ')';                   
	ss << "-(" << TgtVxRect.tail[0]
		<< ',' << TgtVxRect.tail[1]
		<< ',' << TgtVxRect.tail[2]
		<< ')';
	ss << " size:";
	ss << "("  << buffer->GetSize()[0]
		<< ',' << buffer->GetSize()[1]
		<< ',' << buffer->GetSize()[2]
		<< ',' << buffer->GetSize()[3]
		<< ')';
	ss << " {";
	for (int e = 0; e < aryN.size(); e++) { if (aryN[e]) ss << e << ','; }
	ss << '}';
	
	return ss.str();
}


