//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#include <sstream>
#include <iomanip>

#include <stdlib.h>

#include <Foundation/const.h>
#include <Foundation/error.h>
#include <Foundation/util.h>

#include "SklSolverFoundation.h"

unsigned int SklSolverFoundation::init_BoundaryConditionTbl()
{
#ifndef _NO_MESSAGE
	MESSAGE("Load Boundary Condition Settings ...");
#else
	MESSAGE("Load Boundary Condition Settings ...",0);
#endif //_NO_MESSAGE

	vector<BoundaryCond> tmp;

	// default condition
	tmp.push_back(BoundaryCond());
	tmp[0].cid   = 0;
	tmp[0].type  = BoundaryCond::NEUMANN;
	tmp[0].value = SKL_REAL(0.0);

	int maxid = 0;

	bool found_cid0 = false;
	const CfgElem* outerbnd = SklCfgGetTop(OUTERBND);
	// for (const CfgElem* bnd = outerbnd->GetElemFirst();
	//		bnd; bnd = outerbnd->GetElemNext(bnd)) {
	for (const CfgElem* bnd = outerbnd->GetElemFirst("BoundaryCond");
                        bnd; bnd = outerbnd->GetElemNext(bnd, "BoundaryCond")) {

                if (!(bnd->isSetID())) { continue; }

                int   type;
                float value;
                if (bnd->GetValue("type", &type) && bnd->GetValue("value", &value)) {
                        BoundaryCond bc;
                        bc.cid  = bnd->GetID();
			if( bc.cid >= maxid ) maxid = bc.cid;

                        bc.type =       (type == 0)? BoundaryCond::DIRICHLET:
                                                (type == 1)? BoundaryCond::NEUMANN:
                                                             BoundaryCond::UNKNOWN;
                        bc.value = value;

                        TRY((bc.type != BoundaryCond::UNKNOWN),
                                        "Unkown boundary condition type was specified.");

			bc.face = BoundaryCond::UNSET;
			bc.mat  = NULL;
			bc.mol  = NULL;

                        found_cid0 |= (bc.cid == 0);
                        tmp.push_back(bc);
                } else {
                        ERROR("Param element 'type' and 'value' be must specified");
                }
        }

	if (found_cid0) { tmp.erase(tmp.begin()); } // erase default condition


	for (const CfgElem* bnd = outerbnd->GetElemFirst("FaceBC");
			bnd; bnd = outerbnd->GetElemNext(bnd, "FaceBC")) {

		// if (!(bnd->isSetID())) { continue; }

		int   type;
		float value;
		const char* face;
		const char* mat;
		const char* mol;
		if (bnd->GetValue("type", &type) && bnd->GetValue("value", &value) && bnd->GetValue("face", &face)) {
			BoundaryCond bc;
			maxid++;
			bc.cid  = maxid;
			bc.type = 	(type == 0)? BoundaryCond::DIRICHLET:
						(type == 1)? BoundaryCond::NEUMANN:
						             BoundaryCond::UNKNOWN;
			bc.value = value;

			TRY((bc.type != BoundaryCond::UNKNOWN),
					"Unkown boundary condition type was specified.");

			// face
			bc.face = (strcasecmp(face, "X_MINUS") == 0)? BoundaryCond::XMINUS:
				  (strcasecmp(face, "Y_MINUS") == 0)? BoundaryCond::YMINUS:
				  (strcasecmp(face, "Z_MINUS") == 0)? BoundaryCond::ZMINUS:
				  (strcasecmp(face, "X_PLUS")  == 0)? BoundaryCond::XPLUS:
				  (strcasecmp(face, "Y_PLUS")  == 0)? BoundaryCond::YPLUS:
				  (strcasecmp(face, "Z_PLUS")  == 0)? BoundaryCond::ZPLUS:
				  (strcasecmp(face, "ALLFACE") == 0)? BoundaryCond::ALLFACE:
								      BoundaryCond::UNSET;

			// mat
			if( bnd->GetValue("mat", &mat) ){
				bc.mat = UtilGetMaterialId(string_cast(mat));
			} else {
				bc.mat = NULL;
			}

			//mol
			if( bnd->GetValue("mol", &mol) ){
				bc.mol = UtilGetMolecularId(string_cast(mol));
			} else {
				bc.mol = NULL;
			}

			tmp.push_back(bc);
		} else {
			ERROR("Param element 'type' and 'value' be must specified");
		}
	}


	const size_t bcs = tmp.size();
	cfg.BndCndTbl = new BoundaryCond[bcs];
	for (size_t i = 0; i < bcs; i++) {
		tmp[i].id = i;
		cfg.BndCndTbl[i] = tmp[i];
		cfg.BndCidIdMap.insert(Cid2IdMapT::value_type(tmp[i].cid, i));

		string bcname = 
				(tmp[i].type == BoundaryCond::DIRICHLET)?"Dirichlet":
				(tmp[i].type == BoundaryCond::NEUMANN)?  "Neumann":"";

		string facename =
				(tmp[i].face == BoundaryCond::XMINUS)? "Xminus":
				(tmp[i].face == BoundaryCond::YMINUS)? "Yminus":
				(tmp[i].face == BoundaryCond::ZMINUS)? "Zminus":
				(tmp[i].face == BoundaryCond::XPLUS)?  "Xplus":
				(tmp[i].face == BoundaryCond::YPLUS)?  "Xplus":
				(tmp[i].face == BoundaryCond::ZPLUS)?  "Zplus":
				(tmp[i].face == BoundaryCond::ALLFACE)?"AllFace":"unset";

		string matname = (tmp[i].mat != NULL)? tmp[i].mat->name:"all";
		string molname = (tmp[i].mol != NULL)? tmp[i].mol->name:"all";

		if( tmp[i].face == BoundaryCond::UNSET ) {
			MESSAGE(
				string("Boundary Condition")
					+ string(" ID = ")    + string_cast(i) + ","
					//+ string(" ID = ")    + string_cast(tmp[i].cid) + ","
					+ string(" Type = ")  + bcname         + ","
#ifndef _NO_MESSAGE
					+ string(" Value = ") + string_cast(tmp[i].value));
#else
					+ string(" Value = ") + string_cast(tmp[i].value),0);
#endif
		} else {
			MESSAGE(
				string("Boundary Condition")
					+ string(" ID = ")    + string_cast(i) + ","
					//+ string(" ID = ")    + string_cast(tmp[i].cid) + ","
					+ string(" Type = ")  + bcname         + ","
					+ string(" Value = ") + string_cast(tmp[i].value) + ","
					+ string(" face = ")  + facename       + ","
					+ string(" mat = ")   + matname        + ","
#ifndef _NO_MESSAGE
					+ string(" mol = ")   + molname);
#else
					+ string(" mol = ")   + molname,0);
#endif
		}

	}

	cfg.BndCndCount = bcs;
	return bcs;
}

int SklSolverFoundation::LoadBndCond(const SklVoxDataSet* dSet, const size_t e)
{
	const size_t I = cfg.Pdiv.size[0];
	const size_t J = cfg.Pdiv.size[1];
	const size_t K = cfg.Pdiv.size[2];

	S3D_INT* bcx = dynamic_cast<S3D_INT*>(dSet->GetData(6));
	S3D_INT* bcy = dynamic_cast<S3D_INT*>(dSet->GetData(7));
	S3D_INT* bcz = dynamic_cast<S3D_INT*>(dSet->GetData(8));

	const int tx = 0; //cfg.Pdiv.head[0];
	const int ty = 0; //cfg.Pdiv.head[1];
	const int tz = 0; //cfg.Pdiv.head[2];
	
	// X boundary
	if (bcx) {
	for (size_t k = 0; k < K; k++) {
	for (size_t j = 0; j < J; j++) {
		OutBndId_X->val(0,j,k,e) = bcx->val(tx + 0, ty + j, tz + k);
		OutBndId_X->val(1,j,k,e) = bcx->val(tx + I, ty + j, tz + k);
	}}}
	// Y boundary
	if (bcy) {
	for (size_t k = 0; k < K; k++) {
	for (size_t i = 0; i < I; i++) {
		OutBndId_Y->val(i,0,k,e) = bcy->val(tx + i, ty + 0, tz + k);
		OutBndId_Y->val(i,1,k,e) = bcy->val(tx + i, ty + J, tz + k);
	}}}
	// Z boundary
	if (bcz) {
	for (size_t j = 0; j < J; j++) {
	for (size_t i = 0; i < I; i++) {
		OutBndId_Z->val(i,j,0,e) = bcz->val(tx + i, ty + j, tz + 0);
		OutBndId_Z->val(i,j,1,e) = bcz->val(tx + i, ty + j, tz + K);
	}}}                                                      

	return (bcx || bcy || bcz)? 1: 0;
}                                                            


int SklSolverFoundation::init_FaceBndCond()
{
	const size_t I = cfg.Pdiv.size[0];
	const size_t J = cfg.Pdiv.size[1];
	const size_t K = cfg.Pdiv.size[2];

	int nID[6];
	const int& gid = cfg.grpP.gid;
#ifdef SPHERE175
	nID[0] = para_mng->IsCommID(-1, 0, 0, gid); // -X
	nID[1] = para_mng->IsCommID( 1, 0, 0, gid); // +X
	nID[2] = para_mng->IsCommID( 0,-1, 0, gid); // -Y
	nID[3] = para_mng->IsCommID( 0, 1, 0, gid); // +Y
	nID[4] = para_mng->IsCommID( 0, 0,-1, gid); // -Z
	nID[5] = para_mng->IsCommID( 0, 0, 1, gid); // +Z
#else
	nID[0] = ParaMngr.GetCommID(-1, 0, 0, gid); // -X
	nID[1] = ParaMngr.GetCommID( 1, 0, 0, gid); // +X
	nID[2] = ParaMngr.GetCommID( 0,-1, 0, gid); // -Y
	nID[3] = ParaMngr.GetCommID( 0, 1, 0, gid); // +Y
	nID[4] = ParaMngr.GetCommID( 0, 0,-1, gid); // -Z
	nID[5] = ParaMngr.GetCommID( 0, 0, 1, gid); // +Z
#endif

	for( int n = 0; n < cfg.BndCndCount; n++ ){
		// get id
		size_t id;
		if( cfg.BndCndTbl[n].face != BoundaryCond::UNSET ){
			id = cfg.BndCidIdMap.find(cfg.BndCndTbl[n].cid)->second;
		} else {
			continue;
		}

		// equation
		const Material*  mat = cfg.BndCndTbl[n].mat;
		const Molecular* mol = cfg.BndCndTbl[n].mol;

		// equation id
		vector<size_t> FaceId;
		if( (mat != NULL) && (mol != NULL) ){
			const Equation* eqn = UtilGetEquationId(mol, mat);
			if( eqn ){
				FaceId.push_back( eqn->id );
			}
		}
		else if( (mat == NULL) && (mol != NULL) ){
			for( int m = 0; m < cfg.MatCount; m++ ){
				const Equation* eqn = UtilGetEquationId(mol, cfg.MatTbl[m]);
				if( eqn ){
					FaceId.push_back( eqn->id );
				}
			}
		}
		else if( (mat != NULL) && (mol == NULL) ){
			for( int g = 0; g < cfg.MolCount; g++ ){
				const Equation* eqn = UtilGetEquationId(cfg.MolTbl[g], mat);
				if( eqn ){
					FaceId.push_back( eqn->id );
				}
			}
		}
		else{
			for( int e = 0; e < cfg.EqnCount; e++ ){
				FaceId.push_back( e );
			}
		}

		for( vector<size_t>::iterator it = FaceId.begin(); it != FaceId.end(); it++ ){

			// X Minus
			if( cfg.BndCndTbl[n].face == BoundaryCond::XMINUS ){
				int km = 0, kp = K;
				int jm = 0, jp = J;
				if( nID[4] >= 0 ) km--;
				if( nID[5] >= 0 ) kp++;
				if( nID[2] >= 0 ) jm--;
				if( nID[3] >= 0 ) jp++;

				for( int k = km; k < kp; k++ ){
				for( int j = jm; j < jp; j++ ){
					OutBndId_X->val(0,j,k,*it) = id;
				}}
			}

			// X Plus
			if( cfg.BndCndTbl[n].face == BoundaryCond::XPLUS ){
				int km = 0, kp = K;
				int jm = 0, jp = J;
				if( nID[4] >= 0 ) km--;
				if( nID[5] >= 0 ) kp++;
				if( nID[2] >= 0 ) jm--;
				if( nID[3] >= 0 ) jp++;

				for( int k = km; k < kp; k++ ){
				for( int j = jm; j < jp; j++ ){
					OutBndId_X->val(1,j,k,*it) = id;
				}}
			}

			// Y Minus
			if( cfg.BndCndTbl[n].face == BoundaryCond::YMINUS ){
				int km = 0, kp = K;
				int im = 0, ip = I;
				if( nID[4] >= 0 ) km--;
				if( nID[5] >= 0 ) kp++;
				if( nID[0] >= 0 ) im--;
				if( nID[1] >= 0 ) ip++;

				for( int k = km; k < kp; k++ ){
				for( int i = im; i < ip; i++ ){
					OutBndId_Y->val(i,0,k,*it) = id;
				}}
			}

			// Y Plus
			if( cfg.BndCndTbl[n].face == BoundaryCond::YPLUS ){
				int km = 0, kp = K;
				int im = 0, ip = I;
				if( nID[4] >= 0 ) km--;
				if( nID[5] >= 0 ) kp++;
				if( nID[0] >= 0 ) im--;
				if( nID[1] >= 0 ) ip++;

				for( size_t k = km; k < kp; k++ ){
				for( size_t i = im; i < ip; i++ ){
					OutBndId_Y->val(i,1,k,*it) = id;
				}}
			}

			// Z Minus
			if( cfg.BndCndTbl[n].face == BoundaryCond::ZMINUS ){
				int jm = 0, jp = J;
				int im = 0, ip = I;
				if( nID[2] >= 0 ) jm--;
				if( nID[3] >= 0 ) jp++;
				if( nID[0] >= 0 ) im--;
				if( nID[1] >= 0 ) ip++;

				for( size_t j = jm; j < jp; j++ ){
				for( size_t i = im; i < ip; i++ ){
					OutBndId_Z->val(i,j,0,*it) = id;
				}}
			}

			// Z Plus
			if( cfg.BndCndTbl[n].face == BoundaryCond::ZPLUS ){
				int jm = 0, jp = J;
				int im = 0, ip = I;
				if( nID[2] >= 0 ) jm--;
				if( nID[3] >= 0 ) jp++;
				if( nID[0] >= 0 ) im--;
				if( nID[1] >= 0 ) ip++;

				for( size_t j = jm; j < jp; j++ ){
				for( size_t i = im; i < ip; i++ ){
					OutBndId_Z->val(i,j,1,*it) = id;
				}}
			}

			// All Face
			if( cfg.BndCndTbl[n].face == BoundaryCond::ALLFACE ){
				int km = 0, kp = K;
				int jm = 0, jp = J;
				int im = 0, ip = I;
				if( nID[4] >= 0 ) km--;
				if( nID[5] >= 0 ) kp++;
				if( nID[2] >= 0 ) jm--;
				if( nID[3] >= 0 ) jp++;
				if( nID[0] >= 0 ) im--;
				if( nID[1] >= 0 ) ip++;

				for( size_t k = km; k < kp; k++ ){
				for( size_t j = jm; j < jp; j++ ){
					OutBndId_X->val(0,j,k,*it) = id;
					OutBndId_X->val(1,j,k,*it) = id;
				}}
				for( size_t k = km; k < kp; k++ ){
				for( size_t i = im; i < ip; i++ ){
					OutBndId_Y->val(i,0,k,*it) = id;
					OutBndId_Y->val(i,1,k,*it) = id;
				}}
				for( size_t j = jm; j < jp; j++ ){
				for( size_t i = im; i < ip; i++ ){
					OutBndId_Z->val(i,j,0,*it) = id;
					OutBndId_Z->val(i,j,1,*it) = id;
				}}
			}
		}
	}

	return 1;
}

