//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#include <sstream>
#include <iomanip>

#include <stdlib.h>
#include <math.h>

#include <Foundation/const.h>
#include <Foundation/error.h>
#include <Foundation/util.h>
#include <Foundation/api.h>

#include "SklSolverFoundation.h"

int SklSolverFoundation::LoadMatFile()
{
#ifndef _NO_MESSAGE
	MESSAGE("Load Material Volumes ...");
#else
	MESSAGE("Load Material Volumes ...",0);
#endif //_NO_MESSAGE

	if (cfg.grpAll.gid < 0) { return 1; }

	const size_t mats = cfg.MatCount;
	const size_t cmid = cfg.CmpMatId;
	const size_t vc   = VC;

	for (size_t matid = 0; matid < mats; matid++) {

		const char* name = cfg.MatTbl[matid]->name.c_str();
#ifndef _NO_MESSAGE
		MESSAGE(string("Load Volume data of ") + name);
#else
		MESSAGE(string("Load Volume data of ") + name,0);
#endif //_NO_MESSAGE

#ifdef SPHERE175
		SklVoxDataSet* dSet = LoadFile(para_mng, name, vc, cfg.grpIO.gid);	//2009.12.07
#else
		SklVoxDataSet* dSet = LoadFile(name, vc, cfg.grpIO.gid);
#endif

		if ((!dSet) && (matid == cmid)) { continue; }

		TRY(dSet, string("failed to load ") + name);
		TRY(LoadMatElement(dSet, 0, MatVfFld, matid),
			string("failed to load volume ratio from ") + name);
		TRY(LoadMatElement(dSet, 1, MatAreaX, matid),
			string("failed to load area ratio of X-face from ") + name);
		TRY(LoadMatElement(dSet, 2, MatAreaY, matid),
			string("failed to load area ratio of Y-face from ") + name);
		TRY(LoadMatElement(dSet, 3, MatAreaZ, matid),
			string("failed to load area ratio of Z-face from ") + name);

		DB4(MESSAGE(S4D2string(MatVfFld,matid,true)));
		DB4(MESSAGE(S4D2string(MatAreaX,matid,true)));
		DB4(MESSAGE(S4D2string(MatAreaY,matid,true)));
		DB4(MESSAGE(S4D2string(MatAreaZ,matid,true)));

		// boundary
		const size_t mols = cfg.MolCount;
		for (size_t molid = 0; molid < mols; molid++) {
			const Molecular* g = cfg.MolTbl[molid];
			const Material*  m = cfg.MatTbl[matid];
			const Equation*  e = UtilGetEquationId(g,m);
			if (!e) { continue; }
			LoadBndCond(dSet, e->id);
		}

		delete dSet;
	}

	return 1;
}

int SklSolverFoundation::LoadMembareaFile()
{
#ifndef _NO_MESSAGE
	MESSAGE("Load Membrane Area Volumes ...");
#else
	MESSAGE("Load Membrane Area Volumes ...",0);
#endif //_NO_MESSAGE

	if( cfg.grpAll.gid < 0 ) { return 1; }

	const size_t vc    = VC;

	for( int im = 0; im < cfg.MembareaCount; im++ ){
		const Membarea* area = cfg.MembareaTbl[im];
		if( area == NULL ) { continue; }

		const char* name = area->name.c_str();
#ifndef _NO_MESSAGE
		MESSAGE(string("Load Membrane Area data of ") + name);
#else
		MESSAGE(string("Load Membrane Area data of ") + name,0);
#endif //_NO_MESSAGE

#ifdef SPHERE175
		SklVoxDataSet* dSet = LoadFile(para_mng, name, vc, cfg.grpIO.gid);
#else
		SklVoxDataSet* dSet = LoadFile(name, vc, cfg.grpIO.gid);
#endif
		if( !dSet ) { continue; }

		TRY(dSet, string("failed to load ") + name);
		TRY(LoadMatElement(dSet, 0, MembAreaC, area->id),
			string("failed to load Membrane Area Center from ") + name);
		TRY(LoadMatElement(dSet, 1, MembAreaX, area->id),
			string("failed to load Membrane Area Edge(X) from ") + name);
		TRY(LoadMatElement(dSet, 2, MembAreaY, area->id),
			string("failed to load Membrane Area Edge(Y) from ") + name);
		TRY(LoadMatElement(dSet, 3, MembAreaZ, area->id),
			string("failed to load Membrane Area Edge(Z) from ") + name);

		// MembArea = MembArea * pitch * pitch
		RecoverMembarea();

		DB4(MESSAGE(S4D2string(MembAreaC,area->id,true)));
		DB4(MESSAGE(S4D2string(MembAreaX,area->id,true)));
		DB4(MESSAGE(S4D2string(MembAreaY,area->id,true)));
		DB4(MESSAGE(S4D2string(MembAreaZ,area->id,true)));

		delete dSet;
	}

	return 1;
}

int SklSolverFoundation::RecoverMembarea()
{
	const int vc = VC;
	const int I  = cfg.Pdiv.size[0];
	const int J  = cfg.Pdiv.size[1];
	const int K  = cfg.Pdiv.size[2];

	const SKL_REAL dx = cfg.pitch[0];
	const SKL_REAL dy = cfg.pitch[1];
	const SKL_REAL dz = cfg.pitch[2];
	SKL_REAL coef     = SKL_REAL(0.0);

	for( int im = 0; im < cfg.MembareaCount; im++ ){
		const int id  = cfg.MembareaTbl[im]->id;

		SKL_REAL* SmC = UtilGetS4DN(MembAreaC, id);
		coef = dx * dx;
		ricscpys3dcoef_(SmC,SmC,&I,&J,&K,&vc,&coef);

		SKL_REAL* SmX = UtilGetS4DN(MembAreaX, id);
		coef = dy * dz;
		ricscpys3dcoef_(SmX,SmX,&I,&J,&K,&vc,&coef);

		SKL_REAL* SmY = UtilGetS4DN(MembAreaY, id);
		coef = dx * dz;
		ricscpys3dcoef_(SmY,SmY,&I,&J,&K,&vc,&coef);

		SKL_REAL* SmZ = UtilGetS4DN(MembAreaZ, id);
		coef = dx * dy;
		ricscpys3dcoef_(SmZ,SmZ,&I,&J,&K,&vc,&coef);
	}

	return 1;
}

int SklSolverFoundation::LoadMatElement(                     
		const SklVoxDataSet* dSet, const size_t rec_id,      
		S4D_REAL* dist, const size_t matid)                  
	{
 
	const char* name = cfg.MatTbl[matid]->name.c_str();
	const size_t  vc = VC;

	S3D_FLT* in = dynamic_cast<S3D_FLT*>((dSet)? 
						dSet->GetData(rec_id):
						NULL);

	if (!in) return 0;
	
	const size_t I = cfg.IOdiv.size[0] + 1;
	const size_t J = cfg.IOdiv.size[1] + 1;
	const size_t K = cfg.IOdiv.size[2] + 1;
	for (size_t i = 0; i < I; i++) {
	for (size_t j = 0; j < J; j++) {
	for (size_t k = 0; k < K; k++) {
		dist->val(i,j,k,matid) = static_cast<SKL_REAL>(in->val(i,j,k));
	}}}

	return 1;
}

int SklSolverFoundation::LoadMatElement(
		const SklVoxDataSet* dSet, const size_t rec_id,
		S4D_INT* dist, const size_t matid)
{
	const char* name = cfg.MatTbl[matid]->name.c_str();
	const size_t  vc = VC;

	S3D_INT* in = dynamic_cast<S3D_INT*>((dSet)? 
							dSet->GetData(rec_id):
							NULL);

	if (!in) return 0;

	const size_t I = cfg.IOdiv.size[0] + 1;
	const size_t J = cfg.IOdiv.size[1] + 1;
	const size_t K = cfg.IOdiv.size[2] + 1;
	for (size_t i = 0; i < I; i++) {
	for (size_t j = 0; j < J; j++) {
	for (size_t k = 0; k < K; k++) {
		dist->val(i,j,k,matid) = in->val(i,j,k);
	}}}

	return 1;
}

int SklSolverFoundation::LoadMatElement(
		const SklVoxDataSet* dSet, const size_t rec_id, S3D_INT* dist)
{
	// i am node #0
	S3D_INT* in = dynamic_cast<S3D_INT*>((dSet)? 
						dSet->GetData(rec_id):
						NULL);
	if (!in) return 0;
	SklUtil::cpyS3D(dist, in);

	return 1;
}

int SklSolverFoundation::LoadEqnFile()
{
#ifndef _NO_MESSAGE
	MESSAGE("Load Initial Phi Value from SPH files ...");
#else
	MESSAGE("Load Initial Phi Value from SPH files ...",0);
#endif //_NO_MESSAGE

	// Load each initial phi(l,m)^0
	const size_t eqns = cfg.EqnCount;
	const size_t I    = cfg.Pdiv.size[0] + 1;
	const size_t J    = cfg.Pdiv.size[1] + 1;
	const size_t K    = cfg.Pdiv.size[2] + 1;
	const int    vc   = VC;

	for (size_t e = 0; e < eqns; e++) {

		if (!cfg.eqnIO[e]) { continue; }

		const Equation* eqn = cfg.EqnTbl[e];

		const SKL_REAL amp   = SKL_REAL(1.0);
		const char*    attr  = cfg.EqnTbl[e]->name.c_str();

		UtilLoadSph(attr, EqnP0Fld, e, cfg.grpIO.gid);

#ifdef ECELL_3_1
                // mask by material volume
                for (int k = 0; k < K; k++) {
                for (int j = 0; j < J; j++) {
                for (int i = 0; i < I; i++) {
                        EqnP0Fld->val(i,j,k,e) *= ceil(MatVfFld->val(i,j,k,eqn->mat->id));
                }}}
#else
#ifndef _WITHOUT_ECELL
		// mask by material volume
		for (int k = 0; k < K; k++) {
		for (int j = 0; j < J; j++) {
		for (int i = 0; i < I; i++) {
			EqnP0Fld->val(i,j,k,e) *= ceil(MatVfFld->val(i,j,k,eqn->mat->id));
		}}}
#endif // _WITHOUT_ECELL
#endif //ECELL_3_1
	}

	if((cfg.Restart) && (cfg.time_integ == ADAMS_BASHFORTH)){
		LoadEqnRstFile();
	}

	return 1;
}

int SklSolverFoundation::LoadEqnRstFile()
{
	const size_t eqns = cfg.EqnCount;
	// const size_t I    = cfg.Pdiv.size[0] + 1;
	// const size_t J    = cfg.Pdiv.size[1] + 1;
	// const size_t K    = cfg.Pdiv.size[2] + 1;

	cfg.RhsoInitFlag = 1;

	for(size_t e=0; e<eqns; e++){
		if(!cfg.eqnIO[e]) { continue; }

		const Equation* eqn = cfg.EqnTbl[e];

		string buf = string("RHS_") + eqn->name;
		const char* attr = buf.c_str();

		if(UtilLoadSph(attr, Rhso, e, cfg.grpIO.gid) == 0){
			cfg.RhsoInitFlag = 0;
#ifndef _NO_MESSAGE
			MESSAGE(UtilStrFormat("Sph File is not loaded. : attr = %s", attr));
#else
			MESSAGE(UtilStrFormat("Sph File is not loaded. : attr = %s", attr),0);
#endif //_NO_MESSAGE
		}
	}

	return 1;
}

int SklSolverFoundation::LoadEqnBndCond()
{
#ifndef _NO_MESSAGE
	MESSAGE("Load Boundary Condition ...");
#else
	MESSAGE("Load Boundary Condition ...",0);
#endif //_NO_MESSAGE

	// Load each initial phi(l,m)^0
	const size_t eqns = cfg.EqnCount;
	const int    vc   = VC;

	for (size_t e = 0; e < eqns; e++) {

		const string   attr = cfg.EqnTbl[e]->name + "_bc";

#ifdef SPHERE175
		SklVoxDataSet* dSet = LoadFile(para_mng, attr.c_str(), vc, cfg.grpIO.gid);	//2009.12.07
#else
		SklVoxDataSet* dSet = LoadFile(attr.c_str(), vc, cfg.grpIO.gid);
#endif
		if (!dSet) { continue; }

		// set base time for resart
		SetRestartTime(dSet);

		// set bc from bc element of dSet
		TRY(LoadBndCond(dSet, e), attr + " has no boundary condition record");

		delete dSet;
	}

	return 1;
}

int SklSolverFoundation::LoadDifCoef()
{
#ifndef _NO_MESSAGE
	MESSAGE("Load Coef Value ...");
#else
	MESSAGE("Load Coef Value ...",0);
#endif //_NO_MESSAGE

	const size_t eqns = cfg.EqnCount;
	const size_t I    = cfg.Pdiv.size[0] + 1;
	const size_t J    = cfg.Pdiv.size[1] + 1;
	const size_t K    = cfg.Pdiv.size[2] + 1;
	const int    vc   = VC;

	for (size_t e = 0; e < eqns; e++) {
		const string attr = cfg.EqnTbl[e]->name + "_ccoef";

#ifdef SPHERE175
		SklVoxDataSet* dSet = LoadFile(para_mng, attr.c_str(), vc, cfg.grpIO.gid);	//2009.12.07
#else
		SklVoxDataSet* dSet = LoadFile(attr.c_str(), vc, cfg.grpIO.gid);
#endif
		if (!dSet) { continue; }
#ifndef _NO_MESSAGE
		MESSAGE(UtilStrFormat("Load DifCoef field of %s", attr.c_str()));
#else
		MESSAGE(UtilStrFormat("Load DifCoef field of %s", attr.c_str()),0);
#endif //_NO_MESSAGE

		// copy input data to fld
		if (dSet->GetDataType() == SklVoxDataSet::TYPE_SINGLE) {
			if (dSet->GetSVType() == 1) { // scalar
				S3D_FLT* in = (dSet)? dynamic_cast<S3D_FLT*>(
								dSet->GetData(SklVoxDataSet::SPH_DATA)):
								NULL;
				TRY(in, string("failed to load ") + attr);
				for (size_t k = 0; k < K; k++) {
				for (size_t j = 0; j < J; j++) {
				for (size_t i = 0; i < I; i++) {
					DifCoefX->val(i,j,k,e) = static_cast<SKL_REAL>(in->val(i,j,k));
					DifCoefY->val(i,j,k,e) = static_cast<SKL_REAL>(in->val(i,j,k));
					DifCoefZ->val(i,j,k,e) = static_cast<SKL_REAL>(in->val(i,j,k));
				}}}
			} else if (dSet->GetSVType() == 2) { // vector
				V3D_FLT* in = (dSet)? dynamic_cast<V3D_FLT*>(
								dSet->GetData(SklVoxDataSet::SPH_DATA)):
								NULL;
				TRY(in, string("failed to load ") + attr);
				for (size_t k = 0; k < K; k++) {
				for (size_t j = 0; j < J; j++) {
				for (size_t i = 0; i < I; i++) {
					DifCoefX->val(i,j,k,e) = static_cast<SKL_REAL>(in->u(i,j,k));
					DifCoefY->val(i,j,k,e) = static_cast<SKL_REAL>(in->v(i,j,k));
					DifCoefZ->val(i,j,k,e) = static_cast<SKL_REAL>(in->w(i,j,k));
				}}}
			} else {
				ERROR(string("failed to load ") + attr);
			}
		} else if (dSet->GetDataType() == SklVoxDataSet::TYPE_DOUBLE) {
			if (dSet->GetSVType() == 1) { // scalar
				S3D_DBL* in = (dSet)? dynamic_cast<S3D_DBL*>(
								dSet->GetData(SklVoxDataSet::SPH_DATA)):
								NULL;
				TRY(in, string("failed to load ") + attr);
				for (size_t k = 0; k < K; k++) {
				for (size_t j = 0; j < J; j++) {
				for (size_t i = 0; i < I; i++) {
					DifCoefX->val(i,j,k,e) = static_cast<SKL_REAL>(in->val(i,j,k));
					DifCoefY->val(i,j,k,e) = static_cast<SKL_REAL>(in->val(i,j,k));
					DifCoefZ->val(i,j,k,e) = static_cast<SKL_REAL>(in->val(i,j,k));
				}}}
			} else if (dSet->GetSVType() == 2) { // vector
				V3D_DBL* in = (dSet)? dynamic_cast<V3D_DBL*>(
								dSet->GetData(SklVoxDataSet::SPH_DATA)):
								NULL;
				TRY(in, string("failed to load ") + attr);
				for (size_t k = 0; k < K; k++) {
				for (size_t j = 0; j < J; j++) {
				for (size_t i = 0; i < I; i++) {
					DifCoefX->val(i,j,k,e) = static_cast<SKL_REAL>(in->u(i,j,k));
					DifCoefY->val(i,j,k,e) = static_cast<SKL_REAL>(in->v(i,j,k));
					DifCoefZ->val(i,j,k,e) = static_cast<SKL_REAL>(in->w(i,j,k));
				}}}
			} else {
				ERROR(string("failed to load ") + attr);
			}
		} else {
			ERROR(attr + " data type is unknown");
			delete dSet;
		}

		// set base time for resart
		SetRestartTime(dSet);

		delete dSet;
	}

	return 1;
}


