//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#include <sstream>
#include <iomanip>

#include <stdlib.h>

#include <Foundation/const.h>
#include <Foundation/error.h>
#include <Foundation/util.h>
#include <Foundation/api.h>

#include "SklSolverFoundation.h"

int SklSolverFoundation::InitParallelEnvironment()
{
#ifndef _NO_MESSAGE
	MESSAGE("Initialize Parallel Computing Environment ...");
#else
	MESSAGE("Initialize Parallel Computing Environment ...",0);
#endif //_NO_MESSAGE

	const unsigned int& isz = cfg.size[0];
	const unsigned int& jsz = cfg.size[1];
	const unsigned int& ksz = cfg.size[2];

	const CfgElem* steer = SklCfgGetTop(STEER);
	const CfgElem* para  = steer->GetElemValue("ParallelComputing");

#ifdef SPHERE175
	para_cmp->SklVoxelInit(isz,jsz,ksz,0,0,0,0);	//2009.12.08
	cfg.Parallel = para_cmp->IsParallel();		//2009.12.07
#else
	ParaMngr.SklVoxelInit(isz,jsz,ksz,0,0,0,0);
	cfg.Parallel = ParaMngr.IsParallel();
#endif

#ifndef _NO_MESSAGE
	if (cfg.Parallel) {
		MESSAGE("+ Parallel Computing Now!!");
	} else {
		MESSAGE("+ Serial Computing Now!!");
	}
#else
	if (cfg.Parallel) {
		MESSAGE("+ Parallel Computing Now!!",0);
	} else {
		MESSAGE("+ Serial Computing Now!!",0);
	}
#endif //_NO_MESSAGE

	CfgInfo::Communicator* comm[] = {
							&cfg.grpAll,  &cfg.grpIO,
							&cfg.grpComp, &cfg.grpP,  &cfg.grpE };
	for (size_t i = 0; i < 5; i++) {
		comm[i]->gid = cfg.Parallel? -1: 0;
		comm[i]->nid = cfg.Parallel? -1: 0;
		comm[i]->num = cfg.Parallel? -1: 1;
	}

	////////////////////////////////////////////////////////////////////
	// Input/Output mode
	const CfgElem*  IOnode = para? para->GetElemFirst ("IONodeSetting"): NULL;
	const CfgParam* gthrIO = IOnode? IOnode->GetParamFirst("gather"): NULL;
	const CfgParam* spclIO = IOnode? IOnode->GetParamFirst("specialize"): NULL;
	int tmp = 0;
	cfg.GatherIO       =  gthrIO? gthrIO->GetData(&tmp),tmp: 0;
	cfg.SpecialIONode  =  spclIO? spclIO->GetData(&tmp),tmp: 0;
	cfg.GatherIO       &= cfg.Parallel;
	cfg.SpecialIONode  &= cfg.GatherIO;
	const unsigned int node_offset = (cfg.SpecialIONode)? 1: 0;

#if 1
	cfg.GatherIO      = false;
	cfg.SpecialIONode = false;
#endif

	////////////////////////////////////////////////////////////////////
	// build job assignment table
	JobManageTable jmt;
	if (cfg.Parallel) {
		BuildJobManageTable(jmt);
	} else {
		JobEntry  je;
		MatAssign ma;

		for (int e = 0; e < cfg.EqnCount; e++) {
			je.Eqns.first.push_back(cfg.EqnTbl[e]);
		}
		for (int m = 0; m < cfg.MatCount; m++) {
			ma.first.push_back(cfg.MatTbl[m]);
		}
		for (int i = 0; i < 4; i++) {
			je.Eqns.second[i] = (i == 4)? 1: 0;
			ma.second[i]      = (i == 4)? 1: 0;
		}

		je.Mats.push_back(ma);
		jmt.push_back(je);
	}

	////////////////////////////////////////////////////////////////////
	// check node count
	int Nr  = node_offset;
	int _Nr = node_offset;

	if (cfg.Parallel) {
#ifndef _NO_MESSAGE
		MESSAGE("checking required node count...");
#else
		MESSAGE("checking required node count...",0);
#endif //_NO_MESSAGE
		for (JobManageTable::const_iterator
				je  = jmt.begin();
				je != jmt.end();
				je++) {

			const unsigned int* div   = je->Eqns.second;
			const unsigned int  dsize = (div[3])? div[3]: div[0]*div[1]*div[2];
			Nr += dsize;

			for (MatAssignSet::const_iterator
					ma  = je->Mats.begin();
					ma != je->Mats.end();
					ma++) {

				const unsigned int* div   = ma->second;
				const unsigned int  dsize = div[3]?div[3]:div[0]*div[1]*div[2];
				_Nr += dsize;
			}
		}

#ifndef _NO_MESSAGE
		MESSAGE(string("required computation node count: ")+string_cast(Nr));
#else
		MESSAGE(string("required computation node count: ")+string_cast(Nr),0);
#endif //_NO_MESSAGE

		TRY(
#ifdef SPHERE175
		(para_mng->GetNodeNum(0) >= Nr),
#else
		(ParaMngr.GetNodeNum(0) >= Nr),
#endif
			"Node count is less than parallel configuration requires");
	} else {
		Nr  = 1;
		_Nr = 1;
	}

	////////////////////////////////////////////////////////////////////
	// create io and worker node group
	
	// group all nodes
	cfg.grpAll.gid = 0;

#ifdef SPHERE175
	cfg.grpAll.nid = para_cmp->GetMyID(0);		//2009.12.08
	cfg.grpAll.num = para_mng->GetNodeNum(0);	//2009.12.07
#else
	cfg.grpAll.nid = ParaMngr.GetMyID(0);
	cfg.grpAll.num = ParaMngr.GetNodeNum(0);
#endif

	BuildGroupComputation(jmt);
	BuildGroupPandE(jmt);
	BuildGroupIO(jmt);

	cfg.IONode = cfg.AssIOEqns.size();

	// Assgined Material Array
	cfg.eqnIO   = new char[cfg.EqnCount]; memset(cfg.eqnIO  ,0,cfg.EqnCount);
	cfg.matIO   = new char[cfg.MatCount]; memset(cfg.matIO  ,0,cfg.MatCount);
	cfg.molIO   = new char[cfg.MolCount]; memset(cfg.molIO  ,0,cfg.MolCount);
	cfg.PeqnAss = new char[cfg.EqnCount]; memset(cfg.PeqnAss,0,cfg.EqnCount);
	cfg.PmolAss = new char[cfg.MolCount]; memset(cfg.PmolAss,0,cfg.MolCount);
	cfg.EmatAss = new char[cfg.MatCount]; memset(cfg.EmatAss,0,cfg.MatCount);
	for (EqnSet::const_iterator ei = cfg.AssIOEqns.begin();
			ei != cfg.AssIOEqns.end(); ei++) {

		const size_t e  = (*ei)->id;
		const size_t mt = (*ei)->mat->id;
		const size_t ml = (*ei)->mol->id;
		cfg.eqnIO[e]  = 1;
		cfg.matIO[mt] = 1;
		cfg.molIO[ml] = 1;
	}
	for (EqnSet::const_iterator ei = cfg.AssEqns.begin();
			ei != cfg.AssEqns.end(); ei++) {
		const size_t e  = (*ei)->id;
		const size_t ml = (*ei)->mol->id;
		cfg.PeqnAss[e]  = 1;
		cfg.PmolAss[ml] = 1;
	}
	for (MatSet::const_iterator mi = cfg.AssMats.begin();
			mi != cfg.AssMats.end(); mi++) {
		const size_t mt = (*mi)->id;
		cfg.EmatAss[mt] = 1;
	}

	return 1;
}

int SklSolverFoundation::InitParallelCommLink()
{
	if (!cfg.Parallel) { return 0; }

	// IO-Comm (IO <-> P) Link
//	BuildIOCommLink();

	// P2P-Comm (P <-> P) Link
	BuildP2PBCommLink();
	BuildP2PCCommLink();

	// P <-> E Link
	BuildP2ECommLink();
	BuildE2PCommLink();

	// Vf (P <-> E) Link
	BuildVfP2ECommLink();
	BuildVfE2PCommLink();

	// Membarea (P <-> E) Link
	BuildSmP2ECommLink();
	BuildSmE2PCommLink();
	return 1;
}

int SklSolverFoundation::BuildGroupComputation(const JobManageTable& jmt)
{
	const unsigned int isz = cfg.size[0];
	const unsigned int jsz = cfg.size[1];
	const unsigned int ksz = cfg.size[2];

	const int node_offset = cfg.SpecialIONode? 1: 0;
	const int nodes = cfg.grpAll.num - node_offset;
 	int* node = new int[nodes];
 
	for (int nid = 0; nid < nodes; nid++) { node[nid] = nid + node_offset; }
	int gid = (cfg.Parallel)?

#ifdef SPHERE175
				para_cmp->CreateProcessGroup(nodes, node, cfg.grpAll.gid): 0;	//2009.12.07
#else
				ParaMngr.CreateProcessGroup(nodes, node, cfg.grpAll.gid): 0;
#endif

#ifdef SPHERE175
	para_cmp->SklVoxelInit(isz,jsz,ksz,0,0,0,gid);	//2009.12.08
#else
	ParaMngr.SklVoxelInit(isz,jsz,ksz,0,0,0,gid);
#endif

	cfg.grpComp.gid = gid;

#ifdef SPHERE175
	cfg.grpComp.nid = para_cmp->GetMyID(gid);	//2009.12.08
	cfg.grpComp.num = para_mng->GetNodeNum(gid);	//2009.12.07
#else
	cfg.grpComp.nid = ParaMngr.GetMyID(gid);
	cfg.grpComp.num = ParaMngr.GetNodeNum(gid);
#endif
	delete[] node;

	return gid;
}

int SklSolverFoundation::BuildGroupPandE(const JobManageTable& jmt)
{
	const unsigned int isz = cfg.size[0];
	const unsigned int jsz = cfg.size[1];
	const unsigned int ksz = cfg.size[2];

	unsigned int PVoxDiv[4] = { 0 };
	unsigned int EVoxDiv[4] = { 0 };

	int Pnode_ite = 0;
	for (JobManageTable::const_iterator
			je  = jmt.begin();
			je != jmt.end();
			je++) {

		const unsigned int* Pdiv  = je->Eqns.second;
		const unsigned int  Psz   = Pdiv[3]? Pdiv[3]: Pdiv[0]*Pdiv[1]*Pdiv[2];
		const unsigned int  Pidiv = Pdiv[3]? 0: Pdiv[0];
		const unsigned int  Pjdiv = Pdiv[3]? 0: Pdiv[1];
		const unsigned int  Pkdiv = Pdiv[3]? 0: Pdiv[2];

		// make node list of sub group
		int* Pnodes = new int[Psz];
		for (int nid = 0; nid < Psz; nid++) { Pnodes[nid] = Pnode_ite++; }

		// create sub group
		const int Pgid = (cfg.Parallel)?
#ifdef SPHERE175
					para_cmp->CreateProcessGroup(Psz,Pnodes,cfg.grpComp.gid): 0;	//2009.12.07
#else
					ParaMngr.CreateProcessGroup(Psz,Pnodes,cfg.grpComp.gid): 0;
#endif
		delete[] Pnodes;

		// is this node a member of the new group?
		if (Pgid < 0) { continue; }

#ifdef SPHERE175
		para_cmp->SklVoxelInit(isz,jsz,ksz,Pidiv,Pjdiv,Pkdiv,Pgid);	//2009.12.08
#else
		ParaMngr.SklVoxelInit(isz,jsz,ksz,Pidiv,Pjdiv,Pkdiv,Pgid);
#endif

		cfg.grpP.gid = Pgid;

#ifdef SPHERE175
		cfg.grpP.nid = para_cmp->GetMyID(Pgid);		//2009.12.08
		cfg.grpP.num = para_mng->GetNodeNum(Pgid);	//2009.12.07
#else
		cfg.grpP.nid = ParaMngr.GetMyID(Pgid);
		cfg.grpP.num = ParaMngr.GetNodeNum(Pgid);
#endif

		for (unsigned int d = 0; d < 3; d++) {
			const int gid = cfg.grpP.gid;
			const int nid = cfg.grpP.nid;
#ifdef SPHERE175
			cfg.Pdiv.head[d] = para_mng->GetVoxelHeadIndex(nid, d, gid);	//2009.12.07
			cfg.Pdiv.tail[d] = para_mng->GetVoxelTailIndex(nid, d, gid);	//2009.12.07
			cfg.Pdiv.size[d] = para_mng->GetVoxelSize(gid)[d];		//2009.12.07
#else
			cfg.Pdiv.head[d] = ParaMngr.GetVoxelHeadIndex(nid, d, gid);
			cfg.Pdiv.tail[d] = ParaMngr.GetVoxelTailIndex(nid, d, gid);
			cfg.Pdiv.size[d] = ParaMngr.GetVoxelSize(gid)[d];
#endif
		}
		
		for (EqnSet::const_iterator
				e =  je->Eqns.first.begin();
				e != je->Eqns.first.end();
				e++) {

			cfg.AssEqns.push_back(*e);
		}

		// sub groups for E term computing
		int Enode_ite = 0;
		for (MatAssignSet::const_iterator
				ma  = je->Mats.begin();
				ma != je->Mats.end();
				ma++) {

			const unsigned int* Ediv = ma->second;
			const unsigned int Esz   = Ediv[3]?Ediv[3]:Ediv[0]*Ediv[1]*Ediv[2];
			const unsigned int Eidiv = Ediv[3]? 0: Ediv[0];
			const unsigned int Ejdiv = Ediv[3]? 0: Ediv[1];
			const unsigned int Ekdiv = Ediv[3]? 0: Ediv[2];
 			int* Enodes = new int[Esz];

			for (int nid = 0; nid < Esz; nid++) { Enodes[nid] = Enode_ite++; }
			const int Egid = (cfg.Parallel)?
#ifdef SPHERE175
					para_cmp->CreateProcessGroup(Esz,Enodes,cfg.grpP.gid): 0;	//2009.12.07
#else
					ParaMngr.CreateProcessGroup(Esz,Enodes,cfg.grpP.gid): 0;
#endif
			delete[] Enodes;

			// is this node a member of the new group?
			if (Egid < 0) { continue; }

#ifdef SPHERE175
			para_cmp->SklVoxelInit(isz,jsz,ksz,Eidiv,Ejdiv,Ekdiv,Egid);	//2009.12.08
#else
			ParaMngr.SklVoxelInit(isz,jsz,ksz,Eidiv,Ejdiv,Ekdiv,Egid);
#endif

			cfg.grpE.gid = Egid;

#ifdef SPHERE175
			cfg.grpE.nid = para_cmp->GetMyID(Egid);		//2009.12.08
			cfg.grpE.num = para_mng->GetNodeNum(Egid);	//2009.12.07
#else
			cfg.grpE.nid = ParaMngr.GetMyID(Egid);
			cfg.grpE.num = ParaMngr.GetNodeNum(Egid);
#endif

			for (unsigned int d = 0; d < 3; d++) {
				const int gid = cfg.grpE.gid;
				const int nid = cfg.grpE.nid;
#ifdef SPHERE175
				cfg.Ediv.head[d] = para_mng->GetVoxelHeadIndex(nid, d, gid);	//2009.12.07
				cfg.Ediv.tail[d] = para_mng->GetVoxelTailIndex(nid, d, gid);	//2009.12.07
				cfg.Ediv.size[d] = para_mng->GetVoxelSize(gid)[d];		//2009.12.07
#else
				cfg.Ediv.head[d] = ParaMngr.GetVoxelHeadIndex(nid, d, gid);
				cfg.Ediv.tail[d] = ParaMngr.GetVoxelTailIndex(nid, d, gid);
				cfg.Ediv.size[d] = ParaMngr.GetVoxelSize(gid)[d];
#endif
			}

			// assignment jobs
			for (MatSet::const_iterator
					m =  ma->first.begin();
					m != ma->first.end();
					m++) {
				cfg.AssMats.push_back(*m);
			}
		}
	}

	return cfg.grpP.gid;
}

int SklSolverFoundation::BuildGroupIO(const JobManageTable& jmt)
{
	const unsigned int isz = cfg.size[0];
	const unsigned int jsz = cfg.size[1];
	const unsigned int ksz = cfg.size[2];

	if (cfg.GatherIO) {
		int io_node[1] = { 0 };
		int io_gid = (cfg.Parallel)?
#ifdef SPHERE175
						para_cmp->CreateProcessGroup(1,io_node,cfg.grpAll.gid):		//2009.12.07
						0;
#else
						ParaMngr.CreateProcessGroup(1,io_node,cfg.grpAll.gid):
						0;
#endif

		for (unsigned int d = 0; d < 3; d++) {
			cfg.grpIO.gid = -1;
			cfg.grpIO.nid = -1;
			cfg.grpIO.num = 0;
			cfg.IOdiv.head[d] = 0;
			cfg.IOdiv.tail[d] = cfg.size[d] - 1;
			cfg.IOdiv.size[d] = cfg.size[d];
		}

		if (io_gid > 0) {
			cfg.grpIO.gid = io_gid;
#ifdef SPHERE175
			cfg.grpIO.nid = para_cmp->GetMyID(io_gid);		//2009.12.08
			cfg.grpIO.num = para_mng->GetNodeNum(io_gid);		//2009.12.07
			para_cmp->SklVoxelInit(isz,jsz,ksz,0,0,0,io_gid);	//2009.12.08
#else
			cfg.grpIO.nid = ParaMngr.GetMyID(io_gid);
			cfg.grpIO.num = ParaMngr.GetNodeNum(io_gid);
			ParaMngr.SklVoxelInit(isz,jsz,ksz,0,0,0,io_gid);
#endif

			if (cfg.SpecialIONode) {
				cfg.grpP = cfg.grpIO;
				cfg.grpE = cfg.grpIO;
				cfg.Pdiv = cfg.IOdiv;
				cfg.Ediv = cfg.IOdiv;
			}

			for (int e = 0; e < cfg.EqnCount; e++) {
				cfg.AssIOEqns.push_back(cfg.EqnTbl[e]);
			}
		}
	} else {
		cfg.grpIO = cfg.grpP;
		for (unsigned int d = 0; d < 3; d++) {
			cfg.IOdiv.head[d] = cfg.Pdiv.head[d];
			cfg.IOdiv.tail[d] = cfg.Pdiv.tail[d];
			cfg.IOdiv.size[d] = cfg.Pdiv.size[d];
		}

		cfg.AssIOEqns = cfg.AssEqns;
	}

	return cfg.grpIO.gid;
}

int SklSolverFoundation::BuildIOCommLink()
{
	if (!cfg.GatherIO) { return 0; }

	vector<char> Es(cfg.PeqnAss, cfg.PeqnAss + cfg.MatCount);

	if (cfg.SpecialIONode) {
		const int nodes = cfg.grpAll.num;
		const int gid   = cfg.grpAll.gid;
		for (int n = 0; n < nodes; n++) {
			AreaCommLink* link;
			link = new AreaCommLink(cfg,"IO",n,gid,cfg.Pdiv,cfg.Pdiv,Es);
			if (!link->enable()) { delete link; continue; }
			IOlink.push_back(link);
		}
	} else if (cfg.GatherIO) {
		const int nodes = cfg.grpAll.num;
		const int gid   = cfg.grpAll.gid;
		for (int n = 0; n < nodes; n++) {
			AreaCommLink* link;

			link = new AreaCommLink(cfg,"IO",n,gid,cfg.Pdiv,cfg.IOdiv,Es);
			if ((!link->enable()) || (link->node() != 0)) { delete link; }
			else { IOlink.push_back(link); }

			link = new AreaCommLink(cfg,"IO",n,gid,cfg.IOdiv,cfg.Pdiv,Es);
			if ((!link->enable()) || (cfg.grpAll.nid != 0)) { delete link; }
			else { IOlink.push_back(link); }
		}
	}

	return IOlink.size();
}

int SklSolverFoundation::BuildP2PBCommLink()
{
	const int nodes = cfg.grpComp.num;
	const int gid   = cfg.grpComp.gid;

	vector<char> Es(cfg.EqnCount);
	for (int e = 0; e < Es.size(); e++) { Es[e] = cfg.PeqnAss[e]? 1: 0; }

	for (int n = 0; n < nodes; n++) {
		AreaCommLink* link;
		link = new AreaCommLink(cfg,"P2PB", n, gid, cfg.Pdiv, cfg.Pdiv, Es);
		if (n == cfg.grpComp.nid) { delete link; continue; }
		if (!link->enable())      { delete link; continue; }
		P2PBlink.push_back(link);
	}

	return P2PBlink.size();
}

int SklSolverFoundation::BuildP2PCCommLink()
{
	const int nodes = cfg.grpComp.num;
	const int gid   = cfg.grpComp.gid;

	for (int n = 0; n < nodes; n++) {
 		char* Es = new char[cfg.EqnCount];
		for (int e = 0; e < cfg.EqnCount; e++) {
			Es[e] = cfg.PeqnAss[e]? 1: 0;
		}

#ifdef SPHERE175
		para_mng->Broadcast(Es,cfg.EqnCount,SKL_ARRAY_DTYPE_CHAR,n,gid);	//2009.12.07
#else
		ParaMngr.Broadcast(Es,cfg.EqnCount,SKL_ARRAY_DTYPE_CHAR,n,gid);
#endif

		AreaCommLink* link;
		link = new AreaCommLink(cfg,"P2PC", n, gid, cfg.Pdiv, cfg.Pdiv, vector<char>(Es,Es + cfg.EqnCount));
		if (n == cfg.grpComp.nid) { delete link; continue; }
		if (!link->enable())      { delete link; continue; }
		P2PClink.push_back(link);

		delete[] Es;
	}

	return P2PClink.size();
}

int SklSolverFoundation::BuildP2ECommLink()
{
	const int nodes = cfg.grpP.num;
	const int gid   = cfg.grpP.gid;

	for (int n = 0; n < nodes; n++) {
 		char* Es = new char[cfg.EqnCount];
		for (int e = 0; e < cfg.EqnCount; e++) {
			const size_t m = cfg.EqnTbl[e]->mat->id;
			Es[e] = cfg.EmatAss[m]? 1: 0;
		}
		for (int m = 0; m < cfg.MtbCount; m++) {
			if (!cfg.EmatAss[cfg.MtbTbl[m]->mat->id]) { continue; }
			if (!cfg.MtbTbl[m]->reactsite) { continue; }
			Es[cfg.MtbTbl[m]->LR->id] = 1;
			MESSAGE(cfg.MtbTbl[m]->name
					+ "needs " + cfg.MtbTbl[m]->LR->name
#ifndef _NO_MESSAGE
					+ " by receptor reaction site");
#else
					+ " by receptor reaction site",0);
#endif //_NO_MESSAGE
		}

#ifdef SPHERE175
		para_mng->Broadcast(Es,cfg.EqnCount,SKL_ARRAY_DTYPE_CHAR,n,gid);	//2009.12.07
#else
		ParaMngr.Broadcast(Es,cfg.EqnCount,SKL_ARRAY_DTYPE_CHAR,n,gid);
#endif

		AreaCommLink* link;
		link = new AreaCommLink(cfg,"P2E",n,gid,cfg.Pdiv,cfg.Ediv,vector<char>(Es,Es + cfg.EqnCount));
		if (!link->enable()) { delete link; continue; }
		P2Elink.push_back(link);

		delete[] Es;
	}

	return P2Elink.size();
}

int SklSolverFoundation::BuildE2PCommLink()
{
	const int nodes = cfg.grpP.num;
	const int gid   = cfg.grpP.gid;

	vector<char> Es(cfg.EqnCount);
	for (int e = 0; e < Es.size(); e++) {
		const size_t m = cfg.EqnTbl[e]->mat->id;
		Es[e] = cfg.EmatAss[m]? 1: 0;
	}
	for (int m = 0; m < cfg.MtbCount; m++) {
		if (!cfg.EmatAss[cfg.MtbTbl[m]->mat->id]) { continue; }
		if (!cfg.MtbTbl[m]->reactsite) { continue; }
		Es[cfg.MtbTbl[m]->LR->id] = 1;
	}
	for (int n = 0; n < nodes; n++) {
		AreaCommLink* link;
		link = new AreaCommLink(cfg,"E2P",n,gid,cfg.Ediv,cfg.Pdiv,Es);
		if (!link->enable()) { delete link; continue; }
		E2Plink.push_back(link);
	}

	return E2Plink.size();
}

int SklSolverFoundation::BuildVfP2ECommLink()
{
	const int nodes = cfg.grpP.num;
	const int gid   = cfg.grpP.gid;

	for (int n = 0; n < nodes; n++) {
		const size_t s = cfg.MatCount;
		char* Ms = new char[s];
		for (int m = 0; m < s; m++) { Ms[m] = cfg.EmatAss[m]?1:0; }
		for (int m = 0; m < cfg.MtbCount; m++) {
			if (!cfg.EmatAss[cfg.MtbTbl[m]->mat->id]) { continue; }
			if (!cfg.MtbTbl[m]->reactsite) { continue; }
			Ms[cfg.MtbTbl[m]->LR->mat->id] = 1;
			MESSAGE(cfg.MtbTbl[m]->mat->name
					+ "needs Vf[" + cfg.MtbTbl[m]->LR->mat->name + "]"
#ifndef _NO_MESSAGE
					+ "by receptor reaction site");
#else
					+ "by receptor reaction site",0);
#endif //_NO_MESSAGE
		}

#ifdef SPHERE175
		para_mng->Broadcast(Ms,s,SKL_ARRAY_DTYPE_CHAR,n,gid);	//2009.12.07
#else
		ParaMngr.Broadcast(Ms,s,SKL_ARRAY_DTYPE_CHAR,n,gid);
#endif

		AreaCommLink* link;
		
		link = new AreaCommLink(cfg,"Vf ",n,gid,cfg.Pdiv,cfg.Ediv,
									vector<char>(Ms,Ms + cfg.MatCount));

		delete[] Ms;
		if (!link->enable()) { delete link; continue; }
		VfP2Elink.push_back(link);
	}

	return VfP2Elink.size();
}

int SklSolverFoundation::BuildVfE2PCommLink()
{
	const int nodes = cfg.grpP.num;
	const int gid   = cfg.grpP.gid;

	vector<char> Vf(cfg.MatCount);
	for (size_t m = 0; m < cfg.MatCount; m++) { Vf[m] = cfg.EmatAss[m]? 1: 0; }
	for (size_t m = 0; m < cfg.MtbCount; m++) {
		if (!cfg.EmatAss[cfg.MtbTbl[m]->mat->id]) { continue; }
		if (!cfg.MtbTbl[m]->reactsite) { continue; }
		Vf[cfg.MtbTbl[m]->LR->mat->id] = 1;
	}

	for (int n = 0; n < nodes; n++) {
		
		AreaCommLink* link;
		link = new AreaCommLink(cfg,"Vf ",n,gid,cfg.Ediv,cfg.Pdiv,Vf);
		if (!link->enable()) { delete link; continue; }
		VfE2Plink.push_back(link);
	}

	return VfE2Plink.size();
}

int SklSolverFoundation::BuildSmP2ECommLink()
{
	const int nodes = cfg.grpP.num;
	const int gid   = cfg.grpP.gid;

	if( cfg.MembareaCount == 0 ) return 0;

	for (int n = 0; n < nodes; n++) {
		const size_t s = cfg.MembareaCount;
		char* Ms = new char[s];
		for (int m = 0; m < s; m++) { Ms[m] = 1; }
#ifdef SPHERE175
		para_mng->Broadcast(Ms,s,SKL_ARRAY_DTYPE_CHAR,n,gid);	//2009.12.07
#else
		ParaMngr.Broadcast(Ms,s,SKL_ARRAY_DTYPE_CHAR,n,gid);
#endif

		AreaCommLink* link;
		
		link = new AreaCommLink(cfg,"Sm ",n,gid,cfg.Pdiv,cfg.Ediv,
									vector<char>(Ms,Ms + cfg.MembareaCount));

		delete[] Ms;
		if (!link->enable()) { delete link; continue; }
		SmP2Elink.push_back(link);
	}

	return SmP2Elink.size();
}

int SklSolverFoundation::BuildSmE2PCommLink()
{
	const int nodes = cfg.grpP.num;
	const int gid   = cfg.grpP.gid;

	if( cfg.MembareaCount == 0 ) return 0;

	vector<char> Sm(cfg.MembareaCount);
	for (size_t m = 0; m < cfg.MembareaCount; m++) { Sm[m] = 1; }

	for (int n = 0; n < nodes; n++) {
		
		AreaCommLink* link;
		link = new AreaCommLink(cfg,"Sm ",n,gid,cfg.Ediv,cfg.Pdiv,Sm);
		if (!link->enable()) { delete link; continue; }
		SmE2Plink.push_back(link);
	}

	return SmE2Plink.size();
}


int SklSolverFoundation::SendRecv(
		S4D_REAL* dest,      const CommLinks& recvers,
		const S4D_REAL* src, const CommLinks& senders,
		const bool comm_vc)
{
	if (!cfg.Parallel) {
		// direct copy for serial computing mode
		const int I  = static_cast<int>(dest->GetSize()[0]);
		const int J  = static_cast<int>(dest->GetSize()[1]);
		const int K  = static_cast<int>(dest->GetSize()[2]);
		const int N  = static_cast<int>(dest->GetSize()[3]);
		const int vc = comm_vc?
						lesser(dest->GetVCellSize(),src->GetVCellSize()):
						0;

		for (int n = 0;      n < N;      n++) {
			SKL_REAL       *pdst = UtilGetS4DN(dest,n);
			const SKL_REAL *psrc = UtilGetS4DN(src, n);

			const int dvc = dest->GetVCellSize();
			const int svc = src->GetVCellSize();
			ricscpys3dsendrecv_(
				pdst,psrc,&dvc,&svc,&I,&J,&K,&vc);
			//ricscpys3d_(dst,psrc,&I,&J,&K,&vc);
		}

//		for (int n = 0;      n < N;      n++) {
//		for (int k = 0 - vc; k < K + vc; k++) {
//		for (int j = 0 - vc; j < J + vc; j++) {
//		for (int i = 0 - vc; i < I + vc; i++) {
//			dest->val(i,j,k,n) = src->val(i,j,k,n);
//		}}}}

		return 0;
	} else {
		MEASURE_WAIT_TIME(clock_wait, para_mng->Barrier())  // 2011.05.11 
		BEGIN_CLOCK(comm)
		// transport and copy for parallel computing mode
		int* keys = new int[senders.size()];
		int* k = keys;
		int  c = 0;
		for (CommLinks::const_iterator link = senders.begin();
			 link != senders.end(); link++) {
			
			DB2(MESSAGE(string("Send ") + string_cast(*k), -1, true));
			(*k) = (*link)->Send(src, comm_vc);
			if ((*k) != -1) { k++; c++; }
		}
		for (CommLinks::const_iterator link = recvers.begin();
			 link != recvers.end(); link++){
			DB2(MESSAGE(string("Recv ") + string_cast(*k), -1, true));
			(*link)->Recv(dest, comm_vc);
		}
		END_CLOCK(clock_comm, comm);
		delete[] keys;
	}

	return 0;
}

string SklSolverFoundation::CommLinkString()
{
	stringstream ss;
	vector<CommLinks*> links;
	links.push_back(&IOlink);
	links.push_back(&P2PBlink);
	links.push_back(&P2PClink);
	links.push_back(&P2Elink);
	links.push_back(&E2Plink);
	links.push_back(&VfP2Elink);
	links.push_back(&VfE2Plink);
	for (int i = 0; i < links.size(); i++) {
		for (CommLinks::const_iterator
				link = links[i]->begin(); link != links[i]->end(); link++) {

			ss << (*link)->toString() << endl;
		}
	}

	const string s = ss.str();
	int _l = s.length() + 1;
	int l;

#ifdef SPHERE175
	para_mng->Allreduce(&_l,&l,1,SKL_ARRAY_DTYPE_INT,SKL_MAX,cfg.grpAll.gid);	//2009.12.07
#else
	ParaMngr.Allreduce(&_l,&l,1,SKL_ARRAY_DTYPE_INT,SKL_MAX,cfg.grpAll.gid);
#endif

 	char* buf = new char[l];

#ifdef SPHERE175
	for (int n = 1; n < para_mng->GetNodeNum(0); n++)				//2009.12.07
#else
	for (int n = 1; n < ParaMngr.GetNodeNum(0); n++)
#endif
	{
#ifdef SPHERE175
		if (para_cmp->GetMyID(0))						//2009.12.08
#else
		if (ParaMngr.GetMyID(0))
#endif
		{
#ifdef SPHERE175
			if (para_cmp->GetMyID(0) != n) { continue; }
#else
			if (ParaMngr.GetMyID(0) != n) { continue; }
#endif
			strcpy(buf, s.c_str());
#ifdef SPHERE175
			para_mng->Send(buf, l, SKL_ARRAY_DTYPE_CHAR, 0, cfg.grpAll.gid);	//2009.12.07
#else
			ParaMngr.Send(buf, l, SKL_ARRAY_DTYPE_CHAR, 0, cfg.grpAll.gid);
#endif
		} else {
#ifdef SPHERE175
			para_mng->Recv(buf, l, SKL_ARRAY_DTYPE_CHAR, n, cfg.grpAll.gid);	//2009.12.07
#else
			ParaMngr.Recv(buf, l, SKL_ARRAY_DTYPE_CHAR, n, cfg.grpAll.gid);
#endif
			ss << buf;
		}
	}

	delete[] buf;

	return ss.str();
}

string SklSolverFoundation::JobAssignTableString()
{
	stringstream jaes; // job assign entry string

	const bool hasJob = cfg.AssEqns.size();
	const bool hasIO  = cfg.IONode; // cfg.AssIOEqns.size();
	if (hasIO) {
		jaes << "IO Equations : ";
		for (EqnSet::const_iterator
				e =  cfg.AssIOEqns.begin();
				e != cfg.AssIOEqns.end();
				e++) {

			jaes << " " << (*e)->name << ",";
		}
		jaes << endl;
	}

	if (hasJob) {
		jaes << "Equations : ";
		for (EqnSet::const_iterator
				e =  cfg.AssEqns.begin();
				e != cfg.AssEqns.end();
				e++) {

			jaes << " " << (*e)->name << ",";
		}
		jaes << endl;
		jaes << " @" << cfg.grpP.gid << " #" << cfg.grpP.nid;
		jaes << "  Voxel:"<< '('
				<< cfg.Pdiv.head[0] << ','
				<< cfg.Pdiv.head[1] << ','
				<< cfg.Pdiv.head[2]
				<< ')' << '-' << '('
				<< cfg.Pdiv.tail[0] << ','
				<< cfg.Pdiv.tail[1] << ','
				<< cfg.Pdiv.tail[2] << ')';
		jaes << "  size:" << '('
				<< cfg.Pdiv.size[0] << ','
				<< cfg.Pdiv.size[1] << ','
				<< cfg.Pdiv.size[2] << ')'
				<< endl;

		jaes << "Metabolome : ";
		for (MatSet::const_iterator
				m =  cfg.AssMats.begin();
				m != cfg.AssMats.end();
				m++) {

			jaes << " " << (*m)->name << ",";
		}
		jaes << endl;
		jaes << " @" << cfg.grpE.gid << " #" << cfg.grpE.nid;
		jaes << "  Voxel:" << '('
				<< cfg.Ediv.head[0] << ','
				<< cfg.Ediv.head[1] << ','
				<< cfg.Ediv.head[2]
				<< ')' << '-' << '('
				<< cfg.Ediv.tail[0] << ','
				<< cfg.Ediv.tail[1] << ','
				<< cfg.Ediv.tail[2] << ')';
		jaes << "  size:" << '('
				<< cfg.Ediv.size[0] << ','
				<< cfg.Ediv.size[1] << ','
				<< cfg.Ediv.size[2] << ')' << endl;

	}
	if (!(hasJob || hasIO)) { 
		jaes << " <<No Assigned Job>> ";
	}

	///////////////////////////////////////////////
	// marge inter all nodes
	stringstream jats; // job assign table string

#ifdef SPHERE175
	const int nodes = para_mng->GetNodeNum(0);	//2009.12.07
	const int nid   = para_cmp->GetMyID(0);		//2009.12.08
#else
	const int nodes = ParaMngr.GetNodeNum(0);
	const int nid   = ParaMngr.GetMyID(0);
#endif

	for (int n = 0; n < nodes; n++) {
		const string src = jaes.str();
		size_t len = src.length();

#ifdef SPHERE175
		para_mng->Broadcast(&len, sizeof(size_t), SKL_ARRAY_DTYPE_CHAR, n, 0);	//2009.12.07
#else
		ParaMngr.Broadcast(&len, sizeof(size_t), SKL_ARRAY_DTYPE_CHAR, n, 0);
#endif

		char* buff = new char[len+1];
		if (n == nid) { strcpy(buff, src.c_str()); }

#ifdef SPHERE175
		para_mng->Broadcast(buff, len+1, SKL_ARRAY_DTYPE_CHAR, n, 0);	//2009.12.07
#else
		ParaMngr.Broadcast(buff, len+1, SKL_ARRAY_DTYPE_CHAR, n, 0);
#endif

		jats << endl;
		jats << "Node #" << n << endl;
		jats << buff << endl;

		delete[] buff;
	};

	return jats.str();
}

int SklSolverFoundation::BuildJobManageTable(JobManageTable& jmt)
{
#ifndef _NO_MESSAGE
	MESSAGE("Building Job Management Table, now...");
#else
	MESSAGE("Building Job Management Table, now...",0);
#endif //_NO_MESSAGE

	const CfgElem* steer = SklCfgGetTop(STEER);
	const CfgElem* para  = steer->GetElemValue("ParallelComputing");
	const CfgElem* pjmt  = (para)? para->GetElemValue("JobManTable"): NULL;

	TRY(para, "Not found 'ParallelComputing' element at Conditioanl XML !!");
	TRY(pjmt, "Not found 'JobManTable' element at Conditioanl XML !!");

	for (const CfgElem* pjme = (pjmt)? pjmt->GetElemFirst("JobEntry"): NULL;
		pjme; pjme = pjmt->GetElemNext(pjme, "JobEntry")) {

		jmt.push_back(JobEntry());
		JobEntry& je = jmt[jmt.size() - 1];

		/////////////////////////////////////////
		// Equation Group
		const CfgElem* egrp = pjme->GetElemValue("EquationGroup");
		EqnSet eqns;
		for (const CfgParam*
				eqn = egrp? egrp->GetParamFirst("Equation"): NULL;
				eqn; eqn = egrp->GetParamNext(eqn, "Equation")) {

			const char* p;
			eqn->GetData(&p);

			// check format
			const string name(p);
			const size_t nsep = name.find_first_of(':');
			const string mat  = name.substr(0, nsep);
			const string mol  = name.substr(nsep + 1);

			MatSet matset;
			MolSet molset;

			// Mat Set
			if (mat == "*") {
				// wild-card description - for all materials
				for (size_t m = 0; m < cfg.MatCount; m++) {
					const Material* mt = cfg.MatTbl[m];
					matset.push_back(mt);
				}
			} else if (IdRange::isTuple(mat)) {
				// ID range description
				IdRange range(mat);
				for (int cid = range.begin(); cid <= range.end(); cid++) {
					const size_t id = cfg.MatCidIdMap.find(cid)->second;
					const Material* m =  cfg.MatTbl[id];
					matset.push_back(m);
				}
			} else {
				// regular description
				const size_t matid = cfg.MatNameIdMap.find(mat)->second;
				const Material*  m = cfg.MatTbl[matid];
				matset.push_back(m);
			}

			// Mol Set
			if (mol == "*") {
				// wild-card description - for all moleculars
				for (size_t m = 0; m < cfg.MolCount; m++) {
					const Molecular* ml = cfg.MolTbl[m];
					molset.push_back(ml);
				}
			} else if (IdRange::isTuple(mol)) {
				// ID range description
				IdRange range(mol);
				for (int cid = range.begin(); cid <= range.end(); cid++) {
					const size_t id = cfg.MolCidIdMap.find(cid)->second;
					const Molecular* m =  cfg.MolTbl[id];
					molset.push_back(m);
				}
			} else {
				// regular description
				const size_t molid = cfg.MolNameIdMap.find(mol)->second;
				const Molecular* m = cfg.MolTbl[molid];
				molset.push_back(m);
			}

			for (MatSet::const_iterator
					mt = matset.begin(); mt != matset.end(); mt++) {
			for (MolSet::const_iterator
					ml = molset.begin(); ml != molset.end(); ml++) {

				const Equation* e = UtilGetEquationId((*ml),(*mt));
				if (!e) { continue; }
				eqns.push_back(e);
			}}
		}

		// over load
		for (int i = 0; i < eqns.size(); i++) {
			Equation* e = const_cast<Equation*>(cfg.EqnTbl[(eqns[i]->id)]);
			if (e->je >= 0) {
				EqnSet& es = jmt[e->je].Eqns.first;
				for (EqnSet::iterator
						ei = es.begin(); ei != es.end(); ei++) {

					if ((*ei)->id == e->id) {
						es.erase(ei);
						e->je = -1;
						break;
					}
				}
			}
			e->je = jmt.size() - 1;
		}

		je.Eqns.first = eqns;

		/////////////////////////////////////////
		// Equation Voxel Division
		const CfgElem* vdm = egrp->GetElemValue("VoxelDivisionMethod");
		je.Eqns.second[0]  = (vdm)? vdm->GetIntValue("I"): 0;
		je.Eqns.second[1]  = (vdm)? vdm->GetIntValue("J"): 0;
		je.Eqns.second[2]  = (vdm)? vdm->GetIntValue("K"): 0;
		je.Eqns.second[3]  = (vdm)? vdm->GetIntValue("N"): 0;
		
		/////////////////////////////////////////
		// Metabolism Group
		const CfgElem* mgrp = pjme->GetElemValue("MetabolismGroup");
		je.mgrp = mgrp;
	}

	for (JobManageTable::iterator
			jmi = jmt.begin(); jmi != jmt.end(); jmi++) {

		JobEntry& je = *jmi;
		const CfgElem* mgrp = je.mgrp;
		vector<int> jem; jem.insert(jem.begin(), cfg.MatCount, -1);

		for (const CfgElem* grp = mgrp? mgrp->GetElemFirst("Group"):NULL;
			grp; grp = mgrp->GetElemNext(grp, "Group")) {

			MatAssign s;
			for (const CfgParam* mat = grp->GetParamFirst("material");
				mat; mat = grp->GetParamNext(mat, "material")) {

				const char* p;
				mat->GetData(&p);
				const string name(p);
				MatSet ms;

				if (name == "*") {
					// wild-card description - for all materials
					for (size_t mid = 0; mid < cfg.MatCount; mid++) {
						const Material* m = cfg.MatTbl[mid];
						ms.push_back(m);
					}
				} else if (IdRange::isTuple(name)) {
					// ID range description
					IdRange range(name);
					for (int cid = range.begin(); cid <= range.end(); cid++) {
						const size_t id = cfg.MatCidIdMap.find(cid)->second;
						const Material* m =  cfg.MatTbl[id];
						ms.push_back(m);
					}
				} else {
					// regular description
					const size_t id = cfg.MatNameIdMap.find(name)->second;
					const Material* m = cfg.MatTbl[id];
					ms.push_back(m);
				}

				// select material which is in equations
				for (int i = 0; i < ms.size(); i++) {
					bool find = false;
					for (EqnSet::const_iterator e = je.Eqns.first.begin();
							e != je.Eqns.first.end(); e++) {

						find |= (ms[i]->id == (*e)->mat->id);
					}
					if (find) { s.first.push_back(ms[i]); }
				}

				// over load
				for (int i = 0; i < s.first.size(); i++) {
					const size_t mid = s.first[i]->id;
					const int    at  = jem[mid];
					if ((at >= 0) && (je.Mats.size() > 0)) {
						MatSet& ams = je.Mats[at].first;
						for (int mi = 0; mi < ams.size(); mi++) {

							if (ams[mi]->id == mid) {
								ams.erase(ams.begin() + mi);
								break;
							}
						}
						jem[mid] = je.Mats.size() - 1;
					} else {
						jem[mid] = 0;
					}
				}
			}

			const CfgElem* vdm = grp->GetElemValue("VoxelDivisionMethod");
			s.second[0] = (vdm)? vdm->GetIntValue("I"): 0;
			s.second[1] = (vdm)? vdm->GetIntValue("J"): 0;
			s.second[2] = (vdm)? vdm->GetIntValue("K"): 0;
			s.second[3] = (vdm)? vdm->GetIntValue("N"): 0;
			if (!vdm) {
				// inherit Eqnation level Voxel division
				s.second[0] = je.Eqns.second[0];
				s.second[1] = je.Eqns.second[1];
				s.second[2] = je.Eqns.second[2];
				s.second[3] = je.Eqns.second[3];
			}

			je.Mats.push_back(s);
		}
	}

	return 1;
}

int SklSolverFoundation::TrimAssignedMaterial()
{
	if (!cfg.Parallel) { return 0; }

#ifndef _NO_MESSAGE
	MESSAGE("Trim Assigned Material");
#else
	MESSAGE("Trim Assigned Material",0);
#endif //_NO_MESSAGE

	const EqnSet& eqns = cfg.AssEqns;
	const MatSet& mats = cfg.AssMats;
	MatSet nmats; // new assgined material set
	for (MatSet::const_iterator m = mats.begin(); m != mats.end(); m++) {
		bool used = false;

		for (int i = 0; i < (*m)->MtbTbl.size(); i++) {

			const Metabolism* mtb = (*m)->MtbTbl[i];

			if (!mtb) {
//				MESSAGE(string((*m)->name)+" has no metabolism pathway.");
				continue;
			}

			MESSAGE(UtilStrFormat("Test %s for Trim Assigned Material: %s",
#ifndef _NO_MESSAGE
					mtb->name.c_str(), mtb->eml));
#else
					mtb->name.c_str(), mtb->eml),0);
#endif //_NO_MESSAGE

#ifndef _WITHOUT_ECELL
			EmlModel eml;
			emcsim* sim;

			eml.open(mtb->eml);
			sim = new emcsim();
			sim->loadModel(eml, mtb->sys);
			
			// check: some assigned molecular is used at EML model?
			for (EqnSet::const_iterator e = eqns.begin();
				e != eqns.end(); e++) {

				if (!(*e)->mol->fullid) { continue; }
				const string FullPN = string((*e)->mol->fullid) + ":Value";
				const bool exist = sim->isExistEntityProperty(FullPN);

				if (exist) {
					DB1(MESSAGE(FullPN + " is existing at " + mtb->name));
				} else {
					DB2(MESSAGE(FullPN + " isn't existing at " + mtb->name));
				}

				used |= exist;
			}

			delete sim;
			sim = NULL;
#endif // _WITHOUT_ECELL
		}

		// add the material to new assigned set
		if (used) {
			nmats.push_back(*m);

			MESSAGE(UtilStrFormat("%s is computed on #%d",
#ifndef _NO_MESSAGE
						(*m)->name.c_str(),cfg.grpAll.nid));
#else
						(*m)->name.c_str(),cfg.grpAll.nid),0);
#endif //_NO_MESSAGE
		} else {
			MESSAGE(UtilStrFormat("%s is not computed on #%d",
#ifndef _NO_MESSAGE
						(*m)->name.c_str(),cfg.grpAll.nid));
#else
						(*m)->name.c_str(),cfg.grpAll.nid),0);
#endif //_NO_MESSAGE
		}
	}

	// change assigned material set to new one from old
	cfg.AssMats = nmats;

	// rebuild EmatAss array
	delete[] cfg.EmatAss;
	cfg.EmatAss = new char[cfg.MatCount]; memset(cfg.EmatAss,0,cfg.MatCount);
	for (int m = 0; m < cfg.AssMats.size(); m++) {
		cfg.EmatAss[cfg.AssMats[m]->id] = 1;
	}

	return 1;
}

