//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#include <sstream>
#include <iomanip>
#include <cmath>

#include <stdlib.h>

#include <Foundation/const.h>
#include <Foundation/error.h>
#include <Foundation/util.h>
#include <Foundation/api.h>

#include "SklSolverFoundation.h"

SklSolverFoundation::SklSolverFoundation()
#if ENABLE_WARP && ENABLE_MEMBRANEPOTENTIAL
:mtbsolv(cfg), difsolv(cfg), membsolv(cfg), warpsolv(cfg), potentialsolv(cfg)
#elif ENABLE_WARP
:mtbsolv(cfg), difsolv(cfg), membsolv(cfg), warpsolv(cfg)
#elif ENABLE_MEMBRANEPOTENTIAL
:mtbsolv(cfg), difsolv(cfg), membsolv(cfg), potentialsolv(cfg)
#else
:mtbsolv(cfg), difsolv(cfg), membsolv(cfg)
#endif

#ifdef ENABLE_ADVECTION
, advectionsolv(cfg)
#endif
{
/* USER WRITE
   Please place initialize member variable here.
*/
	clock        = double(0.0);
	clock_all    = double(0.0);
	clock_exptim = double(0.0);
	clock_comm   = double(0.0);
	clock_wait   = double(0.0);
}

SklSolverFoundation::SklSolverFoundation(int sType)
#if ENABLE_WARP && ENABLE_MEMBRANEPOTENTIAL
:mtbsolv(cfg), difsolv(cfg), membsolv(cfg), warpsolv(cfg), potentialsolv(cfg)
#elif ENABLE_WARP
:mtbsolv(cfg), difsolv(cfg), membsolv(cfg), warpsolv(cfg)
#elif ENABLE_MEMBRANEPOTENTIAL
:mtbsolv(cfg), difsolv(cfg), membsolv(cfg), potentialsolv(cfg)
#else
:mtbsolv(cfg), difsolv(cfg), membsolv(cfg)
#endif

#ifdef ENABLE_ADVECTION
, advectionsolv(cfg)
#endif
{
/* USER WRITE
   Please place initialize member variable here.
*/
	clock        = double(0.0);
	clock_all    = double(0.0);
	clock_exptim = double(0.0);
	clock_comm   = double(0.0);
	clock_wait   = double(0.0);
}

SklSolverFoundation::~SklSolverFoundation()
{
/* USER WRITE
   Please place destory member variable here.
*/

}

int SklSolverFoundation::LoadUnitSetting()
{
#ifndef _NO_MESSAGE
	MESSAGE("Load Unit Settings");
#else
	MESSAGE("Load Unit Settings",0);
#endif //_NO_MESSAGE

	const CfgElem* steer = SklCfgGetTop(STEER);

	const CfgParam* lup = steer->GetParamFirst(cfg_param_lu);
	const CfgParam* tup = steer->GetParamFirst(cfg_param_tu);
	const CfgParam* nap = steer->GetParamFirst("NA");
	const CfgParam* kap = steer->GetParamFirst("K_NA");

	SKL_REAL lu, tu, na, ka;
	cfg.Ls  = (lup && lup->GetData(&lu) && (lu > 0))? lu: 1e0;
	cfg.Ts  = (tup && tup->GetData(&tu) && (tu > 0))? tu: 1e0;
	cfg.NA  = (nap && nap->GetData(&na) && (na > 0))? na: default_NA;
	cfg.Kna = (kap && kap->GetData(&ka) && (ka > 0))? ka: 1e0;

	return 0;
}

int SklSolverFoundation::LoadTimeStepSetting()
{
#ifndef _NO_MESSAGE
	MESSAGE("Loading CourantNumber or Immediated DeltaT ...");
#else
	MESSAGE("Loading CourantNumber or Immediated DeltaT ...",0);
#endif //_NO_MESSAGE

	const CfgElem* steer = SklCfgGetTop(STEER);

	const CfgParam* cnp = steer->GetParamFirst("CourantNumber");
	const CfgParam* dtp = steer->GetParamFirst(cfg_param_dt);
	SKL_REAL dt = SKL_REAL(0.0);
	SKL_REAL Cn = SKL_REAL(0.0);

	cfg.fixdt = dtp && dtp->GetData(&dt); 
	cfg.dt    = cfg.fixdt? dt: SKL_REAL(0.0);
	Cn        = (cnp && cnp->GetData(&Cn))? Cn: SKL_REAL(0.8);
	cfg.CourantNumber = Cn;

	// check
	if (cfg.fixdt) {
#ifndef _NO_MESSAGE
		MESSAGE(string("Found DeltaT. dt = ") + string_cast(cfg.dt));
#else
		MESSAGE(string("Found DeltaT. dt = ") + string_cast(cfg.dt),0);
#endif //_NO_MESSAGE
		TRY(cfg.dt > SKL_REAL(0.0), "'dt' must be greater than 0.0");
		SklSetDeltaT(cfg.dt);
	} else {
#ifndef _NO_MESSAGE
		MESSAGE(string("Found Courant Number. Cn = ") + string_cast(Cn));
#else
		MESSAGE(string("Found Courant Number. Cn = ") + string_cast(Cn),0);
#endif //_NO_MESSAGE
		MSG_WARNING("Courant Number is greater than 1.0",Cn > 1.0,0);
		TRY(Cn > SKL_REAL(0.0), "Courant Number must be greater than 0.0");
	}

	// check point
	const CfgParam* chkintvl = steer->GetParamFirst("checkpoint_interval");
	chkpoint_interval = (chkintvl && chkintvl->GetData(&chkpoint_interval))?
						chkpoint_interval: 0;

	// time integration
	const char* t_integ = NULL;
	const CfgParam* time_integ = steer->GetParamFirst("time_integration");
	if(time_integ && time_integ->GetData(&t_integ) && (strcasecmp(t_integ, "Euler") == 0)){
		cfg.time_integ = EULER;
	}
	else if(time_integ && time_integ->GetData(&t_integ) && (strcasecmp(t_integ, "2nd Adams-Bashforth") == 0)){
		cfg.time_integ = ADAMS_BASHFORTH;
	}
	else if(time_integ && time_integ->GetData(&t_integ) && (strcasecmp(t_integ, "2nd Runge-Kutta") == 0)){
		cfg.time_integ = RUNGE_KUTTA_2;
	}
	else if(time_integ && time_integ->GetData(&t_integ) && (strcasecmp(t_integ, "4th Runge-Kutta") == 0)){
		cfg.time_integ = RUNGE_KUTTA_4;
	}
	else{
		cfg.time_integ = EULER;
	}

	// log
	if(cfg.time_integ == EULER){
#ifndef _NO_MESSAGE
		MESSAGE(string("TimeIntegration. Euler : time_integ = ") + string_cast(cfg.time_integ));
#else
		MESSAGE(string("TimeIntegration. Euler : time_integ = ") + string_cast(cfg.time_integ),0);
#endif //_NO_MESSAGE
	}
	else if(cfg.time_integ == ADAMS_BASHFORTH){
#ifndef _NO_MESSAGE
		MESSAGE(string("TimeIntegration. 2nd Adams-Bashforth : time_integ = ") + string_cast(cfg.time_integ));
#else
		MESSAGE(string("TimeIntegration. 2nd Adams-Bashforth : time_integ = ") + string_cast(cfg.time_integ),0);
#endif //_NO_MESSAGE
	}
	else if(cfg.time_integ == RUNGE_KUTTA_2){
#ifndef _NO_MESSAGE
		MESSAGE(string("TimeIntegration. 2nd Runge-Kutta : time_integ = ") + string_cast(cfg.time_integ));
#else
		MESSAGE(string("TimeIntegration. 2nd Runge-Kutta : time_integ = ") + string_cast(cfg.time_integ),0);
#endif //_NO_MESSAGE
	}
	else{
#ifndef _NO_MESSAGE
		MESSAGE(string("TimeIntegration. 4th Runge-Kutta : time_integ = ") + string_cast(cfg.time_integ));
#else
		MESSAGE(string("TimeIntegration. 4th Runge-Kutta : time_integ = ") + string_cast(cfg.time_integ),0);
#endif //_NO_MESSAGE
	}

	return 0;
}

SKL_REAL SklSolverFoundation::CalcDeltaT()
{
	const int gid = cfg.grpComp.gid;
	SKL_REAL rcc;
	SKL_REAL cc = difsolv.CourantCond();
	if (cfg.Parallel) {

#ifdef SPHERE175
		para_cmp->Allreduce(&cc,&rcc,1,SKL_ARRAY_DTYPE_REAL,SKL_MIN,gid);	//2009.12.07
#else
		ParaMngr.Allreduce(&cc,&rcc,1,SKL_ARRAY_DTYPE_REAL,SKL_MIN,gid);
#endif

	} else {
		rcc = cc;
	}
	cfg.dt = (cfg.fixdt)? cfg.dt: rcc * cfg.CourantNumber;

	return cfg.dt;
}

int SklSolverFoundation::init_Domain()
{
	// DomainSettings
	float origin[3];
	float width[3];
	float pitch[3];
	bool o = SklCfgGetVoxelOrigin(origin[0],  origin[1],  origin[2]);
	bool w = SklCfgGetVoxelWidth (width[0],   width[1],   width[2]);
	bool p = SklCfgGetVoxelPitch (pitch[0],   pitch[1],   pitch[2]);
	bool s = SklCfgGetVoxelSize  (cfg.size[0],cfg.size[1],cfg.size[2]);

	for (int i = 0; i < 3; i++) {
		cfg.origin[i] = static_cast<SKL_REAL>(origin[i]);
		cfg.width[i]  = static_cast<SKL_REAL>(width[i]);
		cfg.pitch[i]  = static_cast<SKL_REAL>(pitch[i]);
	}

	//
	TRY(o, "Voxel origin is not found");
	TRY(s, "Voxel size is not found");

	if (w) {
		if ((cfg.width[0] <= SKL_REAL(0.0)) ||
			(cfg.width[1] <= SKL_REAL(0.0)) ||
			(cfg.width[2] <= SKL_REAL(0.0))) {

			ERROR("VoxelWidth is illegal number");
		}
		for (unsigned int i = 0; i < 3; i++) {
			const SKL_REAL w  = static_cast<SKL_REAL>(cfg.width[i]);
			const SKL_REAL sz = static_cast<SKL_REAL>(cfg.size[i]);
			cfg.pitch[i]      = (w / sz);
			cfg.outpitch[i]   = cfg.pitch[i];
		}
	} else if (p) {
		if ((cfg.pitch[0] <= SKL_REAL(0.0)) ||
			(cfg.pitch[1] <= SKL_REAL(0.0)) ||
			(cfg.pitch[2] <= SKL_REAL(0.0))) {

			ERROR("VoxelPitch is illegal number");
		}
		for (unsigned int i = 0; i < 3; i++) {
			const SKL_REAL p  = static_cast<SKL_REAL>(cfg.pitch[i]);
			const SKL_REAL sz = static_cast<SKL_REAL>(cfg.size[i]);
			cfg.width[i]      = sz * p;
			cfg.outpitch[i]   = p;
		}
	} else {
		ERROR("Voxel pitch or Voxel width are not found.");
	}

	if ((cfg.size[0] <= 0) ||
		(cfg.size[1] <= 0) ||
		(cfg.size[2] <= 0)) {

		ERROR("VoxelSize is illegal number");
	}
#ifndef _NO_MESSAGE
	MESSAGE(string("VoxelSize is ")
			+ string_cast(cfg.size[0]) + " * "
			+ string_cast(cfg.size[1]) + " * "
			+ string_cast(cfg.size[2]));
#else
	MESSAGE(string("VoxelSize is ")
			+ string_cast(cfg.size[0]) + " * "
			+ string_cast(cfg.size[1]) + " * "
			+ string_cast(cfg.size[2]),0);
#endif //_NO_MESSAGE

#ifndef _NO_MESSAGE
	MESSAGE(string("VoxelWidth is ")
			+ string_cast(cfg.width[0]) + " * "
			+ string_cast(cfg.width[1]) + " * "
			+ string_cast(cfg.width[2]));
#else
	MESSAGE(string("VoxelWidth is ")
			+ string_cast(cfg.width[0]) + " * "
			+ string_cast(cfg.width[1]) + " * "
			+ string_cast(cfg.width[2]),0);
#endif //_NO_MESSAGE

#ifndef _NO_MESSAGE
	MESSAGE(string("VoxelPitch is ")
			+ string_cast(cfg.pitch[0]) + " * "
			+ string_cast(cfg.pitch[1]) + " * "
			+ string_cast(cfg.pitch[2]));
#else
	MESSAGE(string("VoxelPitch is ")
			+ string_cast(cfg.pitch[0]) + " * "
			+ string_cast(cfg.pitch[1]) + " * "
			+ string_cast(cfg.pitch[2]),0);
#endif //_NO_MESSAGE


	// Volume size of a cell
	cfg.cvol = cfg.pitch[0] * cfg.pitch[1] * cfg.pitch[2]; // [/m^3]

	return 1;
}

unsigned int SklSolverFoundation::init_MatMap()
{
#ifndef _NO_MESSAGE
	MESSAGE("Load Material Definitions and Settings ...");
#else
	MESSAGE("Load Material Definitions and Settings ...",0);
#endif //_NO_MESSAGE

	typedef std::map<unsigned int, Material*>  MatMapT;
	MatMapT MatMap; // working tempolary mapping object;

	const CfgElem* DomainInfo = SklCfgGetTop(DOMAININFO);
	const CfgElem* cell       = DomainInfo->GetElemFirst("cell");

	cfg.CmpMatId = -1;

	while (cell) {
		const CfgElem* mat = cell->GetElemFirst();
		while (mat) {
			// found a organelle entry
			Material* m = new Material;
			const CfgParam* eml = mat->GetParamFirst("metabolism");
			const CfgParam* sys = mat->GetParamFirst("system");
			const CfgParam* fus = mat->GetParamFirst("fusion");
			const CfgParam* fut = mat->GetParamFirst("fusion_threshold");
			const CfgParam* cmp = mat->GetParamFirst("complete");

			const CfgParam* vu = mat->GetParamFirst("volume_unit");
			const CfgParam* tu = mat->GetParamFirst("time_unit");

			const char* t;  // tmp
			SKL_REAL rt;    // tmp

			TRY(mat->isSetID(), string(m->name) + " has no id");
			m->cid          = mat->GetID();
			m->id           = MatMap.size();
			m->name         = mat->GetName();
			m->xml_elem     = mat;
			m->fus          = (fus && fus->GetData(&t) && !strcmp(t,"true"));
			m->fut          = (fut && fut->GetData(&rt))? rt: SKL_REAL(0.0);
			m->body_mtb.eml = (eml && eml->GetData(&t))? t: NULL;
			DB2(MESSAGE(UtilStrFormat("fus==%d, fut==%lf", m->fus, m->fut)));
			
#ifndef _NO_MESSAGE
			MESSAGE(string("++ Material ") + m->name + " found");
#else
			MESSAGE(string("++ Material ") + m->name + " found",0);
#endif //_NO_MESSAGE

			if (!sys) { m->body_mtb.sys.insert("/"); }
			for ( ;sys; sys = mat->GetParamNext(sys, "system")) {
				// find eml::system selection
				const char* sysname;
				sys->GetData(&sysname);
				m->body_mtb.sys.insert(sysname);
			}

			// is complementaly material?
			if (cmp) {
				int f = 0;
				if (cmp->GetData(&f) && (f)) {
					const int cmid = cfg.CmpMatId;
					TRY((cmid < 0), \
						MatMap[cmid]->name + " was Completed, aleady.");
					cfg.CmpMatId = m->id;
				}
			}

			// volume unit
			if (vu) {
				TRY(vu->GetData(&(m->body_mtb.ve)),
					"Illigal E-Cell Volume Unit Setting.")
			} else {
				// defaule volume unit: 1e-3[m3](= L)
				m->body_mtb.ve = SKL_REAL(1e-3);
			}

			// time unit
			if (tu) {
				TRY(tu->GetData(&(m->body_mtb.te)),
					"Illigal E-Cell Time Unit Setting.")
			} else {
				// defaule time unit is 1.0[s]
				m->body_mtb.te = SKL_REAL(1e0);
			}

			// commit
			MatMap.insert(MatMapT::value_type(m->id, m));
    		mat = cell->GetElemNext(mat, NULL);
		};
		cell = DomainInfo->GetElemNext(cell, "cell");
	};

	const size_t mats = MatMap.size();
	TRY(mats, "No cell Material Entries.");
#ifndef _NO_MESSAGE
	MESSAGE(string("total Material Count is ") + string_cast(mats));
#else
	MESSAGE(string("total Material Count is ") + string_cast(mats),0);
#endif //_NO_MESSAGE

	//
	// Transform Material Map to Material Table
	//
#ifndef _NO_MESSAGE
	MESSAGE("Constructing Material Table ...");
#else
	MESSAGE("Constructing Material Table ...",0);
#endif //_NO_MESSAGE
	for (size_t matid = 0; matid < mats; matid++) {
		Material* m = MatMap.find(matid)->second;

		// doubly
		TRY(cfg.MatNameIdMap.find(m->name) == cfg.MatNameIdMap.end(),
			string("cell material '") + m->name + "' is doubly");

		m->id = matid;
		cfg.MatTbl.push_back(m);

#ifndef _NO_MESSAGE
		MESSAGE(string("++ ") + m->name + " id=" + string_cast(m->id));
#else
		MESSAGE(string("++ ") + m->name + " id=" + string_cast(m->id),0);
#endif //_NO_MESSAGE

		const char* name = m->name.c_str();
		cfg.TypMap.insert(TypMapT::value_type(name,TypElmMapT(MATERIAL,m)));
		cfg.TypMap.insert(TypMapT::value_type(string("AX-_") + name,TypElmMapT(FACERATE_MX,  m)));
		cfg.TypMap.insert(TypMapT::value_type(string("AX+_") + name,TypElmMapT(FACERATE_PX,  m)));
		cfg.TypMap.insert(TypMapT::value_type(string("AY-_") + name,TypElmMapT(FACERATE_MY,  m)));
		cfg.TypMap.insert(TypMapT::value_type(string("AY+_") + name,TypElmMapT(FACERATE_PY,  m)));
		cfg.TypMap.insert(TypMapT::value_type(string("AZ-_") + name,TypElmMapT(FACERATE_MZ,  m)));
		cfg.TypMap.insert(TypMapT::value_type(string("AZ+_") + name,TypElmMapT(FACERATE_PZ,  m)));
		cfg.TypMap.insert(TypMapT::value_type(string("Sm_")  + name,TypElmMapT(MEMBAREA,     m)));
		cfg.TypMap.insert(TypMapT::value_type(string("NX_")  + name,TypElmMapT(MEMBNRML_X,   m)));
		cfg.TypMap.insert(TypMapT::value_type(string("NY_")  + name,TypElmMapT(MEMBNRML_Y,   m)));
		cfg.TypMap.insert(TypMapT::value_type(string("NZ_")  + name,TypElmMapT(MEMBNRML_Z,   m)));
        
		cfg.MatNameIdMap.insert(Name2IdMapT::value_type(m->name, matid));
		cfg.MatCidIdMap.insert(Cid2IdMapT::value_type(m->cid, matid));
	}

	cfg.MatCount = mats;
	return mats;
}

unsigned int SklSolverFoundation::init_MolMap()
{
#ifndef _NO_MESSAGE
	MESSAGE("Load Molecular Definitions and Settings ...");
#else
	MESSAGE("Load Molecular Definitions and Settings ...",0);
#endif //_NO_MESSAGE

	typedef std::map<unsigned int, Molecular*>  MolMapT;
	MolMapT MolMap; // working tempolary mapping object;

	const CfgElem* DomainInfo;
	const CfgElem* mole;

	DomainInfo = SklCfgGetTop(DOMAININFO);
	mole = DomainInfo->GetElemFirst("molecular");

	const size_t mats = cfg.MatCount;

	while (mole) {
		const CfgElem* mol = mole->GetElemFirst();
		while (mol) {
			// found a molecular entry
			Molecular* m = new Molecular;
			const CfgParam* fullid = mol->GetParamFirst("FullID");

			const char* t;
			m->xml_elem = mol;
			m->id     = MolMap.size();
			m->name   = mol->GetName();
			TRY(mol->isSetID(), m->name + " has no ID");
			m->cid    = mol->GetID();
			m->fullid = NULL, fullid && fullid->GetData(&(m->fullid));
#ifdef ENABLE_MEMBRANEPOTENTIAL
			const CfgParam* potential = mol->GetParamFirst("Potential");
                        const char* flg=NULL;
                        m->potential = false;
                        m->potential = (potential && potential->GetData(&flg) && 
                                       (strcasecmp(flg, "true") == 0 ));
                                       //(strcmp(flg, "true") == 0 || strcmp(flg, "TRUE") == 0));

                        printf("potential : %d\n",m->potential);
#endif


#ifndef _NO_MESSAGE
			MESSAGE(string("++ Molecular ") + m->name + " found");
#else
			MESSAGE(string("++ Molecular ") + m->name + " found",0);
#endif //_NO_MESSAGE

			// diffusion coef
			for (const CfgParam* ccoef = mol->GetParamFirst("DifCoef");
					ccoef; ccoef = mol->GetParamNext(ccoef, "DifCoef")) {

#ifndef _NO_MESSAGE
				MESSAGE(string("++ Molecular ")+m->name+" has DifCoef tag");
#else
				MESSAGE(string("++ Molecular ")+m->name+" has DifCoef tag",0);
#endif //_NO_MESSAGE

				SKL_REAL coef;
				TRY(ccoef->GetData(&coef),"DifCoef value is not REAL");
				if (ccoef->isSetID()) {
					// material dependent
					const int cid = ccoef->GetID();
					const Cid2IdMapT::const_iterator i = cfg.MatCidIdMap.find(cid);
					if (i == cfg.MatCidIdMap.end()) { continue; }
					const size_t mid = i->second;
					m->ccoefm.insert(Int2RealMap::value_type(mid, coef));
				} else {
					// material independent
					m->ccoef = coef;
				}
			}

			// initial molconc [mol/m3]
			for (const CfgParam* molconc = mol->GetParamFirst("MolConc");
					molconc; molconc = mol->GetParamNext(molconc,"MolConc")) {

				SKL_REAL conc;
				conc = molconc->GetData(&conc)? conc: SKL_REAL(0.0);

#ifndef _NO_MESSAGE
				MESSAGE(string("++ Molecular ")+m->name+" has MolConc tag");
#else
				MESSAGE(string("++ Molecular ")+m->name+" has MolConc tag",0);
#endif //_NO_MESSAGE
				TRY(molconc->isSetID(),string(m->name)+" has no material id")

				const int    cid = molconc->GetID();
				const Cid2IdMapT::const_iterator i = cfg.MatCidIdMap.find(cid);
				if (i == cfg.MatCidIdMap.end()) { continue; }
				const size_t mid = i->second;
				m->molconc.insert(RealMap::value_type(cfg.MatTbl[mid]->name,conc));
			}
			// initial mol number
			for (const CfgParam* molnum = mol->GetParamFirst("Num");
					molnum; molnum = mol->GetParamNext(molnum,"Num")) {

				SKL_REAL n;
				n = molnum->GetData(&n)? n: SKL_REAL(0.0);

#ifndef _NO_MESSAGE
				MESSAGE(string("++ Molecular ")+m->name+" has Num tag");
#else
				MESSAGE(string("++ Molecular ")+m->name+" has Num tag",0);
#endif //_NO_MESSAGE
				TRY(molnum->isSetID(),string(m->name)+" has no material id")

				const int    cid = molnum->GetID();
				const Cid2IdMapT::const_iterator i = cfg.MatCidIdMap.find(cid);
				if (i == cfg.MatCidIdMap.end()) { continue; }
				const size_t mid = i->second;
				m->molnum.insert(RealMap::value_type(cfg.MatTbl[mid]->name,n));
			}

			// commit
			MolMap.insert(MolMapT::value_type(m->id,m));
			mol = mole->GetElemNext(mol, NULL);
		};
		mole = DomainInfo->GetElemNext(mole, "molecular");
	};

	const size_t mols = MolMap.size();
	MSG_WARNING("no molecular entries", !mols, 0);
#ifndef _NO_MESSAGE
	MESSAGE(string("total Molecular Count is ") + string_cast(mols));
#else
	MESSAGE(string("total Molecular Count is ") + string_cast(mols),0);
#endif //_NO_MESSAGE

	//
	// Transform Molecular Map to Molecular Table
	//
#ifndef _NO_MESSAGE
	MESSAGE("Constructing Molecular Table ...");
#else
	MESSAGE("Constructing Molecular Table ...",0);
#endif //_NO_MESSAGE
	for (size_t molid = 0; molid < mols; molid++) {
		Molecular* m = MolMap.find(molid)->second;

		// doubly
		if (cfg.MolNameIdMap.find(m->name) != cfg.MolNameIdMap.end()) {
			stringstream msg;
			msg << "molecular '" << m->name << "' is doubly";
			ERROR(msg.str());
			return 0;
		}

		m->id = cfg.MolTbl.size();
		cfg.MolTbl.push_back(m);

		cfg.TypMap.insert(TypMapT::value_type(m->name,TypElmMapT(MOLECULAR,m)));
		cfg.MolNameIdMap.insert(Name2IdMapT::value_type(m->name, molid));
		cfg.MolCidIdMap.insert(Cid2IdMapT::value_type(m->cid, molid));

#ifndef _NO_MESSAGE
		MESSAGE(string("++ ") + m->name + " id=" + string_cast(m->id));
#else
		MESSAGE(string("++ ") + m->name + " id=" + string_cast(m->id),0);
#endif //_NO_MESSAGE
	}

	cfg.MolCount = mols;
	return mols;
}

unsigned int SklSolverFoundation::init_EqnMap()
{
#ifndef _NO_MESSAGE
	MESSAGE("Build Basic Equation structure ...");
#else
	MESSAGE("Build Basic Equation structure ...",0);
#endif //_NO_MESSAGE

	const unsigned int mats = cfg.MatCount;
	const unsigned int mols = cfg.MolCount;

	// reduction flag
#ifdef _NO_REDUCT_EQUATION_MAP
	vector<bool> tbl(mols*mats, true);
#else // defined _NO_REDUCT_EQUATION_MAP

#ifndef _NO_MESSAGE
	MESSAGE("Check Equation Exist ...");
#else //_NO_MESSAGE
	MESSAGE("Check Equation Exist ...",0);
#endif //_NO_MESSAGE

	vector<bool> tbl(mols*mats, false);

#ifndef _NO_MESSAGE
	MESSAGE("Checking with initial value.");
#else
	MESSAGE("Checking with initial value.",0);
#endif //_NO_MESSAGE
	for (size_t molid = 0; molid < mols; molid++) {
	for (size_t matid = 0; matid < mats; matid++) {
		const int idx = molid + matid * mols;
		const Molecular* ml = cfg.MolTbl[molid];
		const Material*  mt = cfg.MatTbl[matid];

		bool f = false;
		// by molconc
		f |= (ml->molconc.find(mt->name) != ml->molconc.end());

		// by molnum
		f |= (ml->molnum.find(mt->name)  != ml->molnum.end());

		// by FileIO
		const string attr = mt->name + ':' + ml->name;
		f |= SklCfgCheckInFile(attr.c_str());
		f |= SklCfgCheckInFile((attr+"_bc").c_str());
		f |= SklCfgCheckOutFile(attr.c_str());


#ifdef ENABLE_MEMBRANEPOTENTIAL
                if( ml->potential ) f=true;
#endif

		tbl[idx] = f;
#ifndef _NO_MESSAGE
		if (tbl[idx]) { MESSAGE(string("++ ") + attr + " exists."); }
#else
		if (tbl[idx]) { MESSAGE(string("++ ") + attr + " exists.",0); }
#endif //_NO_MESSAGE

	}}

#ifndef _NO_MESSAGE
	MESSAGE("Checking with membrane function.");
#else
	MESSAGE("Checking with membrane function.",0);
#endif //_NO_MESSAGE
	for (int igmn = 0; igmn < cfg.MlMbCount; igmn++) {
		const MolMemb* gmn = cfg.MlMbTbl[igmn];
		const int ig  = gmn->mol->id;
		const int im  = gmn->memb->mat[0]->id;
		const int igm = ig + im * mols;

		const string attr = gmn->memb->mat[0]->name + ':' + gmn->mol->name;
#ifndef _NO_MESSAGE
		MESSAGE(string("++ ") + attr + " exists.");
#else
		MESSAGE(string("++ ") + attr + " exists.",0);
#endif //_NO_MESSAGE
	
		tbl[igm] = true;
	}
#endif // _NO_REDUCT_EQUATION_MAP

#ifndef _NO_MESSAGE
	MESSAGE("Construct Equation Table ...");
#else
	MESSAGE("Construct Equation Table ...",0);
#endif //_NO_MESSAGE
	for (size_t molid = 0; molid < mols; molid++) {
	for (size_t matid = 0; matid < mats; matid++) {
		const int idx = molid + matid * mols;
		if (!tbl[idx]) { continue; }

		const string matname = cfg.MatTbl[matid]->name;
		const string molname = cfg.MolTbl[molid]->name; 
		const string eqnname = matname + ":" + molname; // "[mat]:[mol]"

		Equation* e = new Equation;

		e->id   = cfg.EqnTbl.size();
		e->mat  = cfg.MatTbl[matid];
		e->mol  = cfg.MolTbl[molid];
		e->name = eqnname;
		e->je   = -1;

#ifndef _NO_MESSAGE
		MESSAGE(string("++ ") + e->name + " id=" + string_cast(e->id));
#else
		MESSAGE(string("++ ") + e->name + " id=" + string_cast(e->id),0);
#endif //_NO_MESSAGE

		cfg.EqnTbl.push_back(e);
		cfg.EqnNameIdMap.insert(Name2IdMapT::value_type(e->name, e->id));

		const char* name = e->name.c_str();
		cfg.TypMap.insert(TypMapT::value_type(name,TypElmMapT(EQUATION,e)));
		cfg.TypMap.insert(TypMapT::value_type(string("D_")+name,TypElmMapT(DIFFLUX,e)));
		cfg.TypMap.insert(TypMapT::value_type(string("R_")+name,TypElmMapT(METABOTERM,e)));
		cfg.TypMap.insert(TypMapT::value_type(string("F_")+name,TypElmMapT(MEMBTERM,e)));
		cfg.TypMap.insert(TypMapT::value_type(string("S0_")+name,TypElmMapT(SRCTERM,e)));
		cfg.TypMap.insert(TypMapT::value_type(string("Qm_")+name,TypElmMapT(MEMBEXTERM,e)));
	}}
	
	const int eqns = cfg.EqnTbl.size();
	cfg.EqnCount = eqns;
#ifndef _NO_MESSAGE
	MESSAGE(string("total Equation Count is ") + string_cast(eqns));
#else
	MESSAGE(string("total Equation Count is ") + string_cast(eqns),0);
#endif //_NO_MESSAGE

	return eqns;
}


unsigned int SklSolverFoundation::init_MtbMap()
{
#ifndef _NO_MESSAGE
	MESSAGE("Constructing Metabolism Table ...");
#else
	MESSAGE("Constructing Metabolism Table ...",0);
#endif //_NO_MESSAGE

	// metabolism on material body
	for (int mid = 0; mid < cfg.MatCount; mid++) {

		const Material* m = cfg.MatTbl[mid];
		if (!(m->body_mtb.eml)) {
			const_cast<Material*>(m)->MtbTbl.push_back(NULL);
			continue;
		}

		Metabolism* mtb = new Metabolism();

		mtb->id        = cfg.MtbTbl.size();
		mtb->name      = m->name;
		mtb->eml       = m->body_mtb.eml;
		mtb->sys       = m->body_mtb.sys;
		mtb->Ve        = m->body_mtb.ve;
		mtb->Te        = m->body_mtb.te;
		mtb->mat       = m;
		mtb->reactsite = false;
		mtb->onMemb    = false;
		mtb->contacted = NULL;

#ifndef _NO_MESSAGE
		MESSAGE(UtilStrFormat("++ %s id=%d",m->name.c_str(),mtb->id));
#else
		MESSAGE(UtilStrFormat("++ %s id=%d",m->name.c_str(),mtb->id),0);
#endif //_NO_MESSAGE

		const_cast<Material*>(m)->MtbTbl.push_back(mtb);
		cfg.MtbTbl.push_back(mtb);
		cfg.MtbNameIdMap.insert(Name2IdMapT::value_type(mtb->name, mtb->id));
	}

	// recepter
	for (int mid = 0; mid < cfg.MatCount; mid++) {

		const Material* m = cfg.MatTbl[mid];
		const CfgElem*  mat_xml = m->xml_elem;

		const char* tag = "membrane-metabolism";
		for (const CfgElem* mtb_xml = mat_xml->GetElemFirst(tag);
			mtb_xml; mtb_xml = mat_xml->GetElemNext(mtb_xml,tag)) {

			const CfgParam* name_xml = mtb_xml->GetParamFirst("name");
			const CfgParam* eml_xml  = mtb_xml->GetParamFirst("eml");
			const CfgParam* hg_xml   = mtb_xml->GetParamFirst("height");
			const CfgParam* rs_xml   = mtb_xml->GetParamFirst("reactsite");
			const CfgParam* ve_xml   = mtb_xml->GetParamFirst("volume_unit");
			const CfgParam* te_xml   = mtb_xml->GetParamFirst("time_unit");
			const CfgParam* co_xml   = mtb_xml->GetParamFirst("contacted");

			TRY(name_xml && (name_xml->GetDataType() == _STR),
				"illigal name tag");
			TRY(eml_xml  && (eml_xml->GetDataType()  == _STR),
				"illigal eml tag");
			
			const char* name; name_xml->GetData(&name);
			const char* eml;  eml_xml->GetData(&eml);
			// SKL_REAL hg; hg_xml? hg_xml->GetData(&hg): hg = 30e-9 / cfg.Ls;
			SKL_REAL hg; hg_xml? hg_xml->GetData(&hg): hg = -1.0;
			SKL_REAL ve; ve_xml? ve_xml->GetData(&ve): ve = 1e-3;
			SKL_REAL te; te_xml? te_xml->GetData(&te): te = 1e0;
			int      rs; rs_xml? rs_xml->GetData(&rs): rs = 0;
			
			Metabolism* mtb = new Metabolism();

			mtb->id   = cfg.MtbTbl.size();
			mtb->name = name;
			mtb->eml  = eml;
			mtb->mat  = m;
			mtb->sys.insert("/");

			mtb->reactsite = rs;
			mtb->height    = hg;
			mtb->Ve        = ve;
			mtb->Te        = te;
                           
			mtb->onMemb    = true;

			// contated material with membrane
			mtb->contacted = NULL;
			if (co_xml) {
				TRY(co_xml->GetDataType() == _STR, "illigal contacted tag");
				const char* co_mat; co_xml->GetData(&co_mat);
				const Material* co = UtilGetMaterialId(co_mat);

				TRY(co, UtilStrFormat(
						"contacted tag has unknown material %s", co_mat));
				mtb->contacted = co;
			}

			// the metabolism is Receptor reaction site
			if (mtb->reactsite) {
				const CfgParam* lr_xml = mtb_xml->GetParamFirst("LR");
				TRY(lr_xml && (lr_xml->GetDataType() == _STR),
					"illigal LR tag");
				const char* lr; lr_xml->GetData(&lr);
				mtb->LR = UtilGetEquationId(lr);
				TRY(mtb->LR, UtilStrFormat("LR %s is unknown", lr))
			}

			TRY(cfg.MtbNameIdMap.find(name) == cfg.MtbNameIdMap.end(),
				UtilStrFormat("membrane-metabolism %s is doubly", name)); 

			const_cast<Material*>(m)->MtbTbl.push_back(mtb);
			cfg.MtbTbl.push_back(mtb);
			cfg.MtbNameIdMap.insert(Name2IdMapT::value_type(name,mtb->id));

			if (mtb->contacted) {
#ifndef _NO_MESSAGE
				MESSAGE(UtilStrFormat("++ %s id=%d on %s membrane with %s",
						name, mtb->id, mtb->mat->name.c_str(),
						mtb->contacted->name.c_str()));
#else
				MESSAGE(UtilStrFormat("++ %s id=%d on %s membrane with %s",
						name, mtb->id, mtb->mat->name.c_str(),
						mtb->contacted->name.c_str()),0);
#endif //_NO_MESSAGE
			} else {
#ifndef _NO_MESSAGE
				MESSAGE(UtilStrFormat("++ %s id=%d on %s membrane",
						name, mtb->id, mtb->mat->name.c_str()));
#else
				MESSAGE(UtilStrFormat("++ %s id=%d on %s membrane",
						name, mtb->id, mtb->mat->name.c_str()),0);
#endif //_NO_MESSAGE
			}
		}
	}
	
	const int mtbs = cfg.MtbTbl.size();
	cfg.MtbCount = mtbs;
#ifndef _NO_MESSAGE
	MESSAGE(string("total Metabolism Count is ") + string_cast(mtbs));
#else
	MESSAGE(string("total Metabolism Count is ") + string_cast(mtbs),0);
#endif //_NO_MESSAGE

	return mtbs;
}


unsigned int SklSolverFoundation::init_MembMap()
{
#ifndef _NO_MESSAGE
	MESSAGE("Constructing Membrane Table ...");
#else
	MESSAGE("Constructing Membrane Table ...",0);
#endif //_NO_MESSAGE

	const CfgElem* DomainInfo = SklCfgGetTop(DOMAININFO);
	const CfgElem* membfunc   = DomainInfo->GetElemFirst("membrane");

	////////////////////////////////////////////////////////////////////////
	// loop of Mat*Mat
	for (int im = 0; im < cfg.MatCount; im++) {
	for (int in = 0; in < cfg.MatCount; in++) {
		if (!membfunc) { continue; }
		if (im == in)  { continue; }
		const Material* m = cfg.MatTbl[im];
		const Material* n = cfg.MatTbl[in];
		
		// construct membrane map mn
		const string mnname = m->name + ":" + n->name;
		const string nmname = n->name + ":" + m->name;
		if (membfunc->GetElemFirst(mnname.c_str())
			&& !UtilGetMembraneId(mnname)) {
			//
			Membrane* mn = new Membrane;
			Membrane* nm = new Membrane;
			mn->id = cfg.MembTbl.size();
			nm->id = cfg.MembTbl.size() + 1;
			mn->mat[0] = m; mn->mat[1] = n;
			nm->mat[0] = n; nm->mat[1] = m;
			mn->name = mnname;
			nm->name = nmname;

			cfg.MembTbl.push_back(mn);
			cfg.MembTbl.push_back(nm);
			cfg.MembNameIdMap.insert(Name2IdMapT::value_type(mn->name,mn->id));
			cfg.MembNameIdMap.insert(Name2IdMapT::value_type(nm->name,nm->id));
#ifndef _NO_MESSAGE
			MESSAGE(string("++ ") + mn->name + " id=" + string_cast(mn->id));
			MESSAGE(string("++ ") + nm->name + " id=" + string_cast(nm->id));
#else
			MESSAGE(string("++ ") + mn->name + " id=" + string_cast(mn->id),0);
			MESSAGE(string("++ ") + nm->name + " id=" + string_cast(nm->id),0);
#endif //_NO_MESSAGE
		}
	}}
	cfg.MembCount = cfg.MembTbl.size();
#ifndef _NO_MESSAGE
	MESSAGE("Membrane Table size is " + string_cast(cfg.MembCount));
#else
	MESSAGE("Membrane Table size is " + string_cast(cfg.MembCount),0);
#endif //_NO_MESSAGE

	return cfg.MembCount;
}

unsigned int SklSolverFoundation::init_MlMbMap()
{
#ifndef _NO_MESSAGE
	MESSAGE("Constructing Molecular * Membrane Table ...");
#else
	MESSAGE("Constructing Molecular * Membrane Table ...",0);
#endif //_NO_MESSAGE

	const CfgElem* DomainInfo = SklCfgGetTop(DOMAININFO);
	const CfgElem* membfunc   = DomainInfo->GetElemFirst("membrane");

	////////////////////////////////////////////////////////////////////////
	// loop of membrane
	for (int imn = 0; imn < cfg.MembCount; imn++) {
		if (!membfunc) { continue; }

		const Membrane* mn = cfg.MembTbl[imn];
		const Material* m  = mn->mat[0];
		const Material* n  = mn->mat[1];
		const Membrane* nm = UtilGetMembraneId(n,m);

		for (const CfgElem* mb = membfunc->GetElemFirst(mn->name.c_str());
				mb != NULL; mb = membfunc->GetElemNext(mb,mn->name.c_str())) {

		// loop of molecular
		for (const CfgElem* mol = mb->GetElemFirst(NULL);
				mol != NULL; mol = mb->GetElemNext(mol,NULL)) {

			const std::string name_g = mol->GetName();
			const Molecular* g = UtilGetMolecularId(name_g);
			TRY(g, name_g + " has been undefined.");

			UtilAddNewMolMembId(g,mn);
			UtilAddNewMolMembId(g,nm);
		}}
	}
	cfg.MlMbCount = cfg.MlMbTbl.size();
#ifndef _NO_MESSAGE
	MESSAGE("Mol*Membrane Table size is " + string_cast(cfg.MlMbCount));
#else
	MESSAGE("Mol*Membrane Table size is " + string_cast(cfg.MlMbCount),0);
#endif //_NO_MESSAGE

	return cfg.MlMbCount;
}

unsigned int SklSolverFoundation::init_MembareaMap()
{
#ifndef _NO_MESSAGE
	MESSAGE("Constructing Membrane Area Table ...");
#else
	MESSAGE("Constructing Membrane Area Table ...",0);
#endif //_NO_MESSAGE

	////////////////////
	// loop of Mat*Mat
	for( int im = 0; im < cfg.MatCount; im++ ){
	for( int in = im+1; in < cfg.MatCount; in++ ){
		const Material* m = cfg.MatTbl[im];
		const Material* n = cfg.MatTbl[in];

		const string mnname = m->name + ":" + n->name + "_MembArea";
		const string nmname = n->name + ":" + m->name + "_MembArea";

		const SklCfgInFile* infile = SklCfgGetInFileFirst();
		while( infile ){
			const char* attr = infile->GetAttr();
			if( (strcasecmp( attr, mnname.c_str() ) == 0) ||
			    (strcasecmp( attr, nmname.c_str() ) == 0) )
			{
				if( cfg.MembareaNameIdMap.find(attr) == cfg.MembareaNameIdMap.end() ){
					Membarea* mn = new Membarea;
					mn->id = cfg.MembareaTbl.size();
					mn->mat[0] = m; mn->mat[1] = n;
					mn->name = attr;

					cfg.MembareaTbl.push_back(mn);
					cfg.MembareaNameIdMap.insert(Name2IdMapT::value_type(mnname, mn->id));
					cfg.MembareaNameIdMap.insert(Name2IdMapT::value_type(nmname, mn->id));
#ifndef _NO_MESSAGE
					MESSAGE(string("++ ") + mnname + " id=" + string_cast(mn->id));
					MESSAGE(string("++ ") + nmname + " id=" + string_cast(mn->id));
#else
					MESSAGE(string("++ ") + mnname + " id=" + string_cast(mn->id),0);
					MESSAGE(string("++ ") + nmname + " id=" + string_cast(mn->id),0);
#endif //_NO_MESSAGE
				}
				else{
					MSG_WARNING(string("Duplicate ") + string_cast(attr) + string(" attribute, --Ignored."));
				}
			}

			infile = SklCfgGetInFileNext( infile );
		}
	}}
	cfg.MembareaCount = cfg.MembareaTbl.size();
#ifndef _NO_MESSAGE
	MESSAGE("Membrane Area Table size is " + string_cast(cfg.MembareaCount));
#else
	MESSAGE("Membrane Area Table size is " + string_cast(cfg.MembareaCount),0);
#endif //_NO_MESSAGE

	return cfg.MembareaCount;
}

unsigned int SklSolverFoundation::init_SteadyStateMap()
{
#ifndef _NO_MESSAGE
	MESSAGE("Constructing steady state table ...");
#else
	MESSAGE("Constructing steady state table ...",0);
#endif //_NO_MESSAGE

	// initialize
	cfg.SSinterval = 0;

	const CfgElem* parameter  = SklCfgGetTop(PARAMETER);
	const CfgElem* ss_setting = parameter->GetElemFirst("steady_state_setting");

	while( ss_setting ){
		// log file name
		const char* ssfname = NULL;
		const CfgParam* fname_xml = ss_setting->GetParamFirst("fname");
		if( fname_xml && fname_xml->GetData(&ssfname) ){
			cfg.sslogfilename = ssfname;
		}
		else{
			cfg.sslogfilename = "SteadyState_log.csv";
		}

		// interval
		const CfgParam* interval_xml = ss_setting->GetParamFirst("Interval");
		int interval; interval_xml? interval_xml->GetData(&interval): interval = 1;
		cfg.SSinterval = interval;

		// start step
		const CfgParam* startstep_xml = ss_setting->GetParamFirst("StartStep");
		int startstep; startstep_xml? startstep_xml->GetData(&startstep): startstep = 0;
		cfg.SSstartStep = startstep;

		// solver exec
		const char* sexec = NULL;
		const CfgParam* solvexec_xml  = ss_setting->GetParamFirst("SolvExec");
		if( solvexec_xml && solvexec_xml->GetData(&sexec) && (strcasecmp(sexec, "STOP") == 0) ){
			cfg.SSsolvExec = STOP;
		}
		else{
			cfg.SSsolvExec = CONTINUE;
		}

		// output sph
		const char* osph = NULL;
		const CfgParam* osph_xml = ss_setting->GetParamFirst("OutputSPH");
		if( osph_xml && osph_xml->GetData(&osph) && (strcasecmp(osph, "ON") == 0) ){
			cfg.SSoutSPH = true;
		}
		else{
			cfg.SSoutSPH = false;
		}

		// operation
		const char* ope = NULL;
		const CfgParam* operation_xml = ss_setting->GetParamFirst("operation");
		if( operation_xml && operation_xml->GetData(&ope) && (strcasecmp(ope, "OR") == 0) ){
			cfg.SSope = OR;
		}
		else{
			cfg.SSope = AND;
		}

		const CfgElem* ss_eqn = ss_setting->GetElemFirst();
		while( ss_eqn ){
			// threshold
			SKL_REAL th;
			const CfgParam* threshold_xml = ss_eqn->GetParamFirst("threshold");
			threshold_xml? threshold_xml->GetData(&th): th = 0.1;

			// threshold type
			THTYPE thType;
			const char* th_type = NULL;
			const CfgParam* th_type_xml = ss_eqn->GetParamFirst("th_type");
			if( th_type_xml && th_type_xml->GetData(&th_type) && (strcasecmp(th_type, "RATIO") == 0) ){
				thType = RATIO;
			}
			else{
				thType = ABSOLUTE;
			}

			// compare type
			COMPTYPE compType;
			const char* comp_type = NULL;
			const CfgParam* CompType_xml  = ss_eqn->GetParamFirst("CompType");
			if( CompType_xml && CompType_xml->GetData(&comp_type) && (strcasecmp(comp_type, "LessEqual") == 0) ){
				compType = LE;
			}
			else if( CompType_xml && CompType_xml->GetData(&comp_type) && (strcasecmp(comp_type, "GreaterThan") == 0) ){
				compType = GT;
			}
			else if( CompType_xml && CompType_xml->GetData(&comp_type) && (strcasecmp(comp_type, "GreaterEqual") == 0) ){
				compType = GE;
			}
			else{
				compType = LT;
			}

			// judge number
			unsigned int jnum;
			int gdata;
			const CfgParam* jnum_xml = ss_eqn->GetParamFirst("JudgeNum");
			jnum_xml? jnum_xml->GetData(&gdata): gdata = 1;

			if( gdata < 0 ){
				MSG_WARNING(UtilStrFormat("Judge Number is negative number.(%d)", gdata));
				MSG_WARNING(UtilStrFormat("Judge Number change 1"));
				jnum = 1;
			}
			else{
				jnum = gdata;
			}

			if( jnum == 0 ){
				MSG_WARNING(UtilStrFormat("Judge Number is %d", jnum));
			}

			//////////////////////////////////////////////////
			// equation
			//
			const char* eqnName = ss_eqn->GetName();

			// check name format
			const string name( eqnName );
			const size_t nsep = name.find_first_of(':');
			const string mat  = name.substr(0, nsep);
			const string mol  = name.substr(nsep + 1);

			MatSet matset;
			if( mat == "*" ){
				for( size_t m = 0; m < cfg.MatCount; m++ ){
					const Material* mt = cfg.MatTbl[m];
					matset.push_back(mt);
				}
			} else if ( IdRange::isTuple( mat ) ){
				IdRange range( mat );
				for( int cid = range.begin(); cid <= range.end(); cid++ ) {
					const size_t id = cfg.MatCidIdMap.find(cid)->second;
					const Material* mt = cfg.MatTbl[id];
					matset.push_back(mt);
				}
			} else {
				const size_t matid = cfg.MatNameIdMap.find(mat)->second;
				const Material* mt = cfg.MatTbl[matid];
				matset.push_back(mt);
			}

			MolSet molset;
			if( mol == "*" ){
				for( size_t m = 0; m < cfg.MolCount; m++ ){
					const Molecular* ml = cfg.MolTbl[m];
					molset.push_back(ml);
				}
			} else if ( IdRange::isTuple( mol ) ){
				IdRange range( mol );
				for( int cid = range.begin(); cid <= range.end(); cid++ ){
					const size_t id = cfg.MolCidIdMap.find(cid)->second;
					const Molecular* ml = cfg.MolTbl[id];
					molset.push_back(ml);
				}
			} else {
				const size_t molid = cfg.MolNameIdMap.find(mol)->second;
				const Molecular* ml = cfg.MolTbl[molid];
				molset.push_back(ml);
			}

			for( MatSet::const_iterator mt = matset.begin(); mt != matset.end(); mt++ ){
			for( MolSet::const_iterator ml = molset.begin(); ml != molset.end(); ml++ ){
				const Equation* e = UtilGetEquationId( (*ml), (*mt) );
				if( !e ) { continue; }

				map<int, SteadyState*>::iterator it = cfg.SSmap.find( e->id );
				if( it != cfg.SSmap.end() ){
					// erase
					cfg.SSmap.erase( it );
				}

				SteadyState* ss = new SteadyState;
				ss->eqn = e;
				ss->threshold = th;
				ss->thType = thType;
				ss->compType = compType;
				ss->SScnt = 0;
				ss->SSjnum = jnum;
				cfg.SSmap.insert( map<int, SteadyState*>::value_type(e->id, ss) );

			}}


			ss_eqn = ss_setting->GetElemNext(ss_eqn, NULL);
		}
		ss_setting = parameter->GetElemNext(ss_setting, "steady_state_setting");
	}

	cfg.SSCount = cfg.SSmap.size();

#ifndef _NO_MESSAGE
	MESSAGE("steady state map size is " + string_cast(cfg.SSCount));
#else
	MESSAGE("steady state map size is " + string_cast(cfg.SSCount),0);
#endif //_NO_MESSAGE

	// steady state : false
	if( cfg.SSCount == 0 ) return cfg.SSCount;

	////////////////////
	// output log file
	//
	MESSAGE("steady state setting");
	MESSAGE(UtilStrFormat("+ steady state log file : %s", cfg.sslogfilename));

	MESSAGE(UtilStrFormat("+ interval    : %d", cfg.SSinterval));

	if( cfg.SSsolvExec == STOP ){
		MESSAGE("+ solver exec : STOP");
	} else if( cfg.SSsolvExec == CONTINUE ){
		MESSAGE("+ solver exec : CONTINUE");
	}

	if( cfg.SSoutSPH ){
		MESSAGE("+ Output SPH  : ON");
	} else {
		MESSAGE("+ Output SPH  : OFF");
	}

	if( cfg.SSope == AND ){
		MESSAGE("+ operation   : AND");
	} else if( cfg.SSope == OR ){
		MESSAGE("+ operation   : OR");
	}

	MESSAGE("steady state Equation table ...");
	for( map<int, SteadyState*>::iterator it = cfg.SSmap.begin(); it != cfg.SSmap.end(); it++ ){
		int id = (*it).first;
		const SteadyState* ss = (*it).second;
		MESSAGE(UtilStrFormat("++ map ID         : %d", id));
		MESSAGE(UtilStrFormat("++ Equation       : %s", ss->eqn->name.c_str()));
		MESSAGE(UtilStrFormat("++ threshold      : %e", ss->threshold));

		if( ss->thType == ABSOLUTE ){
			MESSAGE("++ threshold type : ABSOLUTE");
		} else if( ss->thType == RATIO ){
			MESSAGE("++ threshold type : RATIO");
		}

		if( ss->compType == LT ){
			MESSAGE("++ compare type   : LT");
		} else if( ss->compType == LE ){
			MESSAGE("++ compare type   : LE");
		} else if( ss->compType == GT ){
			MESSAGE("++ compare type   : GT");
		} else if( ss->compType == GE ){
			MESSAGE("++ compare type   : GE");
		}

		MESSAGE(UtilStrFormat("++ counter        : %d", ss->SScnt));
		MESSAGE(UtilStrFormat("++ judge number   : %d", ss->SSjnum));
	}

	return cfg.SSCount;
}

int SklSolverFoundation::ChkMatVolumeRatio()
{
	const size_t I    = cfg.Pdiv.size[0];
	const size_t J    = cfg.Pdiv.size[1];
	const size_t K    = cfg.Pdiv.size[2];
	const size_t mats = cfg.MatCount;

	S3D_REAL VolFld(I,J,K,0);
	UtilClearArray(&VolFld,SKL_REAL(0.0));

	bool over = false, under = false;
	for (size_t m = 0; m < mats; m++) {
	for (size_t k = 0; k < K;    k++) {
	for (size_t j = 0; j < J;    j++) {
	for (size_t i = 0; i < I;    i++) {

		const SKL_REAL& Vf = MatVfFld->val(i,j,k,m);
		SKL_REAL&       Vs = VolFld.val(i,j,k);
		
		Vs += Vf;
		under |= (Vs < SKL_REAL(0.00));
		under |= (Vf < SKL_REAL(0.00));
		over  |= (Vf > SKL_REAL(1.00));
		over  |= (Vs > SKL_REAL(1.00));
	}}}}

	MSG_WARNING("some cell volume is over  1.0", over);
	MSG_WARNING("some cell volume is under 0.0", under);

	return 1;
}

int SklSolverFoundation::ExptimEqn(const SKL_REAL dt)
{
	const int vc = VC;
	const int I  = cfg.Pdiv.size[0];
	const int J  = cfg.Pdiv.size[1];
	const int K  = cfg.Pdiv.size[2];

	const SKL_REAL dxi = SKL_REAL(1.0)/cfg.pitch[0];
	const SKL_REAL dyi = SKL_REAL(1.0)/cfg.pitch[1];
	const SKL_REAL dzi = SKL_REAL(1.0)/cfg.pitch[2];

	for (int p = 0; p < cfg.EqnCount; p++) {
		if (!cfg.PeqnAss[p]) { continue; }

		const int mt = cfg.EqnTbl[p]->mat->id;

		SKL_REAL*       P1 = UtilGetS4DN(EqnP1Fld,p);
		const SKL_REAL* P0 = UtilGetS4DN(EqnP0Fld,p);
		const SKL_REAL* fx = UtilGetS4DN(DifFluxX,p);
		const SKL_REAL* fy = UtilGetS4DN(DifFluxY,p);
		const SKL_REAL* fz = UtilGetS4DN(DifFluxZ,p);
		const SKL_REAL* Mt = UtilGetS4DN(MembTrmFld,p);
		const SKL_REAL* Fm = UtilGetS4DN(MembFlux,p);
		const SKL_REAL* Qm = UtilGetS4DN(MembExQ,p);
		const SKL_REAL* R  = UtilGetS4DN(MtbTrmFld,p);
		const SKL_REAL* S0 = UtilGetS4DN(SrcTrmFld,p);
		const SKL_REAL* Vfi= UtilGetS4DN(MatVfInv,mt);
		const SKL_REAL* Sm = UtilGetS4DN(MembArea,mt);

		exptim_(P1,P0,fx,fy,fz,Mt,Fm,Qm,R,S0,
			Vfi,Sm,&dt,&dxi,&dyi,&dzi,&I,&J,&K,&vc);
	}

	return 1;
}

int SklSolverFoundation::ExptimEqn(const SKL_REAL dt, const unsigned int step)
{
	const int vc = VC;
	const int I  = cfg.Pdiv.size[0];
	const int J  = cfg.Pdiv.size[1];
	const int K  = cfg.Pdiv.size[2];

	const SKL_REAL dxi = SKL_REAL(1.0)/cfg.pitch[0];
	const SKL_REAL dyi = SKL_REAL(1.0)/cfg.pitch[1];
	const SKL_REAL dzi = SKL_REAL(1.0)/cfg.pitch[2];

	for (int p = 0; p < cfg.EqnCount; p++) {
		if (!cfg.PeqnAss[p]) { continue; }

		const int mt = cfg.EqnTbl[p]->mat->id;

		SKL_REAL*       P1 = UtilGetS4DN(EqnP1Fld,p);
		const SKL_REAL* P0 = UtilGetS4DN(EqnP0Fld,p);
		const SKL_REAL* fx = UtilGetS4DN(DifFluxX,p);
		const SKL_REAL* fy = UtilGetS4DN(DifFluxY,p);
		const SKL_REAL* fz = UtilGetS4DN(DifFluxZ,p);
		const SKL_REAL* Mt = UtilGetS4DN(MembTrmFld,p);
		const SKL_REAL* Fm = UtilGetS4DN(MembFlux,p);
		const SKL_REAL* Qm = UtilGetS4DN(MembExQ,p);
		const SKL_REAL* R  = UtilGetS4DN(MtbTrmFld,p);
		const SKL_REAL* S0 = UtilGetS4DN(SrcTrmFld,p);
		const SKL_REAL* Vfi= UtilGetS4DN(MatVfInv,mt);
		const SKL_REAL* Sm = UtilGetS4DN(MembArea,mt);


                const char* ename = cfg.EqnTbl[p]->name.c_str();

		if(cfg.time_integ == EULER){

			exptim_(P1,P0,fx,fy,fz,Mt,Fm,Qm,R,S0,
				Vfi,Sm,&dt,&dxi,&dyi,&dzi,&I,&J,&K,&vc);
		}
		else if(cfg.time_integ == ADAMS_BASHFORTH){
			SKL_REAL* rhs = UtilGetS4DN(Rhs,p);
			SKL_REAL* rhso = UtilGetS4DN(Rhso,p);

			// const unsigned int totalstep = SklGetTotalStep();

			// debug write
			//exptimab_(&p,&totalstep,rhs,rhso,P1,P0,fx,fy,fz,Fm,Qm,R,S0,
			//	Vfi,Sm,&dt,&dxi,&dyi,&dzi,&I,&J,&K,&vc);
			exptimab_(&cfg.RhsoInitFlag,rhs,rhso,P1,P0,fx,fy,fz,Mt,Fm,Qm,R,S0,
				Vfi,Sm,&dt,&dxi,&dyi,&dzi,&I,&J,&K,&vc);

			// cfg.RhsoInitFlag = 1;
		}
		else{
			SKL_REAL* rhs = UtilGetS4DN(Rhs,p);

			exptimrk_(rhs,P1,P0,Mt,Fm,Qm,R,S0,
				Vfi,Sm,&dt,&dxi,&dyi,&dzi,&I,&J,&K,&vc);
		}

	}

	cfg.RhsoInitFlag = 1;

        return 1;
}

int SklSolverFoundation::ExptimMat(const SKL_REAL dt)
{
	return 1;
}

int SklSolverFoundation::SetRestartTime(const SklVoxDataSet* dSet)
{
	// set base time for resart
	if (cfg.Restart) {
		SKL_REAL time;
		int      step;
		if (dSet->GetTime(time) && dSet->GetStep(step)) {
			if (time > SklGetBaseTime()) { SklSetBaseTime(time); }
			if (step > SklGetBaseStep()) { SklSetBaseStep(step); }
		}
	}

	return 1;
}

int SklSolverFoundation::SyncRestartTime()
{
	const int gid = cfg.grpAll.gid;

	SKL_REAL time = SklGetBaseTime();
	int      step = SklGetBaseStep();

	SKL_REAL ntime = SKL_REAL(0.0);
	int      nstep = 0;

	// select maximum value inter all nodes
	if (cfg.Restart) {
#ifdef SPHERE175
		//2009.12.07
		para_cmp->Allreduce(&time,&ntime,1,SKL_ARRAY_DTYPE_REAL,SKL_MAX,gid);
		para_cmp->Allreduce(&step,&nstep,1,SKL_ARRAY_DTYPE_INT, SKL_MAX,gid);
#else
		ParaMngr.Allreduce(&time,&ntime,1,SKL_ARRAY_DTYPE_REAL,SKL_MAX,gid);
		ParaMngr.Allreduce(&step,&nstep,1,SKL_ARRAY_DTYPE_INT, SKL_MAX,gid);
#endif
	}
	
	SklSetBaseTime(ntime);
	SklSetBaseStep(nstep);

	return 1;
}
                                                             
int SklSolverFoundation::CalcCmpMat()
{
#ifndef _NO_MESSAGE
	MESSAGE("Complete Material Volumes ...");
#else
	MESSAGE("Complete Material Volumes ...",0);
#endif //_NO_MESSAGE

	const size_t mats = cfg.MatCount;
	const size_t I    = cfg.Pdiv.size[0];
	const size_t J    = cfg.Pdiv.size[1];
	const size_t K    = cfg.Pdiv.size[2];
	const int    vc   = VC;

	const int cmid = cfg.CmpMatId;
	if (cmid < 0) {
#ifndef _NO_MESSAGE
		MESSAGE("+++ Materials were incompleted.");
		MESSAGE("+++ Because material, specified to complete is none.");
#else
		MESSAGE("+++ Materials were incompleted.",0);
		MESSAGE("+++ Because material, specified to complete is none.",0);
#endif //_NO_MESSAGE
		return 0;
	}

#ifndef _NO_MESSAGE
	MESSAGE(string("Complete with ") + cfg.MatTbl[cmid]->name);
#else
	MESSAGE(string("Complete with ") + cfg.MatTbl[cmid]->name,0);
#endif //_NO_MESSAGE

	for (int k = 0; k < K; k++) {
	for (int j = 0; j < J; j++) {
	for (int i = 0; i < I; i++) {
		SKL_REAL Vf = SKL_REAL(0.0);
		for (int m = 0; m < mats; m++) {
			if (m == cmid) { continue; }
			Vf += MatVfFld->val(i,j,k,m);
		}
		if (Vf < SKL_REAL(1.0)) {
			MatVfFld->val(i,j,k,cmid) = SKL_REAL(1.0) - Vf;
		}
	}}}

//add fns 2011/12 strat
        int MatVfGC = MatVfFld->GetVCellSize();
        // printf("MatVfGC : %d \n",MatVfGC);
        MatVfFld->CommBndCell(MatVfGC);
        int nID[6];
        const int& gid = cfg.grpP.gid;
#ifdef SPHERE175
        nID[0] = para_mng->IsCommID(-1, 0, 0, gid); // -X
        nID[1] = para_mng->IsCommID( 1, 0, 0, gid); // +X
        nID[2] = para_mng->IsCommID( 0,-1, 0, gid); // -Y
        nID[3] = para_mng->IsCommID( 0, 1, 0, gid); // +Y
        nID[4] = para_mng->IsCommID( 0, 0,-1, gid); // -Z
        nID[5] = para_mng->IsCommID( 0, 0, 1, gid); // +Z
#else
        nID[0] = ParaMngr.GetCommID(-1, 0, 0, gid); // -X
        nID[1] = ParaMngr.GetCommID( 1, 0, 0, gid); // +X
        nID[2] = ParaMngr.GetCommID( 0,-1, 0, gid); // -Y
        nID[3] = ParaMngr.GetCommID( 0, 1, 0, gid); // +Y
        nID[4] = ParaMngr.GetCommID( 0, 0,-1, gid); // -Z
        nID[5] = ParaMngr.GetCommID( 0, 0, 1, gid); // +Z
#endif
//        printf("nID : %d %d %d %d %d %d \n",nID[0],nID[1],nID[2],nID[3],nID[4],nID[5]);
//-x copy
        if( nID[0] < 0 ) {
           for(int i=0; i < MatVfGC ; i++ ) {
           for(int k=0; k < K;        k++ ) {
           for(int j=0; j < J;        j++ ) {
               MatVfFld->val(0-MatVfGC+i,j,k,cmid)=MatVfFld->val(i,j,k,cmid);
           }}}
        }
//+x copy
        if( nID[1] < 0 ) {
           for(int i=0; i < MatVfGC ; i++ ) {
           for(int k=0; k < K;        k++ ) {
           for(int j=0; j < J;        j++ ) {
               MatVfFld->val(I+i,j,k,cmid)=MatVfFld->val(I-MatVfGC+i,j,k,cmid);
           }}}
        }
//-y copy
        if( nID[2] < 0 ) {
           for(int j=0; j < MatVfGC; j++ ) {
           for(int k=0; k < K;       k++ ) {
           for(int i=0; i < I;       i++ ) {
              MatVfFld->val(i,0-MatVfGC+j,k,cmid)=MatVfFld->val(i,j,k,cmid);
           }}}
         }
//+y copy
         if( nID[3] < 0 ) {
           for(int j=0; j < MatVfGC; j++ ) { 
           for(int k=0; k < K;       k++ ) {
           for(int i=0; i < I;       i++ ) {
              MatVfFld->val(i,J+j,k,cmid)=MatVfFld->val(i,J-MatVfGC+j,k,cmid);
           }}}
         }
//-z copy
         if( nID[4] < 0 ) {
           for(int k=0; k < MatVfGC; k++ ) {
           for(int j=0; j < J;       j++ ) {
           for(int i=0; i < I;       i++ ) { 
              MatVfFld->val(i,j,0-MatVfGC+k,cmid)=MatVfFld->val(i,j,k,cmid);
           }}}
         }
//+z copy
         if( nID[5] < 0 ) {
           for(int k=0; k < MatVfGC; k++ ) {
           for(int j=0; j < J;       j++ ) {
           for(int i=0; i < I;       i++ ) { 
              MatVfFld->val(i,j,K+k,cmid)=MatVfFld->val(i,j,K-MatVfGC+k,cmid);
           }}}
         }
//add fns 2011/12 end

	DB4(MESSAGE(S4D2string(MatVfFld,cmid,false)));

	for (int k = 0; k < K;     k++) {
	for (int j = 0; j < J;     j++) {
	for (int i = 0; i < I + 1; i++) {
//mod fns 2011/12 start
//		SKL_REAL Ax = SKL_REAL(0.0);
//		for (int m = 0; m < mats; m++) {
//			if (m == cmid) { continue; }
//			Ax += MatAreaX->val(i,j,k,m);
//		}
//		if (Ax < SKL_REAL(1.0)) {
//			MatAreaX->val(i,j,k,cmid) = SKL_REAL(1.0) - Ax;
//		}
                if(( MatVfFld->val(i-1,j,k,cmid) == 0.0 ) ||
                   ( MatVfFld->val(i,  j,k,cmid) == 0.0 ) ) {
                        MatAreaX->val(i,j,k,cmid) = SKL_REAL(0.0);
                }
                else if(( MatVfFld->val(i-1,j,k,cmid) == 1.0 ) ||
                  ( MatVfFld->val(i,  j,k,cmid) == 1.0 )) {
                        MatAreaX->val(i,j,k,cmid) = SKL_REAL(1.0);
                }
                else {
                        MatAreaX->val(i,j,k,cmid) = 
                      ( MatVfFld->val(i-1,j,k,cmid)+MatVfFld->val(i,j,k,cmid) )*0.5;
                }
//mod fns 2011/12 end
	}}}
	DB4(MESSAGE(S4D2string(MatAreaX,cmid,false)));

	for (int k = 0; k < K;     k++) {
	for (int j = 0; j < J + 1; j++) {
	for (int i = 0; i < I;     i++) {
//mod fns 2011/12 start
//		SKL_REAL Ay = SKL_REAL(0.0);
//		for (int m = 0; m < mats; m++) {
//			if (m == cmid) { continue; }
//			Ay += MatAreaY->val(i,j,k,m);
//		}
//		if (Ay < SKL_REAL(1.0)) {
//			MatAreaY->val(i,j,k,cmid) = SKL_REAL(1.0) - Ay;
//		}
                if(( MatVfFld->val(i,j-1,k,cmid) == 0.0 ) ||
                   ( MatVfFld->val(i,j  ,k,cmid) == 0.0 )) {
                        MatAreaY->val(i,j,k,cmid) = SKL_REAL(0.0);
                }
                else if(( MatVfFld->val(i,j-1,k,cmid) == 1.0 ) ||
                  ( MatVfFld->val(i,j  ,k,cmid) == 1.0 )) {
                        MatAreaY->val(i,j,k,cmid) = SKL_REAL(1.0);
                }
                else {
                        MatAreaY->val(i,j,k,cmid) =
                      ( MatVfFld->val(i,j-1,k,cmid)+MatVfFld->val(i,j,k,cmid) )*0.5;
                }
//mod fns 2011/12 end
	}}}
	DB4(MESSAGE(S4D2string(MatAreaY,cmid,false)));

	for (int k = 0; k < K + 1; k++) {
	for (int j = 0; j < J;     j++) {
	for (int i = 0; i < I;     i++) { 
//mod fns 2011/12 start
//		SKL_REAL Az = SKL_REAL(0.0);
//		for (int m = 0; m < mats; m++) {
//			if (m == cmid) { continue; }
//			Az += MatAreaZ->val(i,j,k,m);
//		}
//		if (Az < SKL_REAL(1.0)) {
//			MatAreaZ->val(i,j,k,cmid) = SKL_REAL(1.0) - Az;
//		}
                if(( MatVfFld->val(i,j,k-1,cmid) == 0.0 ) ||
                   ( MatVfFld->val(i,j,k  ,cmid) == 0.0 )) {
                        MatAreaZ->val(i,j,k,cmid) = SKL_REAL(0.0);
                }
                else if(( MatVfFld->val(i,j,k-1,cmid) == 1.0 ) ||
                  ( MatVfFld->val(i,j,k  ,cmid) == 1.0 )) {
                        MatAreaZ->val(i,j,k,cmid) = SKL_REAL(1.0);
                }
                else {
                        MatAreaZ->val(i,j,k,cmid) =
                      ( MatVfFld->val(i,j,k-1,cmid)+MatVfFld->val(i,j,k,cmid) )*0.5;
                }
//mod fns 2011/12 end                
	}}}
	DB4(MESSAGE(S4D2string(MatAreaZ,cmid,false)));

	return 1;
}

int SklSolverFoundation::CalcMatVfInv()
{
#ifndef _NO_MESSAGE
	MESSAGE("Calc Material Volumes Inverse ...");
#else
	MESSAGE("Calc Material Volumes Inverse ...",0);
#endif //_NO_MESSAGE

	const int mats = cfg.MatCount;
	const int I    = cfg.Pdiv.size[0];
	const int J    = cfg.Pdiv.size[1];
	const int K    = cfg.Pdiv.size[2];
	const int vc   = VC;

	for (int m = 0;   m < mats;   m++) {
	for (int k = -vc; k < K + vc; k++) {
	for (int j = -vc; j < J + vc; j++) {
	for (int i = -vc; i < I + vc; i++) {
		if (MatVfFld->val(i,j,k,m) != SKL_REAL(0.0)) {
			MatVfInv->val(i,j,k,m) = SKL_REAL(1.0) / MatVfFld->val(i,j,k,m);
		} else {
			MatVfInv->val(i,j,k,m) = SKL_REAL(0.0);
		}
	}}}}

	return 1;
}

int SklSolverFoundation::CalcMembArea()
{
#ifndef _NO_MESSAGE
	MESSAGE("Calculate Membrane size from Material Volumes ...");
#else
	MESSAGE("Calculate Membrane size from Material Volumes ...",0);
#endif //_NO_MESSAGE

	const size_t I    = cfg.Pdiv.size[0];
	const size_t J    = cfg.Pdiv.size[1];
	const size_t K    = cfg.Pdiv.size[2];
	const int    vc   = VC;

	const SKL_REAL dx = cfg.pitch[0];
	const SKL_REAL dy = cfg.pitch[1];
	const SKL_REAL dz = cfg.pitch[2];

	for(size_t m = 0; m < cfg.MatCount; m++) {

		for (size_t k = 0; k < K; k++) {
		for (size_t j = 0; j < J; j++) {
		for (size_t i = 0; i < I; i++) {
			const SKL_REAL& Ax0 = MatAreaX->val(i,  j,k,m);
			const SKL_REAL& Ax1 = MatAreaX->val(i+1,j,k,m);
			const SKL_REAL& Ay0 = MatAreaY->val(i,j,  k,m);
			const SKL_REAL& Ay1 = MatAreaY->val(i,j+1,k,m);
			const SKL_REAL& Az0 = MatAreaZ->val(i,j,k  ,m);
			const SKL_REAL& Az1 = MatAreaZ->val(i,j,k+1,m);
			const SKL_REAL  dAx = (Ax1 - Ax0) * dy * dz;
			const SKL_REAL  dAy = (Ay1 - Ay0) * dx * dz;
			const SKL_REAL  dAz = (Az1 - Az0) * dx * dy;

			// area size Sm
			MembArea->val(i,j,k,m) = std::sqrt(dAx * dAx + dAy * dAy + dAz * dAz);

			// normal vector
			if (MembArea->val(i,j,k,m) > SKL_REAL(0.0)) {
				const SKL_REAL p = (MatVfFld->val(i,j,k,m) == 1.0)? 1.0: 0.5;
				MembNrmlX->val(i,j,k,m) = p * -dAx / MembArea->val(i,j,k,m);
				MembNrmlY->val(i,j,k,m) = p * -dAy / MembArea->val(i,j,k,m);
				MembNrmlZ->val(i,j,k,m) = p * -dAz / MembArea->val(i,j,k,m);
			} else {
				MembNrmlX->val(i,j,k,m) = SKL_REAL(0.0);
				MembNrmlY->val(i,j,k,m) = SKL_REAL(0.0);
				MembNrmlZ->val(i,j,k,m) = SKL_REAL(0.0);
			}
		}}}

		DB4(MESSAGE(S4D2string(MembArea,m,false)));
	}

	return 1;
}

int SklSolverFoundation::InitFieldOut(const CfgElem* output)
{
	const CfgElem*  fldout = (output)?output->GetElemFirst ("field"):   NULL;
	const CfgParam* fldobj = (fldout)?fldout->GetParamFirst("material"):NULL;

	if (fldobj) {
		for (const CfgParam* obj = fldout->GetParamFirst("material");
			obj; obj = fldout->GetParamNext(obj, "material")) {

			const char* name;
			if (!obj->GetData(&name)) { continue; }

			DBX(MESSAGE("material name : "+string(name)));

#ifdef ENABLE_MEMBRANEPOTENTIAL // for membrane potential output
			TypeEnum typ = UNKNOWN;
			const Element* e = NULL;
			if( (typ = potentialsolv.checkPotenType(name)) == POTENTIAL)
			{
				e = potentialsolv.getPotenElem(name);
				if( e != NULL ) DBX(MESSAGE("InitField Potential Eqn nqme:"+string(e->name)));
			}
			else
			{
				typ = cfg.TypMap.find(name)->second.first;
				e   = cfg.TypMap.find(name)->second.second;
			}
#else
			const TypeEnum typ = cfg.TypMap.find(name)->second.first;
			const Element* e   = cfg.TypMap.find(name)->second.second;
#endif

			const bool outflg =
                (typ == MATERIAL)    ? cfg.matIO[cfg.MatNameIdMap[name]]:
                (typ == MEMBAREA)    ? cfg.matIO[cfg.EqnNameIdMap[name]]:
                (typ == EQUATION)    ? cfg.eqnIO[cfg.EqnNameIdMap[name]]:
                (typ == DIFFLUX)     ? cfg.eqnIO[cfg.EqnNameIdMap[name]]:
                (typ == METABOTERM)  ? cfg.eqnIO[cfg.EqnNameIdMap[name]]:
                (typ == MEMBTERM)    ? cfg.eqnIO[cfg.EqnNameIdMap[name]]:
                (typ == SRCTERM)     ? cfg.eqnIO[cfg.EqnNameIdMap[name]]:
                (typ == MEMBEXTERM)  ? cfg.eqnIO[cfg.EqnNameIdMap[name]]:
                (typ == FACERATE_MX) ? cfg.eqnIO[cfg.EqnNameIdMap[name]]: // 2011.06.07
                (typ == FACERATE_PX) ? cfg.eqnIO[cfg.EqnNameIdMap[name]]: // for output opening fraction
                (typ == FACERATE_MY) ? cfg.eqnIO[cfg.EqnNameIdMap[name]]:
                (typ == FACERATE_PY) ? cfg.eqnIO[cfg.EqnNameIdMap[name]]:
                (typ == FACERATE_MZ) ? cfg.eqnIO[cfg.EqnNameIdMap[name]]:
                (typ == FACERATE_PZ) ? cfg.eqnIO[cfg.EqnNameIdMap[name]]:
#ifdef ENABLE_MEMBRANEPOTENTIAL // for membrane potential output
                (typ == POTENTIAL)?  cfg.eqnIO[cfg.EqnNameIdMap[e->name]]:
#endif
					false;
			if (!outflg) { continue; }

			const vxsize_t* sz  = cfg.IOdiv.size;
			const int       gid = cfg.grpIO.gid;
			const SKL_REAL* org = cfg.origin;

			const SklClassType cls = 
                (typ == MATERIAL)    ? SKL_CLASS_SCALAR3D:
                (typ == EQUATION)    ? SKL_CLASS_SCALAR3D:
                (typ == MEMBAREA)    ? SKL_CLASS_VECTOR3D:
                (typ == DIFFLUX)     ? SKL_CLASS_VECTOR3D:
                (typ == METABOTERM)  ? SKL_CLASS_SCALAR3D:
                (typ == MEMBTERM)    ? SKL_CLASS_VECTOR3D:
                (typ == SRCTERM)     ? SKL_CLASS_SCALAR3D:
                (typ == MEMBEXTERM)  ? SKL_CLASS_SCALAR3D:
                (typ == POTENTIAL)   ? SKL_CLASS_SCALAR3D:
                (typ == FACERATE_MX) ? SKL_CLASS_SCALAR3D:  // 2011.06.07
                (typ == FACERATE_PX) ? SKL_CLASS_SCALAR3D:  // for output opening fraction
                (typ == FACERATE_MY) ? SKL_CLASS_SCALAR3D:
                (typ == FACERATE_PY) ? SKL_CLASS_SCALAR3D:
                (typ == FACERATE_MZ) ? SKL_CLASS_SCALAR3D:
                (typ == FACERATE_PZ) ? SKL_CLASS_SCALAR3D:
                SKL_CLASS_SCALAR3D;
			const SklDataType dtype = SKL_ARRAY_DTYPE_REAL;

#ifdef SPHERE175
			SklArayBase* aray = SklAllocateArray(para_mng,name,cls,dtype,sz,0,gid);	//2009.12.07
#else
			SklArayBase* aray = SklAllocateArray(name,cls,dtype,sz,0,gid);
#endif

			MSG_WARNING(string(name) + " is unknown object", !aray, 0);
			if (!aray) continue;
			
#ifdef REAL_IS_DOUBLE
			unsigned long long fsz[3];
#else
			unsigned int fsz[3];
#endif
			for (int i = 0; i < 3; i++) { fsz[i] = sz[i]; }
			SklVoxDataSet* dSet = InitFile(name,fsz,org,cfg.pitch,aray);

			unsigned int s((dSet)?1:0), f;

#ifdef SPHERE175
			para_cmp->Allreduce(&s,&f,1,SKL_ARRAY_DTYPE_UINT,SKL_BAND,cfg.grpP.gid);	//2009.12.07
#else
			ParaMngr.Allreduce(&s,&f,1,SKL_ARRAY_DTYPE_UINT,SKL_BAND,cfg.grpP.gid);
#endif

			stringstream msg;
			msg << "output file initialization `" << name << "` is failed";
			MSG_WARNING(msg.str(), !f, 0);
			if (!f) continue;

			m_OutDataMap.insert(OutDataMapT::value_type(name,dSet));
#ifndef _NO_MESSAGE
			MESSAGE(UtilStrFormat("Set Output attr='%s'", name));
#else
			MESSAGE(UtilStrFormat("Set Output attr='%s'", name),0);
#endif //_NO_MESSAGE
		}

		if(cfg.time_integ == ADAMS_BASHFORTH){
			InitFieldOutRst();
		}
	}

	return 1;
}

int SklSolverFoundation::InitFieldOutRst()
{
		// Rst data InitFieldOut

		// Rhs data
		const SklClassType cls = SKL_CLASS_SCALAR3D;
		const SklDataType dtype = SKL_ARRAY_DTYPE_REAL;
		const vxsize_t* sz  = cfg.IOdiv.size;
		const int       gid = cfg.grpIO.gid;
		const SKL_REAL* org = cfg.origin;
#ifdef REAL_IS_DOUBLE
		unsigned long long fsz[3];
#else
		unsigned int fsz[3];
#endif
		for (int i = 0; i < 3; i++) { fsz[i] = sz[i]; }

		for(int i=0; i<cfg.EqnTbl.size(); i++){
			const Equation* e = cfg.EqnTbl[i];
			string attr = string("RHS_")+e->name;
			if(!SklCfgCheckOutFile(attr.c_str())) continue;

#ifdef SPHERE175
			SklArayBase* aray = SklAllocateArray(para_mng,attr.c_str(),cls,dtype,sz,0,gid);
#else
			SklArayBase* aray = SklAllocateArray(attr.c_str(),cls,dtype,sz,0,gid);
#endif
			SklVoxDataSet* dSet = InitFile(attr.c_str(),fsz,org,cfg.pitch,aray);

			unsigned int s((dSet)?1:0), f;

#ifdef SPHERE175
			para_cmp->Allreduce(&s,&f,1,SKL_ARRAY_DTYPE_UINT,SKL_BAND,cfg.grpP.gid);	//2009.12.07
#else
			ParaMngr.Allreduce(&s,&f,1,SKL_ARRAY_DTYPE_UINT,SKL_BAND,cfg.grpP.gid);
#endif

			stringstream msg;
			msg << "output file initialization `" << attr.c_str() << "` is failed";
#ifdef ECELL_3_1
			MSG_WARNING(msg.str(), !f, 0);
			if (!f) continue;
#else
			MSG_WARNING(msg.str(), !s, 0);
			if (!s) continue;
#endif //ECELL_3_1

			OutRstDataMap.insert(OutDataMapT::value_type(e->name.c_str(),dSet));
#ifndef _NO_MESSAGE
			MESSAGE(UtilStrFormat("Set Output attr='%s'", attr.c_str()));
#else
			MESSAGE(UtilStrFormat("Set Output attr='%s'", attr.c_str()),0);
#endif //_NO_MESSAGE
		}

		return 1;
}


int SklSolverFoundation::InitProbeOut(const CfgElem* output)
{
	m_ProbeListSize = 0;
	if (!output) { return 1; }
	for (const CfgElem* prb = output->GetElemFirst("probe");
			prb; prb = output->GetElemNext(prb, "probe")) {

		m_ProbeListSize++;

		ProbeT probe;
		float coord[3];
		const CfgParam* px = prb->GetParamFirst("px");
		const CfgParam*	py = prb->GetParamFirst("py");
		const CfgParam* pz = prb->GetParamFirst("pz");
		bool flg = ((px)? px->GetData(&coord[0]): false)
				&& ((py)? py->GetData(&coord[1]): false)
				&& ((pz)? pz->GetData(&coord[2]): false);

		MSG_WARNING("probe point cood is illegal", !flg, 0);
		if (!flg) continue;

		probe.x  = coord[0];
		probe.y  = coord[1];
		probe.z  = coord[2];
		probe._i = static_cast<unsigned int>((probe.x - cfg.origin[0]) / cfg.pitch[0]);
		probe._j = static_cast<unsigned int>((probe.y - cfg.origin[1]) / cfg.pitch[1]);
		probe._k = static_cast<unsigned int>((probe.z - cfg.origin[2]) / cfg.pitch[2]);
		
		const int C = (cfg.Parallel)? 1: 0;
		probe.i = probe._i - C * static_cast<int>(cfg.Pdiv.head[0]);
		probe.j = probe._j - C * static_cast<int>(cfg.Pdiv.head[1]);
		probe.k = probe._k - C * static_cast<int>(cfg.Pdiv.head[2]);

		const bool probe_in_domain = (
							(0 <= probe._i) && (probe._i < cfg.size[0])
						 && (0 <= probe._j) && (probe._j < cfg.size[1])
						 && (0 <= probe._k) && (probe._k < cfg.size[2]));
		const bool probe_in_local_domain = (
							(0 <= probe.i) && (probe.i < cfg.Pdiv.size[0])
						 && (0 <= probe.j) && (probe.j < cfg.Pdiv.size[1])
						 && (0 <= probe.k) && (probe.k < cfg.Pdiv.size[2]));

		// exchange node ID whitch has probed cell
		int z = (probe_in_local_domain)? cfg.grpP.nid: 0;

#ifdef SPHERE175
		para_cmp->Allreduce(&z,&(probe.nID),1,SKL_ARRAY_DTYPE_INT,SKL_SUM,cfg.grpP.gid);	//2009.12.07
#else
		ParaMngr.Allreduce(&z,&(probe.nID),1,SKL_ARRAY_DTYPE_INT,SKL_SUM,cfg.grpP.gid);
#endif

		if ((cfg.grpP.nid != probe.nID) && (cfg.grpP.nid != 0)) { continue; }

		MSG_WARNING(UtilStrFormat("probe point (%e,%e,%e) is out of domain",
					probe.x, probe.y, probe.z), !probe_in_domain, 0);
		if (!probe_in_domain) continue;

		int interval;
		const CfgParam* itvl = prb->GetParamFirst("interval");
		probe.interval = ((itvl)&&(itvl->GetData(&interval)))? interval: 1;

		const char* fn = NULL;
		const CfgParam* fname = prb->GetParamFirst("fname");
		probe.fname = ((fname) && (fname->GetData(&fn)))? fn: "probe.csv";

		for (const CfgParam* obj = prb->GetParamFirst("material");
				obj; obj = prb->GetParamNext(obj, "material")) {

			const char* objname = NULL;
			obj->GetData(&objname);
			if (objname && (cfg.TypMap.find(objname) != cfg.TypMap.end())) {
				probe.obj.push_back(objname);
			} else {
				MSG_WARNING(
					UtilStrFormat("probe target \"%s\" is unknown", objname),
					true, 0);
				continue;
			}
		}

		// store probe info
		m_ProbeList.push_back(probe);

		// create file on the node of ID is 0
		if ((cfg.grpP.nid == 0) && (cfg.grpAll.nid == (cfg.SpecialIONode)?1:0)
			&& cfg.InitStart && !cfg.CheckMode) {

			ofstream of(probe.fname.c_str());
			of << "# step, time, x, y, z, i, j, k, ";
			for (vector<string>::iterator i = probe.obj.begin();
					i != probe.obj.end(); i++) {

				const TypeEnum typ = cfg.TypMap.find(*i)->second.first;
				of << '"' << *i << "\", ";
			}
			of << endl;
			of.close();
		}
	}

	return 1;
}

int SklSolverFoundation::OutputField(const bool output_force)
{
	////////////////////////////////////////////////////////////
	// output fields
	const unsigned int I = cfg.Pdiv.size[0];
	const unsigned int J = cfg.Pdiv.size[1];
	const unsigned int K = cfg.Pdiv.size[2];

	const unsigned int step = SklGetTotalStep();
	const SKL_REAL     time = SklGetTotalTime();

	for (OutDataMapT::iterator
	 		of  = m_OutDataMap.begin();
			of != m_OutDataMap.end();
			of++) {

		const size_t   size  = S3D_SIZE(I,J,K,0);
		const string&  attr  = of->first;
		SklVoxDataSet* dSet  = of->second;
		SklArayBase*   ary   = NULL;
		S3D_REAL*      s3d   = dynamic_cast<S3D_REAL*>(dSet->GetData());
		V3D_REAL*      v3d   = dynamic_cast<V3D_REAL*>(dSet->GetData());

#ifdef ENABLE_MEMBRANEPOTENTIAL // for Potential output
		TypeEnum typ = UNKNOWN;
		const Element* e = NULL;
		if( (typ = potentialsolv.checkPotenType(attr.c_str())) == POTENTIAL)
		{
			e = potentialsolv.getPotenElem(attr.c_str());
			if( e != NULL ) DBX(MESSAGE("OutputField Potential Eqn nqme:"+string(e->name)));
		}
		else
		{
			typ   = cfg.TypMap.find(attr)->second.first;
			e     = cfg.TypMap.find(attr)->second.second;
		}
#else
		const TypeEnum typ   = cfg.TypMap.find(attr)->second.first;
		const Element* e     = cfg.TypMap.find(attr)->second.second;
#endif
	 	if (typ == MATERIAL) {
			for (int i = 0; i < I; i++) {
			for (int j = 0; j < J; j++) {
			for (int k = 0; k < K; k++) {
				s3d->val(i,j,k) = MatVfFld->val(i,j,k,e->id);
			}}}
			ary = s3d;
		} else if (typ == MEMBAREA) {
			for (int i = 0; i < I; i++) {
			for (int j = 0; j < J; j++) {
			for (int k = 0; k < K; k++) {
				v3d->u(i,j,k) = MembArea->val(i,j,k,e->id)
								* MembNrmlX->val(i,j,k,e->id);
				v3d->v(i,j,k) = MembArea->val(i,j,k,e->id)
								* MembNrmlY->val(i,j,k,e->id);
				v3d->w(i,j,k) = MembArea->val(i,j,k,e->id)
								* MembNrmlZ->val(i,j,k,e->id);
			}}}
			ary = v3d;
		} else if (typ == EQUATION) {
			for (int i = 0; i < I; i++) {
			for (int j = 0; j < J; j++) {
			for (int k = 0; k < K; k++) {
				s3d->val(i,j,k) = EqnP0Fld->val(i,j,k,e->id);
			}}}
			ary = s3d;
#ifdef ENABLE_MEMBRANEPOTENTIAL // for Potential output
		} else if (typ == POTENTIAL) {
			ary = potentialsolv.getPotenArray(attr.c_str(),s3d,EqnP0Fld,e->id);
			if( !ary ) {
				string n1,m1,m2;
				potentialsolv.splitOutFileName(attr.c_str(),n1,m1,m2);
				DBX(MESSAGE(UtilStrFormat("warning not membranePotential_setting %s:%s"
				                         ,m1.c_str(),m2.c_str())));
				continue;
			}
#endif
		} else if (typ == DIFFLUX) {
			for (int i = 0; i < I; i++) {
			for (int j = 0; j < J; j++) {
			for (int k = 0; k < K; k++) {
				v3d->u(i,j,k) = - (DifFluxX->val(i,j,k,e->id) + DifFluxX->val(i+1,j,k,e->id)) / 2;
				v3d->v(i,j,k) = - (DifFluxY->val(i,j,k,e->id) + DifFluxY->val(i,j+1,k,e->id)) / 2;
				v3d->w(i,j,k) = - (DifFluxZ->val(i,j,k,e->id) + DifFluxZ->val(i,j,k+1,e->id)) / 2;
			}}}
			ary = v3d;
		} else if (typ == MEMBTERM) {
			const Material* m = reinterpret_cast<const Equation*>(e)->mat;
			for (int i = 0; i < I; i++) {
			for (int j = 0; j < J; j++) {
			for (int k = 0; k < K; k++) {
				v3d->u(i,j,k) = MembFlux->val(i,j,k,e->id)
								* MembNrmlX->val(i,j,k,m->id);
				v3d->v(i,j,k) = MembFlux->val(i,j,k,e->id)
								* MembNrmlY->val(i,j,k,m->id);
				v3d->w(i,j,k) = MembFlux->val(i,j,k,e->id)
								* MembNrmlZ->val(i,j,k,m->id);
			}}}
			ary = v3d;
		} else if (typ == METABOTERM) {
			for (int i = 0; i < I; i++) {
			for (int j = 0; j < J; j++) {
			for (int k = 0; k < K; k++) {
				s3d->val(i,j,k) = MtbTrmFld->val(i,j,k,e->id);
			}}}
			ary = s3d;
		} else if (typ == SRCTERM) {
			for (int i = 0; i < I; i++) {
			for (int j = 0; j < J; j++) {
			for (int k = 0; k < K; k++) {
				s3d->val(i,j,k) = SrcTrmFld->val(i,j,k,e->id);
			}}}
			ary = s3d;
		} else if (typ == MEMBEXTERM) {
			for (int i = 0; i < I; i++) {
			for (int j = 0; j < J; j++) {
			for (int k = 0; k < K; k++) {
				s3d->val(i,j,k) = MembExQ->val(i,j,k,e->id);
			}}}
			ary = s3d;
		} else if(typ == FACERATE_MX){ // 2011.06.07 for output opening fraction
			for(int i = 0; i < I; i++){
			for(int j = 0; j < J; j++){
			for(int k = 0; k < K; k++){
				s3d->val(i,j,k) = MatAreaX->val(i,j,k,e->id);
			}}}
			ary = s3d;
        } else if(typ == FACERATE_PX){ // 2011.06.07 for output opening fraction
			for(int i = 0; i < I; i++){
			for(int j = 0; j < J; j++){
			for(int k = 0; k < K; k++){
				s3d->val(i,j,k) = MatAreaX->val(i+1,j,k,e->id);
			}}}
			ary = s3d;
        } else if(typ == FACERATE_MY){ // 2011.06.07 for output opening fraction
			for(int i = 0; i < I; i++){
			for(int j = 0; j < J; j++){
			for(int k = 0; k < K; k++){
				s3d->val(i,j,k) = MatAreaY->val(i,j,k,e->id);
			}}}
			ary = s3d;
        } else if(typ == FACERATE_PY){ // 2011.06.07 for output opening fraction
			for(int i = 0; i < I; i++){
			for(int j = 0; j < J; j++){
			for(int k = 0; k < K; k++){
				s3d->val(i,j,k) = MatAreaY->val(i,j+1,k,e->id);
			}}}
			ary = s3d;
        } else if(typ == FACERATE_MZ){ // 2011.06.07 for output opening fraction
			for(int i = 0; i < I; i++){
			for(int j = 0; j < J; j++){
			for(int k = 0; k < K; k++){
				s3d->val(i,j,k) = MatAreaZ->val(i,j,k,e->id);
			}}}
			ary = s3d;
        } else if(typ == FACERATE_PZ){ // 2011.06.07 for output opening fraction
			for(int i = 0; i < I; i++){
			for(int j = 0; j < J; j++){
			for(int k = 0; k < K; k++){
				s3d->val(i,j,k) = MatAreaZ->val(i,j,k+1,e->id);
			}}}
			ary = s3d;
        }

		// is check point now?
		const bool chkpoint = (chkpoint_interval)?
								(step % chkpoint_interval == 0):
								false;
		// const bool force = chkpoint; // forced output?
		const bool force = (output_force)? true:chkpoint;

		const bool res = WriteFile(dSet, step, time, cfg.grpIO.gid, force);
		DB2(MESSAGE(UtilStrFormat("Output Field attr='%s'",attr.c_str())));
		TRY(res, UtilStrFormat("to output \"%s\" field data was faild",
					of->first.c_str()));
	}

	if(cfg.time_integ == ADAMS_BASHFORTH){
		OutputFieldRst();
	}

	return 1;
}

int SklSolverFoundation::OutputFieldRst(const bool output_force)
{
	// Rst data OutputField

	const unsigned int I = cfg.Pdiv.size[0];
	const unsigned int J = cfg.Pdiv.size[1];
	const unsigned int K = cfg.Pdiv.size[2];

	const unsigned int step = SklGetTotalStep();
	const SKL_REAL     time = SklGetTotalTime();

	// Rhs data
	for (OutDataMapT::iterator
	 		of  = OutRstDataMap.begin();
			of != OutRstDataMap.end();
			of++) {

		const size_t   size  = S3D_SIZE(I,J,K,0);
		const string&  attr  = of->first;
		SklVoxDataSet* dSet  = of->second;
		SklArayBase*   ary   = NULL;
		S3D_REAL*      s3d   = dynamic_cast<S3D_REAL*>(dSet->GetData());
		const TypeEnum typ   = cfg.TypMap.find(attr)->second.first;
		const Element* e     = cfg.TypMap.find(attr)->second.second;

		for(int i=0; i<I; i++){
		for(int j=0; j<J; j++){
		for(int k=0; k<K; k++){
			s3d->val(i,j,k) = Rhs->val(i,j,k,e->id);
		}}}
		ary = s3d;

		// is check point now?
		const bool chkpoint = (chkpoint_interval)?
								(step % chkpoint_interval == 0):
								false;
		// const bool force = chkpoint; // forced output?
		const bool force = (output_force)? true:chkpoint;

		const bool res = WriteFile(dSet, step, time, cfg.grpIO.gid, force);
		DB2(MESSAGE(UtilStrFormat("Output Field attr='%s'",attr.c_str())));
		TRY(res, UtilStrFormat("to output \"%s\" field data was faild",
					of->first.c_str()));
	}

	return 1;
}

int SklSolverFoundation::OutputProbe()
{
	////////////////////////////////////////////////////////////
	// output probes

	if( m_ProbeListSize == 0 ) { return 1; }

	const size_t I = cfg.Pdiv.size[0];
	const size_t J = cfg.Pdiv.size[1];
	const size_t K = cfg.Pdiv.size[2];

	int f = (cfg.grpAll.nid == (cfg.SpecialIONode?1:0))? 1: 0;

	MEASURE_WAIT_TIME(clock_wait, para_cmp->Barrier(cfg.grpP.gid))  // 2011.05.11
	BEGIN_CLOCK(comm)
#ifdef SPHERE175
	para_cmp->Broadcast(&f,1,SKL_ARRAY_DTYPE_INT,0,cfg.grpP.gid);	//2009.12.07
#else
	ParaMngr.Broadcast(&f,1,SKL_ARRAY_DTYPE_INT,0,cfg.grpP.gid);
#endif
	END_CLOCK(clock_comm, comm)

	if (!f) { return 0; }

#ifdef _MEASURE_WAIT  // 2011.05.11
	BEGIN_CLOCK(wlap)
#endif
#ifdef SPHERE175
	para_cmp->Barrier(cfg.grpP.gid);	//2009.12.07
#else
	ParaMngr.Barrier(cfg.grpP.gid);
#endif
#ifdef _MEASURE_WAIT  // 2011.05.11
	END_CLOCK(clock_wait, wlap)
#endif	

	BEGIN_CLOCK(comm2)
	for (ProbeListT::iterator
			prb = m_ProbeList.begin();
	 		prb != m_ProbeList.end();
			prb++) {

		const size_t i = prb->i;
		const size_t j = prb->j;
		const size_t k = prb->k;

		// is check point?
		bool chkpoint = (chkpoint_interval)?
							(SklGetTotalStep() % chkpoint_interval == 0):
							false;

		if ((SklGetTotalStep() % prb->interval) && !chkpoint) { continue; }

		if (cfg.grpP.nid == prb->nID) { // probed cell node

			stringstream log;
			const char sep = ',';
			const int  sp  = 0;
			log << setw(sp) << SklGetTotalStep() << sep
				<< setw(sp) << SklGetTotalTime() << sep
				<< setw(sp) << prb->x  << sep
				<< setw(sp) << prb->y  << sep
				<< setw(sp) << prb->z  << sep
				<< setw(sp) << prb->_i << sep
				<< setw(sp) << prb->_j << sep
				<< setw(sp) << prb->_k << sep;

			for (vector<string>::iterator
					obj = prb->obj.begin();
					obj != prb->obj.end();
					obj++) {

				SKL_REAL value = SKL_REAL(0.0);
				const TypeEnum typ = cfg.TypMap.find(*obj)->second.first;
				const Element* e   = cfg.TypMap.find(*obj)->second.second;

				value =
                    (typ == MATERIAL)    ? MatVfFld->val(i,j,k,e->id):
                    (typ == EQUATION)    ? EqnP0Fld->val(i,j,k,e->id):
                    (typ == FACERATE_MX) ? MatAreaX->val(i,j,k,e->id):
                    (typ == FACERATE_PX) ? MatAreaX->val(i+1,j,k,e->id):
                    (typ == FACERATE_MY) ? MatAreaY->val(i,j,k,e->id):
                    (typ == FACERATE_PY) ? MatAreaY->val(i,j+1,k,e->id):
                    (typ == FACERATE_MZ) ? MatAreaZ->val(i,j,k,e->id):
                    (typ == FACERATE_PZ) ? MatAreaZ->val(i,j,k+1,e->id):
                    (typ == MEMBAREA)    ? MembArea->val(i,j,k,e->id):
                    (typ == MEMBNRML_X)  ? MembNrmlX->val(i,j,k,e->id):
                    (typ == MEMBNRML_Y)  ? MembNrmlY->val(i,j,k,e->id):
                    (typ == MEMBNRML_Z)  ? MembNrmlZ->val(i,j,k,e->id):
                    (typ == METABOTERM)  ? MtbTrmFld->val(i,j,k,e->id):
                    (typ == MEMBTERM)    ? MembFlux->val(i,j,k,e->id):
                    (typ == SRCTERM)     ? SrcTrmFld->val(i,j,k,e->id):
                    (typ == MEMBEXTERM)  ? MembExQ->val(i,j,k,e->id):
                    SKL_REAL(0.0);
                
				log << setw(0) << value << sep;
            }
            
			if (cfg.grpP.nid != 0) {
				// send probe log message to writer node whitch nID is 0
				const string data = log.str();
				size_t s = data.length() + 1;
 				char *buf = new char[s+1];
				strcpy(buf, data.c_str());
#ifdef SPHERE175
				para_cmp->Send(&s, sizeof(size_t),SKL_ARRAY_DTYPE_CHAR,0,cfg.grpP.gid);
				para_cmp->Send(buf,s,SKL_ARRAY_DTYPE_CHAR,0,cfg.grpP.gid);
#else
				ParaMngr.Send(&s, sizeof(size_t),SKL_ARRAY_DTYPE_CHAR,0,cfg.grpP.gid);
				ParaMngr.Send(buf,s,SKL_ARRAY_DTYPE_CHAR,0,cfg.grpP.gid);
#endif
				delete[] buf;

			} else {
				// direct write on nID = 0.
				ofstream of(prb->fname.c_str(), std::ios::out | std::ios::app);
                of << log.str() << endl;
				of.close();
			}
            
		} else if (cfg.grpP.nid == 0) {
			// writer node (ID = 0)
			size_t s;
			char* data = NULL;
            
#ifdef SPHERE175
			para_cmp->Recv(&s,sizeof(size_t),SKL_ARRAY_DTYPE_CHAR,prb->nID,cfg.grpP.gid);	//2009.12.07
#else
			ParaMngr.Recv(&s,sizeof(size_t),SKL_ARRAY_DTYPE_CHAR,prb->nID,cfg.grpP.gid);
#endif

			data = new char[s + 1];

#ifdef SPHERE175
			para_cmp->Recv(data,s,SKL_ARRAY_DTYPE_CHAR,prb->nID,cfg.grpP.gid);	//2009.12.07
#else
			ParaMngr.Recv(data,s,SKL_ARRAY_DTYPE_CHAR,prb->nID,cfg.grpP.gid);
#endif
			ofstream of(prb->fname.c_str(), std::ios::out | std::ios::app);
			of << data << endl;
			of.close();
			delete[] data;
		}
	}
	END_CLOCK(clock_comm, comm2)

	return 1;
}

int SklSolverFoundation::AllocateCompVariable()
{
#ifndef _NO_MESSAGE
	MESSAGE("Memory Allocation ...");
#else
	MESSAGE("Memory Allocation ...",0);
#endif //_NO_MESSAGE

	const vxsize_t PI   = cfg.Pdiv.size[0];
	const vxsize_t PJ   = cfg.Pdiv.size[1];
	const vxsize_t PK   = cfg.Pdiv.size[2];

	const vxsize_t Psz[3]     = { PI, PJ, PK };
	const vxsize_t Psz_mat[4] = { PI, PJ, PK, cfg.MatCount };
	const vxsize_t Psz_mol[4] = { PI, PJ, PK, cfg.MolCount };
	const vxsize_t Psz_eqn[4] = { PI, PJ, PK, cfg.EqnCount };
	const vxsize_t Psz_mem[4] = { PI, PJ, PK, cfg.MembareaCount };
	const vxsize_t Psz_SSF[4] = { PI, PJ, PK, cfg.SSCount };

	const vxsize_t vc = VC;
	const int    Pgid = cfg.grpP.gid;
	const int    Egid = cfg.grpE.gid;

	if (cfg.grpP.nid < 0) { return 0; }

	const SklClassType& S4D  = SKL_CLASS_SCALAR4D;
	const SklClassType& S3D  = SKL_CLASS_SCALAR3D;
	const SklDataType&  REAL = SKL_ARRAY_DTYPE_REAL;
	const SklDataType&  INT  = SKL_ARRAY_DTYPE_INT;

	EqnP0Fld = UtilAllocS4DReal("EqnP0Fld",Psz_eqn,vc,Pgid);
	EqnP1Fld = UtilAllocS4DReal("EqnP1Fld",Psz_eqn,vc,Pgid);

	MatVfFld = UtilAllocS4DReal("MatVfFld",Psz_mat,vc,Pgid);
	MatVfInv = UtilAllocS4DReal("MatVfInv",Psz_mat,vc,Pgid);
	MatAreaX = UtilAllocS4DReal("MatAreaX",Psz_mat,vc,Pgid);
	MatAreaY = UtilAllocS4DReal("MatAreaY",Psz_mat,vc,Pgid);
	MatAreaZ = UtilAllocS4DReal("MatAreaZ",Psz_mat,vc,Pgid);

	DifCoefX  = UtilAllocS4DReal("DifCoefX",Psz_eqn,vc,Pgid);
	DifCoefY  = UtilAllocS4DReal("DifCoefY",Psz_eqn,vc,Pgid);
	DifCoefZ  = UtilAllocS4DReal("DifCoefZ",Psz_eqn,vc,Pgid);

	MembArea  = UtilAllocS4DReal("MembArea", Psz_mat,vc,Pgid);
	MembNrmlX = UtilAllocS4DReal("MembNrmlX",Psz_mat,vc,Pgid);
	MembNrmlY = UtilAllocS4DReal("MembNrmlY",Psz_mat,vc,Pgid);
	MembNrmlZ = UtilAllocS4DReal("MembNrmlZ",Psz_mat,vc,Pgid);

	if( cfg.MembareaCount == 0 ){
		MembAreaC = NULL;
		MembAreaX = NULL;
		MembAreaY = NULL;
		MembAreaZ = NULL;
	}
	else{
		MembAreaC = UtilAllocS4DReal("MembAreaC",Psz_mem,vc,Pgid);
		MembAreaX = UtilAllocS4DReal("MembAreaX",Psz_mem,vc,Pgid);
		MembAreaY = UtilAllocS4DReal("MembAreaY",Psz_mem,vc,Pgid);
		MembAreaZ = UtilAllocS4DReal("MembAreaZ",Psz_mem,vc,Pgid);
	}

	MtbTrmFld = UtilAllocS4DReal("MtbTerm",   Psz_eqn,vc,Pgid);
	DifFluxX  = UtilAllocS4DReal("DifFluxX",  Psz_eqn,vc,Pgid);
	DifFluxY  = UtilAllocS4DReal("DifFluxY",  Psz_eqn,vc,Pgid);
	DifFluxZ  = UtilAllocS4DReal("DifFluxZ",  Psz_eqn,vc,Pgid);
	MembTrmFld= UtilAllocS4DReal("MembTrmFld",Psz_eqn,vc,Pgid);
	MembFlux  = UtilAllocS4DReal("MembFlux",  Psz_eqn,vc,Pgid);
	MembExQ   = UtilAllocS4DReal("MembExQ",   Psz_eqn,vc,Pgid);
	SrcTrmFld = UtilAllocS4DReal("SrcTerm",   Psz_eqn,vc,Pgid);

	ElectricV = UtilAllocS4DReal("ElectricV", Psz_eqn,vc,Pgid);

	Rhs       = UtilAllocS4DReal("Rhs", Psz_eqn,vc,Pgid);
	Rhso      = UtilAllocS4DReal("Rhso", Psz_eqn,vc,Pgid);

	// steady state : false
	if( cfg.SSCount == 0 ){
		SSFld = NULL;
	}
	else{
		SSFld = UtilAllocS4DReal("SSFld", Psz_SSF,0,Pgid);
	}

	OutBndId_X = new S4D_INT(2,PJ,PK,cfg.EqnCount,vc);
	OutBndId_Y = new S4D_INT(PI,2,PK,cfg.EqnCount,vc);
	OutBndId_Z = new S4D_INT(PI,PJ,2,cfg.EqnCount,vc);

	////////////////////////////////////////////////////////////////////////
	
	const vxsize_t EI = cfg.Ediv.size[0];
	const vxsize_t EJ = cfg.Ediv.size[1];
	const vxsize_t EK = cfg.Ediv.size[2];

	const vxsize_t Esz_mat[4] = { EI, EJ, EK, cfg.MatCount };
	const vxsize_t Esz_mol[4] = { EI, EJ, EK, cfg.MolCount };
	const vxsize_t Esz_eqn[4] = { EI, EJ, EK, cfg.EqnCount };
	const vxsize_t Esz_mem[4] = { EI, EJ, EK, cfg.MembareaCount };

	P0e  = UtilAllocS4DReal("P0e",  Esz_eqn, vc, Egid);
	Ee   = UtilAllocS4DReal("Ee",   Esz_eqn, vc, Egid);
	Vfe  = UtilAllocS4DReal("Vfe",  Esz_mat, vc, Egid);
	Vfie = UtilAllocS4DReal("Vfie", Esz_mat, vc, Egid);
	Sme  = UtilAllocS4DReal("Sme",  Esz_mat, vc, Egid);
	Nmxe = UtilAllocS4DReal("Nmxe", Esz_mat, vc, Egid);
	Nmye = UtilAllocS4DReal("Nmye", Esz_mat, vc, Egid);
	Nmze = UtilAllocS4DReal("Nmze", Esz_mat, vc, Egid);

	// P0e  = UtilAllocS4DReal("P0e",  Esz_eqn, 0, Egid);
	// Ee   = UtilAllocS4DReal("Ee",   Esz_eqn, 0, Egid);
	// Vfe  = UtilAllocS4DReal("Vfe",  Esz_mat, 0, Egid);
	// Vfie = UtilAllocS4DReal("Vfie", Esz_mat, 0, Egid);
	// Sme  = UtilAllocS4DReal("Sme",  Esz_mat, 0, Egid);
	// Nmxe = UtilAllocS4DReal("Nmxe", Esz_mat, 0, Egid);
	// Nmye = UtilAllocS4DReal("Nmye", Esz_mat, 0, Egid);
	// Nmze = UtilAllocS4DReal("Nmze", Esz_mat, 0, Egid);

	if( cfg.MembareaCount == 0 ){
		SmCe = NULL;
		SmXe = NULL;
		SmYe = NULL;
		SmZe = NULL;
	}
	else{
		SmCe = UtilAllocS4DReal("SmCe", Esz_mem, vc, Egid);
		SmXe = UtilAllocS4DReal("SmXe", Esz_mem, vc, Egid);
		SmYe = UtilAllocS4DReal("SmYe", Esz_mem, vc, Egid);
		SmZe = UtilAllocS4DReal("SmZe", Esz_mem, vc, Egid);
	}

	////////////////////////////////////////////////////////////////////////

	return 1;
}

int SklSolverFoundation::InitializeCompVariable()
{
#ifndef _NO_MESSAGE
	MESSAGE("Memory Initialization ...");
#else
	MESSAGE("Memory Initialization ...",0);
#endif //_NO_MESSAGE

	const size_t PI = cfg.Pdiv.size[0];
	const size_t PJ = cfg.Pdiv.size[1];
	const size_t PK = cfg.Pdiv.size[2];

	const size_t vc = VC;

	//
	// initialize all fields
	//
	UtilClearRealArray(EqnP0Fld, SKL_REAL(0.0));
	UtilClearRealArray(EqnP1Fld, SKL_REAL(0.0));

	UtilClearRealArray(MatVfFld, SKL_REAL(0.0));
	UtilClearRealArray(MatVfInv, SKL_REAL(0.0));
	UtilClearRealArray(MatAreaX, SKL_REAL(0.0));
	UtilClearRealArray(MatAreaY, SKL_REAL(0.0));
	UtilClearRealArray(MatAreaZ, SKL_REAL(0.0));

	UtilClearRealArray(MtbTrmFld, SKL_REAL(0.0));
	UtilClearRealArray(DifFluxX,  SKL_REAL(0.0));
	UtilClearRealArray(DifFluxY,  SKL_REAL(0.0));
	UtilClearRealArray(DifFluxZ,  SKL_REAL(0.0));
	UtilClearRealArray(MembTrmFld,SKL_REAL(0.0));
	UtilClearRealArray(MembFlux,  SKL_REAL(0.0));
	UtilClearRealArray(MembExQ,   SKL_REAL(0.0));
	UtilClearRealArray(SrcTrmFld, SKL_REAL(0.0));

	UtilClearRealArray(Rhs, SKL_REAL(0.0));
	UtilClearRealArray(Rhso, SKL_REAL(0.0));

	UtilClearRealArray(SSFld, SKL_REAL(0.0));

#ifndef _NO_MESSAGE
	MESSAGE("Set Const Diffusion Coef (Molecular dependent)...");
#else
	MESSAGE("Set Const Diffusion Coef (Molecular dependent)...",0);
#endif //_NO_MESSAGE
	for (int ml = 0; ml < cfg.MolCount; ml++) {
		const Molecular* g = cfg.MolTbl[ml];
#ifndef _NO_MESSAGE
		MESSAGE(UtilStrFormat("++ %s coef=%f", g->name.c_str(), g->ccoef));
#else
		MESSAGE(UtilStrFormat("++ %s coef=%f", g->name.c_str(), g->ccoef),0);
#endif //_NO_MESSAGE

		for (int mt = 0; mt < cfg.MatCount; mt++) {
			const Material*  m = cfg.MatTbl[mt];
			const Equation*  e = UtilGetEquationId(g,m);

			if (!e) { continue; }

			const size_t ei = e->id;

			for (int i = 0; i < PI; i++) {
			for (int j = 0; j < PJ; j++) {
			for (int k = 0; k < PK; k++) {
				DifCoefX->val(i,j,k,ei) = g->ccoef;
				DifCoefY->val(i,j,k,ei) = g->ccoef;
				DifCoefZ->val(i,j,k,ei) = g->ccoef;
			}}}
		}
	}

#ifndef _NO_MESSAGE
	MESSAGE("Set Const Diffusion Coef (Molecular*Material dependent)...");
#else
	MESSAGE("Set Const Diffusion Coef (Molecular*Material dependent)...",0);
#endif //_NO_MESSAGE
	for (int ml = 0; ml < cfg.MolCount; ml++) {
		const Molecular* g = cfg.MolTbl[ml];

		for (int mt = 0; mt < cfg.MatCount; mt++) {
			const Material*  m = cfg.MatTbl[mt];
			const Equation*  e = UtilGetEquationId(g,m);

			if (!e) { continue; }

			const size_t ei = e->id;
			const Int2RealMap::const_iterator it = g->ccoefm.find(m->id);
			if (it == g->ccoefm.end()) { continue; }

			const SKL_REAL c = it->second;

#ifndef _NO_MESSAGE
			MESSAGE(UtilStrFormat("++ %s coef=%f", e->name.c_str(), c));
#else
			MESSAGE(UtilStrFormat("++ %s coef=%f", e->name.c_str(), c),0);
#endif //_NO_MESSAGE

			for (int i = 0; i < PI; i++) {
			for (int j = 0; j < PJ; j++) {
			for (int k = 0; k < PK; k++) {
				DifCoefX->val(i,j,k,ei) = c;
				DifCoefY->val(i,j,k,ei) = c;
				DifCoefZ->val(i,j,k,ei) = c;
			}}}
		}
	}

#ifndef _NO_MESSAGE
	MESSAGE("Set Electric Potential ...");
#else
	MESSAGE("Set Electric Potential ...",0);
#endif //_NO_MESSAGE
	for (int m = 0; m < cfg.MatCount; m++) {
		const Material* mt = cfg.MatTbl[m];
		const CfgParam* ev = mt->xml_elem->GetParamFirst("Ev");

		SKL_REAL e;
		e = (ev && ev->GetData(&e))? e: SKL_REAL(0.0);
		for (int i = 0; i < PI; i++) {
		for (int j = 0; j < PJ; j++) {
		for (int k = 0; k < PK; k++) {
			ElectricV->val(i,j,k,m) = e;
		}}}
	}

	// Boundary -- default BCID = 0
	const int BC = 0;
	UtilClearIntArray(OutBndId_X, BC);
	UtilClearIntArray(OutBndId_Y, BC);
	UtilClearIntArray(OutBndId_Z, BC);
	
	/////////////////////////////////////////////////////////////////////////

	UtilClearRealArray(P0e,  SKL_REAL(0.0));
	UtilClearRealArray(Ee,   SKL_REAL(0.0));
	UtilClearRealArray(Vfe,  SKL_REAL(0.0));
	UtilClearRealArray(Vfie, SKL_REAL(0.0));

	return 1;
}

int SklSolverFoundation::RegistFields()
{
	for (int eid = 0; eid < cfg.EqnCount; eid++) {
		const char* name = cfg.EqnTbl[eid]->name.c_str();

		UtilRegArrayS4DReal(UtilStrFormat("%s",name),     EqnP0Fld,  eid);
		UtilRegArrayS4DReal(UtilStrFormat("SRC_%s",name), SrcTrmFld, eid);
	}

	for (int mid = 0; mid < cfg.MatCount; mid++) {

		const char* name = cfg.MatTbl[mid]->name.c_str();

		UtilRegArrayS4DReal(UtilStrFormat("Vf_%s",name),  MatVfFld,  mid);
		UtilRegArrayS4DReal(UtilStrFormat("Vfi_%s",name), MatVfInv,  mid);
		UtilRegArrayS4DReal(UtilStrFormat("Sm_%s",name),  MembArea,  mid);
		UtilRegArrayS4DReal(UtilStrFormat("Nx_%s",name),  MembNrmlX, mid);
		UtilRegArrayS4DReal(UtilStrFormat("Ny_%s",name),  MembNrmlX, mid);
		UtilRegArrayS4DReal(UtilStrFormat("Nz_%s",name),  MembNrmlX, mid);
	}

	return 1;
}

int SklSolverFoundation::SetMolConc()
{
#ifndef _NO_MESSAGE
	MESSAGE("Set Initial Equation Value from Computing Condition XML ...");
#else
	MESSAGE("Set Initial Equation Value from Computing Condition XML ...",0);
#endif //_NO_MESSAGE

	const size_t I = cfg.Pdiv.size[0];
	const size_t J = cfg.Pdiv.size[1];
	const size_t K = cfg.Pdiv.size[2];

	const SKL_REAL dx  = cfg.pitch[0];
	const SKL_REAL dy  = cfg.pitch[1];
	const SKL_REAL dz  = cfg.pitch[2];
	const SKL_REAL Vol = dx * dy * dz;

	//////////////////////////////////////////////////////////////////////
	// initial value with molconc
	for (size_t e = 0; e < cfg.EqnCount; e++) {
		const size_t ml = cfg.EqnTbl[e]->mol->id;
		const size_t mt = cfg.EqnTbl[e]->mat->id;
		const string& mname = cfg.EqnTbl[e]->mat->name;

		RealMap::const_iterator mc = cfg.MolTbl[ml]->molconc.find(mname);
		if (mc != cfg.MolTbl[ml]->molconc.end()) {
			const SKL_REAL conc = mc->second;
			for (int i = 0; i < I; i++) {
			for (int j = 0; j < J; j++) {
			for (int k = 0; k < K; k++) {
				const SKL_REAL& mol    = conc * MatVfFld->val(i,j,k,mt);
				EqnP0Fld->val(i,j,k,e) = mol  * MatVfInv->val(i,j,k,mt);
			}}}
		}
	}

	//////////////////////////////////////////////////////////////////////
	// sum Vfm
	SKL_REAL* tVfm = new SKL_REAL[cfg.MatCount];
	for (size_t m = 0; m < cfg.MatCount; m++) {
		tVfm[m] = SKL_REAL(0.0);
		for (int i = 0; i < I; i++) {
		for (int j = 0; j < J; j++) {
		for (int k = 0; k < K; k++) {
			tVfm[m] += MatVfFld->val(i,j,k,m);
		}}}
		SKL_REAL aVfm;

#ifdef SPHERE175
		para_cmp->Allreduce(&tVfm[m],&aVfm,1,SKL_ARRAY_DTYPE_REAL,
							SKL_SUM,cfg.grpE.gid);	//2009.12.07
#else
		ParaMngr.Allreduce(&tVfm[m],&aVfm,1,SKL_ARRAY_DTYPE_REAL,
							SKL_SUM,cfg.grpE.gid);
#endif

		tVfm[m] = aVfm;
	}
	//////////////////////////////////////////////////////////////////////
	// initial value with num
	for (size_t e = 0; e < cfg.EqnCount; e++) {
		const size_t   ml    = cfg.EqnTbl[e]->mol->id;
		const size_t   mt    = cfg.EqnTbl[e]->mat->id;
		const string&  mname = cfg.MatTbl[mt]->name;

		const SKL_REAL aVfm  = tVfm[mt];
		const SKL_REAL aVfmi = aVfm > 0.0? SKL_REAL(1.0)/aVfm: SKL_REAL(0.0);
		const SKL_REAL VolKnaNA = Vol * cfg.Kna * cfg.NA;

		RealMap::const_iterator mc = cfg.MolTbl[ml]->molnum.find(mname);
		if (mc != cfg.MolTbl[ml]->molnum.end()) {
			const SKL_REAL& num = mc->second;
			for (int i = 0; i < I; i++) {
			for (int j = 0; j < J; j++) {
			for (int k = 0; k < K; k++) {
				EqnP0Fld->val(i,j,k,e)  = num * aVfmi / VolKnaNA;
			}}}
		}
	}

	delete[] tVfm;

	return 1;
}

bool SklSolverFoundation::chkCfg()
{
#ifndef _NO_MESSAGE
	MESSAGE("**** Configuration of Live Cell Simulator Foundation ****");
#else
	MESSAGE("**** Configuration of Live Cell Simulator Foundation ****",0);
#endif //_NO_MESSAGE

	//////////////////////////////////////////////////////////////////////////
	// basic physical unit
	MESSAGE(UtilStrFormat("length unit  : %1.6e [m]",    cfg.Ls), 0);
	MESSAGE(UtilStrFormat("time unit    : %1.6e [s]",    cfg.Ts), 0);
	MESSAGE(UtilStrFormat("avogadro num : %1.6e [/mol]", cfg.NA), 0);
	if (!cfg.fixdt) {
		MESSAGE(UtilStrFormat("CourantNumber: %f", cfg.CourantNumber), 0);
	}
	MESSAGE(UtilStrFormat("deltaT: %1.6e [s]", cfg.dt * cfg.Ts), 0);

	//////////////////////////////////////////////////////////////////////////
	// Domain Info
	stringstream domain;
	domain << setw(80) << setfill('=') << "" << setfill(' ') << endl;
	domain << setw(18) << left << "Domain Size[m]" << "| ";
	domain << setw(13) << right << showpoint << cfg.size[0] * cfg.pitch[0] * cfg.Ls
			<< setw(13) << right << showpoint << cfg.size[1] * cfg.pitch[1] * cfg.Ls
			<< setw(13) << right << showpoint << cfg.size[2] * cfg.pitch[2] * cfg.Ls
			<< endl;
	domain << setw(18) << left << "Voxel Pitch[m]" << "| ";
	domain	<< setw(13) << right << showpoint << cfg.pitch[0] * cfg.Ls
			<< setw(13) << right << showpoint << cfg.pitch[1] * cfg.Ls
			<< setw(13) << right << showpoint << cfg.pitch[2] * cfg.Ls
			<< endl;
	
	domain << setw(18) << left << "Voxel Size" << "| ";
	domain	<< setw(13) << right << cfg.size[0]
			<< setw(13) << right << cfg.size[1]
			<< setw(13) << right << cfg.size[2]
			<< endl;
	domain << setw(80) << setfill('-') << "" << setfill(' ') << endl;
	domain << setw(18) << left << "Volume per cell" << "| ";
	domain	<< setw(16) << right << cfg.cvol * cfg.Ls * cfg.Ls * cfg.Ls << "[m^3]"
			<< endl;
	domain << setw(80) << setfill('=') << "" << setfill(' ') << endl;

	//////////////////////////////////////////////////////////////////////////
	// Parallelism
	stringstream para;
	para << "Parallel Computing is " << boolalpha << cfg.Parallel << endl;
	if (cfg.Parallel) {
		para << "Parallelism is " << cfg.grpAll.num << endl;
#ifndef _NO_MESSAGE
		para << "Voxel Divison Table" << endl;
		para << setw(80) << setfill('=') << "" << setfill(' ') << endl;
		para << setw(3)  << left << "";
		para << setw(5)  << left << "nID";
		para << setw(3)  << left << "";
		para << setw(16) << left  << "Head Index";
		para << setw(5)  << right << "";
		para << setw(16) << left  << "Tail Index";
		para << endl;
		para << setw(80) << setfill('-') << "" << setfill(' ') << endl;
		for (unsigned int n = 0; n < cfg.grpAll.num; n++) {
			unsigned int rect[12];
			const int Pnid = cfg.grpP.nid;
			const int Pgid = cfg.grpP.gid;
			const int Enid = cfg.grpE.nid;
			const int Egid = cfg.grpE.gid;
			for (int i = 0; i < 3; i++) {
				rect[i]     = cfg.Pdiv.head[i];
				rect[i + 3] = cfg.Pdiv.tail[i];
				rect[i + 6] = cfg.Ediv.head[i];
				rect[i + 9] = cfg.Ediv.tail[i];
			}

#ifdef SPHERE175			
			para_cmp->Broadcast(rect,12,SKL_ARRAY_DTYPE_UINT,n,cfg.grpAll.gid);	//2009.12.07
#else
			ParaMngr.Broadcast(rect,12,SKL_ARRAY_DTYPE_UINT,n,cfg.grpAll.gid);
#endif

			para << setw(3) << left  << "*";
			para << setw(5) << right << n;
			para << setw(3) << left  << "";
			para << setw(1) << left  << '(';
			para << setw(4) << right << rect[0] << ',';
			para << setw(4) << right << rect[1] << ',';
			para << setw(4) << right << rect[2] << ')';
			para << setw(5) << right << " => ";
			para << setw(1) << left  << '(';
			para << setw(4) << right << rect[3] << ',';
			para << setw(4) << right << rect[4] << ',';
			para << setw(4) << right << rect[5] << ')';
			para << endl;
			para << setw(3) << left  << "*";
			para << setw(8) << right << "";
			para << setw(1) << left  << '(';
			para << setw(4) << right << rect[ 6] << ',';
			para << setw(4) << right << rect[ 7] << ',';
			para << setw(4) << right << rect[ 8] << ')';
			para << setw(5) << right << " => ";
			para << setw(1) << left  << '(';
			para << setw(4) << right << rect[ 9] << ',';
			para << setw(4) << right << rect[10] << ',';
			para << setw(4) << right << rect[11] << ')';
			para << endl;
		}
		para << setw(80) << setfill('=') << "" << setfill(' ') << endl;
#endif //_NO_MESSAGE
	}

	//////////////////////////////////////////////////////////////////////////
	// output MatMap
	stringstream mattbl;
	mattbl << "Material Table" << endl;
	mattbl << setw(80) << setfill('=') << "" << setfill(' ') << endl;
	mattbl << setfill(' ')
		<< setw(3)  << left << ' '
		<< setw(10) << left << "name" << "| "
		<< setw(25) << left << "EML"  << "| "
		<< setw(6)  << left << "fus"  << "| "
		<< setw(10) << left << "fut"  << "| "
		<< setw(2)  << left << "system" << endl;
	mattbl << setw(80) << setfill('-') << "" << setfill(' ') << endl;
	for (size_t matid = 0; matid < cfg.MatCount; matid++) {
		const Material* mat = cfg.MatTbl[matid];
		mattbl
			<< setw(3)  << left << '*'
			<< setw(10) << mat->name << "| "
			<< setw(25) << (mat->body_mtb.eml? mat->body_mtb.eml:"") << "| " 
			<< setw(6)  << boolalpha << mat->fus << "| " 
			<< setw(10) << showpoint << mat->fut << "| ";
		mattbl << '{';
			for (set<string>::const_iterator
				i = mat->body_mtb.sys.begin();
				i != mat->body_mtb.sys.end(); i++) {
				mattbl << *i << ',';
			}
		mattbl << '}' << endl;
	}
	mattbl << setw(80) << setfill('=') << "" << setfill(' ') << endl;

	//////////////////////////////////////////////////////////////////////////
	// output MolMap
	stringstream moltbl;
	moltbl << "Molecular Table" << endl;
	moltbl << setw(80) << setfill('=') << "" << setfill(' ') << endl;
	moltbl << setfill(' ')
		<< setw(3)  << left << ' '
		<< setw(10) << left << " name"  << "| "
		<< setw(35) << left << "fullid" << "| "
		<< setw(25) << left << "diffusion" << endl;
	moltbl << setw(80) << setfill('-') << "" << setfill(' ') << endl;
	for (size_t molid = 0; molid < cfg.MolCount; molid++) {
		const Molecular* mol = cfg.MolTbl[molid];
		moltbl
			<< setw(3)  << left << '*'
			<< setw(10) << mol->name   << "| "
			<< setw(35) << (mol->fullid? mol->fullid: "") << "| "
			<< setw(25) << mol->ccoef << endl;
	}
	moltbl << setw(80) << setfill('=') << "" << setfill(' ') << endl;

	//////////////////////////////////////////////////////////////////////////
	// output Restart
	stringstream restart;
	restart << "Start from "
			<< setw(10) << SklGetBaseStep() << "[step]"
			<< setw(10) << SklGetBaseTime() << "[s]"
			<< endl;

	//////////////////////////////////////////////////////////////////////////
	// output Fieldout Setting
	stringstream fldout;
	fldout << "Field Output Setting" << endl;
	fldout << setw(80) << setfill('=') << "" << setfill(' ') << endl;
	fldout << setfill(' ')
		<< setw(3)  << left << ' '
		<< setw(10) << left << "material"  << "| "
		<< setw(30) << left << "File Base Name" << "| "
		<< endl;
	fldout << setw(80) << setfill('-') << "" << setfill(' ') << endl;
	for (OutDataMapT::iterator od = m_OutDataMap.begin();
			od != m_OutDataMap.end(); od++) {

		fldout
			<< setw(3)  << left << '*'
			<< setw(10) << od->first << "| "
			<< setw(30) << od->second->GetFileName() << "| "
			<< endl;
	}
	fldout << setw(80) << setfill('=') << "" << setfill(' ') << endl;

	//////////////////////////////////////////////////////////////////////////
	// output Probeout Setting
	stringstream prbout;
	prbout << "Probe Output Setting" << endl;
	prbout << setw(80) << setfill('=') << "" << setfill(' ') << endl;
	prbout << setw(40) << left << " cood";
	prbout << setw(20) << left << " filename";
	prbout << setw(10) << left << " interval" << endl;
	for (ProbeListT::iterator prb = m_ProbeList.begin();
			prb != m_ProbeList.end(); prb++) {

		prbout
			<< setw(13) << right << showpoint << prb->x
			<< setw(13) << right << showpoint << prb->y
			<< setw(13) << right << showpoint << prb->z
			<< setw(2)  << "";
		prbout
			<< setw(22) << left << prb->fname
			<< setw(10) << left << prb->interval
			<< endl;

		prbout << " Target materials : ";
		for (vector<string>::iterator obj = prb->obj.begin();
				obj != prb->obj.end(); obj++) {
			prbout <<  (*obj) << ", ";
		}
		prbout << endl;
	}
	prbout << setw(80) << setfill('=') << "" << setfill(' ') << endl;
	
	//////////////////////////////////////////////////////////////////////////
#ifndef _NO_MESSAGE
	MESSAGE(domain.str());
	MESSAGE(para.str());
	MESSAGE(mattbl.str());
	MESSAGE(moltbl.str());
	MESSAGE(restart.str());
	MESSAGE(fldout.str());
	MESSAGE(prbout.str());
#else
	MESSAGE(domain.str(),  0);
	MESSAGE(para.str(),    0);
	MESSAGE(mattbl.str(),  0);
	MESSAGE(moltbl.str(),  0);
	MESSAGE(restart.str(), 0);
	MESSAGE(fldout.str(),  0);
	MESSAGE(prbout.str(),  0);
#endif //_NO_MESSAGE

	return true;
}

int SklSolverFoundation::OutputTimeline(const unsigned int step)
{
	if (step % cfg.moninterval == 0) {
		MESSAGE(UtilStrFormat(
#ifdef _COST_BMT
			"* iter=%6d (total iter=%8d) sim time=%5.4e (calc=%5.4e[SEC])",
			step, SklGetTotalStep(), SklGetTotalTime(), clock_all
#else
			"* iter=%6d (total iter=%8d) sim time=%5.4e",
			step, SklGetTotalStep(), SklGetTotalTime()
#endif
		),0,false);
	}

	return 0;
}

int SklSolverFoundation::CalcEqnVol()
{
	const size_t I = cfg.Pdiv.size[0];
	const size_t J = cfg.Pdiv.size[1];
	const size_t K = cfg.Pdiv.size[2];

	const SKL_REAL Vol = cfg.cvol;

	for (size_t e = 0; e < cfg.EqnCount; e++) {
        SKL_REAL vv  = 0.0;
        SKL_REAL phi = 0.0;
		const size_t mt = cfg.EqnTbl[e]->mat->id;
		for (int i = 0; i < I; i++) {
		for (int j = 0; j < J; j++) {
		for (int k = 0; k < K; k++) {
			const SKL_REAL Vf = MatVfFld->val(i,j,k,mt);

			phi += EqnP0Fld->val(i,j,k,e) * Vf * Vol;
			vv  += Vf;
		}}}
#ifndef _NO_MESSAGE
		MESSAGE(cfg.EqnTbl[e]->name
				+ ",phi," + string_cast(phi)
				+ ",vv,"  + string_cast(vv));
#else
		MESSAGE(cfg.EqnTbl[e]->name
				+ ",phi," + string_cast(phi)
				+ ",vv,"  + string_cast(vv),0);
#endif //_NO_MESSAGE
	}

	return 0;
}

int SklSolverFoundation::InitExtParameterMap(const CfgElem* extParam, const string& path)
{
	int any;

	// sub param
	for (const CfgParam* param = extParam->GetParamFirst();
			param != NULL; param = extParam->GetParamNext(param)) {
		//
		TRY(param->GetName(), "Some Extension Parameter has No Name!!");
		
		ExtParam* p = new ExtParam;	
		p->name = path + param->GetName();
		if (p->extid = param->isSetID()) { p->id = param->GetID(); }
		DataType t = param->GetDataType();
		bool b =
			(t == _INT)?  param->GetData(&(p->value.INT)):
			(t == _REAL)? param->GetData(&(p->value.REAL)):
			(t == _STR)?  param->GetData(&(p->value.STR)):
			false;

		TRY(b,"Extension Parameter is Illigal.");
		p->comment = param->GetComment();

		bool isnew = (cfg.extParams.find(p->name) == cfg.extParams.end());
		cfg.extParams.insert(ExtParamMap::value_type(p->name, p));

		if (isnew) {
#ifndef _NO_MESSAGE
			MESSAGE(string("++ ") + p->name);
#else
			MESSAGE(string("++ ") + p->name,0);
#endif //_NO_MESSAGE
			++any;
		}
	}

	// sub elem
	for (const CfgElem* elem = extParam->GetElemFirst();
			elem != NULL; elem = extParam->GetElemNext(elem)) {
		TRY(elem->GetName(), "Some Extension Parameter Elem has No Name!!");
		const string p = path + elem->GetName() + "@";
		any += InitExtParameterMap(elem, p);
	}

	return any;
}

int SklSolverFoundation::InitSteadyState()
{
	// steady state : false
	if( cfg.SSCount == 0 ) return 1;

	const int dvc = 0;
	const int vc  = VC;
	const int I   = cfg.Pdiv.size[0];
	const int J   = cfg.Pdiv.size[1];
	const int K   = cfg.Pdiv.size[2];

	int mapcnt = 0;
	for( map<int, SteadyState*>::iterator it = cfg.SSmap.begin(); it != cfg.SSmap.end(); it++ ){
		const SteadyState* ss = (*it).second;

/*
MESSAGE(UtilStrFormat("=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*="));
MESSAGE(UtilStrFormat("Equation name : %s", ss->eqn->name.c_str()));
MESSAGE(UtilStrFormat("dst id : %d", mapcnt));
MESSAGE(UtilStrFormat("src id : %d", ss->eqn->id));
*/

		SKL_REAL* SSFLD    = UtilGetS4DN(SSFld, mapcnt);
		const SKL_REAL* P0 = UtilGetS4DN(EqnP0Fld, ss->eqn->id);

		ricscpys3dss_(SSFLD,P0,&I,&J,&K,&dvc,&vc);

		mapcnt++;
	}

	return 1;
}

int SklSolverFoundation::CheckSteadyState(const unsigned int step)
{
	// steady state : false
	if( cfg.SSCount == 0 ) return 1;

	// start step check
	if( step < cfg.SSstartStep ) return 1;

	// check interval
	if( (step - cfg.SSstartStep) % cfg.SSinterval != 0 ) return 1;

	// initialize
	if( step == cfg.SSstartStep ){
		InitSteadyState();
		return 1;
	}

	const int dvc = 0;
	const int vc  = VC;
	const int I   = cfg.Pdiv.size[0];
	const int J   = cfg.Pdiv.size[1];
	const int K   = cfg.Pdiv.size[2];

	int mapcnt = 0;
	for( map<int, SteadyState*>::iterator it = cfg.SSmap.begin(); it != cfg.SSmap.end(); it++ ){
		SteadyState* ss = (*it).second;

		const SKL_REAL* SSFLD = UtilGetS4DN(SSFld, mapcnt);
		const SKL_REAL* P0    = UtilGetS4DN(EqnP0Fld, ss->eqn->id);
		const SKL_REAL* VOF   = UtilGetS4DN(MatVfFld, ss->eqn->mat->id);

		int ssflag  = 1;
		SKL_REAL th = ss->threshold;

		//////////////////////////////////////////////////
		// all voxel check
		if( ss->thType == RATIO ){
			switch( ss->compType ){
				case LE:
					sschkratiole_(&ssflag,&th,SSFLD,P0,VOF,&I,&J,&K,&dvc,&vc);
					break;
				case GT:
					sschkratiogt_(&ssflag,&th,SSFLD,P0,VOF,&I,&J,&K,&dvc,&vc);
					break;
				case GE:
					sschkratioge_(&ssflag,&th,SSFLD,P0,VOF,&I,&J,&K,&dvc,&vc);
					break;
				default:
					sschkratiolt_(&ssflag,&th,SSFLD,P0,VOF,&I,&J,&K,&dvc,&vc);
					break;
			}
		} else {
			switch( ss->compType ){
				case LE:
					sschkabsle_(&ssflag,&th,SSFLD,P0,VOF,&I,&J,&K,&dvc,&vc);
					break;
				case GT:
					sschkabsgt_(&ssflag,&th,SSFLD,P0,VOF,&I,&J,&K,&dvc,&vc);
					break;
				case GE:
					sschkabsge_(&ssflag,&th,SSFLD,P0,VOF,&I,&J,&K,&dvc,&vc);
					break;
				default:
					sschkabslt_(&ssflag,&th,SSFLD,P0,VOF,&I,&J,&K,&dvc,&vc);
					break;
			}
		}

		//////////////////////////////////////////////////
		if( cfg.Parallel ){
			int ibuf = ssflag;
#ifdef SPHERE175
			para_cmp->Allreduce(&ibuf,&ssflag,1,SKL_ARRAY_DTYPE_INT,SKL_PROD,cfg.grpP.gid);
#else
			ParaMngr.Allreduce(&ibuf,&ssflag,1,SKL_ARRAY_DTYPE_INT,SKL_PROD,cfg.grpP.gid);
#endif
		}

		// steady state check
		// ssflag == 0 : false
		// ssflag == 1 : true
		if( ssflag == 1 ){
			ss->SScnt++;
		}
		else{
			ss->SScnt = 0;
		}

		// steady state judge number
		if( (cfg.SSoutSPH) && (ss->SScnt == ss->SSjnum) ){
			// force output field data
			OutputField( true );
		}

		mapcnt++;
	}

	// solver forced termination check
	if( cfg.SSsolvExec == STOP ){
		if( cfg.SSope == OR ){
			for( map<int, SteadyState*>::iterator
				it = cfg.SSmap.begin(); it != cfg.SSmap.end(); it++ ){
				SteadyState* ss = (*it).second;
				// forced termination
				if( ss->SScnt >= ss->SSjnum ){
					// force output field data
					OutputField( true );
					MESSAGE("steady state : solver forced termination.");
					return 0;
				}
			}
		}
		else{
			int ssflag = 1;
			for( map<int, SteadyState*>::iterator
				it = cfg.SSmap.begin(); it != cfg.SSmap.end(); it++ ){
				SteadyState* ss = (*it).second;
				if( ss->SScnt >= ss->SSjnum ){
					ssflag = ssflag * 1;
				}
				else{
					ssflag = ssflag * 0;
				}
			}
			// forced termination
			if( ssflag == 1 ){
				// force output field data
				OutputField( true );
				MESSAGE("steady state : solver forced termination.");
				return 0;
			}
		}
	}


	// copy
	InitSteadyState();

	return 1;
}

int SklSolverFoundation::InitSteadyStateLog()
{
	// parallel : node ID 0 only
	if( cfg.grpP.nid != 0 ) return 1;

	// steady state : false
	if( cfg.SSCount == 0 ) return 1;

	// open log file
	ofstream sslog( cfg.sslogfilename );

	// header
	sslog << "step,";

	// header : eqation name
	for( map<int, SteadyState*>::iterator it = cfg.SSmap.begin(); it != cfg.SSmap.end(); it++ ){
		sslog << (*it).second->eqn->name.c_str() << ",";
	}

	sslog << endl;
	sslog.close();

	return 1;
}

int SklSolverFoundation::OutputSteadyStateLog(const unsigned int step)
{
	// parallel : node ID 0 only
	if( cfg.grpP.nid != 0 ) return 1;

	// steady state : false
	if( cfg.SSCount == 0 ) return 1;

	// start step check
	if( step <= cfg.SSstartStep ) return 1;

	// check interval
	if( (step - cfg.SSstartStep) % cfg.SSinterval != 0 ) return 1;

	//////////////////////////////////////////////////
	// log write
	//
	ofstream sslog(cfg.sslogfilename, std::ios::out | std::ios::app);

	sslog << step << ",";

	for( map<int, SteadyState*>::iterator it = cfg.SSmap.begin(); it != cfg.SSmap.end(); it++ ){
		SteadyState* ss = (*it).second;

		if( ss->SScnt >= ss->SSjnum ){
			sslog << "o,";
		}
		else{
			sslog << "x,";
		}
	}

	sslog << endl;
	sslog.close();

	return 1;
}



bool SklSolverFoundation::chkMemSize()
{
	size_t totalMem = 0;
	size_t advMem = 0;
#ifdef ENABLE_ADVECTION
	advMem = advectionsolv.GetMemory();
#endif
	totalMem = cfg.used_mem + advMem;

	MESSAGE("**** Used Memory Size ****");
	MESSAGE(UtilStrFormat(" RICS total : %d[byte]", totalMem));

	return true;
}

