//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#ifndef _SKL_SOLVER_Foundation_CLASS_H_
#define _SKL_SOLVER_Foundation_CLASS_H_

#include <map>
#include <vector>
#include <fstream>
#include <string>

#include <Skl.h>
#include <SklSolverBase.h>
#include <SklUtil.h>

#include <Foundation/datatype.h>
#include <Foundation/AreaCommLink.h>

#include <LocalSolv/metabolism/MetaboSolv.h>
#include <LocalSolv/diffusion/DiffusSolv.h>
#include <LocalSolv/membrane/MembraneSolv.h>
#ifdef ENABLE_WARP
#include <LocalSolv/warp/WarpSolv.h>
#endif

#ifdef ENABLE_MEMBRANEPOTENTIAL
#include <LocalSolv/potential/PotentialSolv.h>
#endif

#ifdef ENABLE_ADVECTION
#include <LocalSolv/advection/AdvectionSolv.h>
#endif

#include "SklSolverFoundationDefine.h"
#include "FortranFuncFoundation.h"


#define RICS_VERSION   "RICS - RIKEN Cell Simulatior (RICS_Solv_5.4.1)"
// #define RICS_VERSION   "RICS - RIKEN Cell Simulatior (rev.20120507)"
#define RICS_COPYRIGHT "Copyright (c) RIKEN, Japan. All right reserved. 2007-2012"
#define RICS_BUILD     UtilStrFormat("build in %s",__DATE__)

#define EULER 0
#define ADAMS_BASHFORTH 1
#define RUNGE_KUTTA 2

class SklSolverFoundation : public SklSolverBase {
private:
	CfgInfo cfg;
	double clock;
	double clock_all;
	double clock_io;
	double clock_exptim;
	double clock_user;
	double clock_comm;
	double clock_wait;

	// Local Solver
	MetaboSolv   mtbsolv;
	DiffusSolv   difsolv;
	MembraneSolv membsolv;
#ifdef ENABLE_WARP
	WarpSolv     warpsolv;
#endif

#ifdef ENABLE_MEMBRANEPOTENTIAL
        PotentialSolv potentialsolv;
#endif

#ifdef ENABLE_ADVECTION
	AdvectionSolv advectionsolv;
#endif

#ifdef SPHERE175
	//ParaComponent
	SklParaComponent *para_cmp;	//2009.12.07

	//ParaManager
	SklParaManager *para_mng;	//2009.12.07
#endif

	// parameter and const. fields
	S4D_REAL* MatVfFld;   // Medium m Volume Ratio
	S4D_REAL* MatVfInv;   // Medium m Volume Ratio - Inverse
	S4D_REAL* MatAreaX;   // Medium m Area Ratio for X direction
	S4D_REAL* MatAreaY;   // Medium m Area Ratio for Y direction
	S4D_REAL* MatAreaZ;   // Medium m Area Ratio for Z direction
	S4D_REAL* DifCoefX;   // Molecular l Diffusional Coef. for X direction
	S4D_REAL* DifCoefY;   // Molecular l Diffusional Coef. for Y direction
	S4D_REAL* DifCoefZ;   // Molecular l Diffusional Coef. for Z direction

	S4D_INT*  OutBndId_X; // OuterBoundary Condition ID for X of l,m
	S4D_INT*  OutBndId_Y; // OuterBoundary Condition ID for Y of l,m
	S4D_INT*  OutBndId_Z; // OuterBoundary Condition ID for Z of l,m
	
	S4D_REAL* MembArea;   // Membrane Area of Material m
	S4D_REAL* MembNrmlX;  // Membrane Normal Vector for X of Material m
	S4D_REAL* MembNrmlY;  // Membrane Normal Vector for Y of Material m
	S4D_REAL* MembNrmlZ;  // Membrane Normal Vector for Z of Material m

	S4D_REAL* MembAreaC;  // Membrane Area of Voxel Center (File Read)
	S4D_REAL* MembAreaX;  // Membrane Area of Voxel Edge X (File Read)
	S4D_REAL* MembAreaY;  // Membrane Area of Voxel Edge Y (File Read)
	S4D_REAL* MembAreaZ;  // Membrane Area of Voxel Edge Z (File Read)

	// equation and term fields ... phi(n+1) = phi(n) + dt * (E + D + M* + S)
	S4D_REAL* EqnP0Fld;   // [mol/m3]   ... equation value phi(n)
	S4D_REAL* EqnP1Fld;   // [mol/m3]   ... equation value phi(n+1)
	S4D_REAL* MtbTrmFld;  // [mol/m3.s] ... metabolism term
	S4D_REAL* DifFluxX;   // [mol/s.m2] ... diffusion flux term for X
	S4D_REAL* DifFluxY;   // [mol/s.m2] ... diffusion flux term for Y
	S4D_REAL* DifFluxZ;   // [mol/s.m2] ... diffusion flux term for Z
	S4D_REAL* MembTrmFld; // [] ... membrane term
	S4D_REAL* MembFlux;   // [mol/s.m2] ... membrane effection
	S4D_REAL* MembExQ;    // [mol/s]    ... membrane effection with except face
	S4D_REAL* SrcTrmFld;  // [mol/m3.s] ... other generic source term

	S4D_REAL* ElectricV;   // [V] ... Electric Potential of Material m.

	S4D_REAL* Rhs;        //(E + D + M* + S)
	S4D_REAL* Rhso;       //Rhs temporary value

	// steady state field
	S4D_REAL* SSFld;

	// special term fields for E term computation
	S4D_REAL* P0e;     // [mol/m3]
	S4D_REAL* Ee;      // [mol/m3.s]
	S4D_REAL* Vfe;     // (no dimension)
	S4D_REAL* Vfie;    // (no dimension)
	S4D_REAL* Sme;     // [m2]
	S4D_REAL* Nmxe;    // (no dimension)
	S4D_REAL* Nmye;    // (no dimension)
	S4D_REAL* Nmze;    // (no dimension)

	S4D_REAL* SmCe;
	S4D_REAL* SmXe;
	S4D_REAL* SmYe;
	S4D_REAL* SmZe;

	// Communication Links for parallel computing
	CommLinks IOlink;
	CommLinks P2PBlink,  P2PClink;
	CommLinks P2Elink,   E2Plink;
	CommLinks VfP2Elink, VfE2Plink;
	CommLinks SmP2Elink, SmE2Plink;

	// output
	typedef std::vector<ProbeT> ProbeListT;
	ProbeListT  m_ProbeList;
	int m_ProbeListSize;
	OutDataMapT m_OutDataMap;
	OutDataMapT OutRstDataMap;
	int chkpoint_interval;

	bool chkCfg();
	bool chkMemSize();

	int LoadUnitSetting();
	int LoadTimeStepSetting();
	
	int init_Domain();
	unsigned int init_MatMap();
	unsigned int init_MolMap();
	unsigned int init_EqnMap();
	unsigned int init_MtbMap();
	unsigned int init_MembMap();
	unsigned int init_MlMbMap();
	unsigned int init_MembareaMap();
	unsigned int init_BoundaryConditionTbl();
	unsigned int init_SteadyStateMap();

	int SetRestartTime(const SklVoxDataSet* dSet);
	int SyncRestartTime();
	std::string JobAssignTableString();

	int InitParallelEnvironment();
	int BuildGroupComputation(const JobManageTable& jmt);
	int BuildGroupPandE(const JobManageTable& jmt);
	int BuildGroupIO(const JobManageTable& jmt);
	int InitParallelCommLink();
	int BuildIOCommLink();
	int BuildP2PBCommLink();
	int BuildP2PCCommLink();
	int BuildP2ECommLink();
	int BuildE2PCommLink();
	int BuildVfP2ECommLink();
	int BuildVfE2PCommLink();
	int BuildSmP2ECommLink();
	int BuildSmE2PCommLink();
	int SendRecv(S4D_REAL* dest, const CommLinks& recvers,
				const S4D_REAL* src, const CommLinks& senders,
				const bool comm_vc);

	std::string CommLinkString();

	int TrimAssignedMaterial();

	int AllocateCompVariable();
	int InitializeCompVariable();
	int RegistFields();

	int LoadMatFile();
	int LoadMembareaFile();
	int RecoverMembarea();
	int LoadEqnFile();
	int LoadEqnRstFile();
	int LoadMatElement(
			const SklVoxDataSet* dSet, const size_t rec_id,
			S4D_REAL* dist, const size_t matid);
	int LoadMatElement(
			const SklVoxDataSet* dSet, const size_t rec_id,
			S4D_INT* dist, const size_t matid);
	int LoadMatElement(
			const SklVoxDataSet* dSet, const size_t rec_id, S3D_INT* dist);

	int LoadDifCoef();
	int LoadEqnBndCond();
	int LoadBndCond(const SklVoxDataSet* dSet, const size_t e);
	int init_FaceBndCond();

	int CalcEqnVol();

	int SetMolConc();
	int CalcMatVfInv();
	int CalcCmpMat();
	int CalcMembArea();

	int ChkMatVolumeRatio();

	int ExptimEqn(const SKL_REAL dt);
	int ExptimEqn(const SKL_REAL dt, const unsigned int step);
	int ExptimMat(const SKL_REAL dt);

	int InitFieldOut(const CfgElem* output);
	int InitFieldOutRst();
	int InitProbeOut(const CfgElem* output);

	int OutputField(const bool output_force = false);
	int OutputFieldRst(const bool output_force = false);
	int OutputProbe();
	int OutputTimeline(const unsigned int step);

	int BuildJobManageTable(JobManageTable& jmt);
	
	// communication linkage method
	int BuildSyncEqnRecverList();
	int BuildSyncEqnSenderList();
	int BuildCommMtbRecverList();
	int BuildCommMtbSenderList();

	SKL_REAL CalcDeltaT();

	// extension parameter
	int InitExtParameterMap(const CfgElem* extParam, const string& path="");
	
	// steady state
	int InitSteadyState();
	int CheckSteadyState(const unsigned int step);
	int InitSteadyStateLog();
	int OutputSteadyStateLog(const unsigned int step);

	// Log File
	int OpenLogFile();
	int CloseLogFile();

#ifdef MAKE_MODEL
	// Make Material function (MakeMat.C)
	int MakeMatByIndex(int xmin, int xmax, int ymin, int ymax,
			   int zmin, int zmax, size_t matid);
	int MakeMatGLC2Hx();
	int MakeMatRect();
#endif // MAKE_MODEL

protected:
	SklSolverFoundation();

public:
	SklSolverFoundation(int sType);
	virtual ~SklSolverFoundation();
	virtual int SklSolverInitialize();
	virtual int SklSolverLoop(unsigned int step);
	virtual bool SklSolverPost();
	virtual bool SklSolverUsage(const char* cmd);

public:
/* USER WRITE
   Please write member variable here.
*/

/* USER WRITE
   Please write class method here.
*/
};

#endif // _SKL_SOLVER_Foundation_CLASS_H_

