//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#include <SklUtil.h>
#include <sstream>

#include <Foundation/ecellif/ecellif.h>
#include <Foundation/util.h>
#include <Foundation/const.h>
#include <Foundation/error.h>
#include <Foundation/api.h>

#include <UserDefined/UserDefinedFunction.h>

#include "SklSolverFoundation.h"

int
SklSolverFoundation::SklSolverInitialize() {
/* USER WRITE
   Please write solver initialize code.
     Return value  ->   -1 : error
                         0 : forced termination
                         1 : normal return
*/
	try {

	////////////////////////////////////////////////////////
	// Init Solver Enviroment
	InitUtilEnv(&cfg);
#ifdef SPHERE175
	para_cmp = SklGetParaComponent();
#endif
	////////////////////////////////////////////////////////
	// open log file
	OpenLogFile();

	////////////////////////////////////////////////////////
	// output rics version info
#ifndef _NO_MESSAGE
	MESSAGE(RICS_VERSION);
	MESSAGE(RICS_COPYRIGHT);
	MESSAGE(RICS_BUILD);
	MESSAGE("");
#else
	MESSAGE(RICS_VERSION,0);
	MESSAGE(RICS_COPYRIGHT,0);
	MESSAGE(RICS_BUILD,0);
	MESSAGE("",0);
#endif //_NO_MESSAGE

	////////////////////////////////////////////////////////
	// Init Solver Enviroment
	InitECellIf();

	////////////////////////////////////////////////////////
	// load settings 
#ifndef _NO_MESSAGE
	MESSAGE("Load and Setting Computing Configuration");
#else
	MESSAGE("Load and Setting Computing Configuration",0);
#endif //_NO_MESSAGE
	const CfgElem* DomainInfo = SklCfgGetTop(DOMAININFO);
	const CfgElem* steer      = SklCfgGetTop(STEER);
	const CfgElem* parameter  = SklCfgGetTop(PARAMETER);

	////////////////////////////////////////////////////////
	LoadUnitSetting();
	LoadTimeStepSetting();

	// Rhs Copy Flag initialize
	cfg.RhsoInitFlag = 0;
	////////////////////////////////////////////////////////
	// Domain Setting
#ifndef _NO_MESSAGE
	MESSAGE("Loading Domain Setting ...");
#else
	MESSAGE("Loading Domain Setting ...",0);
#endif //_NO_MESSAGE
	init_Domain();

	cfg.moninterval = SklCfgGetMonitorInterval();

#ifdef SPHERE175
	//GetParaComponent, GetParaManager		2009.12.07
	para_cmp = SklGetParaComponent();
	para_mng = para_cmp->GetParaManager();
#endif

	cfg.solver    = this;
	cfg.CheckMode = SklIsCheckMode();

#ifdef SPHERE175
	cfg.Parallel  = para_cmp->IsParallel();		//2009.12.07
#else
	cfg.Parallel  = ParaMngr.IsParallel();
#endif

	cfg.InitStart = (SklCfgGetStartType() == Initial);
	cfg.Restart   = (SklCfgGetStartType() == Restart);

	/////////////////////////////////////////////////////////
	// scan cell and molecular settings
	init_MatMap();
	init_MolMap();

	/////////////////////////////////////////////////////////
	// init membrane mapping table
	init_MembMap();
	init_MlMbMap();

	/////////////////////////////////////////////////////////
	// init membrane area mapping table
	init_MembareaMap();

	/////////////////////////////////////////////////////////
	// init equation mapping table
	init_EqnMap();
	TRY((cfg.EqnCount > 0), "Equation Table size is ZERO!");

	/////////////////////////////////////////////////////////
	// init metabolism mapping table
	init_MtbMap();

	/////////////////////////////////////////////////////////
	// init boundary condition table
	// 2011.02.22.
	init_BoundaryConditionTbl();

	/////////////////////////////////////////////////////////
	// init steady state table
	// 2012.04.06.
	init_SteadyStateMap();

	/////////////////////////////////////////////////////////
	// init parallel environment
	InitParallelEnvironment();
	TrimAssignedMaterial();
	InitParallelCommLink();

	if (cfg.Parallel) {
#ifndef _NO_MESSAGE
		MESSAGE("Parallel Computing Environment.");
		MESSAGE(JobAssignTableString());
		MESSAGE(CommLinkString());
#else
		MESSAGE("Parallel Computing Environment.",0);
		// MESSAGE(JobAssignTableString(),0);	//2012.01.20
		// MESSAGE(CommLinkString(),0);		//2012.01.20
#endif //_NO_MESSAGE

#ifdef SPHERE175
		para_cmp->Barrier(cfg.grpAll.gid);	//2009.12.07
#else
		ParaMngr.Barrier(cfg.grpAll.gid);
#endif

	} else {
#ifndef _NO_MESSAGE
		MESSAGE("This Computing is Serial Mode.");
#else
		MESSAGE("This Computing is Serial Mode.",0);
#endif //_NO_MESSAGE
	}

	////////////////////////////////////////////////////////
	// allocate variables
	AllocateCompVariable();
	InitializeCompVariable();  // set default values to variables
	RegistFields();

	////////////////////////////////////////////////////////
	// input files and initialize variables

	// *********************
#ifdef MAKE_MODEL
	MakeMatGLC2Hx();

	// MakeMatRect();
#else
	LoadMatFile();
#endif //MAKE_MODEL
	// *********************

	LoadMembareaFile();
	if( cfg.MembareaCount != 0 ){
		MembAreaC->CommBndCell(MembAreaC->GetVCellSize());
		MembAreaX->CommBndCell(MembAreaX->GetVCellSize());
		MembAreaY->CommBndCell(MembAreaY->GetVCellSize());
		MembAreaZ->CommBndCell(MembAreaZ->GetVCellSize());
	}

	CalcCmpMat();
	CalcMatVfInv();
	CalcMembArea();        // Calculate Memrane Area form MatArea

	LoadDifCoef();
	SetMolConc();
	LoadEqnFile();
	LoadEqnBndCond();

	init_FaceBndCond();

	////////////////////////////////////////////////////////
	// Sync Initial value of  Virtual Cell
	MatVfFld->CommBndCell(MatVfFld->GetVCellSize());
	MatVfInv->CommBndCell(MatVfInv->GetVCellSize());
	MembArea->CommBndCell(MembArea->GetVCellSize());
	MembNrmlX->CommBndCell(MembNrmlX->GetVCellSize());
	MembNrmlY->CommBndCell(MembNrmlY->GetVCellSize());
	MembNrmlZ->CommBndCell(MembNrmlZ->GetVCellSize());

	// MatAreaX->CommBndCell(MatAreaX->GetVCellSize());
	// MatAreaY->CommBndCell(MatAreaY->GetVCellSize());
	// MatAreaZ->CommBndCell(MatAreaZ->GetVCellSize());
	// DifCoefX->CommBndCell(DifCoefX->GetVCellSize());
	// DifCoefY->CommBndCell(DifCoefY->GetVCellSize());
	// DifCoefZ->CommBndCell(DifCoefZ->GetVCellSize());

	////////////////////////////////////////////////////////
	// Sync Initial Field data and condition inter All Nodes
	SendRecv(EqnP0Fld, P2PClink,  EqnP0Fld,  P2PBlink,  false);
	SendRecv(Vfe,      VfE2Plink, MatVfFld,  VfP2Elink, true);
	SendRecv(Vfie,     VfE2Plink, MatVfInv,  VfP2Elink, true);
	SendRecv(Sme,      VfE2Plink, MembArea,  VfP2Elink, true);
	SendRecv(Nmxe,     VfE2Plink, MembNrmlX, VfP2Elink, true);
	SendRecv(Nmye,     VfE2Plink, MembNrmlY, VfP2Elink, true);
	SendRecv(Nmze,     VfE2Plink, MembNrmlZ, VfP2Elink, true);

	if( cfg.MembareaCount != 0 ){
		SendRecv(SmCe,     SmE2Plink, MembAreaC, SmP2Elink, true);
		SendRecv(SmXe,     SmE2Plink, MembAreaX, SmP2Elink, true);
		SendRecv(SmYe,     SmE2Plink, MembAreaY, SmP2Elink, true);
		SendRecv(SmZe,     SmE2Plink, MembAreaZ, SmP2Elink, true);
	}

	////////////////////////////////////////////////////////
	// Time Sync
	SyncRestartTime();

	////////////////////////////////////////////////////////
	// initialize output files
	const CfgElem* output = (parameter)?	
								parameter->GetElemFirst("output"):
								NULL;
	if (output) {
		InitFieldOut(output);
		InitProbeOut(output);
	}

	////////////////////////////////////////////////////////
	// initialize local solvers
	const CfgElem* simprm = (parameter)?
								parameter->GetElemFirst("sim_setting"):
								NULL;
	if (simprm) {
		mtbsolv.MetaboSolvInitialize(simprm,
							Ee,Vfe,Vfie,
							Sme,Nmxe,Nmye,Nmze,
							P0e,
							SmCe,SmXe,SmYe,SmZe);
		difsolv.DiffusionInitialize ( simprm,
							DifFluxX, DifFluxY, DifFluxZ,
							EqnP0Fld,
							DifCoefX, DifCoefY, DifCoefZ,
							MatAreaX, MatAreaY, MatAreaZ,
							OutBndId_X, OutBndId_Y, OutBndId_Z,
							MatVfInv, Rhs, Rhso
						);
		membsolv.MembraneInitialize (
							simprm,
							MembTrmFld, MembFlux, MembExQ,
							EqnP0Fld, MatVfFld, MatVfInv,
							MatAreaX, MatAreaY, MatAreaZ,
							ElectricV,
							MembArea,
							MembNrmlX, MembNrmlY, MembNrmlZ,
							MembAreaC,
							MembAreaX, MembAreaY, MembAreaZ
						);
#ifdef ENABLE_WARP
		const CfgElem* warpElem = (parameter)?
					parameter->GetElemFirst("warp_setting"):
					NULL;
		warpsolv.WarpInitialize ( simprm, warpElem );
#endif

#ifdef ENABLE_MEMBRANEPOTENTIAL
		const CfgElem* hhElem = (parameter)?
					parameter->GetElemFirst("membranePotential_setting"):
					NULL;
		potentialsolv.PotentialInitialize ( simprm, hhElem,
                                                  EqnP1Fld, EqnP0Fld,
                                                  MatAreaX, MatAreaY, MatAreaZ,
                                                  Rhs, Rhso, 
                                                  MembArea, MembNrmlX, MembNrmlY, MembNrmlZ,
                                                  DifFluxX, DifFluxY, DifFluxZ,
						  DifCoefX, DifCoefY, DifCoefZ,
		                                  OutBndId_X, OutBndId_Y, OutBndId_Z,
						  MembAreaC,
						  MembAreaX, MembAreaY, MembAreaZ);
#endif

#ifdef ENABLE_ADVECTION
		const CfgElem* advElem = (parameter)?
					 parameter->GetElemFirst("advection_setting"):
					 NULL;
		advectionsolv.AdvectionInitialize(simprm, advElem,
							EqnP0Fld, MatVfFld, MatVfInv,
							MatAreaX, MatAreaY, MatAreaZ,
							OutBndId_X, OutBndId_Y, OutBndId_Z);
#endif //ENABLE_ADVECTION
	}

	////////////////////////////////////////////////////////
	// Initial DeletaT
	CalcDeltaT();
	SklSetDeltaT(cfg.dt);

	////////////////////////////////////////////////////////
	// Init Ext Parameter
	InitExtParameterMap(parameter);
	
	////////////////////////////////////////////////////////
	// Init steady state
	InitSteadyState();
	InitSteadyStateLog();

	////////////////////////////////////////////////////////
	// check
	// chkMemSize();
	chkCfg();
	mtbsolv.chkCfg();
	difsolv.chkCfg();
	membsolv.chkCfg();
#ifdef ENABLE_WARP
	warpsolv.chkCfg();
#endif

#ifdef ENABLE_MEMBRANEPOTENTIAL
	potentialsolv.chkCfg();
#endif

#ifdef ENABLE_ADVECTION
	advectionsolv.chkCfg();
#endif

	MESSAGE(string().append(80, '='), 0);

	/////////////////////////////////////////////////////////
	UserDefinedFunctionInitialize();

	/////////////////////////////////////////////////////////
 	if (cfg.CheckMode) { return 0; }

	////////////////////////////////////////////////////////
	// Initial Output
	OutputField();
	OutputProbe();
	OutputTimeline(0);

	// reset clocks
	clock        = double(0.0);
	clock_all    = double(0.0);
	clock_io     = double(0.0);
	clock_exptim = double(0.0);
	clock_user   = double(0.0);
	clock_comm   = double(0.0);
	clock_wait   = double(0.0);
	
	return 1;

	////////////////////////////////////////////////////////
	// on error
	} catch (std::exception& e) {
		MSG_ERROR(e.what());
		//SklSolverPost();

#ifdef SPHERE175
		para_cmp->Abort(1);	//2009.12.07
#else
		ParaMngr.Abort(1);
#endif

	}

	return -1;
}

