//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#include <iostream>
#include <iomanip>

#include <Foundation/util.h>
#include <Foundation/const.h>
#include <Foundation/error.h>
#include <UserDefined/UserDefinedFunction.h>

#include "SklSolverFoundation.h"

int
 SklSolverFoundation::SklSolverLoop(const unsigned int step)
{
/* USER WRITE
   Please write solver main code.
	 Return value  ->   -1 : error
						 0 : forced termination
						 1 : normal return
*/
	BEGIN_CLOCK(all)
	int ret = 0;

	try {
	// Step Initialize
	BEGIN_CLOCK(loop)

	///////////////////////////////////////////////////
	// Clear
	UtilClearRealArray(SrcTrmFld, SKL_REAL(0.0));

	///////////////////////////////////////////////////
	// UserDefinedFunction
	BEGIN_CLOCK(userfunc)
	UserDefinedFunctionLoop();
	END_CLOCK(clock_user,userfunc)


	///////////////////////////////////////////////////
	// Sync Virtual Cells
	DB2(MESSAGE("CommBndCell"));
	MEASURE_WAIT_TIME(clock_wait, para_mng->Barrier());  // 2011.05.10
	BEGIN_CLOCK(comm)
	EqnP0Fld->CommBndCell(EqnP0Fld->GetVCellSize());
	MatAreaX->CommBndCell(MatAreaX->GetVCellSize());
	MatAreaY->CommBndCell(MatAreaY->GetVCellSize());
	MatAreaZ->CommBndCell(MatAreaZ->GetVCellSize());
	DifCoefX->CommBndCell(DifCoefX->GetVCellSize());
	DifCoefY->CommBndCell(DifCoefY->GetVCellSize());
	DifCoefZ->CommBndCell(DifCoefZ->GetVCellSize());
	END_CLOCK(clock_comm, comm);

	///////////////////////////////////////////////////
	// Local Simulations
	if (mtbsolv.enable()) {
		DB2(MESSAGE("SendRecv P2E"));
		SendRecv(P0e, E2Plink, EqnP0Fld, P2Elink, true);

		DB1(MESSAGE("Compute Metabo Term"));
		mtbsolv.MetaboSolvLoop(cfg.dt * step, step); // metabolome with E-Cell

		DB2(MESSAGE("SendRecv E2P"));
		SendRecv(MtbTrmFld, P2Elink, Ee, E2Plink, false);
	}

	DB1(MESSAGE("Compute Membrane Term"));
	membsolv.Calc(SKL_REAL(cfg.dt));             // membrane effection

	DB1(MESSAGE("Compute Diffusion Term"));
	difsolv.DiffusionLoop(SKL_REAL(cfg.dt));     // diffusion

#ifdef ENABLE_WARP
	DB1(MESSAGE("Compute Warp Term"));
	warpsolv.WarpLoop();
#endif

#ifdef ENABLE_MEMBRANEPOTENTIAL
        DB1(MESSAGE("Compute Potation Term"));
        potentialsolv.PotentialLoop(SKL_REAL(cfg.dt), step);  // potential
#endif

#ifdef ENABLE_ADVECTION
	if( advectionsolv.GetEnableAdvection() ){
		DB1(MESSAGE("Compute Advection Term"));
		advectionsolv.AdvectionLoop(step);
	}
#endif

	///////////////////////////////////////////////////
	// Coupling and Time-Marching
	// upd := res1 + res2 + ... + org
	DB1(MESSAGE("Compute Time Term"));
	BEGIN_CLOCK(exptim);
	//ExptimEqn(SKL_REAL(cfg.dt));
	ExptimEqn(SKL_REAL(cfg.dt), step);
	ExptimMat(SKL_REAL(cfg.dt));

#ifdef ENABLE_MEMBRANEPOTENTIAL
        potentialsolv.ExptimHH(SKL_REAL(cfg.dt));  // potential
#endif

#ifdef ENABLE_ADVECTION
	if( advectionsolv.GetEnableAdvection() ){
		advectionsolv.ExptimAdv(SKL_REAL(cfg.dt), EqnP1Fld);
	}
#endif

	END_CLOCK(clock_exptim,exptim);

	// apply p1 flds to p0 flds and time marching
	DB1(MESSAGE("Mirror p1 to p0"));
	SKL_REAL* eqn0 = EqnP0Fld->GetData();
	SKL_REAL* eqn1 = EqnP1Fld->GetData();
	SklUtil::cpVar(EqnP0Fld->GetArrayLength(), eqn0, eqn1);
	SklIncrementTime();

	// sync equation values inter all nodes
	DB2(MESSAGE("SendRecv P2P"));
	SendRecv(EqnP0Fld, P2PClink, EqnP0Fld, P2PBlink, false);

	END_CLOCK(clock,loop)

	BEGIN_CLOCK(io)
	// output data
	DB2(MESSAGE("Output Field Files"));
	OutputField();

	DB2(MESSAGE("Output Probe Files"));
	OutputProbe();

	END_CLOCK(clock_io,io)

	// ret = 1;
	ret = CheckSteadyState( step );
	// MESSAGE(UtilStrFormat("ret : %d", ret));

	// output log file (SS)
	BEGIN_CLOCK(io2);
	OutputSteadyStateLog( step );
	END_CLOCK(clock_io,io2);

	} 
	////////////////////////////////////////////////////////////
	// on error
	catch (std::exception& e) {
		MSG_ERROR(e.what());

#ifdef SPHERE175
		para_cmp->Abort(1);	//2009.12.07
#else
		ParaMngr.Abort(1);
#endif

		ret = -1;
		throw;
	}

	END_CLOCK(clock_all,all)
	OutputTimeline(step);

	return ret;
}

