//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#include <cmath>
#include <cstdio>
#include <unistd.h>
#include <assert.h>

#include <Foundation/datatype.h>
#include <Foundation/util.h>

#include "api.h"

int RicsGetExtParamInt(const char* paramName)
{
	return UtilGetExtParamInt(paramName);
}

SKL_REAL RicsGetExtParamReal(const char* paramName)
{
	return UtilGetExtParamReal(paramName);
}

const char* RicsGetExtParamString(const char* paramName)
{
	return UtilGetExtParamString(paramName);
}

bool RicsLoadCSV(const char* fileName, RealPairVector* table)
{
	const char sep = ',';
	return UtilLoadCSV(fileName, table, sep);
}

SKL_REAL RicsGetValueFromTable(const SKL_REAL x, const RealPairVector* table, const SKL_REAL offst)
{
	return UtilGetValueFromTable(x, table, offst);
}

void RicsGetVoxelWidth(int* id, int* jd, int* kd)
{
	if (id) { *id = UtilGetCfg()->size[0]; }
	if (jd) { *jd = UtilGetCfg()->size[1]; }
	if (kd) { *kd = UtilGetCfg()->size[2]; }
}

unsigned int RicsGetCurrentStep()
{
	return UtilGetCfg()->solver->SklGetCurrentStep();
}

SKL_REAL RicsGetDeltaT()
{
	return UtilGetCfg()->solver->SklGetDeltaT();
}

SKL_REAL RicsGetCurrentTime()
{
	return UtilGetCfg()->solver->SklGetCurrentTime();
}

//////////////////////////////////////////////////////////////////////////////
// Vfm
int RicsGetMaterialVolumeRate(const char* name, SKL_REAL* value, int i, int j, int k)
{
	const Material* m  = UtilGetMaterialId(name);

	if (!m) { return -1; }

	int zi,zj,zk;
	const bool in = UtilCheckIdxPdiv(&zi,&zj,&zk,i,j,k,1);

	if (!in) { return 0; }

	S4DRN Vfm = UtilGetCfg()->S4Dmap[UtilStrFormat("Vf_%s",m->name.c_str())];
	*value = Vfm.ary->val(zi,zj,zk,m->id);

	return 1;
}

//////////////////////////////////////////////////////////////////////////////
// phi
int RicsGetPhi(const char* mat, const char* mol, SKL_REAL* conc, int i, int j, int k)
{
	const Material*  m  = UtilGetMaterialId(mat);
	const Molecular* g  = UtilGetMolecularId(mol);
	const Equation*  gm = UtilGetEquationId(g,m);

	if (!(m && g && gm)) { return -1; }

	int zi,zj,zk;
	const bool in = UtilCheckIdxPdiv(&zi,&zj,&zk,i,j,k,1);

	if (!in) { return 0; }

	S4DRN Pgm = UtilGetCfg()->S4Dmap[UtilStrFormat("%s",gm->name.c_str())];
	*conc = Pgm.ary->val(zi,zj,zk,gm->id);

	return 1;
}

int RicsSetPhi(const char* mat, const char* mol, const SKL_REAL conc, int i, int j, int k)
{
	const Material*  m  = UtilGetMaterialId(mat);
	const Molecular* g  = UtilGetMolecularId(mol);
	const Equation*  gm = UtilGetEquationId(g,m);

	if (!(m && g && gm)) { return -1; }

	int zi,zj,zk;
	const bool in = UtilCheckIdxPdiv(&zi,&zj,&zk,i,j,k,1);

	if (!in) { return 0; }

	S4DRN Pgm = UtilGetCfg()->S4Dmap[UtilStrFormat("%s",gm->name.c_str())];
	S4DRN Vfm = UtilGetCfg()->S4Dmap[UtilStrFormat("Vf_%s",m->name.c_str())];
	Pgm.ary->val(zi,zj,zk,gm->id) = conc * std::ceil(Vfm.ary->val(i,j,k,m->id));

	return 1;
}

//////////////////////////////////////////////////////////////////////////////
// FlipFlop
int RicsFlipFlop(const char* mat0, const char* mat1, const char* mol, const SKL_REAL D, int i, int j, int k)
{
	const SKL_REAL dt = RicsGetDeltaT();

	const Material*  m = UtilGetMaterialId(mat0);
	const Material*  n = UtilGetMaterialId(mat1);
	const Molecular* g = UtilGetMolecularId(mol);

	const Equation* gm = UtilGetEquationId(g,m);
	const Equation* gn = UtilGetEquationId(g,n);

	if (!(m && n && g && gm && gn)) { return -1; }

	int zi,zj,zk;
	const bool in = UtilCheckIdxPdiv(&zi,&zj,&zk,i,j,k,1);

	if (!in) { return 0; }

	// Source Term
	S4DRN Sgm = UtilGetCfg()->S4Dmap[UtilStrFormat("SRC_%s",gm->name.c_str())];
	S4DRN Sgn = UtilGetCfg()->S4Dmap[UtilStrFormat("SRC_%s",gn->name.c_str())];

	//////////////////////////////////////////////////////////////////////////
	// Material m side
	S4DRN Pgm = UtilGetCfg()->S4Dmap[UtilStrFormat("%s",gm->name.c_str())];
	Sgm.ary->val(zi,zj,zk,gm->id) -= D * Pgm.ary->val(zi,zj,zk,gm->id) / dt;

	//////////////////////////////////////////////////////////////////////////
	// Material n side
	S4DRN Vfm  = UtilGetCfg()->S4Dmap[UtilStrFormat("Vf_%s", m->name.c_str())];
	S4DRN Vfin = UtilGetCfg()->S4Dmap[UtilStrFormat("Vfi_%s",n->name.c_str())];

	S4DRN Nx = UtilGetCfg()->S4Dmap[UtilStrFormat("Nx_%s",m->name.c_str())];
	S4DRN Ny = UtilGetCfg()->S4Dmap[UtilStrFormat("Ny_%s",m->name.c_str())];
	S4DRN Nz = UtilGetCfg()->S4Dmap[UtilStrFormat("Nz_%s",m->name.c_str())];

	const int di = zi + static_cast<int>(Nx.ary->val(zi,zj,zk,Nx.N));
	const int dj = zj + static_cast<int>(Ny.ary->val(zi,zj,zk,Ny.N));
	const int dk = zk + static_cast<int>(Nz.ary->val(zi,zj,zk,Nz.N));

	if (UtilCheckIdxPdiv(NULL,NULL,NULL,di,dj,dk,0)) {
		const SKL_REAL Rv = Vfm.ary->val(zi,zj,zk,m->id)
							* Vfin.ary->val(di,dj,dk,n->id);
		Sgn.ary->val(di,dj,dk,gn->id) +=
			D * Pgm.ary->val(zi,zj,zk,gm->id) * Rv / dt;
	}

	return 1;
}

//////////////////////////////////////////////////////////////////////////////
// for membrane function
// penetration ///////////////////////////////////////////////////////////////
// Kpene
int RicsGetMembraneKPene(const char* name, const char* mol, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("Pene_%s:%s",mol,name);
	const penetrater* p = UtilGetPenetraterId(pname);

	return UtilGetMembParam("K",p,i,j,k,value);
}


int RicsSetMembraneKPene(const char* name, const char* mol, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("Pene_%s:%s",mol,name);
	const penetrater* p = UtilGetPenetraterId(pname);

	return UtilSetMembParam("K",p,i,j,k,value);
}

// channel ///////////////////////////////////////////////////////////////////
// Sw
int RicsGetMembraneSwCha(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Ch_%s:%s",id,mol,name);
	const ChTranser* p = UtilGetChTranserId(pname);

	return UtilGetMembParam("SW",p,i,j,k,value);
}

int RicsSetMembraneSwCha(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Ch_%s:%s",id,mol,name);
	const ChTranser* p = UtilGetChTranserId(pname);

	return UtilSetMembParam("SW",p,i,j,k,value);
}

// fuzzy
int RicsGetMembraneFuzzyCha(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Ch_%s:%s",id,mol,name);
	const ChTranser* p = UtilGetChTranserId(pname);

	return UtilGetMembParam("Fuzzy",p,i,j,k,value);
}

int RicsSetMembraneFuzzyCha(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Ch_%s:%s",id,mol,name);
	const ChTranser* p = UtilGetChTranserId(pname);

	return UtilSetMembParam("Fuzzy",p,i,j,k,value);
}

// Den
int RicsGetMembraneDenCha(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Ch_%s:%s",id,mol,name);
	const ChTranser* p = UtilGetChTranserId(pname);

	return UtilGetMembParam("Den",p,i,j,k,value);
}

int RicsSetMembraneDenCha(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Ch_%s:%s",id,mol,name);
	const ChTranser* p = UtilGetChTranserId(pname);

	return UtilSetMembParam("Den",p,i,j,k,value);
}

// K
int RicsGetMembraneKCha(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Ch_%s:%s",id,mol,name);
	const ChTranser* p = UtilGetChTranserId(pname);

	return UtilGetMembParam("K",p,i,j,k,value);
}

int RicsSetMembraneKCha(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Ch_%s:%s",id,mol,name);
	const ChTranser* p = UtilGetChTranserId(pname);

	return UtilSetMembParam("K",p,i,j,k,value);
}

// Ke
int RicsGetMembraneKeCha(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Ch_%s:%s",id,mol,name);
	const ChTranser* p = UtilGetChTranserId(pname);

	return UtilGetMembParam("Ke",p,i,j,k,value);
}

int RicsSetMembraneKeCha(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Ch_%s:%s",id,mol,name);
	const ChTranser* p = UtilGetChTranserId(pname);

	return UtilSetMembParam("Ke",p,i,j,k,value);
}

// pump //////////////////////////////////////////////////////////////////////
// Sw
int RicsGetMembraneSwPum(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilGetMembParam("SW",p,i,j,k,value);
}

int RicsSetMembraneSwPum(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilSetMembParam("SW",p,i,j,k,value);
}

// fuzzy
int RicsGetMembraneFuzzyPum(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilGetMembParam("Fuzzy",p,i,j,k,value);
}

int RicsSetMembraneFuzzyPum(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilSetMembParam("Fuzzy",p,i,j,k,value);
}

// Den
int RicsGetMembraneDenPum(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilGetMembParam("Den",p,i,j,k,value);
}

int RicsSetMembraneDenPum(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilSetMembParam("Den",p,i,j,k,value);
}

// VmaxATP
int RicsGetMembraneVmaxAtpPum(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilGetMembParam("VmaxAtp",p,i,j,k,value);
}

int RicsSetMembraneVmaxAtpPum(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilSetMembParam("VmaxAtp",p,i,j,k,value);
}

// KmATP
int RicsGetMembraneKmAtpPum(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilGetMembParam("KmAtp",p,i,j,k,value);
}

int RicsSetMembraneKmAtpPum(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilSetMembParam("KmAtp",p,i,j,k,value);
}

// VmaxGTP
int RicsGetMembraneVmaxGtpPum(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilGetMembParam("VmaxGtp",p,i,j,k,value);
}

int RicsSetMembraneVmaxGtpPum(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilSetMembParam("VmaxGtp",p,i,j,k,value);
}

// KmGTP
int RicsGetMembraneKmGtpPum(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilGetMembParam("KmGtp",p,i,j,k,value);
}

int RicsSetMembraneKmGtpPum(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Pm_%s:%s",id,mol,name);
	const PmTranser* p = UtilGetPmTranserId(pname);

	return UtilSetMembParam("KmGtp",p,i,j,k,value);
}


// symport channel ///////////////////////////////////////////////////////////
// Sw
int RicsGetMembraneSwChaSym(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CS_%s:%s",id,mol,name);
	const CSTranser* p = UtilGetCSTranserId(pname);

	return UtilGetMembParam("SW",p,i,j,k,value);
}

int RicsSetMembraneSwChaSym(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CS_%s:%s",id,mol,name);
	const CSTranser* p = UtilGetCSTranserId(pname);

	return UtilSetMembParam("SW",p,i,j,k,value);
}

// fuzzy
int RicsGetMembraneFuzzyChaSym(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CS_%s:%s",id,mol,name);
	const CSTranser* p = UtilGetCSTranserId(pname);

	return UtilGetMembParam("Fuzzy",p,i,j,k,value);
}

int RicsSetMembraneFuzzyChaSym(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CS_%s:%s",id,mol,name);
	const CSTranser* p = UtilGetCSTranserId(pname);

	return UtilSetMembParam("Fuzzy",p,i,j,k,value);
}

// Den
int RicsGetMembraneDenChaSym(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CS_%s:%s",id,mol,name);
	const CSTranser* p = UtilGetCSTranserId(pname);

	return UtilGetMembParam("Den",p,i,j,k,value);
}

int RicsSetMembraneDenChaSym(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CS_%s:%s",id,mol,name);
	const CSTranser* p = UtilGetCSTranserId(pname);

	return UtilSetMembParam("Den",p,i,j,k,value);
}

// K
int RicsGetMembraneKChaSym(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CS_%s:%s",id,mol,name);
	const CSTranser* p = UtilGetCSTranserId(pname);

	return UtilGetMembParam("K",p,i,j,k,value);
}

int RicsSetMembraneKChaSym(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CS_%s:%s",id,mol,name);
	const CSTranser* p = UtilGetCSTranserId(pname);

	return UtilSetMembParam("K",p,i,j,k,value);
}

// Ke
int RicsGetMembraneKeChaSym(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CS_%s:%s",id,mol,name);
	const CSTranser* p = UtilGetCSTranserId(pname);

	return UtilGetMembParam("Ke",p,i,j,k,value);
}

int RicsSetMembraneKeChaSym(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CS_%s:%s",id,mol,name);
	const CSTranser* p = UtilGetCSTranserId(pname);

	return UtilSetMembParam("Ke",p,i,j,k,value);
}

// symport pump //////////////////////////////////////////////////////////////
// Sw
int RicsGetMembraneSwPumSym(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilGetMembParam("SW",p,i,j,k,value);
}

int RicsSetMembraneSwPumSym(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilSetMembParam("SW",p,i,j,k,value);
}

// fuzzy
int RicsGetMembraneFuzzyPumSym(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilGetMembParam("Fuzzy",p,i,j,k,value);
}

int RicsSetMembraneFuzzyPumSym(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilSetMembParam("Fuzzy",p,i,j,k,value);
}

// Den
int RicsGetMembraneDenPumSym(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilGetMembParam("Den",p,i,j,k,value);
}

int RicsSetMembraneDenPumSym(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilSetMembParam("Den",p,i,j,k,value);
}

// VmaxATP
int RicsGetMembraneVmaxAtpPumSym(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilGetMembParam("VmaxAtp",p,i,j,k,value);
}

int RicsSetMembraneVmaxAtpPumSym(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilSetMembParam("VmaxAtp",p,i,j,k,value);
}

// KmATP
int RicsGetMembraneKmAtpPumSym(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilGetMembParam("KmAtp",p,i,j,k,value);
}

int RicsSetMembraneKmAtpPumSym(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilSetMembParam("KmAtp",p,i,j,k,value);
}

// VmaxGTP
int RicsGetMembraneVmaxGtpPumSym(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilGetMembParam("VmaxGtp",p,i,j,k,value);
}

int RicsSetMembraneVmaxGtpPumSym(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilSetMembParam("VmaxGtp",p,i,j,k,value);
}

// KmGTP
int RicsGetMembraneKmGtpPumSym(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilGetMembParam("KmGtp",p,i,j,k,value);
}

int RicsSetMembraneKmGtpPumSym(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PS_%s:%s",id,mol,name);
	const PSTranser* p = UtilGetPSTranserId(pname);

	return UtilSetMembParam("KmGtp",p,i,j,k,value);
}


// antiport channel ///////////////////////////////////////////////////////////
// Sw
int RicsGetMembraneSwChaAnti(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CA_%s:%s",id,mol,name);
	const CATranser* p = UtilGetCATranserId(pname);

	return UtilGetMembParam("SW",p,i,j,k,value);
}

int RicsSetMembraneSwChaAnti(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CA_%s:%s",id,mol,name);
	const CATranser* p = UtilGetCATranserId(pname);

	return UtilSetMembParam("SW",p,i,j,k,value);
}

// fuzzy
int RicsGetMembraneFuzzyChaAnti(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CA_%s:%s",id,mol,name);
	const CATranser* p = UtilGetCATranserId(pname);

	return UtilGetMembParam("Fuzzy",p,i,j,k,value);
}

int RicsSetMembraneFuzzyChaAnti(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CA_%s:%s",id,mol,name);
	const CATranser* p = UtilGetCATranserId(pname);

	return UtilSetMembParam("Fuzzy",p,i,j,k,value);
}

// Den
int RicsGetMembraneDenChaAnti(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CA_%s:%s",id,mol,name);
	const CATranser* p = UtilGetCATranserId(pname);

	return UtilGetMembParam("Den",p,i,j,k,value);
}

int RicsSetMembraneDenChaAnti(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CA_%s:%s",id,mol,name);
	const CATranser* p = UtilGetCATranserId(pname);

	return UtilSetMembParam("Den",p,i,j,k,value);
}

// K
int RicsGetMembraneKChaAnti(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CA_%s:%s",id,mol,name);
	const CATranser* p = UtilGetCATranserId(pname);

	return UtilGetMembParam("K",p,i,j,k,value);
}

int RicsSetMembraneKChaAnti(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CA_%s:%s",id,mol,name);
	const CATranser* p = UtilGetCATranserId(pname);

	return UtilSetMembParam("K",p,i,j,k,value);
}

// Ke
int RicsGetMembraneKeChaAnti(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CA_%s:%s",id,mol,name);
	const CATranser* p = UtilGetCATranserId(pname);

	return UtilGetMembParam("Ke",p,i,j,k,value);
}

int RicsSetMembraneKeChaAnti(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:CA_%s:%s",id,mol,name);
	const CATranser* p = UtilGetCATranserId(pname);

	return UtilSetMembParam("Ke",p,i,j,k,value);
}

// antiport pump //////////////////////////////////////////////////////////////
// Sw
int RicsGetMembraneSwPumAnti(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilGetMembParam("SW",p,i,j,k,value);
}

int RicsSetMembraneSwPumAnti(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilSetMembParam("SW",p,i,j,k,value);
}

// fuzzy
int RicsGetMembraneFuzzyPumAnti(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilGetMembParam("Fuzzy",p,i,j,k,value);
}

int RicsSetMembraneFuzzyPumAnti(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilSetMembParam("Fuzzy",p,i,j,k,value);
}

// Den
int RicsGetMembraneDenPumAnti(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilGetMembParam("Den",p,i,j,k,value);
}

int RicsSetMembraneDenPumAnti(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilSetMembParam("Den",p,i,j,k,value);
}

// VmaxATP
int RicsGetMembraneVmaxAtpPumAnti(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilGetMembParam("VmaxAtp",p,i,j,k,value);
}

int RicsSetMembraneVmaxAtpPumAnti(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilSetMembParam("VmaxAtp",p,i,j,k,value);
}

// KmATP
int RicsGetMembraneKmAtpPumAnti(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilGetMembParam("KmAtp",p,i,j,k,value);
}

int RicsSetMembraneKmAtpPumAnti(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilSetMembParam("KmAtp",p,i,j,k,value);
}

// VmaxGTP
int RicsGetMembraneVmaxGtpPumAnti(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilGetMembParam("VmaxGtp",p,i,j,k,value);
}

int RicsSetMembraneVmaxGtpPumAnti(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilSetMembParam("VmaxGtp",p,i,j,k,value);
}

// KmGTP
int RicsGetMembraneKmGtpPumAnti(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilGetMembParam("KmGtp",p,i,j,k,value);
}

int RicsSetMembraneKmGtpPumAnti(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:PA_%s:%s",id,mol,name);
	const PATranser* p = UtilGetPATranserId(pname);

	return UtilSetMembParam("KmGtp",p,i,j,k,value);
}


// gap junction //////////////////////////////////////////////////////////////
// Sw
int RicsGetMembraneSwGap(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Gp_%s:%s",id,mol,name);
	const GpTranser* p = UtilGetGpTranserId(pname);

	return UtilGetMembParam("SW",p,i,j,k,value);
}

int RicsSetMembraneSwGap(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Gp_%s:%s",id,mol,name);
	const GpTranser* p = UtilGetGpTranserId(pname);

	return UtilSetMembParam("SW",p,i,j,k,value);
}

// fuzzy
int RicsGetMembraneFuzzyGap(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Gp_%s:%s",id,mol,name);
	const GpTranser* p = UtilGetGpTranserId(pname);

	return UtilGetMembParam("Fuzzy",p,i,j,k,value);
}

int RicsSetMembraneFuzzyGap(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Gp_%s:%s",id,mol,name);
	const GpTranser* p = UtilGetGpTranserId(pname);

	return UtilSetMembParam("Fuzzy",p,i,j,k,value);
}

// Den
int RicsGetMembraneDenGap(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Gp_%s:%s",id,mol,name);
	const GpTranser* p = UtilGetGpTranserId(pname);

	return UtilGetMembParam("Den",p,i,j,k,value);
}

int RicsSetMembraneDenGap(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Gp_%s:%s",id,mol,name);
	const GpTranser* p = UtilGetGpTranserId(pname);

	return UtilSetMembParam("Den",p,i,j,k,value);
}

// K
int RicsGetMembraneKGap(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Gp_%s:%s",id,mol,name);
	const GpTranser* p = UtilGetGpTranserId(pname);

	return UtilGetMembParam("K",p,i,j,k,value);
}

int RicsSetMembraneKGap(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Gp_%s:%s",id,mol,name);
	const GpTranser* p = UtilGetGpTranserId(pname);

	return UtilSetMembParam("K",p,i,j,k,value);
}

// Ke
int RicsGetMembraneKeGap(const char* name, const char* mol, int id, SKL_REAL* value, int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Gp_%s:%s",id,mol,name);
	const GpTranser* p = UtilGetGpTranserId(pname);

	return UtilGetMembParam("Ke",p,i,j,k,value);
}

int RicsSetMembraneKeGap(const char* name, const char* mol, int id, const SKL_REAL value,  int i, int j, int k)
{
	const string pname = UtilStrFormat("%d:Gp_%s:%s",id,mol,name);
	const GpTranser* p = UtilGetGpTranserId(pname);

	return UtilSetMembParam("Ke",p,i,j,k,value);
}

int RicsAddSrc(const char* mat, const char* mol, const SKL_REAL src, int i, int j, int k)
{
	const Material*  m  = UtilGetMaterialId(mat);
	const Molecular* g  = UtilGetMolecularId(mol);
	const Equation*  gm = UtilGetEquationId(g,m);

	if (!(m && g && gm)) { return -1; }

	int zi,zj,zk;
	const bool in = UtilCheckIdxPdiv(&zi,&zj,&zk,i,j,k,1);

	if (!in) { return 0; }

	S4DRN Sgm = UtilGetCfg()->S4Dmap[UtilStrFormat("SRC_%s",gm->name.c_str())];
	S4DRN Vfm = UtilGetCfg()->S4Dmap[UtilStrFormat("Vf_%s",m->name.c_str())];
	Sgm.ary->val(zi,zj,zk,gm->id) += src * std::ceil(Vfm.ary->val(zi,zj,zk,m->id));

	return 1;
}

//////////////////////////////////////////////////////////////////////////////
// memory info from system
int RicsGetSystemMemoryInfo(SystemMemoryInfo& info)
{
	const int nID        = UtilGetCfg()->grpAll.nid;
	const int MAX_STRING = 1024;
	const int BUF_SIZE   = 16;
	
	char statfile[MAX_STRING];
	char   status[MAX_STRING];

	unsigned int VmPeak, VmSize, VmHWM, VmRSS;
	unsigned int size, resident, shared, trs, drs, lrs, dt;

	const pid_t pid = getpid();
	FILE* fp;

	// read status
	sprintf(statfile, "/proc/%u/status", pid);
	fp = fopen(statfile, "rt");
	if(fp == NULL){
		MESSAGE(UtilStrFormat("ID=%d: /proc/%u/stasus is not open", nID, pid));
		return -1;
	}
	fread((void*)status, 1, 1024, fp);
#ifdef _MEMINFO_VERIFY	
	sprintf(statfile, "status-%1d-%03d.txt", nID, RicsGetCurrentStep());
	FILE* fp2 = fopen(statfile, "w");
	fprintf(fp2, "%s", status);
	fclose(fp2);
#endif	
	fclose(fp);

	// parsing status
	char* p;
	char tmp[BUF_SIZE];
	p = strstr(status, "VmPeak");	sscanf(p, "%s %d", tmp, &VmPeak);
	p = strstr(status, "VmSize");   sscanf(p, "%s %d", tmp, &VmSize);
	p = strstr(status, "VmHWM");    sscanf(p, "%s %d", tmp, &VmHWM);
	p = strstr(status, "VmRSS");    sscanf(p, "%s %d", tmp, &VmRSS);
	memset(status, 0, MAX_STRING);

	// read statm
	sprintf(statfile, "/proc/%u/statm", pid);
	DB2(MESSAGE(statfile));
	fp = fopen(statfile, "rt");
	if(fp == NULL){
		MESSAGE(UtilStrFormat("ID=%d: /proc/%u/statm is not open", nID, pid));
		return -1;
	}
	fread((void*)status, 1, MAX_STRING, fp);
#ifdef _MEMINFO_VERIFY	
	sprintf(statfile, "statm-%1d-%03d.txt", nID, RicsGetCurrentStep());
	fp2 = fopen(statfile, "w");
	fprintf(fp2, "%s", status);
	fclose(fp2);
#endif	
	fclose(fp);

	// parsing statm
	sscanf(status, "%d %d %d %d %d %d %d",
		   &size, &resident, &shared, &trs, &drs, &lrs, &dt);
//	size     *= 4;  // convert page to kB
//	resident *= 4;
	shared   *= 4;
//	trs      *= 4;
//	drs      *= 4;
//	lrs      *= 4;
//	dt       *= 4;

	info.ID     = nID;
	info.pid    = pid;
	info.VmPeak = VmPeak;
	info.VmSize = VmSize;
	info.VmHWM  = VmHWM;
	info.VmRSS  = VmRSS;
	info.Shared = shared;

	return 1;
}

int RicsShowSystemMemoryInfo()
{
	SKL_REAL time     = RicsGetCurrentTime();
	unsigned int step = RicsGetCurrentStep();
	
	SystemMemoryInfo info;
	
	if(RicsGetSystemMemoryInfo(info) == 1){
		MESSAGE("-------------------------");
		MESSAGE("System Memory Info");
		MESSAGE((UtilStrFormat("ID=%d, pid=%d, step=%d, time=%lf",
							   info.ID, info.pid, step, time)));
		MESSAGE((UtilStrFormat("VmPeak: %8d [kB]",info.VmPeak)));
		MESSAGE((UtilStrFormat("VmSize: %8d [kB]",info.VmSize)));
		MESSAGE((UtilStrFormat("VmHWM:  %8d [kB]",info.VmHWM)));
		MESSAGE((UtilStrFormat("VmRSS:  %8d [kB]",info.VmRSS)));
		MESSAGE((UtilStrFormat("Share:  %8d [kB]",info.Shared)));
		MESSAGE("-------------------------");
	}else{
		return -1;
	}
	
	return 1;
}

int RicsSaveSystemMemoryInfo(const char* prefix, const char* udstring, const SystemMemoryInfo& info)
{
	assert(prefix   != NULL);
    assert(udstring != NULL);

	const CfgInfo* cfg      = UtilGetCfg();
	const SKL_REAL time     = RicsGetCurrentTime();
	const unsigned int step = RicsGetCurrentStep();

	const CfgInfo::Communicator com = cfg->grpAll;
	const int nodeID = com.nid;
	
	// make file name and
	std::string fname = UtilStrFormat("%s%08d.csv", prefix, nodeID);
	
	// open file
	FILE*fp = fopen(fname.c_str(), "a");
	if(fp == NULL){
		MESSAGE(UtilStrFormat("Error: can't open file %s", fname.c_str()));
		return 0;
	}

	// write header if this is new file
	if(ftell(fp) == 0){
		fprintf(fp, "step,time,pid,udstring,VmPeak[kB],VmSize[kB],");
		fprintf(fp, "VmHWM[kB],VmRSS[kB],Share[kB]\n");
	}

	// write data
	fprintf(fp, "%d,%e,%d,%s,%d,%d,%d,%d,%d\n",
			step, time, info.pid, udstring, info.VmPeak, info.VmSize, info.VmHWM,
			info.VmRSS, info.Shared);
	fclose(fp);

	return 1;
}

//////////////////////////////////////////////////////////////////////////////
// check values
CheckValueElem::CheckValueElem()
{
	ijk[0] = ijk[1] = ijk[2] = 0;
	phi    = vof    = mol    = 0.0;
}

CheckValueElem::CheckValueElem(int i, int j, int k, SKL_REAL p, SKL_REAL v, SKL_REAL m)
{
	ijk[0] = i;
	ijk[1] = j;
	ijk[2] = k;
	
	phi = p;
	vof = v;
	mol = m;
}

int _CheckPhiValueByDivision(const char* mat, const char* mol, CheckValueInfo& info, SKL_REAL bottom, SKL_REAL upper)
{
	assert(mat != NULL);
	assert(mol != NULL);

	const CfgInfo* cfg = UtilGetCfg();
	const Material*  m = UtilGetMaterialId(mat);
	const Molecular* g = UtilGetMolecularId(mol);
	const Equation* gm = UtilGetEquationId(g, m);
	if(m == NULL || g == NULL || gm == NULL){
		MESSAGE(UtilStrFormat("Error: Can't get eq.%s-%s", mat, mol));
		return 0;
	}

	// check wether equation "gm" assiend in this node or not.
	const EqnSet& AssEqns = cfg->AssEqns;
	if(find(AssEqns.begin(), AssEqns.end(), gm) == AssEqns.end()){
//		MESSAGE(UtilStrFormat("Don't be assigned eq.%s-%s", mat, mol));
		return 0;
	}
	
	const SKL_REAL K = cfg->cvol * pow(cfg->Ls, 3) * cfg->Kna * cfg->NA;
	const int gmid = gm->id;
	const int  mid =  m->id;
	
	const std::string eq   = UtilStrFormat("%s",   gm->name.c_str());
	const std::string ma   = UtilStrFormat("Vf_%s", m->name.c_str());
	const S4D_REAL* arrphi = UtilGetCfg()->S4Dmap[eq].ary;
	const S4D_REAL* arrvof = UtilGetCfg()->S4Dmap[ma].ary;
	
	SKL_REAL minimum =  REAL_MAX;
	SKL_REAL maximum = -REAL_MAX;

	// get local ijk
	const int procGrp = cfg->grpP.gid;
#ifdef SPHERE175
	const SklParaComponent* para_cmp = cfg->solver->SklGetParaComponent();
	const SklParaManager*   para_mng = para_cmp->GetParaManager();
	const unsigned int*     ijkmax   = para_mng->GetVoxelSize(procGrp);
#else
	const unsigned int* ijkmax = cfg->solver->ParaMngr.GetVoxelSize(procGrp);
#endif
	if(ijkmax == NULL){
		MESSAGE("Error: GetVoxelSize returns NULL!");
		return 0;
	}
	const int ihd = cfg->Pdiv.head[0]; // for convert local ijk to global ijk
	const int jhd = cfg->Pdiv.head[1];
	const int khd = cfg->Pdiv.head[2];

	CheckValueElem& min         = info.min; // for simply access
	CheckValueElem& max         = info.max;
	vector<CheckValueElem>& ovf = info.overflow;
	vector<CheckValueElem>& neg = info.negative;
	vector<CheckValueElem>& out = info.outrange;

	// 
	// loop on local ijk space
	//
	SKL_REAL phi, vof, nmo;
	for(int k = 0; k < ijkmax[2]; k++){
	for(int j = 0; j < ijkmax[1]; j++){
	for(int i = 0; i < ijkmax[0]; i++){
		phi = arrphi->val(i, j, k, gmid);

		if(phi < minimum){  // check min
			vof     = arrvof->val(i, j, k, mid);
			nmo     = phi * vof * K;
			minimum = phi;
			min     = CheckValueElem(i+ihd, j+jhd, k+khd, phi, vof, nmo);
		}

		if(phi > maximum){  // check max
			vof     = arrvof->val(i, j, k, mid);
			nmo     = phi * vof * K;
			maximum = phi;
			max     = CheckValueElem(i+ihd, j+jhd, k+khd, phi, vof, nmo);
		}

		if(phi < 0.0){  // check negative
			vof = arrvof->val(i, j, k, mid);
			nmo = phi * vof * K;
			neg.push_back(CheckValueElem(i+ihd, j+jhd, k+khd, phi, vof, nmo));
		}

		if(REAL_INVALID(phi)){  // check overflow
			vof = arrvof->val(i, j, k, mid);
			nmo = phi * vof * K;
			ovf.push_back(CheckValueElem(i+ihd, j+jhd, k+khd, phi, vof, nmo));
		}

		if(!(bottom <= phi && phi <= upper)){  // check out of range
			vof = arrvof->val(i, j, k, mid);
			nmo = phi * vof * K;
			out.push_back(CheckValueElem(i+ihd, j+jhd, k+khd, phi, vof, nmo));
		}
	}}}

	info.mat = mat;
	info.mol = mol;
	return 1;
}

int RicsCheckPhiValue(const char* mat, const char* mol, CheckValueInfo& info, SKL_REAL bottom, SKL_REAL upper)
{
	assert(mat != NULL);
	assert(mol != NULL);

	// in case of serial, division is only entire region.
	const CfgInfo* cfg = UtilGetCfg();
	if(!cfg->Parallel){
		return _CheckPhiValueByDivision(mat, mol, info, bottom, upper);
	}

	// check value on each divided region
	// collect check state and error amount
	CheckValueInfo er;
	int state[4];
	state[0] = _CheckPhiValueByDivision(mat, mol, er, bottom, upper);
	state[1] = er.overflow.size();
	state[2] = er.negative.size();
	state[3] = er.outrange.size();

#ifdef SPHERE175
	SklParaComponent* para_cmp = cfg->solver->SklGetParaComponent();
	SklParaManager*   para_mng = para_cmp->GetParaManager();
#endif
	const CfgInfo::Communicator com = cfg->grpP;
	const int nodeID   = com.nid;
	const int procGrp  = com.gid;
	const int num_node = com.num;

	//
	// collect array "state", local min and local max into node-nID
	//
	const size_t sz = sizeof(CheckValueElem);
	vector<int> arrState;
	vector<CheckValueElem> min_list;
	vector<CheckValueElem> max_list;
	if(nodeID == 0){
		arrState.resize(num_node * 4);
		min_list.resize(num_node);
		max_list.resize(num_node);
	}
	para_mng->Gather((void*)state,       4, SKL_ARRAY_DTYPE_INT,
					 &arrState.front(),  4, SKL_ARRAY_DTYPE_INT,  0, procGrp);
	para_mng->Gather(&er.min,           sz, SKL_ARRAY_DTYPE_CHAR, 
					 &min_list.front(), sz, SKL_ARRAY_DTYPE_CHAR, 0, procGrp);
	para_mng->Gather(&er.max,           sz, SKL_ARRAY_DTYPE_CHAR, 
					 &max_list.front(), sz, SKL_ARRAY_DTYPE_CHAR, 0, procGrp);
	
	//
	// collect global overflow, negative, outrange info into node-nID
	//
	int noverflow = 0;
	int nnegative = 0;
	int noutrange = 0;
	if(nodeID == 0){
		for(int i = 0; i < num_node; i++){ // count total number of 
			noverflow += arrState[i*4+1];  // overflow, negative, outrange
			nnegative += arrState[i*4+2];  
			noutrange += arrState[i*4+3];
		}

		// allocate global info array, and copy nID having local info into global
		info.overflow.resize(noverflow);
		info.negative.resize(nnegative);
		info.outrange.resize(noutrange);
		const int nov = er.overflow.size();
		const int nne = er.negative.size();
		const int nor = er.outrange.size();
		for(int i = 0; i < nov; i++) info.overflow[i] = er.overflow[i];
		for(int i = 0; i < nne; i++) info.negative[i] = er.negative[i];
		for(int i = 0; i < nor; i++) info.outrange[i] = er.outrange[i];

		// receive overflow from other node
		int offset = nov;
		for(int src = 0; src < num_node; src++){
			const int src_nov = arrState[src*4+1];
			if(src == 0 || src_nov == 0) continue;
			para_cmp->Recv(&(info.overflow[offset]),
						   sizeof(CheckValueElem) * src_nov,
						   SKL_ARRAY_DTYPE_CHAR, src, procGrp);
			offset += src_nov;
		}

		// receive negative from other node
		offset = nne;
		for(int src = 0; src < num_node; src++){
			const int src_nne = arrState[src*4+2];
			if(src == 0 || src_nne == 0) continue;
			para_cmp->Recv(&(info.negative[offset]),
						   sizeof(CheckValueElem) * src_nne,
						   SKL_ARRAY_DTYPE_CHAR, src, procGrp);
			offset += src_nne;
		}

		// receive outrange from other node
		offset = nor;
		for(int src = 0; src < num_node; src++){
			const int src_nor = arrState[src*4+3];
			if(src == 0 || src_nor == 0) continue;
			para_cmp->Recv(&(info.outrange[offset]),
						   sizeof(CheckValueElem) * src_nor,
						   SKL_ARRAY_DTYPE_CHAR, src, procGrp);
			offset += src_nor;
		}

		// find global minimum and maximum
		info.min.phi =  REAL_MAX;
		info.max.phi = -REAL_MAX;
		for(int i = 0; i < num_node; i++){
			if(arrState[i*4] == 1){
				const CheckValueElem& max = max_list[i];
				const CheckValueElem& min = min_list[i];
				if(max.phi > info.max.phi) info.max = max;
				if(min.phi < info.min.phi) info.min = min;
			}
		}

	// sender nodes
	}else{

		// send overflow info
		if(er.overflow.size() != 0)
			para_cmp->Send(&(er.overflow.front()),
						   sizeof(CheckValueElem) * er.overflow.size(),
						   SKL_ARRAY_DTYPE_CHAR, 0, procGrp);

		// send negative info
		if(er.negative.size() != 0)
			para_cmp->Send(&(er.negative.front()),
						   sizeof(CheckValueElem) * er.negative.size(),
						   SKL_ARRAY_DTYPE_CHAR, 0, procGrp);

		// send outrange info
		if(er.outrange.size() != 0)
			para_cmp->Send(&(er.outrange.front()),
						   sizeof(CheckValueElem) * er.outrange.size(),
						   SKL_ARRAY_DTYPE_CHAR, 0, procGrp);
	}

	info.mat = mat;
	info.mol = mol;
	return 1;
}

int RicsSaveCheckValueInfo(const char* prefix, const char* udstring, const CheckValueInfo& info)
{
	assert(prefix           != NULL);
    assert(udstring         != NULL);
	assert(info.mat.c_str() != NULL);
	assert(info.mol.c_str() != NULL);

	const CfgInfo* cfg      = UtilGetCfg();
	const SKL_REAL time     = RicsGetCurrentTime();
	const unsigned int step = RicsGetCurrentStep();

	const CfgInfo::Communicator com = cfg->grpAll;
	const int nodeID = com.nid;

	if(nodeID == 0){

		// make file name and
		std::string fname = UtilStrFormat("%s%s%s.csv", prefix,
										  info.mat.c_str(), info.mol.c_str());
		// open file
		FILE*fp = fopen(fname.c_str(), "a");
		if(fp == NULL){
			MESSAGE(UtilStrFormat("Error: can't open file %s", fname.c_str()));
			return 0;
		}

		// write header if this is new file
		if(ftell(fp) == 0){
			fprintf(fp, "step,time,udstring,state,i,j,k,phi,vof,");
			fprintf(fp, "molecules[pieces]\n");
		}

		// arrange write data
		typedef pair<const CheckValueElem*, const char*> column_info;
		vector<column_info> arrInfo;
		arrInfo.push_back(column_info(&info.max, "max"));
		arrInfo.push_back(column_info(&info.min, "min"));
		for(int i = 0; i < info.overflow.size(); i++)
			arrInfo.push_back(column_info(&info.overflow[i], "overflow"));
		for(int i = 0; i < info.negative.size(); i++)
			arrInfo.push_back(column_info(&info.negative[i], "negative"));
		for(int i = 0; i < info.outrange.size(); i++)
			arrInfo.push_back(column_info(&info.outrange[i], "out of range"));
		
		// write file
		for(int i = 0; i < arrInfo.size(); i++){
			const CheckValueElem* pele = arrInfo[i].first;
			const char*           type = arrInfo[i].second;
			
			fprintf(fp, "%d,%e,%s,%s,%d,%d,%d,% e,%e,% e\n",
					step, time, udstring, type, 
					pele->ijk[0], pele->ijk[1], pele->ijk[2],
					pele->phi, pele->vof, pele->mol);
		}

		fclose(fp);
	}
	return 1;
}

#if 1
int TestSetPhi(const char* mat, const char* mol, const SKL_REAL conc, int i, int j, int k)
{
	const Material*  m  = UtilGetMaterialId(mat);
	const Molecular* g  = UtilGetMolecularId(mol);
	const Equation*  gm = UtilGetEquationId(g, m);
	if(m == NULL || g == NULL || gm == NULL){
		MESSAGE(UtilStrFormat("Error: Can't get eq.%s-%s", mat, mol));
		return 0;
	}

	int zi, zj, zk;
	const bool in = UtilCheckIdxPdiv(&zi, &zj, &zk, i, j, k, 1);
	if(!in){
		return 0;
	}

	S4DRN Pgm = UtilGetCfg()->S4Dmap[UtilStrFormat("%s",  gm->name.c_str())];
	S4DRN Vfm = UtilGetCfg()->S4Dmap[UtilStrFormat("Vf_%s",m->name.c_str())];
	Pgm.ary->val(zi,zj,zk,gm->id) = conc;

	return 1;
}
#endif

/////////////////////////////////////////////////////////////////////////////
// molecules number 
int RicsGetNumberOfMolecules(const char* mat, const char* mol, SKL_REAL* nmols, int i, int j, int k)
{
	const Material*  m  = UtilGetMaterialId(mat);
	const Molecular* g  = UtilGetMolecularId(mol);
	const Equation*  gm = UtilGetEquationId(g, m);
	if(!(m && g && gm)) { return -1; }

	int zi,zj,zk;
	const bool in = UtilCheckIdxPdiv(&zi,&zj,&zk,i,j,k,1);
	if (!in) { return 0; }

	const CfgInfo* cfg = UtilGetCfg();
	const SKL_REAL K   = cfg->cvol * pow(cfg->Ls, 3) * cfg->Kna * cfg->NA;
	
	S4DRN Pgm = UtilGetCfg()->S4Dmap[UtilStrFormat("%s",gm->name.c_str())];
	*nmols = Pgm.ary->val(zi,zj,zk,gm->id) * K;

	return 1;
}

int RicsGetMoleNumber(const char* mat, const char* mol, SKL_REAL* nmols, int i, int j, int k)
{
	const Material*  m  = UtilGetMaterialId(mat);
	const Molecular* g  = UtilGetMolecularId(mol);
	const Equation*  gm = UtilGetEquationId(g, m);
	if(!(m && g && gm)) { return -1; }

	int zi,zj,zk;
	const bool in = UtilCheckIdxPdiv(&zi,&zj,&zk,i,j,k,1);
	if (!in) { return 0; }

	const CfgInfo* cfg = UtilGetCfg();
	const SKL_REAL V   = cfg->cvol * pow(cfg->Ls, 3);
	
	S4DRN Pgm = UtilGetCfg()->S4Dmap[UtilStrFormat("%s",gm->name.c_str())];
	*nmols = Pgm.ary->val(zi,zj,zk,gm->id) * V;

	return 1;
}

int _GetTotalMoleculesByDivision(const char*mat, const char* mol, MOLECULES_UNIT unit, SKL_REAL* nmols)
{
	assert(mat != NULL);
	assert(mol != NULL);
	assert(unit == PIECES || unit == MOLE);

	// check whether given equation is valid or not.
	const CfgInfo*   cfg = UtilGetCfg();
	const Material*  m   = UtilGetMaterialId(mat);
	const Molecular* g   = UtilGetMolecularId(mol);
	const Equation*  gm  = UtilGetEquationId(g, m);
	if(m == NULL || g == NULL || gm == NULL){
		MESSAGE(UtilStrFormat("Error: Can't get eq.%s-%s", mat, mol));
		return 0;
	}

	// check wether equation "gm" assiend in this node or not.
	const EqnSet& AssEqns = cfg->AssEqns;
	if(find(AssEqns.begin(), AssEqns.end(), gm) == AssEqns.end()){
//		MESSAGE(UtilStrFormat("Don't be assigned eq.%s-%s", mat, mol));
		return 0;
	}

	SKL_REAL K;
	if(unit == PIECES) K = cfg->cvol * pow(cfg->Ls, 3) * cfg->Kna * cfg->NA;
	if(unit == MOLE)   K = cfg->cvol * pow(cfg->Ls, 3);
	
	const int gmid = gm->id;
	const int mid  =  m->id;
	
	const std::string eq     = UtilStrFormat("%s",   gm->name.c_str());
	const std::string ma     = UtilStrFormat("Vf_%s", m->name.c_str());
	const S4D_REAL*   arrphi = UtilGetCfg()->S4Dmap[eq].ary;
	const S4D_REAL*   arrvof = UtilGetCfg()->S4Dmap[ma].ary;
	
	// get local ijk
	const CfgInfo::Communicator com = cfg->grpP;
	const int procGrp = com.gid;
	const int nodeID  = com.nid;
#ifdef SPHERE175
	const SklParaComponent* para_cmp = cfg->solver->SklGetParaComponent();
	const SklParaManager*   para_mng = para_cmp->GetParaManager();
	const unsigned int*     ijkmax   = para_mng->GetVoxelSize(procGrp);
#else
	const unsigned int* ijkmax = cfg->solver->ParaMngr.GetVoxelSize(procGrp);
#endif
	if(ijkmax == NULL){
		MESSAGE("Error: GetVoxelSize returns NULL!");
		return 0;
	}
	
	// 
	// loop on local ijk space
	//
	SKL_REAL total = 0.0;
	for(int k = 0; k < ijkmax[2]; k++){
	for(int j = 0; j < ijkmax[1]; j++){
	for(int i = 0; i < ijkmax[0]; i++){
		total += arrphi->val(i, j, k, gmid) * arrvof->val(i, j, k, mid) * K;
	}}}
	*nmols = total;
	return 1;
}

int _GetTotalMolecules(const char*mat, const char* mol, MOLECULES_UNIT unit, SKL_REAL* nmols)
{
	assert(mat != NULL);
	assert(mol != NULL);

	// in case of serial, division is only entire region.
	const CfgInfo* cfg = UtilGetCfg();
	if(!cfg->Parallel){
		return _GetTotalMoleculesByDivision(mat, mol, unit, nmols);
	}

#ifdef SPHERE175
	SklParaComponent* para_cmp = cfg->solver->SklGetParaComponent();
	SklParaManager*   para_mng = para_cmp->GetParaManager();
#endif

	// get local total on each divided region.
	SKL_REAL local_total = 0.0;
	_GetTotalMoleculesByDivision(mat, mol, unit, &local_total);

	// sum local total value into node-nID;
	SKL_REAL global_total = 0.0;
	para_mng->Allreduce((void*)&local_total, (void*)&global_total, 1,
						SKL_ARRAY_DTYPE_REAL, SKL_SUM, cfg->grpAll.gid);
	*nmols = global_total;
	return 1;
}

int RicsGetTotalNumberOfMolecules(const char* mat, const char* mol, SKL_REAL* total)
{
	return _GetTotalMolecules(mat, mol, PIECES, total);
}

int RicsGetTotalMoleNumber(const char* mat, const char*mol, SKL_REAL* total)
{
	return _GetTotalMolecules(mat, mol, MOLE, total);
}
