/*
 * Copyright (c) RIKEN, Japan. All right reserved. 2007-2009
 */

#ifndef __DATA_TYPE_H__
#define __DATA_TYPE_H__

#include <SklSolverBase.h>
#include <iomanip>
#include <set>
#include <utility>
#include <vector>

#include <fstream>
#include <iostream>
#include <sstream>

#include <Foundation/element.h>

#define S3D_SIZE(I,J,K,vc) \
	(((I)+2*(vc))*((J)+2*(vc))*((K)+2*(vc)))
#define S3D_INDEX(I,J,K,i,j,k,vc) \
	  (((I)+2*(vc))*((J)+2*(vc))*((k)+(vc)) \
	+ ((I) +2*(vc))*((j)+(vc)) \
	+ ((i) +(vc)))

#define EULER 0
#define ADAMS_BASHFORTH 1
#define RUNGE_KUTTA_2 2
#define RUNGE_KUTTA_4 3

typedef std::pair<SKL_REAL,SKL_REAL> RealPair;
typedef std::vector<RealPair> RealPairVector;

struct IOtupple {
	const char* InFileAttr;   // input file attribute
	const char* OutFileAttr;  // output file attribute
	const char* DataLabel;    // aray data label for DataManager
};

enum TypeEnum {
	UNKNOWN,
	MATERIAL,
	MOLECULAR,
	EQUATION,
	MEMBAREA,
	MEMBNRML_X,
	MEMBNRML_Y,
	MEMBNRML_Z,
	FACERATE_MX,
	FACERATE_PX,
	FACERATE_MY,
	FACERATE_PY,
	FACERATE_MZ,
	FACERATE_PZ,
	DIFFLUX,
	METABOTERM,
	MEMBTERM,
	SRCTERM,
	MEMBEXTERM,
    POTENTIAL,
};

/////////////////////////////////////////////////
// Extension Parameter
union ExtParamValue {
	int INT;
	SKL_REAL REAL;
	const char* STR;
};

struct ExtParam {
	string name;
	ExtParamValue value;
	int id;
	bool extid;
	const char* comment;
};

typedef map<string, const ExtParam*> ExtParamMap;

/////////////////////////////////////////////////
// Array Map
struct S4DRN {
	S4D_REAL* ary;
	int N;
};
typedef map<string,S4DRN> S4DMapT;


//////////////////////////////////////////////////
// JobManageTable

typedef vector<const Equation*>  EqnSet;
typedef vector<const Material*>  MatSet;
typedef vector<const Molecular*> MolSet;

//#if defined(__INTEL_COMPILER) || defined(__GNUC__)
//typedef pair<EqnSet, unsigned int[4]>  EqnAssign;
//typedef pair<MatSet, unsigned int[4]>  MatAssign;
//#else // __INTEL_COMPILER
template<typename T, int s>
class static_array {
	T val[s];
public:
	      T& operator[] (const int& i)       { return val[i]; }
	const T& operator[] (const int& i) const { return val[i]; }
	operator       T*()       { return static_cast<T*>(val); }
	operator const T*() const { return static_cast<const T*>(val); }
};

typedef pair<EqnSet, static_array<unsigned int, 4> >  EqnAssign;
typedef pair<MatSet, static_array<unsigned int, 4> >  MatAssign;
//#endif // __INTEL_COMPILER

typedef vector<MatAssign>        MatAssignSet;

struct JobEntry {
	EqnAssign    Eqns;
	MatAssignSet Mats;

	const CfgElem* mgrp; // for working
};

typedef vector<JobEntry> JobManageTable;

//////////////////////////////////////////////////

typedef std::pair<TypeEnum, const Element*>  TypElmMapT;
typedef std::map<string, TypElmMapT>         TypMapT;
typedef std::map<unsigned int, const string> Id2NameMapT;
typedef std::map<string, unsigned int>       Name2IdMapT;
typedef std::map<int, unsigned int>          Cid2IdMapT;
typedef std::map<string, SklVoxDataSet*>     OutDataMapT;

struct CfgInfo {
	SklSolverBase* solver;
	ofstream logfile;

	// domain
	bool     fixdt;
	SKL_REAL CourantNumber;
	SKL_REAL dt;
	int time_integ;  // TimeIntegration : Euler, AdamsBashforth, Runge-Kutta
	int RhsoInitFlag; // case AdamsBashforth ...
			  // 0 : Initialize rhso. (copy to rhso from rhs)
			  // 1 : Don't Initialize rhso.
//	SKL_REAL lu;  // Length Unit
//	SKL_REAL tu;  // Time Unit
	SKL_REAL Ls;  // Length Unit of solver
	SKL_REAL Ts;  // Time Unit of solver
	SKL_REAL NA;  // Avogadro's Number
	SKL_REAL Kna; // supplementary unit of Avogadro's Number

	unsigned int size[3];     // size of overall domain
	SKL_REAL     width[3];
	SKL_REAL     origin[3];
	SKL_REAL     pitch[3];
	SKL_REAL     cvol;        // control volume size as cubic meter
//	SKL_REAL     cvol_L;      // control volume size as litter
	SKL_REAL     outpitch[3];
	unsigned int moninterval;

	// mode
	bool CheckMode;
	bool InitStart;
	bool Restart;
	bool Parallel;
	bool SpecialIONode;
	bool GatherIO;
	bool IONode;

	///////////////////////////////////////////////////////////////
	// Assgined Voxel Index
	struct VoxelDivInfo {
//		unsigned int head[3];  //
//		unsigned int tail[3];  //
//		unsigned int size[3];  //
		vxsize_t head[3];  //
		vxsize_t tail[3];  //
		vxsize_t size[3];  //
	};
	VoxelDivInfo Pdiv, Ediv, IOdiv;

	// process group, node id at each groups
	struct Communicator {
		int gid; // group ID
		int nid; // node ID
		int num; // number of nodes
	};
	Communicator grpAll, grpIO, grpComp, grpP, grpE;
	
	///////////////////////////////////////////////////////////////
	EqnSet AssIOEqns; // Equations are assigned  to this node for IO
	EqnSet AssEqns;   // Equations are assigned  to this node for P
	MatSet AssMats;   // Metabolism are assigned to this node for E
	char *eqnIO;      // IO flag set
	char *matIO;      // IO flag set
	char *molIO;      // IO flag set
	char *PeqnAss;    // Eqn Assign flag set for P layer
	char *PmolAss;    // Mol Assign flag set for P layer
	char *EmatAss;    // Mat Assign flag set for E layer

	///////////////////////////////////////////////////////////////
	// material info
	vector<const Material*> MatTbl;
	Name2IdMapT MatNameIdMap;
	Cid2IdMapT  MatCidIdMap;
	size_t      MatCount;
	int         CmpMatId;

	// molecular info
	vector<const Molecular*> MolTbl;
	Name2IdMapT MolNameIdMap;
	Cid2IdMapT  MolCidIdMap;
	size_t      MolCount;

	// equation info
	vector<const Equation*> EqnTbl;
	Name2IdMapT EqnNameIdMap;
	size_t      EqnCount;

	// metabolism info
	vector<const Metabolism*> MtbTbl;
	Name2IdMapT MtbNameIdMap;
	size_t      MtbCount;

	// boundary condition info
	BoundaryCond* BndCndTbl;
	Cid2IdMapT    BndCidIdMap;
	size_t        BndCndCount;

	// membrane info
	vector<const Membrane*> MembTbl;
	Name2IdMapT MembNameIdMap;
	size_t      MembCount;

	// mol-mebm info
	vector<const MolMemb*> MlMbTbl;
	Name2IdMapT MlMbNameIdMap;
	size_t      MlMbCount;

	// memb area info
	vector<const Membarea*> MembareaTbl;
	Name2IdMapT MembareaNameIdMap;
	size_t      MembareaCount;

	// penetration
	vector<const penetrater*> PeneTbl;
	Name2IdMapT PeneNameIdMap;
	size_t PeneCount;

	// channel
	vector<const ChTranser*> ChTrTbl;
	Name2IdMapT ChTrNameIdMap;
	size_t ChTrCount;

	// pump
	vector<const PmTranser*> PmTrTbl;
	Name2IdMapT PmTrNameIdMap;
	size_t PmTrCount;

	// Symport channel
	vector<const CSTranser*> CSTrTbl;
	Name2IdMapT CSTrNameIdMap;
	size_t CSTrCount;

	// Symport pump
	vector<const PSTranser*> PSTrTbl;
	Name2IdMapT PSTrNameIdMap;
	size_t PSTrCount;

	// Antiport channel
	vector<const CATranser*> CATrTbl;
	Name2IdMapT CATrNameIdMap;
	size_t CATrCount;

	// Antiport pump
	vector<const PATranser*> PATrTbl;
	Name2IdMapT PATrNameIdMap;
	size_t PATrCount;

	// Gap Janction
	vector<const GpTranser*> GpTrTbl;
	Name2IdMapT GpTrNameIdMap;
	size_t GpTrCount;

	///////////////////////////////////////////////////////////////
	TypMapT TypMap;

	// Extension Parameter
	ExtParamMap extParams;

	// Used Memory Size
	size_t used_mem;

	///////////////////////////////////////////////////////////////
	// array map
	S4DMapT S4Dmap;

	// steady state info : SS
	int SSinterval;
	int SSstartStep;
	EXEC SSsolvExec;
	bool SSoutSPH;
	OPETYPE SSope;
	map<int, SteadyState*> SSmap;
	size_t SSCount;
	const char* sslogfilename;
};

struct ProbeT {
	string fname;
	SKL_REAL x, y, z;
	int _i, _j, _k; // index for overall domain
	int  i,  j,  k; // index for local domain
	int nID;        // node ID which has probed cell
	unsigned int interval;
	vector<string> obj;
};

struct SystemMemoryInfo {
	unsigned int ID;      // ID
	pid_t        pid;     // process id
	unsigned int VmPeak;  // peak virtual memory [kB]
	unsigned int VmSize;  // current virtual memory [kB]
	unsigned int VmHWM;   // peak hardware memory [kB]
	unsigned int VmRSS;   // current hardware memory [kB]
	unsigned int Shared;  // shared (for example, shared with *.so) [kB]
};

struct CheckValueElem {
	int ijk[3];    // ijk index of interest cell
	SKL_REAL phi;  // concentration (mol/volume)
	SKL_REAL vof;  // volume fraction of material (ratio of volume)
	SKL_REAL mol;  // number of molecules (pieces)

	CheckValueElem();
	CheckValueElem(int i, int j, int k, SKL_REAL p, SKL_REAL v, SKL_REAL m);
};

struct CheckValueInfo {
	std::string mat;
	std::string mol;

	// min, max value info
	CheckValueElem min;
	CheckValueElem max;

	vector<CheckValueElem> overflow;  // info. for overflow value voxel
	vector<CheckValueElem> negative;  // info. for negative value voxel
	vector<CheckValueElem> outrange;  // info. for out of range value voxel
};

enum MOLECULES_UNIT {
	PIECES,
	MOLE,
};

#endif // __DATA_TYPE_H__

