//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#ifndef _WITHOUT_ECELL

#include <Foundation/error.h>

// EmlModel header
#include "ecellif.h"
#include "EmlModel.h"

// libxml2 headers
#include <libxml/parser.h>
#include <libxml/tree.h>

// libecs headers
//#include <libecs/libecs.hpp>
//#include <libecs/FullID.hpp>
//#include <libecs/Model.hpp>
//#include <libecs/Entity.hpp>

// STL
#include <string>
#include <sstream>
using std::string;
using std::stringstream;

#include <iostream>
using std::cerr;
using std::endl;
#ifdef _DEBUG
#endif

USE_LIBECS;

inline void print_poly(const libecs::Polymorph& v)
{
#ifdef ECELL_3_1
	if (v.getType() == libecs::Polymorph::POLYMORPH_VECTOR) {
#else
        if (v.getType() == libecs::PolymorphValue::TUPLE) {
#endif //ECELL_3_1
		const libecs::PolymorphVector vec = v.asPolymorphVector();
		cerr << "[";
		for (libecs::PolymorphVector::const_iterator i = vec.begin();
				i != vec.end(); i++) {
#ifdef ECELL_3_1
			print_poly(vec);
#else
			print_poly(*i);
#endif //ECELL_3_1
			cerr << ",";
		}
		cerr << "]";
	} else {
		cerr << v.asString();
	}
}

inline static const xmlChar* toxmlChar(const char* str)
{
	return reinterpret_cast<const xmlChar*>(str);
}

inline static const char* tochar(const xmlChar* str)
{
	return reinterpret_cast<const char*>(str);
}

inline static const string tostring(const xmlChar* str)
{
	return string(tochar(str));
}

inline static string trim(const string& s)
{
	string::const_iterator top  = s.begin();
	string::const_iterator tail = s.end();
	while ((top  != s.end())   && isspace(*top))        top++;
	while ((tail != s.begin()) && isspace(*(tail - 1))) tail--;

	if (top > tail) { return string(); }
	return string(top, tail);
}

inline static string trim(const xmlChar* str)
{
	return trim(tochar(str));
}

inline static string tagname(const xmlNodePtr n)
{
	return tostring(n->name);
}

bool _lesser(const EmlModel::entity& e0, const EmlModel::entity& e1)
{
	const string fullid0 = makeid(e0.type, e0.path, e0.id);
	const string fullid1 = makeid(e1.type, e1.path, e1.id);

	return fullid0.compare(fullid1) < 0;
}

static const char* tag_eml       = "eml";
static const char* tag_variable  = "variable";
static const char* tag_process   = "process";
static const char* tag_system    = "system";
static const char* tag_stepper   = "stepper";
static const char* tag_property  = "property";
static const char* tag_value     = "value";
static const char* tag_text      = "text";

static const xmlChar* prop_id    = toxmlChar("id");
static const xmlChar* prop_name  = toxmlChar("name");
static const xmlChar* prop_class = toxmlChar("class");

static const char* empty         = "";
static const char* type_stepper  = "Stepper";
static const char* type_variable = "Variable";
static const char* type_process  = "Process";
static const char* type_system   = "System";

EmlModel::EmlModel()
{
}

EmlModel::EmlModel(const char* emlfile)
{
	open(emlfile);
}

EmlModel::~EmlModel()
{
}

const char* EmlModel::getEmlFileName() const
{
	return filename.c_str();
}

void EmlModel::open(const char* emlfile)
{
	xmlDocPtr doc;
	xmlNodePtr cur, root;

	doc = xmlParseFile(emlfile);
	cur = xmlDocGetRootElement(doc);

	TRY((cur && (tagname(cur) == tag_eml)),
		string("File ") + emlfile + " is not EML");
	find_eml(cur);
	xmlFreeDoc(doc);

	filename = emlfile;

	// sort
	// sort(system_list.begin(),   system_list.end(),   _lesser);
	// sort(variable_list.begin(), variable_list.end(), _lesser);
	// sort(process_list.begin(),  process_list.end(),  _lesser);

	EntityList* el[3] = { &system_list, &variable_list, &process_list };
	for (int i = 0; i < 3; i++) {
	for (EntityList::iterator e = el[i]->begin(); e != el[i]->end(); e++) {
		entity_list.push_back(*e);
	}}
}

void EmlModel::find_eml(xmlNodePtr& eml)
{
	const string pathbase = empty;
	for (xmlNodePtr cur = eml->children; cur; cur = cur->next) {
		const string tag = tagname(cur);
		if (tag == tag_system) {
			find_entity(cur, pathbase);
		} else if (tag == tag_stepper) {
			find_stepper(cur);
		}
	}
}


void EmlModel::find_stepper(xmlNodePtr& stp)
{
	stepper s;
	s.id  = tochar(xmlGetProp(stp, prop_id));
	s.cls = tochar(xmlGetProp(stp, prop_class));
	stepper_list.push_back(s);

	// child elements
	for (xmlNodePtr cur = stp->children; cur; cur = cur->next) {
		const string ctag = tagname(cur);
		if (ctag == tag_property) {
			find_property(cur, makeid(type_stepper, s.id));
		}
	}
}

void EmlModel::find_entity(xmlNodePtr& ent, const string& path)
{
	const string tag = tagname(ent);
	const bool
		ent_is_system   = (tag == tag_system),
		ent_is_process  = (tag == tag_process),
		ent_is_variable = (tag == tag_variable);

	// Entity
	entity e;
	e.path = path;
	e.cls  = tochar(xmlGetProp(ent, prop_class));
	e.id   = tochar(xmlGetProp(ent, prop_id));
	e.type =
		ent_is_process?  type_process:
		ent_is_variable? type_variable:
		ent_is_system?   type_system:
						 empty;

	const string chpath = ent_is_system? e.id: path;
	if (ent_is_system) {
		// "System::/xxx/yyy/zzz" => "System:/xxx/yyy:zzz"
		const string& oid = chpath; //e.id;
		const bool oid_is_root = (oid == "/");
		const size_t
			osize = oid.size(),
			lsep  = oid.find_last_of(path_sep),
			ptail = lsep + ((!oid_is_root && (lsep == 0))?     1: 0),
			idtop = lsep + ((!oid_is_root && (lsep != osize))? 1: 0);
		e.path = string(oid.begin(), oid.begin() + ptail);
		e.id   = string(oid.begin() + idtop, oid.end());
	}
	(ent_is_system?   &system_list:
	 ent_is_process?  &process_list:
	 ent_is_variable? &variable_list:
	 					NULL)->push_back(e);

	// child elements
	for (xmlNodePtr cur = ent->children; cur; cur = cur->next) {
		const string ctag = tagname(cur);
		if ((ctag == tag_process) ||
			(ctag == tag_system)  ||
			(ctag == tag_variable)) {
			find_entity(cur, chpath);
		} else if (ctag == tag_stepper) {
			find_stepper(cur);
		} else if (ctag == tag_property) {
			find_property(cur, makeid(e.type, e.path, e.id));
		}
	}
}

void EmlModel::find_property(xmlNodePtr& p, const string& id)
{
	PolymorphVector pv;
	const string pn  = makeid(id, tochar(xmlGetProp(p, prop_name)));
	const string ppn = tochar(xmlGetProp(p, prop_name));
	
	for (xmlNodePtr cur = p->children; cur; cur = cur->next) { 
		const string tag = tagname(cur);
		if (tag == tag_value) {
			pv.push_back(find_value(cur, makeid(id, tag)));
		}
	}

	if ((pv.size() == 1) && (ppn != "VariableReferenceList")) {
		property_map.insert(PropertyMap::value_type(pn, pv[0]));
	} else {
#ifdef ECELL_3_1
		property_map.insert(PropertyMap::value_type(pn, pv));
#else
		property_map.insert(PropertyMap::value_type(pn, Polymorph(pv)));
#endif //ECELL_3_1
	}
}

Polymorph EmlModel::find_value(xmlNodePtr& p, const string& id)
{
	PolymorphVector pv;

	for (xmlNodePtr cur = p->children; cur; cur = cur->next) {
		const string tag = tagname(cur);
		if (tag == tag_value) {
			pv.push_back(find_value(cur, makeid(id, tag)));
		} else if (tag == tag_text) {
			const string c = trim(cur->content);
			if (!c.empty()) { 
				// convert the path to absolution path
				const size_t i = c.find_first_of(id_sep);
				if (c[i + 1] == '.') {
					const string::const_iterator
						p = id.begin() + id.find_first_of(id_sep) + 1,
						t = p + string(p, id.end()).find_first_of(id_sep);
					const string
						v = makeid(
								type_variable,
								string(p, t),
								string(c.begin() + i + 3, c.end()));
#ifdef ECELL_3_1
					pv.push_back(v);
#else
					pv.push_back(Polymorph(v));
#endif //ECELL_3_1
				} else {
#ifdef ECELL_3_1
					pv.push_back(c);
#else
					pv.push_back(Polymorph(c));
#endif //ECELL_3_1
				}
			}
		}
	}

	return (pv.size() == 1)?  pv[0]: Polymorph(pv);
}

void EmlModel::Dump() const
{
	// Stepper
	for (StepperList::const_iterator s = stepper_list.begin();
			s != stepper_list.end(); s++) {
		//
		MESSAGE(UtilStrFormat("%s %s",
					s->cls.c_str(),
					s->id.c_str()));
	}

	// Entity
	for (EntityList::const_iterator e = entity_list.begin();
			e != entity_list.end(); e++) {
		//
		MESSAGE(UtilStrFormat("%s %s:%s:%s",
					e->type.c_str(),
					e->cls.c_str(),
					e->path.c_str(),
					e->id.c_str()));
	}

	// Property
	for (PropertyMap::const_iterator p = property_map.begin();
			p != property_map.end(); p++) {
		//
		const string    pn = p->first;
		const Polymorph pv = p->second;

		MESSAGE(UtilStrFormat("%s = %s",
				pn.c_str(), poly2string(pv).c_str()));
	}
}

EmlModel::operator const StepperList&() const { return stepper_list; }
EmlModel::operator const EntityList&()  const { return entity_list;  }
EmlModel::operator const PropertyMap&() const { return property_map; }


#endif // _WITHOUT_ECELL

