//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#ifndef _WITHOUT_ECELL
#ifndef USE_EMC
#include <libecs/Model.hpp>

#include <libecs/Entity.hpp>
#include <libecs/LoggerBroker.hpp>
#include <libecs/SystemStepper.hpp>
#endif //USE_EMC

#include <Foundation/error.h>
#include <Foundation/util.h>

#include <sstream>
#include <iostream>

#include "ecellif.h"
#include "emcsim.h"
#include "EmlModel.h"

#ifndef ECELL_3_1
#include <cstdlib>
#endif //ECELL_3_1

using std::string;
using std::stringstream;
using std::set;

emcsim::emcsim() 

#ifndef USE_EMC
	:
#ifdef ECELL_3_1
#if ECELL_VERSION_INT >= VERSION_INT(3,1,106)
	propertiedObjectMaker(),
#endif //ECELL_VERSION_INT
#else
	propertiedObjectMaker(libecs::createDefaultModuleMaker()),
#endif //ECELL_3_1
	running(false),
	dirty(true),
	eventCheckInterval(20),
	eventChecker(),
	eventHandler()
#endif //USE_EMC

{
	_hasModel = false;

#ifndef USE_EMC

#ifdef ECELL_3_1
#if ECELL_VERSION_INT >= VERSION_INT(3,1,106)
	model = new libecs::Model(propertiedObjectMaker);
#else  //ECELL_VERSION_INT
	model = new libecs::Model();
#endif //ECELL_VERSION_INT
#else
	model = new libecs::Model(*propertiedObjectMaker);

	 const char* const path(std::getenv("LD_LIBRARY_PATH")); //01
	model->libecs::Model::setup();

	model->libecs::Model::setDMSearchPath(path ? path : "");

#endif //ECELL_3_1

	clearEventChecker();
#endif //USE_EMC

}

emcsim::~emcsim() {

#ifndef USE_EMC
#ifdef ECELL_3_1
	delete model;
#endif //ECELL_3_1

	//delete &model;
#endif //USE_EMC

}

#ifndef USE_EMC
#ifdef ECELL_3_1
void emcsim::setEventChecker(EventCheckerSharedPtrCref anEventChecker) {
#else
void emcsim::setEventChecker(boost::shared_ptr<EventChecker> const& anEventChecker) {
#endif //ECELL_3_1
	eventChecker = anEventChecker;
}

void emcsim::clearEventChecker() {
#ifdef ECELL_3_1
	setEventChecker(EventCheckerSharedPtr(new DefaultEventChecker()));
#else
	setEventChecker(boost::shared_ptr<EventChecker>(new DefaultEventChecker()));
#endif //ECELL_3_1
}


#ifdef ECELL_3_1
libecs::ModelRef emcsim::getModel() {
#else
libecs::Model& emcsim::getModel() {
#endif //ECELL_3_1
	return *model;
}

void emcsim::initialize() const {
	const_cast<emcsim *>(this)->model->initialize();
}

const bool emcsim::isDirty() const {
#ifdef ECELL_3_1
	return dirty;
#else
	return model->isDirty();
#endif //ECELL_3_1
}

void emcsim::clearDirty() const {
	if(emcsim::isDirty()){
		emcsim::initialize();
		//interruptAll();

		dirty = false;
	}
}

void emcsim::setDirty() {
#ifdef ECELL_3_1
	dirty = true;
#else
	model->markDirty();
#endif //ECELL_3_1
}


#ifdef ECELL_3_1
void emcsim::createStepper(libecs::StringCref aClassname, libecs::StringCref anId) {
#else
void emcsim::createStepper(libecs::String const& aClassname, libecs::String const& anId) {
#endif //ECELL_3_1
	if(running){
		THROW_EXCEPTION(libecs::Exception, "Cannot create a Stepper while running.");
	}

	emcsim::setDirty();
	model->createStepper(aClassname, anId);
}

const libecs::Polymorph emcsim::getStepperList() const {
#ifdef ECELL_3_1
	libecs::StepperMapCref stepperMap(const_cast<emcsim *>(this)->model->getStepperMap());
#else
	libecs::Model::StepperMap const& stepperMap(const_cast<emcsim *>(this)->model->getStepperMap());
#endif //ECELL_3_1

	libecs::PolymorphVector polymorphVector;
	polymorphVector.reserve(stepperMap.size());

#ifdef ECELL_3_1
	for(libecs::StepperMapConstIterator i(stepperMap.begin()); i != stepperMap.end(); ++i){
		polymorphVector.push_back(libecs::String((*i).first));
#else
	for(libecs::Model::StepperMap::const_iterator i(stepperMap.begin()); i != stepperMap.end(); ++i){
		polymorphVector.push_back(libecs::Polymorph((*i).first));
#endif //ECELL_3_1
	}

#ifdef ECELL_3_1
	return polymorphVector;
#else
	return libecs::Polymorph( polymorphVector );
#endif //ECELL_3_1
}

#ifdef ECELL_3_1
const libecs::Polymorph emcsim::getStepperPropertyList(libecs::StringCref aStepperID) const {
	libecs::StepperPtr stepperPtr(const_cast<emcsim *>(this)->model->getStepper(aStepperID));

	return stepperPtr->getPropertyList();
#else
const libecs::Polymorph emcsim::getStepperPropertyList(libecs::String const& aStepperID) const {
	libecs::Stepper* stepperPtr(const_cast<emcsim *>(this)->model->getStepper(aStepperID));

	return libecs::Polymorph(stepperPtr->getPropertyList());
#endif //ECELL_3_1
}

#ifdef ECELL_3_1
const libecs::Polymorph emcsim::getStepperProperty(libecs::StringCref aStepperID, libecs::StringCref aPropertyName) const {
	libecs::StepperCptr stepperPtr(const_cast<emcsim *>(this)->model->getStepper(aStepperID));
#else
const libecs::Polymorph emcsim::getStepperProperty(libecs::String const& aStepperID, libecs::String const& aPropertyName) const {
	libecs::Stepper const* stepperPtr(const_cast<emcsim *>(this)->model->getStepper(aStepperID));
#endif //ECELL_3_1

#ifdef ECELL_3_1
	emcsim::clearDirty();
#endif //ECELL_3_1


	return stepperPtr->getProperty(aPropertyName);
}

#ifdef ECELL_3_1
void emcsim::loadStepperProperty(libecs::StringCref aStepperID, libecs::StringCref aPropertyName, libecs::PolymorphCref aValue) {
	libecs::StepperPtr stepperPtr(model->getStepper(aStepperID));
#else
void emcsim::loadStepperProperty(libecs::String const& aStepperID, libecs::String const& aPropertyName, libecs::Polymorph const & aValue) {
	libecs::Stepper* stepperPtr(model->getStepper(aStepperID));
#endif //ECELL_3_1

	emcsim::setDirty();
	stepperPtr->loadProperty(aPropertyName, aValue);
}

#ifdef ECELL_3_1
void emcsim::createEntity(libecs::StringCref aClassname, libecs::StringCref aFullIDString) {
#else
void emcsim::createEntity(libecs::String const& aClassname, libecs::String const& aFullIDString) {
#endif //ECELL_3_1
	if(running){
		THROW_EXCEPTION(libecs::Exception, "Cannot create an Entity while running.");
	}

	emcsim::setDirty();
	model->createEntity(aClassname, libecs::FullID(aFullIDString));
}

#ifdef ECELL_3_1
const libecs::Polymorph emcsim::getEntityPropertyList(libecs::StringCref aFullIDString) const {
	libecs::FullID fullID(aFullIDString);
	libecs::EntityCptr entityPtr(const_cast<emcsim *>(this)->model->getEntity(fullID));

	return entityPtr->getPropertyList();
}
#else
const libecs::Polymorph emcsim::getEntityPropertyList(libecs::String const& aFullIDString) const {
	libecs::FullID fullID(aFullIDString);
	libecs::Entity const* entityPtr(const_cast<emcsim *>(this)->model->getEntity(fullID));

	return libecs::Polymorph(entityPtr->getPropertyList());
}
#endif //ECELL_3_1

#ifdef ECELL_3_1
void emcsim::loadEntityProperty(libecs::StringCref aFullPNString, libecs::PolymorphCref aValue) {
	libecs::FullPN fullPN(aFullPNString);
	libecs::EntityPtr entityPtr(model->getEntity(fullPN.getFullID()));
#else
void emcsim::loadEntityProperty(libecs::String const& aFullPNString, libecs::Polymorph const & aValue) {
	libecs::FullPN fullPN(aFullPNString);
	libecs::Entity* entityPtr(model->getEntity(fullPN.getFullID()));
#endif //ECELL_3_1

	emcsim::setDirty();
	entityPtr->loadProperty(fullPN.getPropertyName(), aValue);
}

#ifdef ECELL_3_1
void emcsim::setEntityProperty(libecs::StringCref aFullPNString, libecs::PolymorphCref aValue) {
	libecs::FullPN fullPN(aFullPNString);
	libecs::EntityPtr entityPtr(model->getEntity(fullPN.getFullID()));
#else
void emcsim::setEntityProperty(libecs::String const& aFullPNString, libecs::Polymorph const & aValue) {
	libecs::FullPN fullPN(aFullPNString);
	libecs::Entity* entityPtr(model->getEntity(fullPN.getFullID()));
#endif //ECELL_3_1

	emcsim::setDirty();
	entityPtr->setProperty(fullPN.getPropertyName(), aValue);
}

#ifdef ECELL_3_1
const libecs::Polymorph emcsim::getEntityProperty(libecs::StringCref aFullPNString) const {
	libecs::FullPN fullPN( aFullPNString );
	//libecs::EntityCptr entityPtr(model->getEntity(fullPN.getFullID()));
	libecs::EntityCptr entityPtr(const_cast<emcsim *>(this)->model->getEntity(fullPN.getFullID()));
#else
const libecs::Polymorph emcsim::getEntityProperty(libecs::String const& aFullPNString) const {
	libecs::FullPN fullPN( aFullPNString );
	//libecs::Entity const* entityPtr(model->getEntity(fullPN.getFullID()));
	libecs::Entity const* entityPtr(const_cast<emcsim *>(this)->model->getEntity(fullPN.getFullID()));
#endif //ECELL_3_1
#ifdef ECELL_3_1
	emcsim::clearDirty();
#endif //ECELL_3_1
	return entityPtr->getProperty(fullPN.getPropertyName());
}

#ifdef ECELL_3_1
void emcsim::createLogger(libecs::StringCref aFullPNString) {
#else
void emcsim::createLogger(libecs::String const& aFullPNString) {
#endif //ECELL_3_1
#ifdef ECELL_3_1
	libecs::PolymorphVector vector;
	vector.push_back(libecs::Integer(1));
	vector.push_back(libecs::Real(0.0));
	vector.push_back(libecs::Integer(0));
	vector.push_back(libecs::Integer(0));
	emcsim::createLogger(aFullPNString, libecs::Polymorph(vector));
#else
	emcsim::createLogger(aFullPNString, boost::make_tuple( 1l, 0.0, 0l, 0l ) );
#endif //ECELL_3_1
}

#ifdef ECELL_3_1
void emcsim::createLogger(libecs::StringCref aFullPNString, libecs::Polymorph aParamList) {
#else
void emcsim::createLogger(libecs::String const& aFullPNString, libecs::Polymorph aParamList) {
#endif //ECELL_3_1
#ifdef ECELL_3_1
	if(running){
		THROW_EXCEPTION(libecs::Exception, "Cannot create a Logger while running.");
	}

	if(aParamList.getType() != libecs::Polymorph::POLYMORPH_VECTOR){
		THROW_EXCEPTION(libecs::Exception, "2nd argument of createLogger must be a list.");
	}

	libecs::FullPN fullPN(aFullPNString);

	emcsim::clearDirty();

	model->getLoggerBroker().createLogger(fullPN, aParamList.asPolymorphVector());

	emcsim::setDirty();
#else
	typedef libecs::PolymorphValue::Tuple Tuple;

	if(running){
		THROW_EXCEPTION(libecs::Exception, "cannot create a Logger during simulator");
	}

	if(aParamList.getType() != libecs::PolymorphValue::TUPLE
           || aParamList.as< Tuple const& >().size() != 4){
		THROW_EXCEPTION(libecs::Exception, "second argument must be a tuple of 4 items");
        }

	model->getLoggerBroker().createLogger(
		libecs::FullPN(aFullPNString),
		libecs::Logger::Policy(
			aParamList.as< Tuple const& >()[0].as< libecs::Integer >(),
			aParamList.as< Tuple const& >()[1].as< libecs::Real >(),
			aParamList.as< Tuple const& >()[2].as< libecs::Integer >(),
			aParamList.as< Tuple const& >()[3].as< libecs::Integer >() != 0 ? true : false));

	emcsim::setDirty();
#endif //ECELL_3_1
}

#ifdef ECELL_3_1
libecs::LoggerPtr emcsim::getLogger(libecs::StringCref aFullPNString) const {
#else
libecs::Logger* emcsim::getLogger(libecs::String const& aFullPNString) const {
#endif //ECELL_3_1
	return const_cast<emcsim *>(this)->model->getLoggerBroker().getLogger(aFullPNString);
}

#ifdef ECELL_3_1
const libecs::DataPointVectorSharedPtr emcsim::getLoggerData(libecs::StringCref aFullPNString) const {
#else
//const boost::shared_ptr<libecs::DataPointVector> emcsim::getLoggerData(libecs::StringCref aFullPNString) const{
const boost::shared_ptr<libecs::DataPointVector> emcsim::getLoggerData(libecs::String const& aFullPNString) const{
#endif //ECELL_3_1
	return emcsim::getLogger(aFullPNString)->getData();
}

void emcsim::start(){
#ifdef ECELL_3_1
	emcsim::clearDirty();
#endif //ECELL_3_1
	running = true;
}

void emcsim::stop() {
	running = false;

	model->flushLoggers();
}

#ifdef ECELL_3_1
void emcsim::handleEvent() {
	if((*eventChecker)()) {
		do {
			(*eventHandler)();
		} while((*eventChecker)());

		clearDirty();
	}
}
#else
inline void emcsim::handleEvent() {
	while((*eventChecker)()){
		(*eventHandler)();
	}
}
#endif //ECELL_3_1

#ifdef ECELL_3_1
void emcsim::runWithEvent() {
	libecs::StepperCptr const systemStepper(model->getSystemStepper());
#else
void emcsim::runWithEvent(libecs::Real const stopTime) {
#endif //ECELL_3_1

	do {
		unsigned int counter(eventCheckInterval);
		do {
#ifdef ECELL_3_1
			if(model->getTopEvent().getStepper() == systemStepper) {
				model->step();
#else
			if(model->getTopEvent().getTime() > stopTime) {
#endif //ECELL_3_1
				emcsim::stop();
				return;
			}

			model->step();

			--counter;
		} while(counter != 0);

		emcsim::handleEvent();

	} while(running);

#ifdef ECELL_3_1
	return;	//the exit
#endif //ECELL_3_1
}

#ifdef ECELL_3_1
void emcsim::runWithoutEvent() {
#else
void emcsim::runWithoutEvent(libecs::Real const stopTime) {
#endif //ECELL_3_1
#ifdef ECELL_3_1
	libecs::StepperCptr const systemStepper(model->getSystemStepper());
#else
	libecs::Stepper const* const systemStepper(model->getSystemStepper());
#endif //ECELL_3_1

	do {
#ifdef ECELL_3_1
		if(model->getTopEvent().getStepper() == systemStepper) {
			model->step();
#else
		if(model->getTopEvent().getTime() > stopTime) {
#endif //ECELL_3_1
			emcsim::stop();
			return;	// the only exit
		}

		model->step();

	} while(1);
}

void emcsim::run(const libecs::Real aDuration) {
	if(aDuration <= 0.0){
#ifdef ECELL_3_1
		THROW_EXCEPTION( libecs::Exception, "run(1) : 1 must be > 0.0. (" +
				libecs::stringCast(aDuration) + " given.)");
#else
		THROW_EXCEPTION( libecs::Exception,
				"duration must be greater than 0 ("
				+ libecs::stringCast(aDuration) + " given)");
#endif //ECELL_3_1
	}

	emcsim::start();

#ifdef ECELL_3_1
	const libecs::Real stopTime(model->getCurrentTime() + aDuration);
#else
	const libecs::Real aCurrentTime(model->getCurrentTime());
	const libecs::Real stopTime(aCurrentTime + aDuration);
#endif //ECELL_3_1

	// setup SystemStepper to step at stopTime
	
	// FIXME: dirty, ugly!
	
	//libecs::StepperPtr systemStepper(model->getSystemStepper());
#ifdef ECELL_3_1
	libecs::SystemStepperPtr systemStepper(model->getSystemStepper());
	systemStepper->setCurrentTime(stopTime);
	systemStepper->setStepInterval(0.0);
#else
	//libecs::StepperPtr systemStepper(model->getSystemStepper());
	libecs::Stepper* systemStepper(model->getSystemStepper());
	systemStepper->setCurrentTime(aCurrentTime);
	systemStepper->setNextTime(stopTime);
#endif //ECELL_3_1

	model->getScheduler().updateEvent(0, stopTime);


#ifdef ECELL_3_1
	if((typeid(*eventChecker) != typeid(DefaultEventChecker)) && (eventHandler.get() != NULLPTR)) {
#else
	if((typeid(*eventChecker) != typeid(DefaultEventChecker)) && (eventHandler.get() != 0)) {
#endif //ECELL_3_1
#ifdef ECELL_3_1
		runWithEvent();
#else
		runWithEvent(stopTime);
#endif //ECELL_3_1
	}
	else {
#ifdef ECELL_3_1
		runWithoutEvent();
#else
		runWithoutEvent(stopTime);
#endif //ECELL_3_1
	}
}


#endif //USE_EMC

bool emcsim::isExistEntityProperty(const std::string& pn) const
{
	bool f = true;
	try { getEntityProperty(pn); }
	catch (...) { f = false; }
	return f;
}

int  emcsim::testModel(const EmlModel& eml)
{
	loadModel(eml, true);

	return 1;
}

void emcsim::loadModel(const EmlModel& eml, const bool msg)
{
	set<string> e;
	e.insert("/");
	loadModel(eml, e, msg);
}

void emcsim::loadModel(const EmlModel& eml, const set<string>& enable, const bool msg)
{
	const char* emlname = eml.getEmlFileName();

	if (!(static_cast<const StepperList&>(eml).size() &&
			static_cast<const EntityList&>(eml).size())) {
		//
		MSG_WARNING(string(emlname) + "' has no stepper or entity.");
		return;
	}

	// apply stepper
	if (msg) {
		DB1(MESSAGE(string("Load Stepper Setting from ") + emlname));
	}
	const StepperList& sl = eml;
	for (StepperList::const_iterator s = sl.begin(); s != sl.end(); s++) {
		DB3(MESSAGE("create " + s->id));
		createStepper(s->cls, s->id);
		StepperNameList.push_back(s->id);
	}
	
	// apply entity
	if (msg) {
		DB1(MESSAGE(string("Load Entities from ") + emlname));
	}
	const EntityList& el = eml;
	for (EntityList::const_iterator e = el.begin(); e != el.end(); e++) {
		const std::string fid = makeid(e->type, e->path, e->id);
		const std::string type = fid.substr(0, fid.find_first_of(id_sep));
		const std::string id   = fid.substr(type.length() + 1);
		const std::string name = fid.substr(fid.find_last_of(id_sep) + 1);

		// select System Entity
		bool entity_enable =  (type == "System")
							| (fid  == "Variable:/:SIZE");
		for (set<string>::const_iterator
				sys = enable.begin(); sys != enable.end(); sys++) { 
			//  is the entity enable?
			entity_enable |= (id.find(*sys,0) == 0);
		}
		if (!entity_enable) { continue; }

		DB2(MESSAGE("create " + fid));

		// Create Entities fid
		if (fid != "System::/") { createEntity(e->cls, fid); }

		// Regist the Entity fid
		if (type == "System")        { SystemNameList.push_back(fid);   }
		else if (type == "Variable") { VariableNameList.push_back(fid); }
		else if (type == "Process")  { ProcessNameList.push_back(fid);  }
	}

	// apply property
	const PropertyMap& pm = eml;
	DB1(MESSAGE(string("Apply Poperties from ") + emlname));

	// stepper
	for (PropertyMap::const_iterator p = pm.begin(); p != pm.end(); p++) {

		const std::string& fullpn = p->first;
		const size_t fsep = fullpn.find_first_of(id_sep);
		const size_t lsep = fullpn.find_last_of(id_sep);
		const std::string cls = fullpn.substr(0, fsep);
		const std::string id  = fullpn.substr(fsep + 1, lsep - fsep - 1);
		const std::string pn  = fullpn.substr(lsep + 1);

		if (cls != "Stepper") { continue; }

		if (msg) { DB2(MESSAGE(fullpn + " = " + p->second.asString())); }
		loadStepperProperty(id, pn, p->second);
	}

	// variable size (heighest priority in each entities)
	for (PropertyMap::const_iterator p = pm.begin(); p != pm.end(); p++) {

		const std::string& fullpn = p->first;
		const size_t fsep = fullpn.find_first_of(id_sep);
		const size_t lsep = fullpn.find_last_of(id_sep);
		const std::string cls = fullpn.substr(0, fsep);
		const std::string id  = fullpn.substr(fsep + 1, lsep - fsep - 1);
		const std::string pn  = fullpn.substr(lsep + 1);

		const std::string name = id.substr(id.find_last_of(id_sep) + 1);
		if ((cls != "Variable") || (name != "SIZE")) { continue;}

		if (msg) { DB2(MESSAGE(fullpn + " = " + p->second.asString())); }
		loadEntityProperty(fullpn, p->second);

		if (pn == "Value") {
#ifdef ECELL_3_1
			const libecs::Real size = p->second.asReal();
#else
			const libecs::Real size = p->second.as<libecs::Real>();
#endif //ECELL_3_1
			TRY((size > libecs::Real(0.0)),
				std::string(eml.getEmlFileName())+':'+fullpn+" is 0.0 or lesser");
			OriginalSize.insert(OriginalSizeT::value_type(fullpn,size));
		}
	}

	// system
	for (PropertyMap::const_iterator p = pm.begin(); p != pm.end(); p++) {

		const std::string& fullpn = p->first;
		const size_t fsep = fullpn.find_first_of(id_sep);
		const size_t lsep = fullpn.find_last_of(id_sep);
		const std::string cls = fullpn.substr(0, fsep);
		const std::string id  = fullpn.substr(fsep + 1, lsep - fsep - 1);
		const std::string pn  = fullpn.substr(lsep + 1);

		if (cls != "System")  { continue; }

		// Entity Property
		bool entity_enable = (cls == "System");
		for (set<string>::const_iterator
				sys = enable.begin(); sys != enable.end(); sys++) { 
			//  is the entity enable?
			entity_enable |= (id.find(*sys,0) == 0);
		}
		if (!entity_enable) { continue; }
		if (msg) { DB2(MESSAGE(fullpn + " = " + p->second.asString())); }
		loadEntityProperty(fullpn, p->second);
	}


	// other properties
	for (PropertyMap::const_iterator p = pm.begin(); p != pm.end(); p++) {

		const std::string& fullpn = p->first;
		const size_t fsep = fullpn.find_first_of(id_sep);
		const size_t lsep = fullpn.find_last_of(id_sep);
		const std::string cls = fullpn.substr(0, fsep);
		const std::string id  = fullpn.substr(fsep + 1, lsep - fsep - 1);
		const std::string pn  = fullpn.substr(lsep + 1);

		const std::string name = id.substr(id.find_last_of(id_sep) + 1);
		if ((cls == "Variable") && (name == "SIZE")) { continue;}
		if (cls == "Stepper") { continue; }
		if (cls == "System")  { continue; }

		// Entity Property
		if (msg) { DB2(MESSAGE(fullpn + " = " + p->second.asString())); }
		loadEntityProperty(fullpn, p->second);
	}

	_hasModel = true;
}

void emcsim::setSize(const libecs::Real rootSize)
{
	// resize E-Cell model with rootSize
	const libecs::Real rtSize = OriginalSize["Variable:/:SIZE:Value"];
	const libecs::Real scale = rootSize / rtSize;
	for (OriginalSizeT::iterator s = OriginalSize.begin();
			s != OriginalSize.end(); s++) {

		const std::string& ppn  = s->first;
		const libecs::Real size = s->second * scale;
#ifdef ECELL_3_1
		loadEntityProperty(ppn, size);
#else
		loadEntityProperty(ppn, libecs::Polymorph(size));
#endif //ECELL_3_1
	}
}

bool emcsim::hasModel() const { return _hasModel; }

void emcsim::Dump() {
	// stepprt
	const libecs::Polymorph sl = getStepperList();
#ifdef ECELL_3_1
	const libecs::PolymorphVector sv = sl.asPolymorphVector();
#else
	const libecs::PolymorphVector sv = sl.as<libecs::PolymorphVector>();
#endif //ECELL_3_1
	
	for (int i = 0; i < sv.size(); i++) {
#ifdef ECELL_3_1
		const string sn = sv[i].asString();
		const libecs::Polymorph spl = getStepperPropertyList(sn);
		const libecs::PolymorphVector spv = spl.asPolymorphVector();
#else
		const string sn = sv[i].as<libecs::String>();
		const libecs::Polymorph spl = getStepperPropertyList(sn);
		const libecs::PolymorphVector spv = spl.as<libecs::PolymorphVector>();
#endif //ECELL_3_1

		for (int j = 0; j < spv.size(); j++) {
#ifdef ECELL_3_1
			const string spn = spv[j].asString();
#else
			const string spn = spv[j].as<libecs::String>();
#endif //ECELL_3_1
			try {
				const libecs::Polymorph pv = getStepperProperty(sn,spn);
				MESSAGE(UtilStrFormat("%s:%s = %s",
							sn.c_str(), spn.c_str(),
							poly2string(pv).c_str()));
			} catch (...) { }
		}
	}

	// entity
	const EntityNameListT* el[3] = {
		&SystemNameList, &VariableNameList, &ProcessNameList
	};
	for (int i = 0; i < 3; i++) {
	for (EntityNameListT::const_iterator
			e = el[i]->begin(); e != el[i]->end(); e++) {

		const string fullid = *e;
		const libecs::Polymorph epl = getEntityPropertyList(fullid);
#ifdef ECELL_3_1
		const libecs::PolymorphVector epv = epl.asPolymorphVector();
#else
		const libecs::PolymorphVector epv = epl.as<libecs::PolymorphVector>();
#endif //ECELL_3_1

		for (int j = 0; j < epv.size(); j++) {
#ifdef ECELL_3_1
			const string epn    = epv[j].asString();
#else
			const string epn    = epv[j].as<libecs::String>();
#endif //ECELL_3_1
			const string fullpn = UtilStrFormat("%s:%s",
									fullid.c_str(),epn.c_str());
			if (isExistEntityProperty(fullpn)) {
				const libecs::Polymorph pv = getEntityProperty(fullpn);
				MESSAGE(UtilStrFormat("%s = %s",
						fullpn.c_str(),
						poly2string(pv).c_str()));
			}
		}
	}}
}

const emcsim::StepperNameListT& emcsim::getStepperNameList() const
{ return StepperNameList; }

const emcsim::EntityNameListT&  emcsim::getSystemNameList() const
{ return SystemNameList; }

const emcsim::EntityNameListT&  emcsim::getVariableNameList() const
{ return VariableNameList; }

const emcsim::EntityNameListT&  emcsim::getProcessNameList() const
{ return ProcessNameList; }

void change_PolymorphType(libecs::Polymorph& p, const libecs::Polymorph& t)
{
	const libecs::Polymorph::Type type = t.getType();
#ifdef ECELL_3_1
	if (type == libecs::Polymorph::NONE) { return; }
#else
	if (type == libecs::PolymorphValue::NONE) { return; }
#endif //ECELL_3_1
#ifdef ECELL_3_1
	if (type == libecs::Polymorph::POLYMORPH_VECTOR) {
#else
	if (type == libecs::PolymorphValue::TUPLE) {
#endif //ECELL_3_1
		if (p.getType() != type) { return; }
#ifdef ECELL_3_1
		libecs::PolymorphVector pv = p.asPolymorphVector();
		const libecs::PolymorphVector tv = t.asPolymorphVector();
#else
		libecs::PolymorphVector pv = p.as<libecs::PolymorphVector>();
		const libecs::PolymorphVector tv = t.as<libecs::PolymorphVector>();
#endif //ECELL_3_1
		if (pv.size() != tv.size()) { return; }
		const size_t s = pv.size();
		for (int i = 0; i < s; i++) {
			change_PolymorphType(pv[i], tv[i]);
		}
	} else {
#ifdef ECELL_3_1
		p.changeType(type);
#else
		if(type == libecs::PolymorphValue::REAL) {
			p = p.as<libecs::Real>();
			std::cout << (p.getType() == libecs::PolymorphValue::REAL) << std::endl;
		}
		if(type == libecs::PolymorphValue::INTEGER) {
			p = p.as<libecs::Integer>();
			std::cout << (p.getType() == libecs::PolymorphValue::INTEGER) << std::endl;
		}
		if(type == libecs::PolymorphValue::STRING) {
			p = p.as<libecs::String>();
			std::cout << (p.getType() == libecs::PolymorphValue::STRING) << std::endl;
		}
#endif //ECELL_3_1
	}
}

#endif // _WITHOUT_ECELL

