//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#ifndef __METABO_SOLV_H__
#define __METABO_SOLV_H__

#ifndef _WITHOUT_ECELL
#include "ecellif.h"

#ifndef USE_EMC
#include <libecs/FullID.hpp>

#ifdef ECELL_3_1
#if ECELL_VERSION_INT >= VERSION_INT(3,1,106)
#include <libecs/PropertiedObjectMaker.hpp>
#endif //ECELL_VERSION_INT
#endif //ECELL_3_1

#else
#include <libemc/libemc.hpp>
#ifndef ECELL_3_1
#include <libemc/Simulator.hpp>
#endif //ECELL_3_1

#endif //USE_EMC

#include <libecs/libecs.hpp>
#ifdef ECELL_3_1
#include <libemc/Simulator.hpp>
#endif //ECELL_3_1
#include <vector>
#include <string>
#include <set>
#include <map>
#include "EmlModel.h"

#ifndef ECELL_3_1
#include <libecs/Model.hpp>
#include <boost/scoped_ptr.hpp>
#endif //ECELL_3_1

#ifndef USE_EMC
#ifdef ECELL_3_1
DECLARE_CLASS(EventChecker);
DECLARE_CLASS(EventHandler);

DECLARE_SHAREDPTR(EventChecker);
DECLARE_SHAREDPTR(EventHandler);
#endif //ECELL_3_1

class EventHandler
	:
	public std::unary_function<void, void>
{
	public:
	EventHandler() {}
	virtual ~EventHandler() {}

	virtual void operator() (void) const = 0;
};

class EventChecker
	:
	public std::unary_function<bool, void>
{
	public:
	EventChecker() {}
	virtual ~EventChecker() {}

	virtual bool operator() (void) const = 0;
};

class DefaultEventChecker
	:
	public EventChecker
{
	public:
	DefaultEventChecker() {}
	// virtual ~DefaultEventChecker() {}

	virtual bool operator() (void) const {
		return false;
	}
};

#endif //USE_EMC

#ifndef USE_EMC
class emcsim {

#else
class emcsim : public libemc::Simulator {

#endif //USE_EMC

public:
	typedef std::vector<std::string> StepperNameListT;
	typedef std::vector<std::string> EntityNameListT;

#ifndef USE_EMC
	libecs::Model *model;

	bool running;
	mutable bool dirty;
	libecs::Integer eventCheckInterval;
#ifdef ECELL_3_1
	EventCheckerSharedPtr eventChecker;
	EventHandlerSharedPtr eventHandler;
#else
	boost::shared_ptr< EventChecker > eventChecker;
	boost::shared_ptr< EventHandler > eventHandler;
#endif //ECELL_3_1

#ifdef ECELL_3_1
#if ECELL_VERSION_INT >= VERSION_INT(3,1,106)
	libecs::PropertiedObjectMaker propertiedObjectMaker;
#endif //ECELL_VERSION_INT
#else
	boost::scoped_ptr<ModuleMaker< libecs::EcsObject > > propertiedObjectMaker;
#endif //ECELL_3_1

#endif //USE_EMC

protected:
	typedef EmlModel::StepperList StepperList;
	typedef EmlModel::EntityList  EntityList;
	typedef EmlModel::PropertyMap PropertyMap;

	StepperNameListT StepperNameList;
	EntityNameListT  SystemNameList;
	EntityNameListT  VariableNameList;
	EntityNameListT  ProcessNameList;

	typedef std::map<std::string, libecs::Real> OriginalSizeT;
	OriginalSizeT OriginalSize;

	bool _hasModel;

public:
	emcsim();
	virtual ~emcsim();

	virtual bool isExistEntityProperty(const std::string& pn) const;

	virtual int  testModel(const EmlModel& eml);
	virtual void loadModel(const EmlModel& eml, const bool msg = false);
	virtual void loadModel(const EmlModel& eml, const std::set<std::string>& enable, const bool msg = false);
	virtual bool hasModel() const;

	virtual void Dump();

	virtual void setSize(const libecs::Real rootSize);

	virtual const StepperNameListT& getStepperNameList()  const;
	virtual const EntityNameListT&  getSystemNameList()   const;
	virtual const EntityNameListT&  getVariableNameList() const;
	virtual const EntityNameListT&  getProcessNameList()  const;

#ifndef USE_EMC
	virtual void clearEventChecker();
#ifdef ECELL_3_1
	virtual void setEventChecker(EventCheckerSharedPtrCref anEventChecker);

	virtual void createStepper(libecs::StringCref aClassname, libecs::StringCref anId);
	virtual const libecs::Polymorph getStepperList() const;
	virtual const libecs::Polymorph getStepperPropertyList(libecs::StringCref aStepperID) const;
	virtual const libecs::Polymorph getStepperProperty(libecs::StringCref aStepperID, libecs::StringCref aPropertyName) const;
	virtual void loadStepperProperty(libecs::StringCref aStepperID, libecs::StringCref aPropertyName, libecs::PolymorphCref aValue);
	virtual void createEntity(libecs::StringCref aClassname, libecs::StringCref aFullIDString);
	virtual const libecs::Polymorph getEntityPropertyList(libecs::StringCref aFullIDString) const;
	virtual void loadEntityProperty(libecs::StringCref aFullPNString, libecs::PolymorphCref aValue);
	virtual void setEntityProperty(libecs::StringCref aFullPNString, libecs::PolymorphCref aValue);
	virtual const libecs::Polymorph getEntityProperty(libecs::StringCref aFullPNString) const;

	virtual void createLogger(libecs::StringCref aFullPNString);
	virtual void createLogger(libecs::StringCref aFullPNString, libecs::Polymorph aParamList);
	virtual const libecs::DataPointVectorSharedPtr getLoggerData(libecs::StringCref aFullPNString) const;
#else
	virtual void setEventChecker(boost::shared_ptr<EventChecker> const& anEventChecker);

	virtual void createStepper(libecs::String const& aClassname, libecs::String const& anId);
	virtual const libecs::Polymorph getStepperList() const;
	virtual const libecs::Polymorph getStepperPropertyList(libecs::String const& aStepperID) const;
	virtual const libecs::Polymorph getStepperProperty(libecs::String const& aStepperID, libecs::String const& aPropertyName) const;
	virtual void loadStepperProperty(libecs::String const& aStepperID, libecs::String const& aPropertyName, libecs::Polymorph const & aValue);
	virtual void createEntity(libecs::String const& aClassname, libecs::String const& aFullIDString);
	virtual const libecs::Polymorph getEntityPropertyList(libecs::String const& aFullIDString) const;
	virtual void loadEntityProperty(libecs::String const& aFullPNString, libecs::Polymorph const & aValue);
	virtual void setEntityProperty(libecs::String const& aFullPNString, libecs::Polymorph const & aValue);
	virtual const libecs::Polymorph getEntityProperty(libecs::String const& aFullPNString) const;

	virtual void createLogger(libecs::String const& aFullPNString);
	virtual void createLogger(libecs::String const& aFullPNString, libecs::Polymorph aParamList);
	virtual const boost::shared_ptr< libecs::DataPointVector > getLoggerData(libecs::String const& aFullPNString) const;
#endif //ECELL_3_1

	virtual void stop();
	virtual void run(const libecs::Real aDuration);
#endif //USE_EMC

protected:
#ifndef USE_EMC
#ifdef ECELL_3_1
	libecs::ModelRef getModel();
#else
	libecs::Model& getModel();
#endif //ECELL_3_1
	void initialize() const;
#ifdef ECELL_3_1
	libecs::LoggerPtr getLogger(libecs::StringCref aFullPNString) const;
#else
	libecs::Logger* getLogger(libecs::String const& aFullPNString) const;
#endif //ECELL_3_1
	void setDirty();
	const bool isDirty() const;
	void clearDirty() const;
	void handleEvent();
	void start();
#ifdef ECELL_3_1
	void runWithEvent();
	void runWithoutEvent();
#else
	void runWithEvent(libecs::Real const stopTime);
	void runWithoutEvent(libecs::Real const stopTime);
#endif //ECELL_3_1
#endif //USE_EMC
};

void change_PolymorphType(libecs::Polymorph& p, const libecs::Polymorph& t);

#endif // _WITHOUT_ECELL
#endif // __METABO_SOLV_H__


