//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#include <Foundation/datatype.h>
#include <Foundation/util.h>

void membfunction::set(
	const size_t _id, const CfgElem* f, const string& _name, const MolMemb* _gmn)
{
	id   = static_cast<int>(_id);
	pid  = f->isSetID()? f->GetID(): 0;
	name = _name;
	gmn = _gmn;
	g   = gmn? gmn->mol: NULL;
	m   = gmn? gmn->memb->mat[0]: NULL;
	n   = gmn? gmn->memb->mat[1]: NULL;
	gm  = UtilGetEquationId(g,m);
	gn  = UtilGetEquationId(g,n);
	mn  = UtilGetMembraneId(m,n);
	nm  = UtilGetMembraneId(n,m);
	gnm = UtilGetMolMembId(g,nm);
}

void penetrater::set(
	const size_t _id, const CfgElem* f, const string& _name, const MolMemb* _gmn)
{
	membfunction::set(_id,f,_name,_gmn);
	e_k = SKL_REAL(0.0);
	for (const CfgParam* K = f->GetParamFirst("K");
		K != NULL; K = f->GetParamNext(K,"K")) {
		K->GetData(&e_k);
	}
}

void Transporter::set(
	const size_t _id, const CfgElem* f, const string& _name, const MolMemb* _gmn)
{
	membfunction::set(_id,f,_name,_gmn);

	e_SW    = SKL_REAL(0.0);
	e_Fuzzy = SKL_REAL(0.0);
	e_Den   = SKL_REAL(0.0);

	for (const CfgParam* SW = f->GetParamFirst("SW");
		SW != NULL; SW = f->GetParamNext(SW,"SW")) {
		SW->GetData(&e_SW);
	}
	for (const CfgParam* Fuzzy = f->GetParamFirst("Fuzzy");
		Fuzzy != NULL; Fuzzy = f->GetParamNext(Fuzzy,"Fuzzy")) {
		Fuzzy->GetData(&e_Fuzzy);
	}
	for (const CfgParam* Den = f->GetParamFirst("Den");
		Den != NULL; Den = f->GetParamNext(Den,"Den")) {
		Den->GetData(&e_Den);
	}
}

void ChTranser::set(
	const size_t _id, const CfgElem* f, const string& _name, const MolMemb* _gmn)
{
	Transporter::set(_id,f,_name,_gmn);
	e_K  = SKL_REAL(0.0);
	e_Ke = SKL_REAL(0.0);

	for (const CfgParam* K = f->GetParamFirst("K");
		K != NULL; K = f->GetParamNext(K,"K")) {
		K->GetData(&e_K);
	}
	for (const CfgParam* Ke = f->GetParamFirst("Ke");
		Ke != NULL; Ke = f->GetParamNext(Ke,"Ke")) {
		Ke->GetData(&e_Ke);
	}
}

void PmTranser::set(
	const size_t _id, const CfgElem* f, const string& _name, const MolMemb* _gmn)
{
	Transporter::set(_id,f,_name,_gmn);

	e_VmaxAtp = SKL_REAL(0.0);
	e_VmaxGtp = SKL_REAL(0.0);
	e_KmAtp = SKL_REAL(0.0);
	e_KmGtp = SKL_REAL(0.0);
	TRatp = SKL_REAL(0.0);
	TRgtp = SKL_REAL(0.0);

	ATPdriven = false;
	GTPdriven = false;

	// ATP
	for (const CfgParam* VmaxATP = f->GetParamFirst("VmaxATP");
		VmaxATP != NULL; VmaxATP = f->GetParamNext(VmaxATP,"VmaxATP")) {
		VmaxATP->GetData(&e_VmaxAtp);
	}
	for (const CfgParam* KmATP = f->GetParamFirst("KmATP");
		KmATP != NULL; KmATP = f->GetParamNext(KmATP,"KmATP")) {
		KmATP->GetData(&e_KmAtp);
	}
	for (const CfgParam* TRATP = f->GetParamFirst("TRATP");
		TRATP != NULL; TRATP = f->GetParamNext(TRATP,"TRATP")) {
		TRATP->GetData(&TRatp);
		ATPdriven = true;
	}
	// GTP
	for (const CfgParam* VmaxGTP = f->GetParamFirst("VmaxGTP");
		VmaxGTP != NULL; VmaxGTP = f->GetParamNext(VmaxGTP,"VmaxGTP")) {
		VmaxGTP->GetData(&e_VmaxAtp);
	}
	for (const CfgParam* KmGTP = f->GetParamFirst("KmGTP");
		KmGTP != NULL; KmGTP = f->GetParamNext(KmGTP,"KmGTP")) {
		KmGTP->GetData(&e_KmAtp);
	}
	for (const CfgParam* TRGTP = f->GetParamFirst("TRGTP");
		TRGTP != NULL; TRGTP = f->GetParamNext(TRGTP,"TRGTP")) {
		TRGTP->GetData(&TRgtp);
		GTPdriven = true;
	}

	if (ATPdriven) {
		const Molecular* atp = UtilGetMolecularId("ATP");
		const Molecular* adp = UtilGetMolecularId("ADP");
		const Molecular* pi  = UtilGetMolecularId("Pi");

		TRY(atp, _name + " is driven with ATP, but ATP is no existing");
		TRY(adp, _name + " is driven with ATP, but ADP is no existing");
		TRY(pi,  _name + " is driven with ATP, but Pi is no existing");

		atpm  = UtilGetEquationId(atp,m);
		adpm  = UtilGetEquationId(adp,m);
		pim   = UtilGetEquationId(pi,m);
		atpmn = UtilAddNewMolMembId(atp,mn);
		adpmn = UtilAddNewMolMembId(adp,mn);
		pimn  = UtilAddNewMolMembId(pi,mn);
	}

	if (GTPdriven) {
		const Molecular* gtp = UtilGetMolecularId("GTP");
		const Molecular* gdp = UtilGetMolecularId("GDP");
		const Molecular* pi  = UtilGetMolecularId("Pi");

		TRY(gtp, _name + " is driven with GTP, but GTP is no existing");
		TRY(gdp, _name + " is driven with GTP, but GDP is no existing");
		TRY(pi,  _name + " is driven with GTP, but Pi is no existing");

		gtpm  = UtilGetEquationId(gtp,m);
		gdpm  = UtilGetEquationId(gdp,m);
		pim   = UtilGetEquationId(pi,m);
		gtpm  = UtilGetEquationId(gtp,m);
		gtpmn = UtilAddNewMolMembId(gtp,mn);
		gdpmn = UtilAddNewMolMembId(gdp,mn);
		pimn  = UtilAddNewMolMembId(pi,mn);
	}
}
