//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#ifndef __ELEMENT_H__
#define __ELEMENT_H__

#include <SklSolverBase.h>
#include <map>
#include <set>
#include <string>

typedef unsigned int vxsize_t;

typedef SklScalar4D<SKL_REAL>     S4D_REAL;
typedef SklScalar4D<float>        S4D_FLT;
typedef SklScalar4D<double>       S4D_DBL;
typedef SklScalar4D<unsigned int> S4D_UINT;
typedef SklScalar4D<int>          S4D_INT;
typedef SklScalar4D<void*>        S4D_VOIDP;
typedef SklScalar4D<bool>         S4D_BOOL;

typedef SklScalar3D<SKL_REAL>     S3D_REAL;
typedef SklScalar3D<float>        S3D_FLT;
typedef SklScalar3D<double>       S3D_DBL;
typedef SklScalar3D<unsigned int> S3D_UINT;
typedef SklScalar3D<int>          S3D_INT;
typedef SklScalar3D<void*>        S3D_VOIDP;

typedef SklVector3D<SKL_REAL>     V3D_REAL;
typedef SklVector3D<float>        V3D_FLT;
typedef SklVector3D<double>       V3D_DBL;

typedef map<string, SKL_REAL> RealMap;
typedef map<int, SKL_REAL>    Int2RealMap;
typedef map<int, SKL_REAL*>   Int2RealpMap;

struct Element {
	int    id;
	string name;
};

struct Metabolism;

struct Material : Element {
	int          cid;   // id attribute value at configuration XML

	// metabolism
	struct mtbT {
		const char*  eml;   // name of path to EML file
		set<string>  sys;
		SKL_REAL ve;
		SKL_REAL te;
	} body_mtb;
	// fusion
	bool         fus;
	SKL_REAL     fut;

	//
	vector<const Metabolism*> MtbTbl;

	// CfgElem
	const CfgElem* xml_elem;
};

struct Molecular : Element {
	int          cid;     // id attribute value at configuration XML
	const char*  fullid;  // FullID for link with E-Cell
	SKL_REAL     ccoef;   // Diffusional Coef. C
	RealMap      molconc; // Molecular Concentration in each mediums
	RealMap      molnum;  // Molecular number in each mediums

	RealMap      trans;   // Trancemittance from m to n with membrane
	Int2RealMap  ccoefm;  // Diffusional Coef. Cm : material dependent

	// CfgElem
	const CfgElem* xml_elem;
#ifdef ENABLE_MEMBRANEPOTENTIAL
        bool         potential; // Memblen Potential Flag
#endif
};

struct Equation : Element {
	const Material*  mat;
	const Molecular* mol;
	//
	int          je;       // Assigend JobEntry index at JobManTable
};

struct Metabolism : Element {
	const char*  eml;   // name of path to EML file
	set<string>  sys;
	
	// unit
	SKL_REAL Ve; // Volume Unit
	SKL_REAL Te; // Time Unit

	const Material* mat;

	// on membrane?
	bool onMemb;
	SKL_REAL height;
	const Material* contacted; // contacted material with membrane

	// receptor
	bool  reactsite;
	const Equation* LR;

	// CfgElem
	const CfgElem* xml_elem;
};

struct Membrane : Element {
	const Material* mat[2];
};

struct MolMemb : Element {
	const Molecular* mol;
	const Membrane*  memb;
};

struct Membarea : Element {
	const Material* mat[2];
};

struct BoundaryCond {
	enum TYPE {
		UNKNOWN,
		DIRICHLET,  // Dirichlet Condition
		NEUMANN,    // Neumann Condition
	};

	enum BCFace {
		UNSET,
		XMINUS,
		YMINUS,
		ZMINUS,
		XPLUS,
		YPLUS,
		ZPLUS,
		ALLFACE,
	};

	int      id;    // id of inner use
	int      cid;   // cid specified at configuration
	TYPE     type;  // type of boundary condition
	SKL_REAL value; // value of boundaty condition

	BCFace   face;
	const Material*  mat;
	const Molecular* mol;
};

struct membfunction : Element {
	int pid;

	const Molecular* g;
	const Material*  m;
	const Material*  n;
	const Equation*  gm;
	const Equation*  gn;
	const Membrane*  mn;
	const Membrane*  nm;
	const MolMemb*   gmn;
	const MolMemb*   gnm;

	void set(const size_t _id, const CfgElem* f, 
				const string& _name, const MolMemb* _gmn);
};

struct penetrater : membfunction {
	SKL_REAL e_k;

	void set(const size_t _id, const CfgElem* f, 
				const string& _name, const MolMemb* _gmn);
};

struct Transporter : membfunction {

	SKL_REAL e_SW;
	SKL_REAL e_Fuzzy;
	SKL_REAL e_Den;
	//
	struct carriee {
		SKL_REAL R;
		const Molecular* h;
		const Equation* hm;
		const Equation* hn;
		const MolMemb* hmn;
		const MolMemb* hnm;
	};
	typedef std::vector<carriee> CarrieeSet;
	CarrieeSet carriees;

	void set(const size_t _id, const CfgElem* f, 
				const string& _name, const MolMemb* _gmn);
};

struct ChTranser : Transporter {
	SKL_REAL e_K;
	SKL_REAL e_Ke;
	//
	SKL_REAL* K;
	SKL_REAL* Ke;

	void set(const size_t _id, const CfgElem* f, 
				const string& _name, const MolMemb* _gmn);
};

struct PmTranser : Transporter {
	SKL_REAL e_VmaxAtp;
	SKL_REAL e_KmAtp;
	SKL_REAL e_VmaxGtp;
	SKL_REAL e_KmGtp;
	SKL_REAL TRatp;
	SKL_REAL TRgtp;

	const Equation* atpm;
	const Equation* gtpm;
	const Equation* adpm;
	const Equation* gdpm;
	const Equation* pim;
	const MolMemb*  atpmn;
	const MolMemb*  gtpmn;
	const MolMemb*  adpmn;
	const MolMemb*  gdpmn;
	const MolMemb*  pimn;

	bool ATPdriven;
	bool GTPdriven;

	void set(const size_t _id, const CfgElem* f, 
				const string& _name, const MolMemb* _gmn);
};

struct CSTranser : ChTranser { };
struct CATranser : ChTranser { };
struct PSTranser : PmTranser { };
struct PATranser : PmTranser { };
struct GpTranser : ChTranser { };

// steady state
enum EXEC {
	CONTINUE,
	STOP,
};

enum OPETYPE {
	AND,
	OR,
};

enum THTYPE {
	ABSOLUTE,
	RATIO,
};

enum COMPTYPE {
	LT,
	LE,
	GT,
	GE,
};

struct SteadyState {
	const Equation* eqn;
	SKL_REAL threshold;
	THTYPE thType;
	COMPTYPE compType;

	// steady state counter (steady state flag)
	unsigned int SScnt;

	// steady state judge number
	unsigned int SSjnum;
};

#endif  // __ELEMENT_H__

