//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#include <cstdlib>
#include <cstdio>
#include <cmath>

#include "util.h"

using std::string;

CfgInfo* cfg;

void InitUtilEnv(CfgInfo* _cfg)
{
	cfg = _cfg;
}

CfgInfo* UtilGetCfg()
{
	return cfg;
}


int UtilGetExtParamInt(const char* paramName)
{
	ExtParamMap::const_iterator i = cfg->extParams.find(paramName);
	if (i == cfg->extParams.end()) { return 0; }
	const ExtParam* p = i->second;
	return p->value.INT;
}

SKL_REAL UtilGetExtParamReal(const char* paramName)
{
	ExtParamMap::const_iterator i = cfg->extParams.find(paramName);
	if (i == cfg->extParams.end()) { return 0; }
	const ExtParam* p = i->second;
	return p->value.REAL;
}

const char* UtilGetExtParamString(const char* paramName)
{
	ExtParamMap::const_iterator i = cfg->extParams.find(paramName);
	if (i == cfg->extParams.end()) { return 0; }
	const ExtParam* p = i->second;
	return p->value.STR;
}

std::string UtilStrFormat(const char* format, ... )
{
	static std::vector<char> tmp(1000);

	va_list args;
	va_start(args, format);

	while(vsnprintf(&tmp[0],tmp.size(),format,args) == -1) {
		tmp.resize(tmp.size() * 2);
	}
	tmp.push_back('\0');

	string ret = &(tmp[0]);
	va_end(args);

	return ret;
}

SKL_REAL* UtilGetS4DN(S4D_REAL* s4d, const int n)
{
	if (!s4d) { return NULL; }
	SKL_REAL* p    = s4d->GetData();
	const int vc   = s4d->GetVCellSize();
	const int sz[] = {
		s4d->GetSize()[0] + vc * 2,
		s4d->GetSize()[1] + vc * 2,
		s4d->GetSize()[2] + vc * 2,
	};
	const int idx = (sz[0] * sz[1] * sz[2]) * n;
	return &p[idx];
}

const SKL_REAL* UtilGetS4DN(const S4D_REAL* s4d, const int n)
{
	if (!s4d) { return NULL; }
	const SKL_REAL* p = s4d->GetData();
	const int vc      = s4d->GetVCellSize();
	const int sz[]    = {
		s4d->GetSize()[0] + vc * 2,
		s4d->GetSize()[1] + vc * 2,
		s4d->GetSize()[2] + vc * 2,
	};
	const int idx = (sz[0] * sz[1] * sz[2]) * n;
	return &p[idx];
}

S4D_REAL* UtilAllocS4DReal(const char* label, const vxsize_t* size, const int vc, const int gid)
{
	TRY((size[0] > 0)&&(size[1] > 0)&&(size[2] > 0)&&(size[3] > 0),
		UtilStrFormat("Allocation size of %s is zero", label));

#ifdef SPHERE175
	SklParaComponent *para_cmp = UtilGetCfg()->solver->SklGetParaComponent();	//2009.12.07
	SklParaManager *para_mng = para_cmp->GetParaManager();				//2009.12.07
#endif

#ifdef SPHERE175
	S4D_REAL* ary = dynamic_cast<S4D_REAL*>(
			cfg->solver->SklAllocateArray(
				para_mng, label, SKL_CLASS_SCALAR4D, SKL_ARRAY_DTYPE_REAL,	//2009.12.07
				size, vc, gid
			));
#else
	S4D_REAL* ary = dynamic_cast<S4D_REAL*>(
			cfg->solver->SklAllocateArray(
				label, SKL_CLASS_SCALAR4D, SKL_ARRAY_DTYPE_REAL,
				size, vc, gid
			));
#endif

	TRY(ary, UtilStrFormat("Failed to allocate %s",label));

	size_t memsize = ary->GetArrayLength() * sizeof(SKL_REAL);
	cfg->used_mem += memsize;
	MESSAGE(UtilStrFormat(
			"++ Allocate Scalar4D_REAL '%s'\t%d*%d*%d*%d:%d[byte]",
			label,ary->_GetSize()[0],ary->_GetSize()[1],ary->_GetSize()[2],
#ifndef _NO_MESSAGE
			ary->_GetSize()[3],memsize));
#else
			ary->_GetSize()[3],memsize),0);
#endif //_NO_MESSAGE

	UtilClearRealArray(ary, SKL_REAL(0.0));
	return ary;
}

S4D_INT* UtilAllocS4DInt(const char* label, const vxsize_t* size, const int vc, const int gid)
{
	TRY((size[0] > 0)&&(size[1] > 0)&&(size[2] > 0)&&(size[3] > 0),
		UtilStrFormat("Allocation size of %s is zero", label));

#ifdef SPHERE175
	SklParaComponent *para_cmp = UtilGetCfg()->solver->SklGetParaComponent();	//2009.12.07
	SklParaManager *para_mng = para_cmp->GetParaManager();				//2009.12.07
#endif

#ifdef SPHERE175
	S4D_INT* ary = dynamic_cast<S4D_INT*>(
			cfg->solver->SklAllocateArray(
				para_mng, label, SKL_CLASS_SCALAR4D, SKL_ARRAY_DTYPE_INT,	//2009.12.07
				size, vc, gid
			));
#else
	S4D_INT* ary = dynamic_cast<S4D_INT*>(
			cfg->solver->SklAllocateArray(
				label, SKL_CLASS_SCALAR4D, SKL_ARRAY_DTYPE_INT,
				size, vc, gid
			));
#endif

	TRY(ary, UtilStrFormat("Failed to allocate %s",label));

	size_t memsize = ary->GetArrayLength() * sizeof(int);
	cfg->used_mem += memsize;
	MESSAGE(UtilStrFormat(
			"++ Allocate Scalar4D_REAL '%s'\t%d*%d*%d*%d:%d[byte]",
			label,ary->_GetSize()[0],ary->_GetSize()[1],ary->_GetSize()[2],
#ifndef _NO_MESSAGE
			ary->_GetSize()[3],memsize));
#else
			ary->_GetSize()[3],memsize),0);
#endif //_NO_MESSAGE

	UtilClearIntArray(ary, 0);
	return ary;
}

S3D_REAL* UtilAllocS3DReal(const char* label, const vxsize_t* size, const int vc, const int gid)
{
	TRY((size[0] > 0)&&(size[1] > 0)&&(size[2] > 0),
		UtilStrFormat("Allocation size of %s is zero", label));

#ifdef SPHERE175
	SklParaComponent *para_cmp = UtilGetCfg()->solver->SklGetParaComponent();	//2009.12.07
	SklParaManager *para_mng = para_cmp->GetParaManager();				//2009.12.07
#endif

#ifdef SPHERE175
	S3D_REAL* ary = dynamic_cast<S3D_REAL*>(
			cfg->solver->SklAllocateArray(
				para_mng, label, SKL_CLASS_SCALAR3D, SKL_ARRAY_DTYPE_REAL,	//2009.12.07
				size, vc, gid
			));
#else
	S3D_REAL* ary = dynamic_cast<S3D_REAL*>(
			cfg->solver->SklAllocateArray(
				label, SKL_CLASS_SCALAR3D, SKL_ARRAY_DTYPE_REAL,
				size, vc, gid
			));
#endif

	TRY(ary, UtilStrFormat("Failed to allocate %s",label));

	size_t memsize = ary->GetArrayLength() * sizeof(SKL_REAL);
	cfg->used_mem += memsize;
	MESSAGE(UtilStrFormat(
			"++ Allocate Scalar4D_REAL '%s'\t%d*%d*%d*%d:%d[byte]",
			label,ary->_GetSize()[0],ary->_GetSize()[1],ary->_GetSize()[2],
#ifndef _NO_MESSAGE
			memsize));
#else
			memsize),0);
#endif //_NO_MESSAGE

	UtilClearRealArray(ary, SKL_REAL(0.0));
	return ary;
}

S4D_UINT* UtilAllocS4DUInt(const char* label, const vxsize_t* size, const int vc, const int gid)
{
	TRY((size[0] > 0)&&(size[1] > 0)&&(size[2] > 0)&&(size[3] > 0),
		UtilStrFormat("Allocation size of %s is zero", label));

#ifdef SPHERE175
	SklParaComponent *para_cmp = UtilGetCfg()->solver->SklGetParaComponent();	//2009.12.07
	SklParaManager *para_mng = para_cmp->GetParaManager();				//2009.12.07
#endif

#ifdef SPHERE175
	S4D_UINT* ary = dynamic_cast<S4D_UINT*>(
			cfg->solver->SklAllocateArray(
				para_mng, label, SKL_CLASS_SCALAR4D, SKL_ARRAY_DTYPE_UINT,	//2009.12.07
				size, vc, gid
			));
#else
	S4D_UINT* ary = dynamic_cast<S4D_UINT*>(
			cfg->solver->SklAllocateArray(
				label, SKL_CLASS_SCALAR4D, SKL_ARRAY_DTYPE_UINT,
				size, vc, gid
			));
#endif

	TRY(ary, UtilStrFormat("Failed to allocate %s",label));

	size_t memsize = ary->GetArrayLength() * sizeof(int);
	cfg->used_mem += memsize;
	MESSAGE(UtilStrFormat(
			"++ Allocate Scalar4D_REAL '%s'\t%d*%d*%d*%d:%d[byte]",
			label,ary->_GetSize()[0],ary->_GetSize()[1],ary->_GetSize()[2],
#ifndef _NO_MESSAGE
			ary->_GetSize()[3],memsize));
#else
			ary->_GetSize()[3],memsize),0);
#endif //_NO_MESSAGE

	UtilClearUIntArray(ary, 0);
	return ary;
}

int UtilLoadSph(const string& attr, S4D_REAL* ary, const int N, const int gid)
{
	return UtilLoadSph(attr.c_str(), ary, N, gid);
}

int UtilLoadSph(const char* attr, S4D_REAL* ary, const int N, const int gid)
{
	const int vc = ary->GetVCellSize();
	const int I  = ary->_GetSize()[0];
	const int J  = ary->_GetSize()[1];
	const int K  = ary->_GetSize()[2];

#ifdef SPHERE175
	SklParaComponent *para_cmp = UtilGetCfg()->solver->SklGetParaComponent();	//2009.12.07
	SklParaManager *para_mng = para_cmp->GetParaManager();				//2009.12.07
#endif

#ifdef SPHERE175
	SklVoxDataSet* dSet = cfg->solver->LoadFile(para_mng, attr, vc, gid);		//2009.12.07
#else
	SklVoxDataSet* dSet = cfg->solver->LoadFile(attr, vc, gid);
#endif

	if (!dSet) { return 0; }

	if (dSet->GetDataType() == SklVoxDataSet::TYPE_SINGLE) {
		S3D_FLT* in = (dSet)? dynamic_cast<S3D_FLT*>(
						dSet->GetData(SklVoxDataSet::SPH_DATA)):
						NULL;

		TRY(in, string("failed to load ") + attr);
		for (size_t k = 0; k < K; k++) {
		for (size_t j = 0; j < J; j++) {
		for (size_t i = 0; i < I; i++) {
			ary->_val(i,j,k,N) = static_cast<SKL_REAL>(in->_val(i,j,k));
		}}}
	} else if (dSet->GetDataType() == SklVoxDataSet::TYPE_DOUBLE) {
		S3D_DBL* in = (dSet)? dynamic_cast<S3D_DBL*>(
						dSet->GetData(SklVoxDataSet::SPH_DATA)):
						NULL;

		TRY(in, string("failed to load ") + attr);
		for (size_t k = 0; k < K; k++) {
		for (size_t j = 0; j < J; j++) {
		for (size_t i = 0; i < I; i++) {
			ary->_val(i,j,k,N) = static_cast<SKL_REAL>(in->_val(i,j,k));
		}}}
	} else {
		delete dSet;
		ERROR(string(attr) + " data type is unknown");
	}

	// Set Restart Time
	if (cfg->Restart) {
		SKL_REAL time;
		int      step;
		if (dSet->GetTime(time) && dSet->GetStep(step)) {
			if (time > cfg->solver->SklGetBaseTime()) {
				cfg->solver->SklSetBaseTime(time);
			}
			if (step > cfg->solver->SklGetBaseStep()) {
				cfg->solver->SklSetBaseStep(step);
			}
		}
	}

	delete dSet;

#ifndef _NO_MESSAGE
	MESSAGE(UtilStrFormat("Sph File is loaded as attr=%s", attr));
#else
	MESSAGE(UtilStrFormat("Sph File is loaded as attr=%s", attr),0);
#endif //_NO_MESSAGE
	return 1;
}

int UtilLoadSph(const string& attr, S3D_REAL* ary, const int gid)
{
	return UtilLoadSph(attr.c_str(), ary, gid);
}

int UtilLoadSph(const char* attr, S3D_REAL* ary, const int gid)
{
	const int vc = ary->GetVCellSize();
	const int I  = ary->_GetSize()[0];
	const int J  = ary->_GetSize()[1];
	const int K  = ary->_GetSize()[2];

#ifdef SPHERE175
	SklParaComponent *para_cmp = UtilGetCfg()->solver->SklGetParaComponent();	//2009.12.07
	SklParaManager *para_mng = para_cmp->GetParaManager();				//2009.12.07
#endif

#ifdef SPHERE175
	SklVoxDataSet* dSet = cfg->solver->LoadFile(para_mng, attr, vc, gid);		//2009.12.07
#else
	SklVoxDataSet* dSet = cfg->solver->LoadFile(attr, vc, gid);
#endif

	if (!dSet) { return 0; }

	if (dSet->GetDataType() == SklVoxDataSet::TYPE_SINGLE) {
		S3D_FLT* in = (dSet)? dynamic_cast<S3D_FLT*>(
						dSet->GetData(SklVoxDataSet::SPH_DATA)):
						NULL;

		TRY(in, string("failed to load ") + attr);
		for (size_t k = 0; k < K; k++) {
		for (size_t j = 0; j < J; j++) {
		for (size_t i = 0; i < I; i++) {
			ary->_val(i,j,k) = static_cast<SKL_REAL>(in->_val(i,j,k));
		}}}
	} else if (dSet->GetDataType() == SklVoxDataSet::TYPE_DOUBLE) {
		S3D_DBL* in = (dSet)? dynamic_cast<S3D_DBL*>(
						dSet->GetData(SklVoxDataSet::SPH_DATA)):
						NULL;

		TRY(in, string("failed to load ") + attr);
		for (size_t k = 0; k < K; k++) {
		for (size_t j = 0; j < J; j++) {
		for (size_t i = 0; i < I; i++) {
			ary->_val(i,j,k) = static_cast<SKL_REAL>(in->_val(i,j,k));
		}}}
	} else {
		delete dSet;
		ERROR(string(attr) + " data type is unknown");
	}

	// Set Restart Time
	if (cfg->Restart) {
		SKL_REAL time;
		int      step;
		if (dSet->GetTime(time) && dSet->GetStep(step)) {
			if (time > cfg->solver->SklGetBaseTime()) {
				cfg->solver->SklSetBaseTime(time);
			}
			if (step > cfg->solver->SklGetBaseStep()) {
				cfg->solver->SklSetBaseStep(step);
			}
		}
	}

	delete dSet;

#ifndef _NO_MESSAGE
	MESSAGE(UtilStrFormat("Sph File is loaded as attr=%s", attr));
#else
	MESSAGE(UtilStrFormat("Sph File is loaded as attr=%s", attr),0);
#endif //_NO_MESSAGE
	return 1;
}

int UtilLoadSph(const char* attr, V3D_REAL* ary, const int gid);

const Material* UtilGetMaterialId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->MatNameIdMap.find(name);
	return (e != cfg->MatNameIdMap.end())? cfg->MatTbl[e->second]: NULL;
}

const Molecular* UtilGetMolecularId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->MolNameIdMap.find(name);
	return (e != cfg->MolNameIdMap.end())? cfg->MolTbl[e->second]: NULL;
}

const Equation* UtilGetEquationId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->EqnNameIdMap.find(name);
	return (e != cfg->EqnNameIdMap.end())? cfg->EqnTbl[e->second]: NULL;
}

const Metabolism* UtilGetMetabolismId(const string& name)
{
	Name2IdMapT::const_iterator m = cfg->MtbNameIdMap.find(name);
	return (m != cfg->MtbNameIdMap.end())? cfg->MtbTbl[m->second]: NULL;
}

const Membrane* UtilGetMembraneId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->MembNameIdMap.find(name);
	return (e != cfg->MembNameIdMap.end())? cfg->MembTbl[e->second]: NULL;
}

const MolMemb* UtilGetMolMembId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->MlMbNameIdMap.find(name);
	return (e != cfg->MlMbNameIdMap.end())? cfg->MlMbTbl[e->second]: NULL;
}

const Membarea* UtilGetMembareaId(const string& name)
{
	const string mnname = name + "_MembArea";

	Name2IdMapT::const_iterator e = cfg->MembareaNameIdMap.find(mnname);
	return (e != cfg->MembareaNameIdMap.end())? cfg->MembareaTbl[e->second]: NULL;
}

const Equation* UtilGetEquationId(const Molecular* ml, const Material* mt)
{
	if (!ml) { return NULL; }
	if (!mt) { return NULL; }

	const string& mlname = ml->name;
	const string& mtname = mt->name;
	const string& eqname = mtname + ':' + mlname;
	Name2IdMapT::const_iterator e = cfg->EqnNameIdMap.find(eqname);
	return (e != cfg->EqnNameIdMap.end())? cfg->EqnTbl[e->second]: NULL;
}

const Membrane* UtilGetMembraneId(const Material* mid, const Material* nid)
{
	if (!mid) { return NULL; }
	if (!nid) { return NULL; }

	const string& mname  = mid->name;
	const string& nname  = nid->name;
	const string  mnname = mname + ':' + nname;
	Name2IdMapT::const_iterator e = cfg->MembNameIdMap.find(mnname);
	return (e != cfg->MembNameIdMap.end())? cfg->MembTbl[e->second]: NULL;
}

const MolMemb* UtilGetMolMembId(const Molecular* ml, const Membrane* mb)
{
	if (!ml) { return NULL; }
	if (!mb) { return NULL; }

	const string& mlname = ml->name;
	const string& mbname = mb->name;
	const string  name   = mlname + ':' + mbname;
	Name2IdMapT::const_iterator e = cfg->MlMbNameIdMap.find(name);
	return (e != cfg->MlMbNameIdMap.end())? cfg->MlMbTbl[e->second]: NULL;
}

const Membarea* UtilGetMembareaId(const Material* mid, const Material* nid)
{
	if( !mid ){ return NULL; }
	if( !nid ){ return NULL; }

	const string& mname  = mid->name;
	const string& nname  = nid->name;
	const string  mnname = mname + ":" + nname + "_MembArea";
	Name2IdMapT::const_iterator e = cfg->MembareaNameIdMap.find(mnname);
	return (e != cfg->MembareaNameIdMap.end())? cfg->MembareaTbl[e->second]: NULL;
}

const MolMemb* UtilAddNewMolMembId(const Molecular* g, const Membrane* mn)
{
	const std::string name = g->name + ":" + mn->name;
	const MolMemb* r = UtilGetMolMembId(name);
	if (!r) {
		MolMemb* gmn = new MolMemb;
		gmn->id = cfg->MlMbTbl.size();
		gmn->mol  = g;
		gmn->memb = mn;
		gmn->name = name;

		const int id = gmn->id;
		const std::string name = gmn->name;
#ifndef _NO_MESSAGE
		MESSAGE(string("++ ") + name +" id=" + string_cast(id));
#else
		MESSAGE(string("++ ") + name +" id=" + string_cast(id),0);
#endif //_NO_MESSAGE

		cfg->MlMbTbl.push_back(gmn);
		cfg->MlMbNameIdMap.insert(Name2IdMapT::value_type(name,id));

		r = gmn;
	}
	cfg->MlMbCount = cfg->MlMbTbl.size();

	return r;
}

const penetrater* UtilGetPenetraterId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->PeneNameIdMap.find(name);
	return (e != cfg->PeneNameIdMap.end())? cfg->PeneTbl[e->second]: NULL;
}

const ChTranser*  UtilGetChTranserId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->ChTrNameIdMap.find(name);
	return (e != cfg->ChTrNameIdMap.end())? cfg->ChTrTbl[e->second]: NULL;
}

const PmTranser*  UtilGetPmTranserId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->PmTrNameIdMap.find(name);
	return (e != cfg->PmTrNameIdMap.end())? cfg->PmTrTbl[e->second]: NULL;
}

const CSTranser*  UtilGetCSTranserId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->CSTrNameIdMap.find(name);
	return (e != cfg->CSTrNameIdMap.end())? cfg->CSTrTbl[e->second]: NULL;
}

const PSTranser*  UtilGetPSTranserId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->PSTrNameIdMap.find(name);
	return (e != cfg->PSTrNameIdMap.end())? cfg->PSTrTbl[e->second]: NULL;
}

const CATranser*  UtilGetCATranserId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->CATrNameIdMap.find(name);
	return (e != cfg->CATrNameIdMap.end())? cfg->CATrTbl[e->second]: NULL;
}

const PATranser*  UtilGetPATranserId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->PATrNameIdMap.find(name);
	return (e != cfg->PATrNameIdMap.end())? cfg->PATrTbl[e->second]: NULL;
}

const GpTranser*  UtilGetGpTranserId(const string& name)
{
	Name2IdMapT::const_iterator e = cfg->GpTrNameIdMap.find(name);
	return (e != cfg->GpTrNameIdMap.end())? cfg->GpTrTbl[e->second]: NULL;
}

const ExtParam* UtilGetExtParam(const string& name)
{
	ExtParamMap::const_iterator p = cfg->extParams.find(name);

	const ExtParam* ep = NULL;
	if (p != cfg->extParams.end()) {
		ep = p->second;
	}

	return ep;
}

bool UtilRegArrayS4DReal(const string& name, S4D_REAL* ary, int N)
{
	if (!ary) { return false; }

	S4DRN p;
	p.ary = ary;
	p.N   = N;

	cfg->S4Dmap.insert(S4DMapT::value_type(name, p));
	MESSAGE(UtilStrFormat("Regist %d in %s as '%s'",
#ifndef _NO_MESSAGE
							N, ary->GetLabel(), name.c_str()));
#else
							N, ary->GetLabel(), name.c_str()),0);
#endif //_NO_MESSAGE

	return true;
}

S4DRN UtilGetRegistedArrayS4DReal(const string& name)
{
	S4DRN p;
	S4DMapT::iterator i = UtilGetCfg()->S4Dmap.find(name);

	if (i != UtilGetCfg()->S4Dmap.end()) {
		p = i->second;
	} else {
		p.ary = NULL;
		p.N   = -1;
	}

	return p;
}

int UtilGetExtParamInt(const string& name)
{
	int i = 0;
	const ExtParam* p = UtilGetExtParam(name);
	if (p) { i = p->value.INT; }

	return i;
}

SKL_REAL UtilGetExtParamReal(const string& name)
{
	SKL_REAL r = SKL_REAL(0.0);
	const ExtParam* p = UtilGetExtParam(name);
	if (p) { r = p->value.REAL; }

	return r;
}

const char* UtilGetExtParamStr(const string& name)
{
	const char* s = NULL;
	const ExtParam* p = UtilGetExtParam(name);
	if (p) { s = p->value.STR; }

	return s;
}

int UtilGetExtParamID(const string& name)
{
	int id = 0;
	const ExtParam* p = UtilGetExtParam(name);
	if (p && (p->extid)) { id = p->id; }

	return id;
}

bool UtilClearRealArray(SklAray<SKL_REAL>* ary, const SKL_REAL val)
{
	const int size = ary? ary->GetArrayLength(): 0;
	if (!size) { return true; }
	return SklSolverBase::SklInitializeSKL_REAL(ary->GetData(),val,size);
}

bool UtilClearIntArray(SklAray<int>* ary, const int val)
{
	const int size = ary? ary->GetArrayLength(): 0;
	if (!size) { return true; }
	return SklSolverBase::SklInitializeInt(ary->GetData(),val,size);
}

bool UtilClearUIntArray(SklAray<unsigned int>* ary, const unsigned int val)
{
	return UtilClearArray(ary,val);
}

bool UtilCheckIdxPdiv(int* zi, int* zj, int* zk, const int i, const int j, const int k, const int vc)
{
	const int pi = i - UtilGetCfg()->Pdiv.head[0];
	const int pj = j - UtilGetCfg()->Pdiv.head[1];
	const int pk = k - UtilGetCfg()->Pdiv.head[2];

	bool in =  (-vc <= pi) && (pi < UtilGetCfg()->Pdiv.size[0] + vc)
			&& (-vc <= pj) && (pj < UtilGetCfg()->Pdiv.size[1] + vc)
			&& (-vc <= pk) && (pk < UtilGetCfg()->Pdiv.size[2] + vc);

	if (zi) { *zi = pi; }
	if (zj) { *zj = pj; }
	if (zk) { *zk = pk; }

	return in;
}

int UtilSetMembParam(const string& fname, const membfunction* p, const int i, const int j, const int k, const SKL_REAL value)
{
	if (!p) { return -1; }

	int zi, zj, zk;
	bool in = UtilCheckIdxPdiv(&zi,&zj,&zk,i,j,k,1);

	if (!in) { return 0; }

	const string name = UtilStrFormat("%s:%s",fname.c_str(),p->name.c_str());
	const Material* m = p->m;

	S4DRN K = UtilGetCfg()->S4Dmap[name];
	S4D_REAL* ary = K.ary;
	const int n   = K.N;

	// inner material
	ary->val(zi,zj,zk,n) = value;

	// outer material
	S4DRN Nx = UtilGetCfg()->S4Dmap[UtilStrFormat("Nx_%s",m->name.c_str())];
	S4DRN Ny = UtilGetCfg()->S4Dmap[UtilStrFormat("Ny_%s",m->name.c_str())];
	S4DRN Nz = UtilGetCfg()->S4Dmap[UtilStrFormat("Nz_%s",m->name.c_str())];

	const int di = zi + static_cast<int>(Nx.ary->val(zi,zj,zk,Nx.N));
	const int dj = zj + static_cast<int>(Ny.ary->val(zi,zj,zk,Ny.N));
	const int dk = zk + static_cast<int>(Nz.ary->val(zi,zj,zk,Nz.N));

	if (UtilCheckIdxPdiv(NULL,NULL,NULL,di,dj,dk,0)) {
		ary->val(di,dj,dk,n) = value;
	}

	return 1;
}

int UtilGetMembParam(const string& fname, const membfunction* p, const int i, const int j, const int k, SKL_REAL* value)
{
	if (!p) { return -1; }

	int zi, zj, zk;
	bool in = UtilCheckIdxPdiv(&zi,&zj,&zk,i,j,k,1);

	if (!in) { return 0; }

	const string name = UtilStrFormat("%s:%s",fname.c_str(),p->name.c_str());
	S4DRN K = UtilGetCfg()->S4Dmap[name];
	*value = K.ary->val(zi,zj,zk,K.N);

	return 1;
}

SKL_REAL UtilGetFieldSum(const S3D_REAL* s3d)
{
	const int I = s3d->GetSize()[0];
	const int J = s3d->GetSize()[1];
	const int K = s3d->GetSize()[2];

	SKL_REAL sum = SKL_REAL(0.0);
	for (int k = 0; k < K; k++) {
	for (int j = 0; j < J; j++) {
	for (int i = 0; i < I; i++) {
		sum += s3d->val(i,j,k);
	}}}

	return sum;
}

SKL_REAL UtilGetFieldSum(const S4D_REAL* s4d, const int N)
{
	const int I = s4d->GetSize()[0];
	const int J = s4d->GetSize()[1];
	const int K = s4d->GetSize()[2];

	SKL_REAL sum = SKL_REAL(0.0);
	for (int k = 0; k < K; k++) {
	for (int j = 0; j < J; j++) {
	for (int i = 0; i < I; i++) {
		sum += s4d->val(i,j,k,N);
	}}}

	return sum;
}

SKL_REAL UtilGetMassSum(const Equation* e, const S4D_REAL* p, const S4D_REAL* vf)
{
	const int I = p->GetSize()[0];
	const int J = p->GetSize()[1];
	const int K = p->GetSize()[2];

	const int eid = e->id;
	const int mid = e->mat->id;

	SKL_REAL mass = SKL_REAL(0.0);
	for (int k = 0; k < K; k++) {
	for (int j = 0; j < J; j++) {
	for (int i = 0; i < I; i++) {
		mass += p->val(i,j,k,eid) * vf->val(i,j,k,mid);
	}}}

	return mass;
}

const char* UtilGetExtParamComment(const string& name)
{
	const char* c = 0;
	const ExtParam* p = UtilGetExtParam(name);
	if (c) { c = p->comment; }

	return c;
}

bool UtilLoadCSV(const string& filename, RealPairVector* table, const char sep)
{
	bool ret = true;

	ifstream ifs;

	try {
		ifs.open(filename.c_str());

		string line;
		while (getline(ifs, line)) {
			size_t p1 = line.find_first_of(sep);
			size_t p2 = line.find_first_of(sep, p1 + 1);
			const string f = line.substr(0,p1);
			const string s = line.substr(p1 + 1, p2 - p1 -1);

			const SKL_REAL v1 = static_cast<SKL_REAL>(atof(f.c_str()));
			const SKL_REAL v2 = static_cast<SKL_REAL>(atof(s.c_str()));
			table->push_back(RealPair(v1,v2));
		};
	} catch(...) {
		ret = false;
	}

	ifs.close();

	return ret;
}

SKL_REAL UtilLinearInterpolation(const SKL_REAL x, const SKL_REAL x1, const SKL_REAL y1, const SKL_REAL x2, const SKL_REAL y2)
{
	const SKL_REAL dy = y2 - y1;
	const SKL_REAL dx = x2 - x1;
	const SKL_REAL k = dy / dx;
	return k * (x - x1) + y1;
}

SKL_REAL UtilGetValueFromTable(const SKL_REAL x, const RealPairVector* table, const SKL_REAL offst)
{
	const SKL_REAL x0 = x - offst;

	// binary tree search
	int d = 2;
	int i = table->size() / 2;
	while (1) {
		const int oi = i;
		const bool smaller = (table->at(i).first >= x0);

		// move
		d *= 2;
		i += (smaller? -1: 1) * (table->size() / d) + (smaller? 0: 1);

		// converge ?
		if (i == oi) break;
	}

	// interpolation
	const SKL_REAL x1 = table->at(i - 1).first;
	const SKL_REAL y1 = table->at(i - 1).second;
	const SKL_REAL x2 = table->at(i).first;
	const SKL_REAL y2 = table->at(i).second;

	// check: x in range?
	if (x0 < x1) return 0.0;
	if (x2 < x0) return 0.0;

	return UtilLinearInterpolation(x0, x1, y1, x2, y2);
}

IdRange::IdRange()
{
	_begin = 0;
	_end   = 0;
}

IdRange::IdRange(const string& tuplestr)
{
	const size_t langle    = tuplestr.find_first_of('(');
	const size_t rangle    = tuplestr.find_last_of(')');
	const string rangestr  = tuplestr.substr(langle+1, rangle-1);

	const size_t sep       = rangestr.find_first_of(',');
	const string begin_str = rangestr.substr(0, sep);
	const string end_str   = rangestr.substr(sep + 1);

	_begin = atoi(begin_str.c_str());
	_end   = atoi(end_str.c_str());

	if (_begin > _end) {
		// swap(_begin,_end)
		_begin ^= _end;
		_end   ^= _begin;
		_begin ^= _end;
	}
}

int  IdRange::begin() const
{
	return _begin;
}

int  IdRange::end() const
{
	return _end;
}

bool IdRange::isWithin(const int id) const
{
	return ((_begin >= id) && (id >= _end));
}


