/*
 * Copyright (c) RIKEN, Japan. All right reserved. 2008-2009
 *
 */

#include "AdvectionSolv.h"

#include <Foundation/SklSolverFoundation.h>

//////////////////////////////////////////////////
// Constractor
//
AdvectionSolv::AdvectionSolv(const CfgInfo& _cfg)
: cfg(_cfg), vc(VC)
{
	clock    = double(0.0);
	comm_clk = double(0.0);
	wait_clk = double(0.0);
}

//////////////////////////////////////////////////
// Destractor
//
AdvectionSolv::~AdvectionSolv()
{
}

//////////////////////////////////////////////////
// Getter method

// Get enableAdvection
bool AdvectionSolv::GetEnableAdvection()
{
	return enableAdvection;
}

//////////////////////////////////////////////////
// Initialize Main
//
int AdvectionSolv::AdvectionInitialize(const CfgElem* simElem, const CfgElem* advectionElem,
					const S4D_REAL* p0, const S4D_REAL* vf, const S4D_REAL* vfi,
					const S4D_REAL* ax, const S4D_REAL* ay, const S4D_REAL* az,
					const S4D_INT* obx, const S4D_INT* oby, const S4D_INT* obz)
{
#ifndef _NO_MESSAGE
	MESSAGE("Initialize Advection Solver.");
#else
	MESSAGE("Initialize Advection Solver.", 0);
#endif

	// Advection Simulation execution check
	const CfgElem* advection = simElem? simElem->GetElemFirst("advection"): NULL;
	const CfgParam* exe = advection? advection->GetParamFirst("execution"): NULL;

	const char* exc = NULL;
	enableAdvection = (exe && exe->GetData(&exc) && (strcmp(exc, "true") == 0));

	if( !enableAdvection ){
		return 1;
	}

	// GetParaComponent and GetParaManager
	para_cmp = cfg.solver->SklGetParaComponent();
	para_mng = para_cmp->GetParaManager();

	// data class
	phi0= p0;
	Vf  = vf;
	Vfi = vfi;
	Ax  = ax;
	Ay  = ay;
	Az  = az;
	BCx = obx;
	BCy = oby;
	BCz = obz;

	// Parse Velocity file Table
	ParseVelocityTable();


	// Allocate
	total_mem = 0;
	const size_t mats = cfg.MatCount;
	const size_t eqns = cfg.EqnCount;
	AllocateVoxel(total_mem, mats, eqns);

	int i;

	// nID
	for( i=0; i<6; i++ ) m_nID[i] = -1;

#ifdef SPHERE175
	m_nID[X_MINUS] = para_mng->GetCommID(-1, 0, 0, cfg.grpP.gid);
	m_nID[Y_MINUS] = para_mng->GetCommID( 0,-1, 0, cfg.grpP.gid);
	m_nID[Z_MINUS] = para_mng->GetCommID( 0, 0,-1, cfg.grpP.gid);
	m_nID[X_PLUS]  = para_mng->GetCommID( 1, 0, 0, cfg.grpP.gid);
	m_nID[Y_PLUS]  = para_mng->GetCommID( 0, 1, 0, cfg.grpP.gid);
	m_nID[Z_PLUS]  = para_mng->GetCommID( 0, 0, 1, cfg.grpP.gid);
#else
	m_nID[X_MINUS] = ParaMngr.GetCommID(-1, 0, 0, gid);	// -X
	m_nID[Y_MINUS] = ParaMngr.GetCommID( 0,-1, 0, gid);	// -Y
	m_nID[Z_MINUS] = ParaMngr.GetCommID( 0, 0,-1, gid);	// -Z
	m_nID[X_PLUS]  = ParaMngr.GetCommID( 1, 0, 0, gid);	// +X
	m_nID[Y_PLUS]  = ParaMngr.GetCommID( 0, 1, 0, gid);	// +Y
	m_nID[Z_PLUS]  = ParaMngr.GetCommID( 0, 0, 1, gid);	// +Z

#endif



	return 1;
}

//////////////////////////////////////////////////
// Loop Main
//
int AdvectionSolv::AdvectionLoop(const unsigned int step)
{
	DB1(MESSAGE("AdvectionLoop"));

	BEGIN_CLOCK(loop);

	int I = cfg.Pdiv.size[0];
	int J = cfg.Pdiv.size[1];
	int K = cfg.Pdiv.size[2];


	if( !LoadVel() ){ return -1; }

	CopyP0();
	SetBndCell();

	for( int e=0; e<cfg.EqnCount; e++ ){
		const int m = cfg.EqnTbl[e]->mat->id;

		const SKL_REAL *Lphi0 = UtilGetS4DN(dc_P0, e);
		SKL_REAL *Ad  = UtilGetS4DN(dc_ad, e);

		const SKL_REAL *uvel= UtilGetS4DN(dc_uVel,m);
		const SKL_REAL *vvel= UtilGetS4DN(dc_vVel,m);
		const SKL_REAL *wvel= UtilGetS4DN(dc_wVel,m);
		const SKL_REAL *vf  = UtilGetS4DN(Vf,  m);
		const SKL_REAL *vfi = UtilGetS4DN(Vfi, m);
		const SKL_REAL *ax  = UtilGetS4DN(Ax,  m);
		const SKL_REAL *ay  = UtilGetS4DN(Ay,  m);
		const SKL_REAL *az  = UtilGetS4DN(Az,  m);

		calc_advection_(Lphi0,Ad,uvel,vvel,wvel,vfi,ax,ay,az,
				&I,&J,&K,&vc,&cfg.pitch[0],&cfg.pitch[1],&cfg.pitch[2]);
	}

	END_CLOCK(clock, loop);

	return 1;
}

//////////////////////////////////////////////////
// Post Main
//
int AdvectionSolv::AdvectionPost()
{
	return 1;
}


//////////////////////////////////////////////////
// Check config
//
bool AdvectionSolv::chkCfg()
{
	if( cfg.grpAll.nid == 0 ){
#ifndef _NO_MESSAGE
		MESSAGE("**** Configuration of Advection Solver ****");
		MESSAGE(string("Advection Simulation is ")
			+ string_cast(enableAdvection, ios::boolalpha));
#else
		MESSAGE("**** Configuration of Advection Solver ****", 0);
		MESSAGE(string("Advection Simulation is ")
			+ string_cast(enableAdvection, ios::boolalpha),0);
#endif //_NO_MESSAGE
	}

	if( !enableAdvection ) { return true; }

	return true;
}

//////////////////////////////////////////////////
// Parse Velocity Table
//
bool AdvectionSolv::ParseVelocityTable()
{

	velFileSize = 0;

	const CfgElem* DomainInfo = cfg.solver->SklCfgGetTop(DOMAININFO);
	const CfgElem* cell       = DomainInfo->GetElemFirst("cell");

	while( cell ){
		const CfgElem* mat = cell->GetElemFirst();
		while( mat ){
			const char* mName = mat->GetName();

			const CfgElem* vel = mat->GetElemFirst("velocity");
			while( vel ){
				int velFileCnt = 0;
				const CfgElem* table = vel->GetElemFirst("table");
				while( table ){
					VelFile* v = new VelFile;

					// Material Name
					v->matName = mName;

					const CfgParam* param;

					// time
					float f_time = 0.0;
					if( param = table->GetParamFirst("time") ){
						if( param->GetData(&f_time) ){
							v->time = f_time;
						}
					}

					// attr
					const char* c_attr = NULL;
					if( param = table->GetParamFirst("attr") ){
						if( param->GetData(&c_attr) ){
							v->attr = c_attr;
						}
					}

					// error check
					if( c_attr == NULL ){
						MSG_WARNING("Velocity file Table attr is NULL...");
					}
					else{
						// Velocity file Table Parameter is OK.
						VelFileTbl.push_back(v);
						velFileCnt++;
					}

					table = vel->GetElemNext(table, "table");
				}

				// Velocity file is only one file in Material.
				if( velFileCnt == 0 ){
					VelFile* v = new VelFile;
					v->matName = mName;
					v->time    = 0.0;
					v->attr    = string_cast(mName) + ":Vel";
					VelFileTbl.push_back(v);
					velFileCnt++;
				}
				velFileSize += velFileCnt;

				vel = mat->GetElemNext(vel, "velocity");
			}

			mat = cell->GetElemNext(mat, NULL);
		}

		cell = DomainInfo->GetElemNext(cell, "cell");
	}

	if( velFileSize == 0 ){
		MSG_WARNING("Velocity file count is 0 ...");
	}


	// debug write
	for(int i=0; i<velFileSize; i++){
#ifndef _NO_MESSAGE
		MESSAGE("  Velocity file table status");
		MESSAGE(UtilStrFormat("    mat  : %s", VelFileTbl[i]->matName.c_str()));
		MESSAGE(UtilStrFormat("    time : %e", VelFileTbl[i]->time));
		MESSAGE(UtilStrFormat("    attr : %s", VelFileTbl[i]->attr.c_str()));
#else
		MESSAGE("  Velocity file table status", 0);
		MESSAGE(UtilStrFormat("    mat  : %s", VelFileTbl[i]->matName.c_str()), 0);
		MESSAGE(UtilStrFormat("    time : %e", VelFileTbl[i]->time), 0);
		MESSAGE(UtilStrFormat("    attr : %s", VelFileTbl[i]->attr.c_str()), 0);
#endif //_NO_MESSAGE
	}


	return true;
}


//////////////////////////////////////////////////
// CalcBndCell
//
SKL_REAL AdvectionSolv::CalcBndCell(const int bc,
				    const SKL_REAL c,
				    const SKL_REAL dx,
				    const int face)
{
	const BoundaryCond& cond = cfg.BndCndTbl[bc];
	const SKL_REAL&     val  = cond.value;
	const SKL_REAL      sign = static_cast<SKL_REAL>(face);

	SKL_REAL conc =
		(cond.type == BoundaryCond::DIRICHLET)? val:
		// (cond.type == BoundaryCond::DIRICHLET)? SKL_REAL(2.0) * val - c:
		(cond.type == BoundaryCond::NEUMANN)?   c + sign * val * dx:
							0;

	return conc;
}

//////////////////////////////////////////////////
// SetBndCell
//
bool AdvectionSolv::SetBndCell()
{
	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];

	const SKL_REAL& dx = cfg.pitch[0];
	const SKL_REAL& dy = cfg.pitch[1];
	const SKL_REAL& dz = cfg.pitch[2];

	for( int e = 0; e < cfg.EqnCount; e++ ){
		// Set outer boundary coef
		// X boundary
		if( m_nID[X_MINUS] < 0 ){
			for( int k=0; k<K; k++ ){
			for( int j=0; j<J; j++ ){
				const int      bc = BCx->val(0,j,k,e);
				const SKL_REAL p  = dc_P0     ->val(0,j,k,e);
				dc_P0->val(-1,j,k,e) = CalcBndCell(bc,p,dx,-1);
			}}
		}
		if( m_nID[X_PLUS] < 0 ){
			for( int k=0; k<K; k++ ){
			for( int j=0; j<J; j++ ){
				const int      bc = BCx->val(  1,j,k,e);
				const SKL_REAL p  = dc_P0     ->val(I-1,j,k,e);
				dc_P0->val(I,j,k,e)  = CalcBndCell(bc,p,dx,1);
			}}
		}
		// Y boundary
		if( m_nID[Y_MINUS] < 0 ){
			for( int k=0; k<K; k++ ){
			for( int i=0; i<I; i++ ){
				const int      bc = BCy->val(i,0,k,e);
				const SKL_REAL p  = dc_P0     ->val(i,0,k,e);
				dc_P0->val(i,-1,k,e) = CalcBndCell(bc,p,dy,-1);
			}}
		}
		if( m_nID[Y_PLUS] < 0 ){
			for( int k=0; k<K; k++ ){
			for( int i=0; i<I; i++ ){
				const int      bc = BCy->val(i,  1,k,e);
				const SKL_REAL p  = dc_P0     ->val(i,J-1,k,e);
				dc_P0->val(i,J,k,e)  = CalcBndCell(bc,p,dy,1);
			}}
		}
		// Z boundary
		if( m_nID[Z_MINUS] < 0 ){
			for( int j=0; j<J; j++ ){
			for( int i=0; i<I; i++ ){
				const int      bc = BCz->val(i,j,0,e);
				const SKL_REAL p  = dc_P0     ->val(i,j,0,e);
				dc_P0->val(i,j,-1,e) = CalcBndCell(bc,p,dz,-1);
			}}
		}
		if( m_nID[Z_PLUS] < 0 ){
			for( int j=0; j<J; j++ ){
			for( int i=0; i<I; i++ ){
				const int      bc = BCz->val(i,j,  1,e);
				const SKL_REAL p  = dc_P0     ->val(i,j,K-1,e);
				dc_P0->val(i,j,K,e)  = CalcBndCell(bc,p,dz,1);
			}}
		}
	}

	return true;
}


//////////////////////////////////////////////////
// Exptim Advection
//
int AdvectionSolv::ExptimAdv(const SKL_REAL dt,
			     SklScalar4D<SKL_REAL> *RICS_P1)
{
	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];

	for( int e = 0; e < cfg.EqnCount; e++ ){
		SKL_REAL *P1  = UtilGetS4DN(RICS_P1,e);
		SKL_REAL *Ad  = UtilGetS4DN(dc_ad,e);

		exptimadv_(P1,Ad,&dt,&I,&J,&K,&vc);
	}

	return 1;
}

//////////////////////////////////////////////////
// CopyP0
//
//bool AdvectionSolv::CopyP0(SklScalar4D<SKL_REAL> *RICS_P0)
bool AdvectionSolv::CopyP0()
{
	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];

	int vcell = vc;

	for( int e = 0; e < cfg.EqnCount; e++ ){
		// const SKL_REAL *src = UtilGetS4DN(RICS_P0,e);
		const SKL_REAL *src = UtilGetS4DN(phi0,e);
		const SKL_REAL *dst = UtilGetS4DN(dc_P0,e);

		ricscpys3d_(dst,src,&I,&J,&K,&vcell);
	}

	return true;
}

