/*
 * Copyright (c) RIKEN, Japan. All right reserved. 2008-2009
 *
 */

#include <SklUtil.h>

#include <iostream>
#include <cmath>
#include <cstring>
#include <map>
#include <cfloat>

#include <Foundation/const.h>
#include <Foundation/util.h>
#include <Foundation/error.h>
#include <Foundation/datatype.h>
#include <Foundation/SklSolverFoundation.h>


#include "DiffusSolv.h"
#include "FortranFuncDiffEq.h"

#define D_DIFF_BASIC 0
#define D_DIFF_26WAY 1

DiffusSolv::DiffusSolv(const CfgInfo& _cfg)
: cfg(_cfg), vc(VC)
{
	clock    = double(0.0);
	comm_clk = double(0.0);
	wait_clk = double(0.0);
}

int DiffusSolv::DiffusionInitialize(
				const CfgElem* sim,
				S4D_REAL* fx, S4D_REAL* fy, S4D_REAL* fz,
				const S4D_REAL* p0,
				S4D_REAL* dcX, S4D_REAL* dcY, S4D_REAL* dcZ,
				const S4D_REAL* baX, const S4D_REAL* baY, const S4D_REAL* baZ,
				const S4D_INT*  obX, const S4D_INT*  obY, const S4D_INT*  obZ,
				S4D_REAL* matVfInv, S4D_REAL* rhs, S4D_REAL* rhso
			)
{
#ifdef SPHERE175
	//GetParaComponent, GetParaManager	2009.12.07
	para_cmp = cfg.solver->SklGetParaComponent();
	para_mng = para_cmp->GetParaManager();
#endif

	P0    = p0;
	Ax    = baX;
	Ay    = baY;
	Az    = baZ;
	Cx    = dcX;
	Cy    = dcY;
	Cz    = dcZ;
	FluxX = fx;
	FluxY = fy;
	FluxZ = fz;

	BCx = obX;
	BCy = obY;
	BCz = obZ;

	MatVfInv = matVfInv;
	Rhs  = rhs;
	Rhso = rhso;

	const CfgElem*  dif = sim? sim->GetElemFirst ("diffusion"): NULL;
	const CfgParam* exe = dif? dif->GetParamFirst("execution"): NULL;
        const CfgParam* diff_type = dif? dif->GetParamFirst("diff_type"): NULL;

	// set isenable metabolism simulation
	const char* exc = NULL;
	enableDiffusion = (exe && exe->GetData(&exc) && (strcmp(exc, "true") == 0));
	if (!enableDiffusion) { return true; }

	// set flag diffusion type
	exc = NULL;
	if(diff_type && diff_type->GetData(&exc) && (strcmp(exc, "basic") == 0)){
		diffTypeFlag = D_DIFF_BASIC;
#ifndef _NO_MESSAGE
		MESSAGE(string("DiffusionType. basic diffTypeFlag : ") + string_cast(diffTypeFlag));
#else
		MESSAGE(string("DiffusionType. basic diffTypeFlag : ") + string_cast(diffTypeFlag),0);
#endif //_NO_MESSAGE
	}
	else{
		diffTypeFlag = D_DIFF_26WAY;
#ifndef _NO_MESSAGE
		MESSAGE(string("DiffusionType. 26way diffTypeFlag : ") + string_cast(diffTypeFlag));
#else
		MESSAGE(string("DiffusionType. 26way diffTypeFlag : ") + string_cast(diffTypeFlag),0);
#endif //_NO_MESSAGE
	}

	////////////////////////////////////////////////////////////////////////
	// check next node id
	const int& gid = cfg.grpP.gid;
//2009.12.08

#ifdef SPHERE175
	nID[0] = para_mng->IsCommID(-1, 0, 0, gid); // -X
	nID[1] = para_mng->IsCommID( 1, 0, 0, gid); // +X
	nID[2] = para_mng->IsCommID( 0,-1, 0, gid); // -Y
	nID[3] = para_mng->IsCommID( 0, 1, 0, gid); // +Y
	nID[4] = para_mng->IsCommID( 0, 0,-1, gid); // -Z
	nID[5] = para_mng->IsCommID( 0, 0, 1, gid); // +Z
#else
	nID[0] = ParaMngr.GetCommID(-1, 0, 0, gid); // -X
	nID[1] = ParaMngr.GetCommID( 1, 0, 0, gid); // +X
	nID[2] = ParaMngr.GetCommID( 0,-1, 0, gid); // -Y
	nID[3] = ParaMngr.GetCommID( 0, 1, 0, gid); // +Y
	nID[4] = ParaMngr.GetCommID( 0, 0,-1, gid); // -Z
	nID[5] = ParaMngr.GetCommID( 0, 0, 1, gid); // +Z
#endif

	////////////////////////////////////////////////////////////////////////
	// init fields
	Allocate();


	SetBndCellCenterCoef();

	return 1;
}

int DiffusSolv::Allocate()
{
#ifdef SPHERE175
	//GetParaComponent, GetParaManager	2009.12.07
	para_cmp = cfg.solver->SklGetParaComponent();
	para_mng = para_cmp->GetParaManager();
#endif

	const size_t& I    = cfg.Pdiv.size[0];
	const size_t& J    = cfg.Pdiv.size[1];
	const size_t& K    = cfg.Pdiv.size[2];
	const size_t& mols = cfg.MolCount;
	const size_t& mats = cfg.MatCount;
	const size_t& eqns = cfg.EqnCount;

	const size_t sz_eqn[4] = { I, J, K, eqns };
	const size_t sz_mol[4] = { I, J, K, mols };
	const size_t sz_mat[4] = { I, J, K, mats };

	const int& gid = cfg.grpP.gid;

	//2009.12.07
#ifdef SPHERE175
	p0 = dynamic_cast<S4D_REAL*>(
				cfg.solver->SklAllocateArray(para_mng, "P0", SKL_CLASS_SCALAR4D,
								SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
	cx = dynamic_cast<S4D_REAL*>(
				cfg.solver->SklAllocateArray(para_mng, "cx", SKL_CLASS_SCALAR4D,
								SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
	cy = dynamic_cast<S4D_REAL*>(
				cfg.solver->SklAllocateArray(para_mng, "cy", SKL_CLASS_SCALAR4D,
								SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
	cz = dynamic_cast<S4D_REAL*>(
				cfg.solver->SklAllocateArray(para_mng, "cz", SKL_CLASS_SCALAR4D,
								SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
#else
	p0 = dynamic_cast<S4D_REAL*>(
				cfg.solver->SklAllocateArray("P0", SKL_CLASS_SCALAR4D,
								SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
	cx = dynamic_cast<S4D_REAL*>(
				cfg.solver->SklAllocateArray("cx", SKL_CLASS_SCALAR4D,
								SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
	cy = dynamic_cast<S4D_REAL*>(
				cfg.solver->SklAllocateArray("cy", SKL_CLASS_SCALAR4D,
								SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
	cz = dynamic_cast<S4D_REAL*>(
				cfg.solver->SklAllocateArray("cz", SKL_CLASS_SCALAR4D,
								SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
#endif

	return 1;
}


int DiffusSolv::chkCfg()
{
	if (cfg.grpAll.nid != 0) { return 0; }

#ifndef _NO_MESSAGE
	MESSAGE("**** Configuration of Diffusion Solver ****");
	MESSAGE(string("Diffusion Simulation is ")
					+ string_cast(enableDiffusion, ios::boolalpha));
#else
	MESSAGE("**** Configuration of Diffusion Solver ****", 0);
	MESSAGE(string("Diffusion Simulation is ")
					+ string_cast(enableDiffusion, ios::boolalpha),0);
#endif //_NO_MESSAGE

	if (!enableDiffusion) { return 1; }

	return 1;
}

int DiffusSolv::SetBndCellCenterCoef()
{
	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];

	for( int e = 0; e < cfg.EqnCount; e++ ){
		if (!cfg.PeqnAss[e]) { continue; }

		SKL_REAL* _Cx  = Cx->GetData();
		SKL_REAL* _Cy  = Cy->GetData();
		SKL_REAL* _Cz  = Cz->GetData();
		const int eqns = cfg.EqnCount;

		setbndcoef_(_Cx,_Cy,_Cz,nID,&e,&eqns,&I,&J,&K,&vc);
	}

	Cx->CommBndCell(Cx->GetVCellSize());
	Cy->CommBndCell(Cy->GetVCellSize());
	Cz->CommBndCell(Cz->GetVCellSize());

	return 1;
}

SKL_REAL DiffusSolv::CourantCond()
{
	// courant condition
	const int I = cfg.Pdiv.size[0]; 
	const int J = cfg.Pdiv.size[1]; 
	const int K = cfg.Pdiv.size[2]; 

	const SKL_REAL& dx = cfg.pitch[0];
	const SKL_REAL& dy = cfg.pitch[1];
	const SKL_REAL& dz = cfg.pitch[2];

	SKL_REAL Cxm = SKL_REAL(REAL_MIN);
	SKL_REAL Cym = SKL_REAL(REAL_MIN);
	SKL_REAL Czm = SKL_REAL(REAL_MIN);

	for (int e = 0; e < cfg.EqnCount; e++) {
		if (!cfg.PeqnAss[e]) { continue; }
		for (int i = 0; i < I; i++) {
		for (int j = 0; j < J; j++) {
		for (int k = 0; k < K; k++) {
			Cxm = (Cxm < Cx->val(i,j,k,e))? Cx->val(i,j,k,e): Cxm;
			Cym = (Cym < Cy->val(i,j,k,e))? Cy->val(i,j,k,e): Cym;
			Czm = (Czm < Cz->val(i,j,k,e))? Cz->val(i,j,k,e): Czm;
		}}}
	}
	
	const SKL_REAL  CCx = dx * dx / (SKL_REAL(2.0) * Cxm);
	const SKL_REAL  CCy = dy * dy / (SKL_REAL(2.0) * Cym);
	const SKL_REAL  CCz = dz * dz / (SKL_REAL(2.0) * Czm);

	const SKL_REAL  CC  = ((CCx < CCy) && (CCx < CCz))? CCx:
						  ((CCy < CCx) && (CCy < CCz))? CCy:
						                                CCz;
	return CC;
}

int DiffusSolv::CalcCoef()
{
	//
	//  center definition to Boundary definition.
	//
	DB2(MESSAGE("CalcCoef"));

	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];

	for (int e = 0; e < cfg.EqnCount; e++) {
		if (!cfg.PeqnAss[e]) { continue; }

		SKL_REAL* _cx  = cx->GetData();
		SKL_REAL* _cy  = cy->GetData();
		SKL_REAL* _cz  = cz->GetData();
		SKL_REAL* _Cx  = Cx->GetData();
		SKL_REAL* _Cy  = Cy->GetData();
		SKL_REAL* _Cz  = Cz->GetData();
		const int eqns = cfg.EqnCount;

		setbndcoef_(_Cx,_Cy,_Cz,nID,&e,&eqns,&I,&J,&K,&vc);
		calccoef_(_cx,_cy,_cz,_Cx,_Cy,_Cz,nID,&e,&eqns,&I,&J,&K,&vc);
	}


	// cx->CommBndCell(cx->GetVCellSize());
	// cy->CommBndCell(cy->GetVCellSize());
	// cz->CommBndCell(cz->GetVCellSize());

	return 1;
}

SKL_REAL DiffusSolv::CalcBndCell(
	const int bc, const SKL_REAL c, const SKL_REAL dx, const int face)
{
	const BoundaryCond& cond = cfg.BndCndTbl[bc];
	const SKL_REAL&     val  = cond.value;
	const SKL_REAL      sign = static_cast<SKL_REAL>(face);

	SKL_REAL conc = 
		(cond.type == BoundaryCond::DIRICHLET)? val:
		// (cond.type == BoundaryCond::DIRICHLET)? SKL_REAL(2.0) * val - c:
		(cond.type == BoundaryCond::NEUMANN)?   c + sign * val * dx:
		                                        0;

	return conc;
}

int DiffusSolv::SetBndCell()
{
	DB2(MESSAGE("SetBndCell "));

	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];

	int vcd = 0;
	if( diffTypeFlag == D_DIFF_26WAY ){
		vcd = 1;
	}

	for (int e = 0; e < cfg.EqnCount; e++) {
		if (!cfg.PeqnAss[e]) { continue; }

		DB3(MESSAGE(string("SetBndCell ")+cfg.EqnTbl[e]->name));

		const size_t    m  = cfg.EqnTbl[e]->mat->id;
		const SKL_REAL& dx = cfg.pitch[0];
		const SKL_REAL& dy = cfg.pitch[1];
		const SKL_REAL& dz = cfg.pitch[2];

		// set outter boundary coef.
		// X boundary
		if (nID[0] < 0) {
			for (int k = (0-vcd); k < (K+vcd); k++) {
			for (int j = (0-vcd); j < (J+vcd); j++) {
				const int      bc = BCx->val(0,j,k,e);
				const SKL_REAL p  = p0 ->val(0,j,k,e);
				p0->val(-1,j,k,e) = CalcBndCell(bc,p,dx,-1);
			}}
		}
		if (nID[1] < 0) {
			for (int k = (0-vcd); k < (K+vcd); k++) {
			for (int j = (0-vcd); j < (J+vcd); j++) {
				const int      bc = BCx->val(  1,j,k,e);
				const SKL_REAL p  = p0 ->val(I-1,j,k,e);
				p0->val(I,j,k,e)  = CalcBndCell(bc,p,dx, 1);
			}}
		}
		// Y boudary
		if (nID[2] < 0) {
			for (int k = (0-vcd); k < (K+vcd); k++) {
			for (int i = (0-vcd); i < (I+vcd); i++) {
				const int      bc = BCy->val(i,0,k,e);
				const SKL_REAL p  = p0 ->val(i,0,k,e);
				p0->val(i,-1,k,e) = CalcBndCell(bc,p,dy,-1);
			}}
		}
		if (nID[3] < 0) {
			for (int k = (0-vcd); k < (K+vcd); k++) {
			for (int i = (0-vcd); i < (I+vcd); i++) {
				const int      bc = BCy->val(i,  1,k,e);
				const SKL_REAL p  = p0 ->val(i,J-1,k,e);
				p0->val(i,J,k,e)  = CalcBndCell(bc,p,dy, 1);
			}}
		}
		// Z boundary
		if (nID[4] < 0) {
			for (int j = (0-vcd); j < (J+vcd); j++) {
			for (int i = (0-vcd); i < (I+vcd); i++) {
				const int      bc = BCz->val(i,j,0,e);
				const SKL_REAL p  = p0 ->val(i,j,0,e);
				p0->val(i,j,-1,e) = CalcBndCell(bc,p,dz,-1);
			}}
		}
		if (nID[5] < 0) {
			for (int j = (0-vcd); j < (J+vcd); j++) {
			for (int i = (0-vcd); i < (I+vcd); i++) {
				const int      bc = BCz->val(i,j,  1,e);
				const SKL_REAL p  = p0 ->val(i,j,K-1,e);
				p0->val(i,j,K,e)  = CalcBndCell(bc,p,dz, 1);
			}}
		}

#if 0
		DB5(MESSAGE(S4D2string(p0,e)));
#endif
	}

	// MEASURE_WAIT_TIME(wait_clk, para_mng->Barrier(cfg.grpP.gid)); // 2011.05.11
	// BEGIN_CLOCK(comm)
	// p0->CommBndCell(p0->GetVCellSize());
	// END_CLOCK(comm_clk, comm);

	return 1;
}

int DiffusSolv::DiffusionLoop(const SKL_REAL time)
{
	if((cfg.time_integ == EULER) || (cfg.time_integ == ADAMS_BASHFORTH)){
		return DiffusionLoopEuler(time);
	}
	else if((cfg.time_integ == RUNGE_KUTTA_2) || (cfg.time_integ == RUNGE_KUTTA_4)){
		return DiffusionLoopRK(time);
	}
	else{
		return DiffusionLoopEuler(time);
	}
}

int DiffusSolv::DiffusionLoopEuler(const SKL_REAL time)
{
	if (!enableDiffusion) { return 0; }

	BEGIN_CLOCK(loop);

	const SKL_REAL& dx = cfg.pitch[0], dxi = SKL_REAL(1.0)/dx;
	const SKL_REAL& dy = cfg.pitch[1], dyi = SKL_REAL(1.0)/dy;
	const SKL_REAL& dz = cfg.pitch[2], dzi = SKL_REAL(1.0)/dz;

	const SKL_REAL 	Vol  = dx * dy * dz;
	const SKL_REAL 	Voli = SKL_REAL(1.0) / Vol;

	const int eqns = cfg.EqnCount;
	const int mols = cfg.MolCount;
	const int mats = cfg.MatCount;

	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];

	const int vc = lesser(p0->GetVCellSize(), P0->GetVCellSize());

	// set dc_P
	for (int e = 0; e < eqns; e++) {
		if (!cfg.PeqnAss[e]) { continue; }

#if ENABLE_MEMBRANEPOTENTIAL 
                if( cfg.EqnTbl[e]->mol->potential ) { continue; }
#endif

		DB3(MESSAGE(string("set dc_P ")+cfg.EqnTbl[e]->name));

		const SKL_REAL *dst = UtilGetS4DN(p0,e);
	        const SKL_REAL *src = UtilGetS4DN(P0,e);

		ricscpys3d_(dst,src,&I,&J,&K,&vc);
#if 0
		DB5(MESSAGE(S4D2string(p0,e)));
#endif
	}

	// calc Coef
	CalcCoef();

	// set boundary
	SetBndCell();

	for (int e = 0; e < cfg.EqnCount; e++) {
		if (!cfg.PeqnAss[e]) { continue; }

#if ENABLE_MEMBRANEPOTENTIAL 
                if( cfg.EqnTbl[e]->mol->potential ) { continue; }
#endif
		DB3(MESSAGE(string("Diffusion ") + cfg.EqnTbl[e]->name));

		const int ml = cfg.EqnTbl[e]->mol->id;
		const int mt = cfg.EqnTbl[e]->mat->id;

		const SKL_REAL *p   = UtilGetS4DN(p0,e);
		const SKL_REAL *ax  = UtilGetS4DN(Ax,mt);
		const SKL_REAL *ay  = UtilGetS4DN(Ay,mt);
		const SKL_REAL *az  = UtilGetS4DN(Az,mt);
		const SKL_REAL *cfx = UtilGetS4DN(cx,e);
		const SKL_REAL *cfy = UtilGetS4DN(cy,e);
		const SKL_REAL *cfz = UtilGetS4DN(cz,e);
		SKL_REAL       *dfx = UtilGetS4DN(FluxX,e);
		SKL_REAL       *dfy = UtilGetS4DN(FluxY,e);
		SKL_REAL       *dfz = UtilGetS4DN(FluxZ,e);

		differential_(&diffTypeFlag,p,cfx,cfy,cfz,ax,ay,az,dfx,dfy,dfz,
							&I,&J,&K,&vc,&dxi,&dyi,&dzi);

		DB5(MESSAGE(string("P0 of ")+cfg.EqnTbl[e]->name));
		DB5(MESSAGE(S4D2string(p0, e)));

		DB5(MESSAGE(string("Area for X of ")+cfg.MatTbl[mt]->name));
		DB5(MESSAGE(S4D2string(Ax,mt)));
		DB5(MESSAGE(string("Area for Y of ")+cfg.MatTbl[mt]->name));
		DB5(MESSAGE(S4D2string(Ay,mt)));
		DB5(MESSAGE(string("Area for Z of ")+cfg.MatTbl[mt]->name));
		DB5(MESSAGE(S4D2string(Az,mt)));

		DB5(MESSAGE(string("VC CoefX of ")+cfg.EqnTbl[e]->name));
		DB5(MESSAGE(S4D2string(Cx,e)));
		DB5(MESSAGE(string("VC CoefY of ")+cfg.EqnTbl[e]->name));
		DB5(MESSAGE(S4D2string(Cy,e)));
		DB5(MESSAGE(string("VC CoefZ of ")+cfg.EqnTbl[e]->name));
		DB5(MESSAGE(S4D2string(Cz,e)));

		DB5(MESSAGE(string("VC CoefX of ")+cfg.EqnTbl[e]->name));
		DB5(MESSAGE(S4D2string(cx,e,true)));
		DB5(MESSAGE(string("VC CoefY of ")+cfg.EqnTbl[e]->name));
		DB5(MESSAGE(S4D2string(cy,e,true)));
		DB5(MESSAGE(string("VC CoefZ of ")+cfg.EqnTbl[e]->name));
		DB5(MESSAGE(S4D2string(cz,e,true)));

		DB5(MESSAGE(string("Flux for X of ")+cfg.EqnTbl[e]->name));
		DB5(MESSAGE(S4D2string(FluxX,e,true)));
		DB5(MESSAGE(string("Flux for Y of ")+cfg.EqnTbl[e]->name));
		DB5(MESSAGE(S4D2string(FluxY,e,true)));
		DB5(MESSAGE(string("Flux for Z of ")+cfg.EqnTbl[e]->name));
		DB5(MESSAGE(S4D2string(FluxZ,e,true)));
	}

	END_CLOCK(clock, loop);

	return 1;
}

int DiffusSolv::DiffusionLoopRK(const SKL_REAL time)
{
	if (!enableDiffusion) { return 0; }

        BEGIN_CLOCK(loop);

        const SKL_REAL& dx = cfg.pitch[0], dxi = SKL_REAL(1.0)/dx;
        const SKL_REAL& dy = cfg.pitch[1], dyi = SKL_REAL(1.0)/dy;
        const SKL_REAL& dz = cfg.pitch[2], dzi = SKL_REAL(1.0)/dz;

        const SKL_REAL  Vol  = dx * dy * dz;
        const SKL_REAL  Voli = SKL_REAL(1.0) / Vol;

        const int eqns = cfg.EqnCount;
        const int mols = cfg.MolCount;
        const int mats = cfg.MatCount;

        const int I = cfg.Pdiv.size[0];
        const int J = cfg.Pdiv.size[1];
        const int K = cfg.Pdiv.size[2];

        const int vc = lesser(p0->GetVCellSize(), P0->GetVCellSize());

	int e;
        // set dc_P
	for (e = 0; e < eqns; e++) {
                if (!cfg.PeqnAss[e]) { continue; }
#if ENABLE_MEMBRANEPOTENTIAL 
                if( cfg.EqnTbl[e]->mol->potential ) { continue; }
#endif
                DB3(MESSAGE(string("set dc_P ")+cfg.EqnTbl[e]->name));

                const SKL_REAL *dst = UtilGetS4DN(p0,e);
                const SKL_REAL *src = UtilGetS4DN(P0,e);

                ricscpys3d_(dst,src,&I,&J,&K,&vc);
#if 0
		DB5(MESSAGE(S4D2string(p0,e)));
#endif
	}

	// calc Coef
	CalcCoef();

	//Runge-Kutta 2 or 4
	int n = 0;
	if(cfg.time_integ == RUNGE_KUTTA_2){
		n = 2;	//Runge-Kutta-2
	}
	else{
		n = 4;	//Runge-Kutta-4
	}

	SKL_REAL c1[4];
	c1[0] = 1.0 / 6.0;
	c1[1] = 1.0 / 3.0;
	c1[2] = 1.0 / 3.0;
	c1[3] = 1.0 / 6.0;

	SKL_REAL c2[3];
	if(n == 2){
		c2[0] = 0.5;
		c2[1] = 0.0;
		c2[2] = 0.0;
	}
	else{
		c2[0] = 0.5;
		c2[1] = 0.5;
		c2[2] = 1.0;
	}

	for(int i = 0; i < n; i++){

		// set boundary
		SetBndCell();

		for (e = 0; e < cfg.EqnCount; e++) {
	                if (!cfg.PeqnAss[e]) { continue; }
#if ENABLE_MEMBRANEPOTENTIAL 
									if( cfg.EqnTbl[e]->mol->potential ) { continue; }
#endif

	                DB3(MESSAGE(string("Diffusion ") + cfg.EqnTbl[e]->name));

	                const int ml = cfg.EqnTbl[e]->mol->id;
	                const int mt = cfg.EqnTbl[e]->mat->id;

	                const SKL_REAL *p   = UtilGetS4DN(p0,e);
	                const SKL_REAL *ax  = UtilGetS4DN(Ax,mt);
	                const SKL_REAL *ay  = UtilGetS4DN(Ay,mt);
	                const SKL_REAL *az  = UtilGetS4DN(Az,mt);
	                const SKL_REAL *cfx = UtilGetS4DN(cx,e);
	                const SKL_REAL *cfy = UtilGetS4DN(cy,e);
	                const SKL_REAL *cfz = UtilGetS4DN(cz,e);
	                SKL_REAL       *dfx = UtilGetS4DN(FluxX,e);
	                SKL_REAL       *dfy = UtilGetS4DN(FluxY,e);
	                SKL_REAL       *dfz = UtilGetS4DN(FluxZ,e);

			const SKL_REAL *vfi= UtilGetS4DN(MatVfInv,mt);
			const SKL_REAL *P  = UtilGetS4DN(P0,e);
			SKL_REAL *rr;
			SKL_REAL *rrtmp;

			if(n == 2){
				rr    = UtilGetS4DN(Rhs,e);
				rrtmp = NULL;
			}
			else{
				rr    = UtilGetS4DN(Rhso,e);
				rrtmp = UtilGetS4DN(Rhs,e);
			}

	                differential_(&diffTypeFlag,p,cfx,cfy,cfz,ax,ay,az,dfx,dfy,dfz,
	                                                        &I,&J,&K,&vc,&dxi,&dyi,&dzi);

			calcrhs_(rr,vfi,dfx,dfy,dfz,&I,&J,&K,&vc,&dxi,&dyi,&dzi);

			if(rrtmp){
				if(i == 0){
					ricscpys3dcoef_(rrtmp,rr,&I,&J,&K,&vc,&c1[i]);
				}
				else{
					ricsadds3dcoef_(rrtmp,rr,&I,&J,&K,&vc,&c1[i]);
				}
			}

			if(i == (n - 1)) continue;

			if(i > 0){
				ricscpys3d_(p,P,&I,&J,&K,&vc);
			}

			exptimrk_diff_(p,rr,&time,&c2[i],&I,&J,&K,&vc);

			DB5(MESSAGE(string("P0 of ")+cfg.EqnTbl[e]->name));
	                DB5(MESSAGE(S4D2string(p0, e)));

	                DB5(MESSAGE(string("Area for X of ")+cfg.MatTbl[mt]->name));
	                DB5(MESSAGE(S4D2string(Ax,mt)));
	                DB5(MESSAGE(string("Area for Y of ")+cfg.MatTbl[mt]->name));
	                DB5(MESSAGE(S4D2string(Ay,mt)));
	                DB5(MESSAGE(string("Area for Z of ")+cfg.MatTbl[mt]->name));
	                DB5(MESSAGE(S4D2string(Az,mt)));

	                DB5(MESSAGE(string("VC CoefX of ")+cfg.EqnTbl[e]->name));
	                DB5(MESSAGE(S4D2string(Cx,e)));
	                DB5(MESSAGE(string("VC CoefY of ")+cfg.EqnTbl[e]->name));
	                DB5(MESSAGE(S4D2string(Cy,e)));
	                DB5(MESSAGE(string("VC CoefZ of ")+cfg.EqnTbl[e]->name));
	                DB5(MESSAGE(S4D2string(Cz,e)));

	                DB5(MESSAGE(string("VC CoefX of ")+cfg.EqnTbl[e]->name));
	                DB5(MESSAGE(S4D2string(cx,e,true)));
	                DB5(MESSAGE(string("VC CoefY of ")+cfg.EqnTbl[e]->name));
	                DB5(MESSAGE(S4D2string(cy,e,true)));
	                DB5(MESSAGE(string("VC CoefZ of ")+cfg.EqnTbl[e]->name));
	                DB5(MESSAGE(S4D2string(cz,e,true)));

	                DB5(MESSAGE(string("Flux for X of ")+cfg.EqnTbl[e]->name));
	                DB5(MESSAGE(S4D2string(FluxX,e,true)));
	                DB5(MESSAGE(string("Flux for Y of ")+cfg.EqnTbl[e]->name));
	                DB5(MESSAGE(S4D2string(FluxY,e,true)));
	                DB5(MESSAGE(string("Flux for Z of ")+cfg.EqnTbl[e]->name));
	                DB5(MESSAGE(S4D2string(FluxZ,e,true)));
	        }
		
		// 2011.05.11
		MEASURE_WAIT_TIME(wait_clk, para_mng->Barrier(cfg.grpP.gid));
		BEGIN_CLOCK(comm)
		p0->CommBndCell(p0->GetVCellSize());
		END_CLOCK(comm_clk, comm);
	}
	
	END_CLOCK(clock, loop);

	return 1;
}

int DiffusSolv::DiffusionPost()
{
	return 1;
}

