//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#include <iostream>
#include <map>
#include <cmath>
#include <cstring>
#include <functional>

#include <SklUtil.h>

#include <Foundation/const.h>
#include <Foundation/util.h>
#include <Foundation/error.h>
#include <Foundation/datatype.h>
#include <Foundation/api.h>
#include <Foundation/SklSolverFoundation.h>

#include "FortranFuncMembEq.h"
#include "MembraneSolv.h"
#include "SwController.h"

MembraneSolv::MembraneSolv(CfgInfo& _cfg)
:cfg(_cfg), vc(VC)
{
	comm_clk = double(0.0);
}

int MembraneSolv::MembraneInitialize(
				const CfgElem* sim,
				S4D_REAL* memtrm,
				S4D_REAL* flux,
				S4D_REAL* q,
				const S4D_REAL* p0,
				const S4D_REAL* vf, const S4D_REAL* vfi,
				const S4D_REAL* ax, const S4D_REAL* ay, const S4D_REAL* az,
				const S4D_REAL* ev,
				const S4D_REAL* sm,
				const S4D_REAL* nmx, const S4D_REAL* nmy, const S4D_REAL* nmz,
				const S4D_REAL* smc,
				const S4D_REAL* smx, const S4D_REAL* smy, const S4D_REAL* smz)
{
#ifndef _NO_MESSAGE
	MESSAGE("Initialize Membrane Solver.");
#else
	MESSAGE("Initialize Membrane Solver.",0);
#endif //_NO_MESSAGE

	clock    = 0.0;
	comm_clk = 0.0;
	wait_clk = 0.0;  // 2011.05.11

	bM = false;

	// set shared field from platform
	P0     = p0;
	MemTrm = memtrm;
	Flux   = flux;
	Q      = q;
	Vf     = vf;
	Vfi    = vfi;
	Sm     = sm;
	Nx     = nmx;
	Ny     = nmy;
	Nz     = nmz;
	Ev     = ev;
	Ax     = ax;
	Ay     = ay;
	Az     = az;
	SmC    = smc;
	SmX    = smx;
	SmY    = smy;
	SmZ    = smz;

	const CfgElem*  memb = sim?  sim->GetElemFirst("membrane")     : NULL;
	const CfgParam* exe  = memb? memb->GetParamFirst("execution")  : NULL;

	// set flag for membrane simulation
	const char* exc = NULL;
	bM = (cfg.MlMbCount > 0)
			&& exe && exe->GetData(&exc) && (strcmp(exc,"true") == 0);

	////////////////////////////////////////////////////////////////////////
	// load membrane function setting
	LoadFunctionSetting();

	bPe = bM && (cfg.PeneCount);
	bCh = bM && (cfg.ChTrCount);
	bPm = bM && (cfg.PmTrCount);
	bCS = bM && (cfg.CSTrCount);
	bPS = bM && (cfg.PSTrCount);
	bCA = bM && (cfg.CATrCount);
	bPA = bM && (cfg.PATrCount);
	bGp = bM && (cfg.GpTrCount);

	if (!bM) { return 0; }

	////////////////////////////////////////////////////////////////////////
	// allocate membrane solver local fields
	Allocate();
	RegistFields();
	SetInitialValue();

	////////////////////////////////////////////////////////////////////////
	// calc membrane area
	CalcMembPairArea();  // Paired Area

	////////////////////////////////////////////////////////////////////////
	construct_ExceptMembraneList();

	return 1;
}

int MembraneSolv::construct_ExceptMembraneList()
{
	const size_t I = cfg.Pdiv.size[0];
	const size_t J = cfg.Pdiv.size[1];
	const size_t K = cfg.Pdiv.size[2];

	const SKL_REAL dx = cfg.pitch[0];
	const SKL_REAL dy = cfg.pitch[1];
	const SKL_REAL dz = cfg.pitch[2];

	for (int im = 0; im < cfg.MatCount; im++) {
		const Material* m = cfg.MatTbl[im];
		for(int k = 0; k < K; k++) {
		for(int j = 0; j < J; j++) {
		for(int i = 0; i < I; i++) {
			if (Vf->val(i,j,k,m->id) >= 1.0) {

				int c = ExceptFaceList.size();
				ExceptFace f;
				f.m  = m;
				f.i = i; f.j = j; f.k = k;

				// -X
				if (Ax->val(i,j,k,m->id) == SKL_REAL(0.0)) {
					f.nx = -1.0; f.ny = 0.0; f.nz = 0.0;
					f.S  = dy * dz;
					ExceptFaceList.push_back(f);
				}

				// +X
				if (Ax->val(i+1,j,k,m->id) == SKL_REAL(0.0)) {
					f.nx = +1.0; f.ny = 0.0; f.nz = 0.0;
					f.S  = dy * dz;
					ExceptFaceList.push_back(f);
				}

				// -Y
				if (Ay->val(i,j,k,m->id) == SKL_REAL(0.0)) {
					f.nx = 0.0; f.ny = -1.0; f.nz = 0.0;
					f.S  = dx * dz;
					ExceptFaceList.push_back(f);
				}

				// +Y
				if (Ay->val(i,j+1,k,m->id) == SKL_REAL(0.0)) {
					f.nx = 0.0; f.ny = +1.0; f.nz = 0.0;
					f.S  = dx * dz;
					ExceptFaceList.push_back(f);
				}

				// -Z
				if (Az->val(i,j,k,m->id) == SKL_REAL(0.0)) {
					f.nx = 0.0; f.ny = 0.0; f.nz = -1.0;
					f.S  = dx * dy;
					ExceptFaceList.push_back(f);
				}

				// +Z
				if (Az->val(i,j,k+1,m->id) == SKL_REAL(0.0)) {
					f.nx = 0.0; f.ny = 0.0; f.nz = 1.0;
					f.S  = dx * dy;
					ExceptFaceList.push_back(f);
				}

				const int l = ExceptFaceList.size() - c;

				// there're more 2 faces?
				if (l >= 2) {
					// clear imitation membrane
					const_cast<S4D_REAL*>(Sm)->val(i,j,k,m->id) = 0.0;
					const_cast<S4D_REAL*>(Nx)->val(i,j,k,m->id) = 0.0;
					const_cast<S4D_REAL*>(Ny)->val(i,j,k,m->id) = 0.0;
					const_cast<S4D_REAL*>(Nz)->val(i,j,k,m->id) = 0.0;

					MESSAGE(UtilStrFormat(
						"Membrane of %s @%d,%d,%d on more 2 voxel faces.",
#ifndef _NO_MESSAGE
						f.m->name.c_str(),f.i,f.j,f.k));
#else
						f.m->name.c_str(),f.i,f.j,f.k),0);
#endif //_NO_MESSAGE
				} else {
					// roll back list
					for (int d = 0; d < l; d++) { ExceptFaceList.pop_back(); }
				}
			}
		}}}
	}

	return ExceptFaceList.size();
}

int MembraneSolv::MembranePost()
{
	return 1;
}

int MembraneSolv::Allocate()
{
#ifndef _NO_MESSAGE
	MESSAGE("Allocation for Membrane Functions");
#else
	MESSAGE("Allocation for Membrane Functions",0);
#endif //_NO_MESSAGE
	const vxsize_t vc = VC;
	const vxsize_t I = cfg.Pdiv.size[0];
	const vxsize_t J = cfg.Pdiv.size[1];
	const vxsize_t K = cfg.Pdiv.size[2];
	const vxsize_t sz_eqn[4] = { I, J, K, cfg.EqnCount };
	const vxsize_t sz_mat[4] = { I, J, K, cfg.MatCount };
	const vxsize_t sz_mol[4] = { I, J, K, cfg.MolCount };
	const vxsize_t sz_mba[4] = { I, J, K, cfg.MembCount };
	const vxsize_t sz_mbt[4] = { I, J, K, cfg.MlMbCount };
	const vxsize_t sz_pe[4]  = { I, J, K, cfg.PeneCount };
	const vxsize_t sz_ch[4]  = { I, J, K, cfg.ChTrCount };
	const vxsize_t sz_pm[4]  = { I, J, K, cfg.PmTrCount };
	const vxsize_t sz_cs[4]  = { I, J, K, cfg.CSTrCount };
	const vxsize_t sz_ps[4]  = { I, J, K, cfg.PSTrCount };
	const vxsize_t sz_ca[4]  = { I, J, K, cfg.CATrCount };
	const vxsize_t sz_pa[4]  = { I, J, K, cfg.PATrCount };
	const vxsize_t sz_gp[4]  = { I, J, K, cfg.GpTrCount };
	const vxsize_t sz_3d[3]  = { I, J, K };

	const int gid = cfg.grpP.gid;

	// share
	Smn     = UtilAllocS4DReal("Smn",     sz_mba, vc, gid);
	FluxM   = UtilAllocS4DReal("FluxM",   sz_eqn, vc, gid);
	FluxMN  = UtilAllocS4DReal("FluxMN",  sz_mbt, vc, gid);
	Epsilon = UtilAllocS3DReal("Epsilon", sz_3d,  vc, gid);

	// pene
	Kpene  = bPe? UtilAllocS4DReal("Kpene",  sz_pe, vc, gid): NULL;

	// channel
	SWcha    = bCh? UtilAllocS4DReal("SWcha",    sz_ch, vc, gid): NULL;
	DenCha   = bCh? UtilAllocS4DReal("DenCha",   sz_ch, vc, gid): NULL;
	FuzzyCha = bCh? UtilAllocS4DReal("FuzzyCha", sz_ch, vc, gid): NULL;
	KCha     = bCh? UtilAllocS4DReal("KCha",     sz_ch, vc, gid): NULL;
	KeCha    = bCh? UtilAllocS4DReal("KeCha",    sz_ch, vc, gid): NULL;

	// pump
	SWpum      = bPm? UtilAllocS4DReal("SWpum",      sz_pm,vc,gid): NULL;
	DenPum     = bPm? UtilAllocS4DReal("DenPum",     sz_pm,vc,gid): NULL;
	FuzzyPum   = bPm? UtilAllocS4DReal("FuzzyPum",   sz_pm,vc,gid): NULL;
	VmaxPumAtp = bPm? UtilAllocS4DReal("VmaxPumAtp", sz_pm,vc,gid): NULL;
	VmaxPumGtp = bPm? UtilAllocS4DReal("VmaxPumGtp", sz_pm,vc,gid): NULL;
	KmPumAtp   = bPm? UtilAllocS4DReal("KmPumAtp",   sz_pm,vc,gid): NULL;
	KmPumGtp   = bPm? UtilAllocS4DReal("KmPumGtp",   sz_pm,vc,gid): NULL;

	// Symport channel
	SWchaSym    = bCS? UtilAllocS4DReal("SWchaSym",   sz_cs,vc,gid): NULL;
	DenChaSym   = bCS? UtilAllocS4DReal("DenChaSym",  sz_cs,vc,gid): NULL;
	FuzzyChaSym = bCS? UtilAllocS4DReal("FuzzyChaSym",sz_cs,vc,gid): NULL;
	KChaSym     = bCS? UtilAllocS4DReal("KChaSym",    sz_cs,vc,gid): NULL;
	KeChaSym    = bCS? UtilAllocS4DReal("KeChaSym",   sz_cs,vc,gid): NULL;

	// Symport pump
	SWpumSym      = bPS? UtilAllocS4DReal("SWpumSym",     sz_ps,vc,gid): NULL;
	DenPumSym     = bPS? UtilAllocS4DReal("DenPumSym",    sz_ps,vc,gid): NULL;
	FuzzyPumSym   = bPS? UtilAllocS4DReal("FuzzyPumSym",  sz_ps,vc,gid): NULL;
	VmaxPumSymAtp = bPS? UtilAllocS4DReal("VmaxPumSymAtp",sz_ps,vc,gid): NULL;
	VmaxPumSymGtp = bPS? UtilAllocS4DReal("VmaxPumSymGtp",sz_ps,vc,gid): NULL;
	KmPumSymAtp   = bPS? UtilAllocS4DReal("KmPumSymAtp",  sz_ps,vc,gid): NULL;
	KmPumSymGtp   = bPS? UtilAllocS4DReal("KmPumSymGtp",  sz_ps,vc,gid): NULL;

	// Antiport channel
	SWchaAnti    = bCA? UtilAllocS4DReal("SWchaAnti",   sz_ca,vc,gid): NULL;
	DenChaAnti   = bCA? UtilAllocS4DReal("DenChaAnti",  sz_ca,vc,gid): NULL;
	FuzzyChaAnti = bCA? UtilAllocS4DReal("FuzzyChaAnti",sz_ca,vc,gid): NULL;
	KChaAnti     = bCA? UtilAllocS4DReal("KChaAnti",    sz_ca,vc,gid): NULL;
	KeChaAnti    = bCA? UtilAllocS4DReal("KeChaAnti",   sz_ca,vc,gid): NULL;

	// Antiport pump
	SWpumAnti      = bPA?UtilAllocS4DReal("SWpumAnti",     sz_pa,vc,gid):NULL;
	DenPumAnti     = bPA?UtilAllocS4DReal("DenPumAnti",    sz_pa,vc,gid):NULL;
	FuzzyPumAnti   = bPA?UtilAllocS4DReal("FuzzyPumAnti",  sz_pa,vc,gid):NULL;
	VmaxPumAntiAtp = bPA?UtilAllocS4DReal("VmaxPumAntiAtp",sz_pa,vc,gid):NULL;
	VmaxPumAntiGtp = bPA?UtilAllocS4DReal("VmaxPumAntiGtp",sz_pa,vc,gid):NULL;
	KmPumAntiAtp   = bPA?UtilAllocS4DReal("KmPumAntiAtp",  sz_pa,vc,gid):NULL;
	KmPumAntiGtp   = bPA?UtilAllocS4DReal("KmPumAntiGtp",  sz_pa,vc,gid):NULL;

	// Gap Janction
	SWgap    = bGp? UtilAllocS4DReal("SWgap",   sz_gp,vc,gid): NULL;
	DenGap   = bGp? UtilAllocS4DReal("DenGap",  sz_gp,vc,gid): NULL;
	FuzzyGap = bGp? UtilAllocS4DReal("FuzzyGap",sz_gp,vc,gid): NULL;
	KGap     = bGp? UtilAllocS4DReal("KGap",    sz_gp,vc,gid): NULL;
	KeGap    = bGp? UtilAllocS4DReal("KeGap",   sz_gp,vc,gid): NULL;

	return 1;
}

int MembraneSolv::SetInitialValue()
{
	const int gid = cfg.grpP.gid;

	// penetration
	for (int ip = 0; ip < cfg.PeneCount; ip++) {
		const penetrater& p = *cfg.PeneTbl[ip];
		const string& name = p.name;
		UtilLoadSphOrEmmToS4DN(Kpene, "K:"+name, p.e_k, ip,gid);
	}

	// channel
#ifdef MAKE_MODEL
	for (int ip = 0; ip < cfg.ChTrCount; ip++) {
		const ChTranser& p = *cfg.ChTrTbl[ip];
		const string& name = p.name;

		UtilLoadSphOrEmmToS4DN(SWcha,   "SW:"+name,   p.e_SW,   ip,gid);
		//UtilLoadSphOrEmmToS4DN(DenCha,  "Den:"+name,  p.e_Den,  ip,gid);
		MakeChannelDen();
		UtilLoadSphOrEmmToS4DN(FuzzyCha,"Fuzzy:"+name,p.e_Fuzzy,ip,gid);
		UtilLoadSphOrEmmToS4DN(KCha,    "K:"+name,    p.e_K,    ip,gid);
		UtilLoadSphOrEmmToS4DN(KeCha,   "Ke:"+name,   p.e_Ke,   ip,gid);
	}
#else
	for (int ip = 0; ip < cfg.ChTrCount; ip++) {
		const ChTranser& p = *cfg.ChTrTbl[ip];
		const string& name = p.name;

		UtilLoadSphOrEmmToS4DN(SWcha,   "SW:"+name,   p.e_SW,   ip,gid);
		UtilLoadSphOrEmmToS4DN(DenCha,  "Den:"+name,  p.e_Den,  ip,gid);
		UtilLoadSphOrEmmToS4DN(FuzzyCha,"Fuzzy:"+name,p.e_Fuzzy,ip,gid);
		UtilLoadSphOrEmmToS4DN(KCha,    "K:"+name,    p.e_K,    ip,gid);
		UtilLoadSphOrEmmToS4DN(KeCha,   "Ke:"+name,   p.e_Ke,   ip,gid);
	}
#endif // MAKE_MODEL

	// pump
	for (int ip = 0; ip < cfg.PmTrCount; ip++) {
		const PmTranser& p = *cfg.PmTrTbl[ip];
		const string& name = p.name;

		UtilLoadSphOrEmmToS4DN(SWpum,     "SW:"+name,     p.e_SW,     ip,gid);
		UtilLoadSphOrEmmToS4DN(DenPum,    "Den:"+name,    p.e_Den,    ip,gid);
		UtilLoadSphOrEmmToS4DN(FuzzyPum,  "Fuzzy:"+name,  p.e_Fuzzy,  ip,gid);
		UtilLoadSphOrEmmToS4DN(KmPumAtp,  "KmAtp:"+name,  p.e_KmAtp,  ip,gid);
		UtilLoadSphOrEmmToS4DN(KmPumGtp,  "KmGtp:"+name,  p.e_KmGtp,  ip,gid);
		UtilLoadSphOrEmmToS4DN(VmaxPumAtp,"VmaxAtp:"+name,p.e_VmaxAtp,ip,gid);
		UtilLoadSphOrEmmToS4DN(VmaxPumGtp,"VmaxGtp:"+name,p.e_VmaxGtp,ip,gid);
	}

	// channel symport
	for (int ip = 0; ip < cfg.CSTrCount; ip++) {
		const CSTranser& p = *cfg.CSTrTbl[ip];
		const string& name = p.name;

		UtilLoadSphOrEmmToS4DN(SWchaSym,   "SW:"+name,   p.e_SW,   ip,gid);
		UtilLoadSphOrEmmToS4DN(DenChaSym,  "Den:"+name,  p.e_Den,  ip,gid);
		UtilLoadSphOrEmmToS4DN(FuzzyChaSym,"Fuzzy:"+name,p.e_Fuzzy,ip,gid);
		UtilLoadSphOrEmmToS4DN(KChaSym,    "K:"+name,    p.e_K,    ip,gid);
		UtilLoadSphOrEmmToS4DN(KeChaSym,   "Ke:"+name,   p.e_Ke,   ip,gid);
	}

	// pump symport
	for (int ip = 0; ip < cfg.PSTrCount; ip++) {
		const PSTranser& p = *cfg.PSTrTbl[ip];
		const string& name = p.name;

		UtilLoadSphOrEmmToS4DN(SWpumSym,     "SW:"+name,     p.e_SW,     ip,gid);
		UtilLoadSphOrEmmToS4DN(DenPumSym,    "Den:"+name,    p.e_Den,    ip,gid);
		UtilLoadSphOrEmmToS4DN(FuzzyPumSym,  "Fuzzy:"+name,  p.e_Fuzzy,  ip,gid);
		UtilLoadSphOrEmmToS4DN(KmPumSymAtp,  "KmAtp:"+name,  p.e_KmAtp,  ip,gid);
		UtilLoadSphOrEmmToS4DN(KmPumSymGtp,  "KmGtp:"+name,  p.e_KmGtp,  ip,gid);
		UtilLoadSphOrEmmToS4DN(VmaxPumSymAtp,"VmaxAtp:"+name,p.e_VmaxAtp,ip,gid);
		UtilLoadSphOrEmmToS4DN(VmaxPumSymGtp,"VmaxGtp:"+name,p.e_VmaxGtp,ip,gid);
	}

	// channel antiport
	for (int ip = 0; ip < cfg.CATrCount; ip++) {
		const CATranser& p = *cfg.CATrTbl[ip];
		const string& name = p.name;

		UtilLoadSphOrEmmToS4DN(SWchaAnti,   "SW:"+name,   p.e_SW,   ip,gid);
		UtilLoadSphOrEmmToS4DN(DenChaAnti,  "Den:"+name,  p.e_Den,  ip,gid);
		UtilLoadSphOrEmmToS4DN(FuzzyChaAnti,"Fuzzy:"+name,p.e_Fuzzy,ip,gid);
		UtilLoadSphOrEmmToS4DN(KChaAnti,    "K:"+name,    p.e_K,    ip,gid);
		UtilLoadSphOrEmmToS4DN(KeChaAnti,   "Ke:"+name,   p.e_Ke,   ip,gid);
	}

	// pump antiport
	for (int ip = 0; ip < cfg.PATrCount; ip++) {
		const PATranser& p = *cfg.PATrTbl[ip];
		const string& name = p.name;

		UtilLoadSphOrEmmToS4DN(SWpumAnti,     "SW:"+name,     p.e_SW,     ip,gid);
		UtilLoadSphOrEmmToS4DN(DenPumAnti,    "Den:"+name,    p.e_Den,    ip,gid);
		UtilLoadSphOrEmmToS4DN(FuzzyPumAnti,  "Fuzzy:"+name,  p.e_Fuzzy,  ip,gid);
		UtilLoadSphOrEmmToS4DN(KmPumAntiAtp,  "KmAtp:"+name,  p.e_KmAtp,  ip,gid);
		UtilLoadSphOrEmmToS4DN(KmPumAntiGtp,  "KmGtp:"+name,  p.e_KmGtp,  ip,gid);
		UtilLoadSphOrEmmToS4DN(VmaxPumAntiAtp,"VmaxAtp:"+name,p.e_VmaxAtp,ip,gid);
		UtilLoadSphOrEmmToS4DN(VmaxPumAntiGtp,"VmaxGtp:"+name,p.e_VmaxGtp,ip,gid);
	}

	// gap junction
	for (int ip = 0; ip < cfg.GpTrCount; ip++) {
		const GpTranser& p = *cfg.GpTrTbl[ip];
		const string& name = p.name;

		UtilLoadSphOrEmmToS4DN(SWgap,   "SW:"+name,   p.e_SW,   ip,gid);
		UtilLoadSphOrEmmToS4DN(DenGap,  "Den:"+name,  p.e_Den,  ip,gid);
		UtilLoadSphOrEmmToS4DN(FuzzyGap,"Fuzzy:"+name,p.e_Fuzzy,ip,gid);
		UtilLoadSphOrEmmToS4DN(KGap,    "K:"+name,    p.e_K,    ip,gid);
		UtilLoadSphOrEmmToS4DN(KeGap,   "Ke:"+name,   p.e_Ke,   ip,gid);
	}

	return 1;
}

int MembraneSolv::chkCfg()
{
	if (cfg.grpAll.nid != 0) { return 0; }

#ifndef _NO_MESSAGE
	MESSAGE("**** Configuration of Membrane Solver ****");
	MESSAGE(UtilStrFormat("Membrane Simulation is %s",(bM?"true":"false")));
#else
	MESSAGE("**** Configuration of Membrane Solver ****",0);
	MESSAGE(UtilStrFormat("Membrane Simulation is %s",(bM?"true":"false")),0);
#endif //_NO_MESSAGE

	return 1;
}

int MembraneSolv::CalcFaceFlux(int* mcoef)
{
	// zero clear incremental fields
	UtilClearRealArray(FluxMN);
	// UtilClearRealArray(Q);

	// SW Control
	SwControl(SwChCtrlMap, SWcha,     mcoef);
	SwControl(SwPmCtrlMap, SWpum,     mcoef);
	SwControl(SwCSCtrlMap, SWchaSym,  mcoef);
	SwControl(SwPSCtrlMap, SWpumSym,  mcoef);
	SwControl(SwCACtrlMap, SWchaAnti, mcoef);
	SwControl(SwPACtrlMap, SWpumAnti, mcoef);
	SwControl(SwGpCtrlMap, SWgap,     mcoef);

	//////////////////////////////////////////////////////////
	// calc flux f(g,m,n)
	for (int i = 0; i < cfg.PeneCount; i++) { FluxPene(cfg.PeneTbl[i],    mcoef); }
	for (int i = 0; i < cfg.ChTrCount; i++) { FluxCha(cfg.ChTrTbl[i],     mcoef); }
	for (int i = 0; i < cfg.PmTrCount; i++) { FluxPum(cfg.PmTrTbl[i],     mcoef); }
	for (int i = 0; i < cfg.CSTrCount; i++) { FluxChaSym(cfg.CSTrTbl[i],  mcoef); }
	for (int i = 0; i < cfg.PSTrCount; i++) { FluxPumSym(cfg.PSTrTbl[i],  mcoef); }
	for (int i = 0; i < cfg.CATrCount; i++) { FluxChaAnti(cfg.CATrTbl[i], mcoef); }
	for (int i = 0; i < cfg.PATrCount; i++) { FluxPumAnti(cfg.PATrTbl[i], mcoef); }
	for (int i = 0; i < cfg.GpTrCount; i++) { FluxGap(cfg.GpTrTbl[i],     mcoef); }

	// calc membrane term field
	for( int ig = 0; ig < cfg.MolCount; ig++ ){
		const Molecular* g = cfg.MolTbl[ig];

		for( int im = 0; im < cfg.MatCount; im++ ){
			const Material* m  = cfg.MatTbl[im];
			const Equation* gm = UtilGetEquationId(g,m);
			if( !gm ){ continue; }

			CalcMembTrmField( gm, mcoef );
		}
	}

	return 1;
}

int MembraneSolv::Calc(const SKL_REAL dt)
{
	if (!bM) { return 0; }

	BEGIN_CLOCK(clk);

	if ((cfg.MembCount == 0) || (cfg.MlMbCount == 0)) { return 0; }

	//////////////////////////////////////////////////
	// zero clear incremental fields
	UtilClearRealArray(MemTrm);
	UtilClearRealArray(Q);

	int mcoef[3];

	//////////////////////////////////////////////////
	// read MembraneAreaFile
	if( cfg.MembareaCount != 0 ){

		//////////////////////////////////////////////////
		// Membrane Area X
		// Nx, Ny, Nz : index coef
		mcoef[0] = 1;
		mcoef[1] = 0;
		mcoef[2] = 0;

		CalcFaceFlux( mcoef );

		mcoef[0] = -1;
		mcoef[1] =  0;
		mcoef[2] =  0;

		CalcFaceFlux( mcoef );

		//////////////////////////////////////////////////
		// Membrane Area Y
		// Nx, Ny, Nz : index coef
		mcoef[0] = 0;
		mcoef[1] = 1;
		mcoef[2] = 0;

		CalcFaceFlux( mcoef );

		mcoef[0] =  0;
		mcoef[1] = -1;
		mcoef[2] =  0;

		CalcFaceFlux( mcoef );

		//////////////////////////////////////////////////
		// Membrane Area Z
		// Nx, Ny, Nz : index coef
		mcoef[0] = 0;
		mcoef[1] = 0;
		mcoef[2] = 1;

		CalcFaceFlux( mcoef );

		mcoef[0] =  0;
		mcoef[1] =  0;
		mcoef[2] = -1;

		CalcFaceFlux( mcoef );

	}

	//////////////////////////////////////////////////
	// Membrane Area Center & Can't read MembraneAreaFile
	// Nx, Ny, Nz : index coef
	mcoef[0] = 0;
	mcoef[1] = 0;
	mcoef[2] = 0;

	// zero clear incremental fields
	UtilClearRealArray(Epsilon);
	UtilClearRealArray(Flux);
	UtilClearRealArray(FluxM);
	UtilClearRealArray(FluxMN);
	// UtilClearRealArray(Q);

	// SW Control
	SwControl(SwChCtrlMap, SWcha,     mcoef);
	SwControl(SwPmCtrlMap, SWpum,     mcoef);
	SwControl(SwCSCtrlMap, SWchaSym,  mcoef);
	SwControl(SwPSCtrlMap, SWpumSym,  mcoef);
	SwControl(SwCACtrlMap, SWchaAnti, mcoef);
	SwControl(SwPACtrlMap, SWpumAnti, mcoef);
	SwControl(SwGpCtrlMap, SWgap,     mcoef);

	//////////////////////////////////////////////////////////
	// calc flux f(g,m,n)
	for (int i = 0; i < cfg.PeneCount; i++) { FluxPene(cfg.PeneTbl[i],    mcoef); }
	for (int i = 0; i < cfg.ChTrCount; i++) { FluxCha(cfg.ChTrTbl[i],     mcoef); }
	for (int i = 0; i < cfg.PmTrCount; i++) { FluxPum(cfg.PmTrTbl[i],     mcoef); }
	for (int i = 0; i < cfg.CSTrCount; i++) { FluxChaSym(cfg.CSTrTbl[i],  mcoef); }
	for (int i = 0; i < cfg.PSTrCount; i++) { FluxPumSym(cfg.PSTrTbl[i],  mcoef); }
	for (int i = 0; i < cfg.CATrCount; i++) { FluxChaAnti(cfg.CATrTbl[i], mcoef); }
	for (int i = 0; i < cfg.PATrCount; i++) { FluxPumAnti(cfg.PATrTbl[i], mcoef); }
	for (int i = 0; i < cfg.GpTrCount; i++) { FluxGap(cfg.GpTrTbl[i],     mcoef); }

	//////////////////////////////////////////////////////////
	// calc epsilon and modified flux
	for (int ig = 0; ig < cfg.MolCount; ig++) {
		const Molecular* g = cfg.MolTbl[ig];

		// reduce flux fgmn to fgm
		for (int im = 0; im < cfg.MatCount; im++) {
			const Material* m = cfg.MatTbl[im];
			const Equation* gm = UtilGetEquationId(g,m);
			if (!gm) { continue; }

			CalcFluxM(gm);
		}

		// epsilon
		CalcEpsilon(g);
#if 0
		// example for output epsilon value.
		// if you want to watch epsilon, please code here
		// like a example.

		// ex)
		MESSAGE(UtilStrFormat(
					"e(%s) = %f",
					g->name.c_str(),
					Epsilon->val(ti,tj,tk)));
#endif

		// modified flux fgm with epsilon
		for (int im = 0; im < cfg.MatCount; im++) {
			const Material* m = cfg.MatTbl[im];
			const Equation* gm = UtilGetEquationId(g,m);
			if (!gm) { continue; }

			CalcFlux(gm);
		}
	}

	// calc membrane term field
	for( int ig = 0; ig < cfg.MolCount; ig++ ){
		const Molecular* g = cfg.MolTbl[ig];

		for( int im = 0; im < cfg.MatCount; im++ ){
			const Material* m  = cfg.MatTbl[im];
			const Equation* gm = UtilGetEquationId(g,m);
			if( !gm ){ continue; }

			CalcMembTrmField( gm, mcoef );
		}
	}


	// --- d w s
/*
	stringstream msg;
	const Equation* eqn = UtilGetEquationId("M2:Hx");

	msg << "debug write k=11" << endl;
	for( int j=0; j<cfg.Pdiv.size[1]; j++ ){
		for( int i=0; i<cfg.Pdiv.size[0]; i++ ){
			msg << string_cast(MemTrm->val(i,j,11,eqn->id)) << ",";
		}
		msg << endl;
	}
	MESSAGE(msg.str());
*/
	// --- d w e

/*
#define PRINT_VAL(_N,_A,_I,_J,_K,_F) \
			MESSAGE(UtilStrFormat("%s(%d,%d,%d) = %e", \
						_N,(_I),(_J),(_K), \
						_A->val((_I),(_J),(_K),(_F(_N))->id)));

	MESSAGE("phi");
	PRINT_VAL("M1:GLC",P0,6,2,3,UtilGetEquationId);
	PRINT_VAL("M1:ATP",P0,6,2,3,UtilGetEquationId);
	PRINT_VAL("M2:GLC",P0,6,2,3,UtilGetEquationId);
	PRINT_VAL("M2:ATP",P0,6,2,3,UtilGetEquationId);
	PRINT_VAL("M3:GLC",P0,6,1,3,UtilGetEquationId);
	PRINT_VAL("M3:ATP",P0,6,1,3,UtilGetEquationId);

	MESSAGE("Nx");
	PRINT_VAL("M1",Nx,6,2,3,UtilGetMaterialId);
	PRINT_VAL("M2",Nx,6,2,3,UtilGetMaterialId);
	PRINT_VAL("M3",Nx,6,1,3,UtilGetMaterialId);

	MESSAGE("Ny");
	PRINT_VAL("M1",Ny,6,2,3,UtilGetMaterialId);
	PRINT_VAL("M2",Ny,6,2,3,UtilGetMaterialId);
	PRINT_VAL("M3",Ny,6,1,3,UtilGetMaterialId);

	MESSAGE("Nz");
	PRINT_VAL("M1",Nz,6,2,3,UtilGetMaterialId);
	PRINT_VAL("M2",Nz,6,2,3,UtilGetMaterialId);
	PRINT_VAL("M3",Nz,6,1,3,UtilGetMaterialId);

	MESSAGE("Flux");
	PRINT_VAL("M1:GLC",Flux,6,2,3,UtilGetEquationId);
	PRINT_VAL("M2:GLC",Flux,6,2,3,UtilGetEquationId);
	PRINT_VAL("M3:GLC",Flux,6,1,3,UtilGetEquationId);
*/

	END_CLOCK(clock, clk);
	return 1;
}

int MembraneSolv::CalcFlux(const Equation* gm)
{
	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];

	const int igm = gm->id;
	const int m   = gm->mat->id;

	for (int k = 0; k < K; k++) {
	for (int j = 0; j < J; j++) {
	for (int i = 0; i < I; i++) {
		SKL_REAL ssf = SKL_REAL(0.0);
		for (int ign = 0; ign < cfg.EqnCount; ign++ ) {
			const Equation* gn = cfg.EqnTbl[ign];
			const Material* n  = gn->mat;
			if (gn->mol != gm->mol) { continue; }
			ssf += abs(Sm->val(i,j,k,n->id) * FluxM->val(i,j,k,ign));
		}

		const SKL_REAL fe = abs(FluxM->val(i,j,k,igm)) * Epsilon->val(i,j,k);

		const int di = static_cast<int>(Nx->val(i,j,k,m));
		const int dj = static_cast<int>(Ny->val(i,j,k,m));
		const int dk = static_cast<int>(Nz->val(i,j,k,m));

		if (di || dj || dk) {
			Flux->val(i,j,k,igm) = FluxM->val(i,j,k,igm);
		} else {
			Flux->val(i,j,k,igm)
				= (ssf != SKL_REAL(0.0))?
					FluxM->val(i,j,k,igm) - (fe / ssf):
					Flux->val(i,j,k,igm) = SKL_REAL(0.0);
		}
	}}}

	return 0;
}

int MembraneSolv::CalcFluxM(const Equation* gm)
{
	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];
	
	const Molecular* g = gm->mol;
	const Material*  m = gm->mat;

	for (int igmn = 0; igmn < cfg.MlMbCount; igmn++) {
		const MolMemb* gmn = cfg.MlMbTbl[igmn];
		if (gmn->memb->mat[0] != m) { continue; }
		if (gmn->memb->mat[1] == m) { continue; }
		if (gmn->mol != g)          { continue; }

		const Membrane* mn = gmn->memb;
		const int imn = mn->id;
		const int igm = gm->id;
		const int im  = m->id;

		int mflag;
		CheckMembArea(&mflag, mn->mat[0], mn->mat[1]);

		if( mflag == 1 ){
			for (int k = 0; k < K; k++) {
			for (int j = 0; j < J; j++) {
			for (int i = 0; i < I; i++) {
				const SKL_REAL SS = (Sm->val(i,j,k,im) != SKL_REAL(0.0))?
								Smn->val(i,j,k,imn) / Sm->val(i,j,k,im):
								SKL_REAL(0.0); 
				FluxM->val(i,j,k,igm) += SS * FluxMN->val(i,j,k,igmn);
			}}}
		}
	}

	return 0;
}

int MembraneSolv::CalcEpsilon(const Molecular* g)
{
	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];

	UtilClearArray(Epsilon, SKL_REAL(0.0));

	for (int igm = 0; igm < cfg.EqnCount; igm++) {
		const Equation*  gm = cfg.EqnTbl[igm];
		if (gm->mol != g) { continue; }

		const Material* m  = gm->mat;
		const SKL_REAL* sm = UtilGetS4DN(Sm,m->id);
		const SKL_REAL* fgm = UtilGetS4DN(FluxM,gm->id);
		SKL_REAL* e = Epsilon->GetData();
		calcepsilon_(e,sm,fgm,&I,&J,&K,&VC);
	}

	return 0;
}

int MembraneSolv::CalcMembTrmField(const Equation* gm, int* mcoef)
{
	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];
	
	const Molecular* g = gm->mol;
	const Material*  m = gm->mat;

	// file read : OK ( Membrane Area Center )
	for (int igmn = 0; igmn < cfg.MlMbCount; igmn++) {
		const MolMemb* gmn = cfg.MlMbTbl[igmn];
		if (gmn->memb->mat[0] != m) { continue; }
		if (gmn->memb->mat[1] == m) { continue; }
		if (gmn->mol != g)          { continue; }

		const Membrane* mn = gmn->memb;
		const int imn = mn->id;
		const int igm = gm->id;
		const int im  = m->id;

		int mflag;
		const Membarea* area = CheckMembArea(&mflag, mn->mat[0], mn->mat[1]);

		if( mflag == 0 ){
			if( mcoef[0] == 1 ){
				for (int k = 0; k < K; k++) {
				for (int j = 0; j < J; j++) {
				for (int i = 0; i < I; i++) {
					MemTrm->val(i,j,k,igm) +=
						FluxMN->val(i,j,k,igmn) * SmX->val(i+1,j,k,area->id);
				}}}
			}
			else if( mcoef[0] == -1 ){
				for (int k = 0; k < K; k++) {
				for (int j = 0; j < J; j++) {
				for (int i = 0; i < I; i++) {
					MemTrm->val(i,j,k,igm) +=
						FluxMN->val(i,j,k,igmn) * SmX->val(i,j,k,area->id);
				}}}
			}
			else if( mcoef[1] == 1 ){
				for (int k = 0; k < K; k++) {
				for (int j = 0; j < J; j++) {
				for (int i = 0; i < I; i++) {
					MemTrm->val(i,j,k,igm) +=
						FluxMN->val(i,j,k,igmn) * SmY->val(i,j+1,k,area->id);
				}}}
			}
			else if( mcoef[1] == -1 ){
				for (int k = 0; k < K; k++) {
				for (int j = 0; j < J; j++) {
				for (int i = 0; i < I; i++) {
					MemTrm->val(i,j,k,igm) +=
						FluxMN->val(i,j,k,igmn) * SmY->val(i,j,k,area->id);
				}}}
			}
			else if( mcoef[2] == 1 ){
				for (int k = 0; k < K; k++) {
				for (int j = 0; j < J; j++) {
				for (int i = 0; i < I; i++) {
					MemTrm->val(i,j,k,igm) +=
						FluxMN->val(i,j,k,igmn) * SmZ->val(i,j,k+1,area->id);
				}}}
			}
			else if( mcoef[2] == -1 ){
				for (int k = 0; k < K; k++) {
				for (int j = 0; j < J; j++) {
				for (int i = 0; i < I; i++) {
					MemTrm->val(i,j,k,igm) +=
						FluxMN->val(i,j,k,igmn) * SmZ->val(i,j,k,area->id);
				}}}
			}
			else{
				for (int k = 0; k < K; k++) {
				for (int j = 0; j < J; j++) {
				for (int i = 0; i < I; i++) {
					MemTrm->val(i,j,k,igm) +=
						FluxMN->val(i,j,k,igmn) * SmC->val(i,j,k,area->id);
				}}}
			}
		}
	}

	// file read : NG
	if( (mcoef[0] == 0) && (mcoef[1] == 0) && (mcoef[2] == 0) ){
		for( int k = 0; k < K; k++ ){
		for( int j = 0; j < J; j++ ){
		for( int i = 0; i < I; i++ ){
			MemTrm->val(i,j,k,gm->id) +=
				Flux->val(i,j,k,gm->id) * Sm->val(i,j,k,gm->mat->id);
		}}}
	}

	return 0;
}

int MembraneSolv::CalcMembPairArea()
{
	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];
	//
	for (int imn = 0; imn < cfg.MembCount; imn++) {

		const Membrane* mn = cfg.MembTbl[imn];
		const Material* m  = mn->mat[0];
		const Material* n  = mn->mat[1];
		const int im  = m->id;
		const int in  = n->id;

		for (int k = 0; k < K; k++) {
		for (int j = 0; j < J; j++) {
		for (int i = 0; i < I; i++) {
			//
			const int id = i + static_cast<int>(Nx->val(i,j,k,im));
			const int jd = j + static_cast<int>(Ny->val(i,j,k,im));
			const int kd = k + static_cast<int>(Nz->val(i,j,k,im));

			SKL_REAL SSl = SKL_REAL(0.0);
			for (int l = 0; l < cfg.MatCount; l++) {
				if (l == im) { continue; }
				SSl += Sm->val(id,jd,kd,l);
			}
			Smn->val(i,j,k,imn) = (SSl != SKL_REAL(0.0))?
				Sm->val(id,jd,kd,in) * Sm->val(i,j,k,im) / SSl:
				SKL_REAL(0.0);
		}}}
	}

	return 0;
}

int MembraneSolv::FluxPene(const penetrater* p, int* mcoef)
{
	DB2(MESSAGE("Calc Flux Penetration..."));
	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];

	const SKL_REAL* pgm = UtilGetS4DN(P0,p->gm->id);
	const SKL_REAL* pgn = UtilGetS4DN(P0,p->gn->id);
	const SKL_REAL* nxm = UtilGetS4DN(Nx,p->m->id);
	const SKL_REAL* nym = UtilGetS4DN(Ny,p->m->id);
	const SKL_REAL* nzm = UtilGetS4DN(Nz,p->m->id);
	const SKL_REAL* nxn = UtilGetS4DN(Nx,p->n->id);
	const SKL_REAL* nyn = UtilGetS4DN(Ny,p->n->id);
	const SKL_REAL* nzn = UtilGetS4DN(Nz,p->n->id);
	const SKL_REAL* kpe = UtilGetS4DN(Kpene,p->id);

	SKL_REAL* fgmn = UtilGetS4DN(FluxMN,p->gmn->id);
	SKL_REAL* fgnm = UtilGetS4DN(FluxMN,p->gnm->id);

	int mflag;
	CheckMembArea(&mflag, p->m, p->n);

	calcfluxpene_(fgmn,pgm,pgn,kpe,nxm,nym,nzm,&I,&J,&K,&vc,&mflag,mcoef);
	calcfluxpene_(fgnm,pgn,pgm,kpe,nxn,nyn,nzn,&I,&J,&K,&vc,&mflag,mcoef);

	if( (mflag == 1) && (mcoef[0] == 0) && (mcoef[1] == 0) && (mcoef[2] == 0) ){
	// if( mflag == 1 ){
		// ExceptMemb
		for (ExceptFaceListT::const_iterator f = ExceptFaceList.begin();
				f != ExceptFaceList.end(); f++) {

			if (p->m != f->m) { continue; }
			const int i = f->i;
			const int j = f->j;
			const int k = f->k;
			const SKL_REAL nx = f->nx;
			const SKL_REAL ny = f->ny;
			const SKL_REAL nz = f->nz;
			const int ni = i + int(nx);
			const int nj = j + int(ny);
			const int nk = k + int(nz);
			const SKL_REAL S = Sm->val(ni,nj,nk,p->n->id);

			Q->val(i,j,k,p->gm->id) += S * 
				fluxpenevoxel_(pgm,pgn,kpe,&nx,&ny,&nz,&i,&j,&k,&I,&J,&K,&vc);
		}
	}

	return 0;
}

int MembraneSolv::FluxChaGeneric(const ChTranser* p,
	const S4D_REAL* SW, const S4D_REAL* Fuzzy, const S4D_REAL* Den,
	const S4D_REAL* K, const S4D_REAL* Ke, const SKL_REAL Rsign,
	int* mcoef)
{
	const int imax = cfg.Pdiv.size[0];
	const int jmax = cfg.Pdiv.size[1];
	const int kmax = cfg.Pdiv.size[2];

	const SKL_REAL* pgm = UtilGetS4DN(P0, p->gm->id);
	const SKL_REAL* pgn = UtilGetS4DN(P0, p->gn->id);
	const SKL_REAL* em  = UtilGetS4DN(Ev, p->m->id);
	const SKL_REAL* en  = UtilGetS4DN(Ev, p->n->id);

	const SKL_REAL* nxm = UtilGetS4DN(Nx, p->m->id);
	const SKL_REAL* nym = UtilGetS4DN(Ny, p->m->id);
	const SKL_REAL* nzm = UtilGetS4DN(Nz, p->m->id);
	const SKL_REAL* nxn = UtilGetS4DN(Nx, p->n->id);
	const SKL_REAL* nyn = UtilGetS4DN(Ny, p->n->id);
	const SKL_REAL* nzn = UtilGetS4DN(Nz, p->n->id);

	const SKL_REAL* sw  = UtilGetS4DN(SW,    p->id);
	const SKL_REAL* fuz = UtilGetS4DN(Fuzzy, p->id);
	const SKL_REAL* den = UtilGetS4DN(Den,   p->id);
	const SKL_REAL* k_  = UtilGetS4DN(K,     p->id);
	const SKL_REAL* ke_ = UtilGetS4DN(Ke,    p->id);

	const SKL_REAL  R   = SKL_REAL(1.0);

	int mflag;
	CheckMembArea(&mflag, p->m, p->n);

	/////////////////////////////////////////////////////////////////////
	// regular membrane
	SKL_REAL* fgmn = UtilGetS4DN(FluxMN, p->gmn->id);
	SKL_REAL* fgnm = UtilGetS4DN(FluxMN, p->gnm->id);

	// main flux
	calcfluxcha_(fgmn,&R,sw,fuz,den,pgm,pgn,em,en,k_,ke_,
				nxm,nym,nzm,&imax,&jmax,&kmax,&vc,&mflag,mcoef);
	calcfluxcha_(fgnm,&R,sw,fuz,den,pgn,pgm,en,em,k_,ke_,
				nxn,nyn,nzn,&imax,&jmax,&kmax,&vc,&mflag,mcoef);

	// carried flux
	for (Transporter::CarrieeSet::const_iterator c = p->carriees.begin();
			c != p->carriees.end(); c++) {

		const SKL_REAL hR  = Rsign * c->R;
		const MolMemb* hmn = c->hmn;
		const MolMemb* hnm = c->hnm;

		SKL_REAL* fhmn = UtilGetS4DN(FluxMN, hmn->id);
		SKL_REAL* fhnm = UtilGetS4DN(FluxMN, hnm->id);

		calcfluxcha_(fhmn,&hR,sw,fuz,den,pgm,pgn,em,en,k_,ke_,
					nxm,nym,nzm,&imax,&jmax,&kmax,&vc,&mflag,mcoef);
		calcfluxcha_(fhnm,&hR,sw,fuz,den,pgn,pgm,en,em,k_,ke_,
					nxn,nyn,nzn,&imax,&jmax,&kmax,&vc,&mflag,mcoef);
	}

	if( (mflag == 1) && (mcoef[0] == 0) && (mcoef[1] == 0) && (mcoef[2] == 0) ){
	// if( mflag == 1 ){
		/////////////////////////////////////////////////////////////////////
		// ExceptMemb
		for (ExceptFaceListT::const_iterator f = ExceptFaceList.begin();
				f != ExceptFaceList.end(); f++) {

			if (p->m != f->m) { continue; }
			const int i = f->i;
			const int j = f->j;
			const int k = f->k;
			const SKL_REAL nx = f->nx;
			const SKL_REAL ny = f->ny;
			const SKL_REAL nz = f->nz;
			const int ni = i + int(nx);
			const int nj = j + int(ny);
			const int nk = k + int(nz);
			const SKL_REAL S = Sm->val(ni,nj,nk,p->n->id);

			// main flux
			Q->val(i,j,k,p->gm->id) += S *
				fluxchannelvoxel_(&R,sw,fuz,den,pgm,pgn,em,en,k_,ke_,
							&nx,&ny,&nz,&i,&j,&k,&imax,&jmax,&kmax,&vc);

			// carried flux
			for (Transporter::CarrieeSet::const_iterator c = p->carriees.begin();
					c != p->carriees.end(); c++) {

				const SKL_REAL  hR = Rsign * c->R;
				const Equation* hm = c->hm;

				Q->val(i,j,k,hm->id) += S *
					fluxchannelvoxel_(&hR,sw,fuz,den,pgm,pgn,em,en,k_,ke_,
								&nx,&ny,&nz,&i,&j,&k,&imax,&jmax,&kmax,&vc);
			}
		}
	}

	return 0;
}

int MembraneSolv::FluxPumGeneric(const PmTranser* p,
	const S4D_REAL* SW, const S4D_REAL* Fuzzy, const S4D_REAL* Den,
	const S4D_REAL* VmaxAtp, const S4D_REAL* KmAtp,
	const S4D_REAL* VmaxGtp, const S4D_REAL* KmGtp,
	const SKL_REAL Rsign,
	int* mcoef)
{
	const int imax = cfg.Pdiv.size[0];
	const int jmax = cfg.Pdiv.size[1];
	const int kmax = cfg.Pdiv.size[2];

	const SKL_REAL* sw  = UtilGetS4DN(SW,    p->id);
	const SKL_REAL* fuz = UtilGetS4DN(Fuzzy, p->id);
	const SKL_REAL* den = UtilGetS4DN(Den,   p->id);
	const SKL_REAL* pgm = UtilGetS4DN(P0, p->gm->id);
	// const SKL_REAL* smn = UtilGetS4DN(Smn, p->mn->id);
	const SKL_REAL* nxm = UtilGetS4DN(Nx, p->m->id);
	const SKL_REAL* nym = UtilGetS4DN(Ny, p->m->id);
	const SKL_REAL* nzm = UtilGetS4DN(Nz, p->m->id);
	const SKL_REAL* nxn = UtilGetS4DN(Nx, p->n->id);
	const SKL_REAL* nyn = UtilGetS4DN(Ny, p->n->id);
	const SKL_REAL* nzn = UtilGetS4DN(Nz, p->n->id);
	const SKL_REAL 	R   = SKL_REAL(1.0);

	int mflag;
	const Membarea* area = CheckMembArea(&mflag, p->m, p->n);

	if( (mflag == 1) && ((mcoef[0] == 1) || (mcoef[1] == 1) || (mcoef[2] == 1)
			|| (mcoef[0] == -1) || (mcoef[1] == -1) || (mcoef[2] == -1)) ){
		return 0;
	}


	const SKL_REAL* smn;

	if( mflag == 1 ){
		smn = UtilGetS4DN(Smn, p->mn->id);
	}
	else{
		if( (mcoef[0] == 1) || (mcoef[0] == -1) ){
			smn = UtilGetS4DN(SmX, area->id);
		}
		else if( (mcoef[1] == 1) || (mcoef[1] == -1) ){
			smn = UtilGetS4DN(SmY, area->id);
		}
		else if( (mcoef[2] == 1) || (mcoef[2] == -1) ){
			smn = UtilGetS4DN(SmZ, area->id);
		}
		else{
			smn = UtilGetS4DN(SmC, area->id);
		}
	}

	/////////////////////////////////////////////////////////////////////
	// regular membrane
	SKL_REAL* fgmn = UtilGetS4DN(FluxMN, p->gmn->id);
	SKL_REAL* fgnm = UtilGetS4DN(FluxMN, p->gnm->id);

	// pump driven with ATP
	if (p->ATPdriven) {
		const SKL_REAL* patpm = UtilGetS4DN(P0, p->atpm->id);
		const SKL_REAL* Vmax  = UtilGetS4DN(VmaxAtp, p->id);
		const SKL_REAL* Km    = UtilGetS4DN(KmAtp,   p->id);
		const SKL_REAL  TR    = p->TRatp;
		const SKL_REAL  iTR   = -TR;

		// main flux with ATP
		SKL_REAL* qatpm = UtilGetS4DN(Q, p->atpm->id);
		SKL_REAL* qadpm = UtilGetS4DN(Q, p->adpm->id);
		SKL_REAL* qpim  = UtilGetS4DN(Q, p->pim->id);

		calcfluxpum_(fgmn,fgnm,&R,sw,fuz,den,pgm,patpm,Vmax,Km,
					nxm,nym,nzm,nxn,nyn,nzn,&imax,&jmax,&kmax,&vc,
					&mflag,mcoef);

		// energy molecular
		calcfluxpumtr_(qatpm,&TR, sw,fuz,den,pgm,patpm,Vmax,Km,
					smn,nxm,nym,nzm,&imax,&jmax,&kmax,&vc,mcoef);
		calcfluxpumtr_(qadpm,&iTR,sw,fuz,den,pgm,patpm,Vmax,Km,
					smn,nxm,nym,nzm,&imax,&jmax,&kmax,&vc,mcoef);
		calcfluxpumtr_(qpim, &iTR,sw,fuz,den,pgm,patpm,Vmax,Km,
					smn,nxm,nym,nzm,&imax,&jmax,&kmax,&vc,mcoef);

		// carried flux
		for (Transporter::CarrieeSet::const_iterator c = p->carriees.begin();
				c != p->carriees.end(); c++) {

			const SKL_REAL hR  = Rsign * c->R;
			const MolMemb* hmn = c->hmn;
			const MolMemb* hnm = c->hnm;

			SKL_REAL* fhmn = UtilGetS4DN(FluxMN, hmn->id);
			SKL_REAL* fhnm = UtilGetS4DN(FluxMN, hnm->id);

			calcfluxpum_(fhmn,fhnm,&hR,sw,fuz,den,pgm,patpm,Vmax,Km,
						nxm,nym,nzm,nxn,nyn,nzn,&imax,&jmax,&kmax,&vc,
						&mflag,mcoef);
		}
	}

	// pump driven with GTP
	if (p->GTPdriven) {
		const SKL_REAL* pgtpm = UtilGetS4DN(P0, p->gtpm->id);
		const SKL_REAL* Vmax  = UtilGetS4DN(VmaxGtp, p->id);
		const SKL_REAL* Km    = UtilGetS4DN(KmGtp,   p->id);
		const SKL_REAL  TR    = p->TRgtp;
		const SKL_REAL  iTR   = -TR;

		// main flux with GTP
		SKL_REAL* qgtpm = UtilGetS4DN(Q, p->gtpm->id);
		SKL_REAL* qgdpm = UtilGetS4DN(Q, p->gdpm->id);
		SKL_REAL* qpim  = UtilGetS4DN(Q, p->pim->id);

		calcfluxpum_(fgmn,fgnm,&R,sw,fuz,den,pgm,pgtpm,Vmax,Km,
					nxm,nym,nzm,nxn,nyn,nzn,&imax,&jmax,&kmax,&vc,
					&mflag,mcoef);

		// energy molecular
		calcfluxpumtr_(qgtpm,&TR, sw,fuz,den,pgm,pgtpm,Vmax,Km,
					smn,nxm,nym,nzm,&imax,&jmax,&kmax,&vc,mcoef);
		calcfluxpumtr_(qgdpm,&iTR,sw,fuz,den,pgm,pgtpm,Vmax,Km,
					smn,nxm,nym,nzm,&imax,&jmax,&kmax,&vc,mcoef);
		calcfluxpumtr_(qpim, &iTR,sw,fuz,den,pgm,pgtpm,Vmax,Km,
					smn,nxm,nym,nzm,&imax,&jmax,&kmax,&vc,mcoef);

		// carried flux
		for (Transporter::CarrieeSet::const_iterator c = p->carriees.begin();
				c != p->carriees.end(); c++) {

			const SKL_REAL hR  = Rsign * c->R;
			const MolMemb* hmn = c->hmn;
			const MolMemb* hnm = c->hnm;

			SKL_REAL* fhmn = UtilGetS4DN(FluxMN, hmn->id);
			SKL_REAL* fhnm = UtilGetS4DN(FluxMN, hnm->id);

			calcfluxpum_(fhmn,fhnm,&hR,sw,fuz,den,pgm,pgtpm,Vmax,Km,
						nxm,nym,nzm,nxn,nyn,nzn,&imax,&jmax,&kmax,&vc,
						&mflag,mcoef);
		}
	}

	if( (mflag == 1) && (mcoef[0] == 0) && (mcoef[1] == 0) && (mcoef[2] == 0) ){
	// if( mflag == 1 ){
		/////////////////////////////////////////////////////////////////////
		// ExceptMemb
		for (ExceptFaceListT::const_iterator f = ExceptFaceList.begin();
				f != ExceptFaceList.end(); f++) {

			if ((p->m != f->m) && (p->n != f->m)) { continue; }

			const int i = f->i;
			const int j = f->j;
			const int k = f->k;
			const SKL_REAL nx = f->nx;
			const SKL_REAL ny = f->ny;
			const SKL_REAL nz = f->nz;
			const int ni = i + int(nx);
			const int nj = j + int(ny);
			const int nk = k + int(nz);
			const SKL_REAL S = Sm->val(ni,nj,nk,p->n->id);

			const SKL_REAL zero = SKL_REAL(0.0);

			// pump driven with ATP
			if (p->ATPdriven) {
				const SKL_REAL* patpm = UtilGetS4DN(P0, p->atpm->id);
				const SKL_REAL* Vmax  = UtilGetS4DN(VmaxAtp, p->id);
				const SKL_REAL* Km    = UtilGetS4DN(KmAtp,   p->id);
				const SKL_REAL  TR    = p->TRatp;
				const SKL_REAL  iTR   = -TR;

				if (f->m == p->m) { // m-side
					// main flux with ATP
					Q->val(i,j,k,p->gm->id) += S *
						fluxpumpvoxel_(&R,sw,fuz,den,pgm,patpm,Vmax,Km,
								&zero,&zero,&zero,&i,&j,&k,&imax,&jmax,&kmax,&vc);

					// energy molecular
					// ATP
					Q->val(i,j,k,p->atpm->id) += S *
						fluxpumptrvoxel_(&TR,sw,fuz,den,pgm,patpm,Vmax,Km,
								&zero,&zero,&zero,&i,&j,&k,&imax,&jmax,&kmax,&vc);
					// ADP
					Q->val(i,j,k,p->adpm->id) += S *
						fluxpumptrvoxel_(&iTR,sw,fuz,den,pgm,patpm,Vmax,Km,
								&zero,&zero,&zero,&i,&j,&k,&imax,&jmax,&kmax,&vc);
					// Pi
					Q->val(i,j,k,p->pim->id)  += S *
						fluxpumptrvoxel_(&iTR,sw,fuz,den,pgm,patpm,Vmax,Km,
								&zero,&zero,&zero,&i,&j,&k,&imax,&jmax,&kmax,&vc);
				}

				if (f->m == p->n) { // n-side
					// main flux with ATP
					Q->val(i,j,k,p->gn->id) -= S *
						fluxpumpvoxel_(&R,sw,fuz,den,pgm,patpm,Vmax,Km,
								&nx,&ny,&nz,&i,&j,&k,&imax,&jmax,&kmax,&vc);
				}

				// carried flux
				for (Transporter::CarrieeSet::const_iterator c = p->carriees.begin();
						c != p->carriees.end(); c++) {

					const SKL_REAL  hR = Rsign * c->R;
					const Equation* hm = c->hm;
					const Equation* hn = c->hn;

					if (f->m == p->m) { // m-side
						Q->val(i,j,k,hm->id) += S *
							fluxpumpvoxel_(&hR,sw,fuz,den,pgm,patpm,Vmax,Km,
									&zero,&zero,&zero,&i,&j,&k,&imax,&jmax,&kmax,&vc);
					}

					if (f->m == p->n) { // n-side
						Q->val(i,j,k,hn->id) += S *
							fluxpumpvoxel_(&hR,sw,fuz,den,pgm,patpm,Vmax,Km,
									&nx,&ny,&nz,&i,&j,&k,&imax,&jmax,&kmax,&vc);
					}
				}
			}

			// pump driven with GTP
			if (p->GTPdriven) {
				const SKL_REAL* pgtpm = UtilGetS4DN(P0, p->gtpm->id);
				const SKL_REAL* Vmax  = UtilGetS4DN(VmaxGtp, p->id);
				const SKL_REAL* Km    = UtilGetS4DN(KmGtp,   p->id);
				const SKL_REAL  TR    = p->TRgtp;
				const SKL_REAL  iTR   = -TR;

				if (f->m == p->m) { // m-side
					// main flux with GTP
					Q->val(i,j,k,p->gm->id) += S *
						fluxpumpvoxel_(&R,sw,fuz,den,pgm,pgtpm,Vmax,Km,
								&zero,&zero,&zero,&i,&j,&k,&imax,&jmax,&kmax,&vc);

					// energy molecular
					// GTP
					Q->val(i,j,k,p->gtpm->id) += S *
						fluxpumptrvoxel_(&TR,sw,fuz,den,pgm,pgtpm,Vmax,Km,
								&zero,&zero,&zero,&i,&j,&k,&imax,&jmax,&kmax,&vc);
					// GDP
					Q->val(i,j,k,p->gdpm->id) += S *
						fluxpumptrvoxel_(&iTR,sw,fuz,den,pgm,pgtpm,Vmax,Km,
								&zero,&zero,&zero,&i,&j,&k,&imax,&jmax,&kmax,&vc);
					// Pi
					Q->val(i,j,k,p->pim->id)  += S *
						fluxpumptrvoxel_(&iTR,sw,fuz,den,pgm,pgtpm,Vmax,Km,
								&zero,&zero,&zero,&i,&j,&k,&imax,&jmax,&kmax,&vc);
				}

				if (f->m == p->n) { // n-side
					// main flux with GTP
					Q->val(i,j,k,p->gn->id) -= S *
						fluxpumpvoxel_(&R,sw,fuz,den,pgm,pgtpm,Vmax,Km,
								&nx,&ny,&nz,&i,&j,&k,&imax,&jmax,&kmax,&vc);
				}

				// carried flux
				for (Transporter::CarrieeSet::const_iterator c = p->carriees.begin();
						c != p->carriees.end(); c++) {

					const SKL_REAL  hR = Rsign * c->R;
					const Equation* hm = c->hm;
					const Equation* hn = c->hn;

					if (f->m == p->m) { // m-side
						Q->val(i,j,k,hm->id) += S *
							fluxpumpvoxel_(&hR,sw,fuz,den,pgm,pgtpm,Vmax,Km,
									&zero,&zero,&zero,&i,&j,&k,&imax,&jmax,&kmax,&vc);
					}

					if (f->m == p->n) { // n-side
						Q->val(i,j,k,hn->id) += S *
							fluxpumpvoxel_(&hR,sw,fuz,den,pgm,pgtpm,Vmax,Km,
									&nx,&ny,&nz,&i,&j,&k,&imax,&jmax,&kmax,&vc);
					}
				}
			}
		}
	}

	return 0;
}

int MembraneSolv::FluxCha(const ChTranser* p, int* mcoef)
{
	DB2(MESSAGE("Calc Flux Channel..."));

	return FluxChaGeneric(p,SWcha,FuzzyCha,DenCha,KCha,KeCha,0.0,mcoef);
}

int MembraneSolv::FluxPum(const PmTranser* p, int* mcoef)
{
	DB2(MESSAGE("Calc Flux Pump..."));
	return FluxPumGeneric(p,
				SWpum,FuzzyPum,DenPum,
				VmaxPumAtp,KmPumAtp,VmaxPumGtp,KmPumGtp,
				0.0,mcoef);
}

int MembraneSolv::FluxChaSym(const CSTranser* p, int* mcoef)
{
	DB2(MESSAGE("Calc Flux Symport Channel ..."));
	return FluxChaGeneric(p,SWchaSym,FuzzyChaSym,DenChaSym,KChaSym,KeChaSym,1.0,mcoef);
}

int MembraneSolv::FluxPumSym(const PSTranser* p, int* mcoef)
{
	DB2(MESSAGE("Calc Flux Symport Pump ..."));
	return FluxPumGeneric(p,
				SWpumSym,FuzzyPumSym,DenPumSym,
				VmaxPumSymAtp,KmPumSymAtp,VmaxPumSymGtp,KmPumSymGtp,
				1.0,mcoef);
}

int MembraneSolv::FluxChaAnti(const CATranser* p, int* mcoef)
{
	DB2(MESSAGE("Calc Flux Antiport Channel ..."));
	return FluxChaGeneric(p,SWchaAnti,FuzzyChaAnti,DenChaAnti,KChaAnti,KeChaAnti,-1.0,mcoef);
}

int MembraneSolv::FluxPumAnti(const PATranser* p, int* mcoef)
{
	DB2(MESSAGE("Calc Flux Antiport Pump ..."));
	return FluxPumGeneric(p,
				SWpumAnti,FuzzyPumAnti,DenPumAnti,
				VmaxPumAntiAtp,KmPumAntiAtp,VmaxPumAntiGtp,KmPumAntiGtp,
				-1.0,mcoef);
}

int MembraneSolv::FluxGap(const GpTranser* p, int* mcoef)
{
	DB2(MESSAGE("Calc Flux Gap Junction..."));
	return FluxChaGeneric(p,SWgap,FuzzyGap,DenGap,KGap,KeGap,0.0,mcoef);
}

int MembraneSolv::LoadFunctionSetting()
{
#ifndef _NO_MESSAGE
	MESSAGE("Load Function Setting from configuration XML");
#else
	MESSAGE("Load Function Setting from configuration XML",0);
#endif //_NO_MESSAGE

	cfg.PeneCount = 0;
	cfg.ChTrCount = 0;
	cfg.PmTrCount = 0;
	cfg.CSTrCount = 0;
	cfg.PSTrCount = 0;
	cfg.CATrCount = 0;
	cfg.PATrCount = 0;
	cfg.GpTrCount = 0;

	const CfgElem* DomainInfo = cfg.solver->SklCfgGetTop(DOMAININFO);
	const CfgElem* membfunc   = DomainInfo->GetElemFirst("membrane");
	for (int igmn = 0; igmn < cfg.MlMbCount; igmn++) {
		const MolMemb*   gmn = cfg.MlMbTbl[igmn];
		const Membrane*  mn  = gmn->memb;
		const Molecular* g   = gmn->mol;

#ifndef _NO_MESSAGE
		MESSAGE(string("Load Membrane Function Setting for ")+gmn->name);
#else
		MESSAGE(string("Load Membrane Function Setting for ")+gmn->name,0);
#endif //_NO_MESSAGE
		for (const CfgElem* mb = membfunc->GetElemFirst(mn->name.c_str());
				mb != NULL; mb = membfunc->GetElemNext(mb,mn->name.c_str())) {
		for (const CfgElem* ml = mb->GetElemFirst(g->name.c_str());
				ml != NULL; ml = mb->GetElemNext(ml,g->name.c_str())) {

			LoadPeSetting(gmn, ml);
			LoadChSetting(gmn, ml);
			LoadPmSetting(gmn, ml);
			LoadCSSetting(gmn, ml);
			LoadPSSetting(gmn, ml);
			LoadCASetting(gmn, ml);
			LoadPASetting(gmn, ml);
			LoadGpSetting(gmn, ml);
		}}
	}

#ifndef _NO_MESSAGE
	MESSAGE("Loaded Membrane Function List");
#else
	MESSAGE("Loaded Membrane Function List",0);
#endif //_NO_MESSAGE
	for (int ip = 0; ip < cfg.PeneCount; ip++) {
		const penetrater* p = cfg.PeneTbl[ip];
#ifndef _NO_MESSAGE
		MESSAGE(string("++ ") + p->name);
#else
		MESSAGE(string("++ ") + p->name,0);
#endif //_NO_MESSAGE
	}
	for (int ip = 0; ip < cfg.ChTrCount; ip++) {
		const ChTranser* p = cfg.ChTrTbl[ip];
#ifndef _NO_MESSAGE
		MESSAGE(string("++ ") + p->name);
#else
		MESSAGE(string("++ ") + p->name,0);
#endif //_NO_MESSAGE
	}
	for (int ip = 0; ip < cfg.PmTrCount; ip++) {
		const PmTranser* p = cfg.PmTrTbl[ip];
#ifndef _NO_MESSAGE
		MESSAGE("++ " + p->name);
#else
		MESSAGE("++ " + p->name,0);
#endif //_NO_MESSAGE
	}
	for (int ip = 0; ip < cfg.CSTrCount; ip++) {
		const CSTranser* p = cfg.CSTrTbl[ip];
#ifndef _NO_MESSAGE
		MESSAGE("++ " + p->name);
#else
		MESSAGE("++ " + p->name,0);
#endif //_NO_MESSAGE
	}
	for (int ip = 0; ip < cfg.PSTrCount; ip++) {
		const PSTranser* p = cfg.PSTrTbl[ip];
#ifndef _NO_MESSAGE
		MESSAGE("++ " + p->name);
#else
		MESSAGE("++ " + p->name,0);
#endif //_NO_MESSAGE
	}
	for (int ip = 0; ip < cfg.CATrCount; ip++) {
		const CATranser* p = cfg.CATrTbl[ip];
#ifndef _NO_MESSAGE
		MESSAGE("++ " + p->name);
#else
		MESSAGE("++ " + p->name,0);
#endif //_NO_MESSAGE
	}
	for (int ip = 0; ip < cfg.PATrCount; ip++) {
		const PATranser* p = cfg.PATrTbl[ip];
#ifndef _NO_MESSAGE
		MESSAGE("++ " + p->name);
#else
		MESSAGE("++ " + p->name,0);
#endif //_NO_MESSAGE
	}
	for (int ip = 0; ip < cfg.GpTrCount; ip++) {
		const GpTranser* p = cfg.GpTrTbl[ip];
#ifndef _NO_MESSAGE
		MESSAGE("++ " + p->name);
#else
		MESSAGE("++ " + p->name,0);
#endif //_NO_MESSAGE
	}

	return 0;
}

int MembraneSolv::LoadPeSetting(const MolMemb* gmn, const CfgElem* ml)
{
	DB2(MESSAGE(string("Load Function Setting of Penetaration for ") + gmn->name));
	const char* fname = "Penetration";

	const Molecular* g = gmn->mol;
	const Material*  m = gmn->memb->mat[0];
	const Material*  n = gmn->memb->mat[1];

	for (const CfgElem* f = ml->GetElemFirst(fname);
			f != NULL; f = ml->GetElemNext(f, fname)) {
		//
		const string name0 = string("Pene_")+g->name+":"+m->name+":"+n->name;
		const string name1 = string("Pene_")+g->name+":"+n->name+":"+m->name;
		penetrater* p = const_cast<penetrater*>(UtilGetPenetraterId(name0));

		if (!p) {
			p = new penetrater;
			p->set(cfg.PeneTbl.size(),f,name0,gmn);

			cfg.PeneTbl.push_back(p);
			cfg.PeneNameIdMap.insert(Name2IdMapT::value_type(name0,p->id));
			cfg.PeneNameIdMap.insert(Name2IdMapT::value_type(name1,p->id));
		}
	}
	cfg.PeneCount = cfg.PeneTbl.size();
	return cfg.PeneCount;
}

int MembraneSolv::LoadChSetting(const MolMemb* gmn, const CfgElem* ml)
{
	DB2(MESSAGE(string("Load Function Setting of Channel for ") + gmn->name));
	const char* fname = "Channel";

	const Molecular* g = gmn->mol;
	const Material*  m = gmn->memb->mat[0];
	const Material*  n = gmn->memb->mat[1];

	for (const CfgElem* f = ml->GetElemFirst(fname);
			f != NULL; f = ml->GetElemNext(f, fname)) {
		//
		TRY(f->isSetID(), gmn->memb->name + "/" + ml->GetName() + " has no ID");

		const int pid = f->GetID();
		const string name_base = string_cast(pid)+string(":Ch_");
		const string name0 = name_base + g->name+":" + m->name + ":" + n->name;
		const string name1 = name_base + g->name+":" + n->name + ":" + m->name;
		if(!UtilGetChTranserId(name0)) {
			ChTranser* p = new ChTranser;
			p->set(cfg.ChTrTbl.size(),f,name0,gmn);

			cfg.ChTrTbl.push_back(p);
			cfg.ChTrNameIdMap.insert(Name2IdMapT::value_type(name0,p->id));
			cfg.ChTrNameIdMap.insert(Name2IdMapT::value_type(name1,p->id));

			// SW Controller
			LoadSwCtrlSetting(SwChCtrlMap, p, f->GetElemFirst("SWCtrl"));
		}
	}
	cfg.ChTrCount = cfg.ChTrTbl.size();
	return cfg.ChTrCount;
}

int MembraneSolv::LoadPmSetting(const MolMemb* gmn, const CfgElem* ml)
{
	DB2(MESSAGE(string("Load Function Setting of Pump for ") + gmn->name));
	const char* fname = "Pump";

	const Molecular* g = gmn->mol;
	const Material*  m = gmn->memb->mat[0];
	const Material*  n = gmn->memb->mat[1];

	for (const CfgElem* f = ml->GetElemFirst(fname);
			f != NULL; f = ml->GetElemNext(f, fname)) {
		//
		TRY(f->isSetID(), gmn->memb->name + "/" + ml->GetName() + " has no ID");

		const int pid = f->GetID();
		const string name_base = string_cast(pid)+string(":Pm_");
		const string name = name_base + g->name+":" + m->name + ":" + n->name;
		if(!UtilGetPmTranserId(name)) {
			PmTranser* p = new PmTranser;
			p->set(cfg.PmTrTbl.size(),f,name,gmn);

			cfg.PmTrTbl.push_back(p);
			cfg.PmTrNameIdMap.insert(Name2IdMapT::value_type(name,p->id));

			// SW Controller
			LoadSwCtrlSetting(SwPmCtrlMap, p, f->GetElemFirst("SWCtrl"));
		}
	}
	cfg.PmTrCount = cfg.PmTrTbl.size();
	return cfg.PmTrCount;
}

int MembraneSolv::LoadCSSetting(const MolMemb* gmn, const CfgElem* ml)
{
	DB2(MESSAGE(string("Load Function Setting of SymportChannel for ")+gmn->name));
	const char* fname = "SymportChannel";

	const Molecular* g = gmn->mol;
	const Material*  m = gmn->memb->mat[0];
	const Material*  n = gmn->memb->mat[1];

	for (const CfgElem* f = ml->GetElemFirst(fname);
			f != NULL; f = ml->GetElemNext(f, fname)) {
		//
		TRY(f->isSetID(), gmn->memb->name + "/" + ml->GetName() + " has no ID");

		const int pid = f->GetID();
		const string name_base = string_cast(pid)+string(":CS_");
		const string name0 = name_base + g->name+":" + m->name + ":" + n->name;
		const string name1 = name_base + g->name+":" + n->name + ":" + m->name;
		if(!UtilGetCSTranserId(name0)) {
			CSTranser* p = new CSTranser;
			p->set(cfg.CSTrTbl.size(),f,name0,gmn);

			cfg.CSTrTbl.push_back(p);
			cfg.CSTrNameIdMap.insert(Name2IdMapT::value_type(name0,p->id));
			cfg.CSTrNameIdMap.insert(Name2IdMapT::value_type(name1,p->id));

			// CSR
			for (const CfgParam* CSR = f->GetParamFirst("CSR");
				CSR != NULL; CSR = f->GetParamNext(CSR,"CSR")) {
				const char* cmt = CSR->GetComment();
				const Molecular* h = UtilGetMolecularId(cmt);

				TRY(cmt, "CSR has no comment");
				TRY(h, UtilStrFormat("Molecular %s is unknown", cmt));

				SKL_REAL R = SKL_REAL(0.0);
				CSR->GetData(&R)? R: SKL_REAL(0.0);
				
				const Equation* hm  = UtilGetEquationId(h,p->m);
				const Equation* hn  = UtilGetEquationId(h,p->n);
				const MolMemb*  hmn = UtilAddNewMolMembId(h,p->mn);
				const MolMemb*  hnm = UtilAddNewMolMembId(h,p->nm);

				Transporter::carriee c = { R, h, hm, hn, hmn, hnm };
				p->carriees.push_back(c);
			}

			// SW Controller
			LoadSwCtrlSetting(SwCSCtrlMap, p, f->GetElemFirst("SWCtrl"));
		}
	}
	cfg.CSTrCount = cfg.CSTrTbl.size();
	return cfg.CSTrCount;
}

int MembraneSolv::LoadPSSetting(const MolMemb* gmn, const CfgElem* ml)
{
	DB2(MESSAGE(string("Load Function Setting of SymportPump for ") + gmn->name));
	const char* fname = "SymportPump";

	const Molecular* g = gmn->mol;
	const Material*  m = gmn->memb->mat[0];
	const Material*  n = gmn->memb->mat[1];

	for (const CfgElem* f = ml->GetElemFirst(fname);
			f != NULL; f = ml->GetElemNext(f, fname)) {
		//
		TRY(f->isSetID(), gmn->memb->name + "/" + ml->GetName() + " has no ID");

		const int pid = f->GetID();
		const string name_base = string_cast(pid)+string(":PS_");
		const string name = name_base + g->name+":" + m->name + ":" + n->name;
		if(!UtilGetPSTranserId(name)) {
			PSTranser* p = new PSTranser;
			p->set(cfg.PSTrTbl.size(),f,name,gmn);

			cfg.PSTrTbl.push_back(p);
			cfg.PSTrNameIdMap.insert(Name2IdMapT::value_type(name,p->id));

			// PSR
			for (const CfgParam* PSR = f->GetParamFirst("PSR");
				PSR != NULL; PSR = f->GetParamNext(PSR,"PSR")) {
				const char* cmt = PSR->GetComment();
				const Molecular* h = UtilGetMolecularId(cmt);
				TRY(cmt, "PSR has no comment");
				TRY(h, UtilStrFormat("Molecular %s is unknown", cmt));

				SKL_REAL R = SKL_REAL(0.0);
				PSR->GetData(&R)? R: SKL_REAL(0.0);

				const Equation* hm  = UtilGetEquationId(h,p->m);
				const Equation* hn  = UtilGetEquationId(h,p->n);
				const MolMemb*  hmn = UtilAddNewMolMembId(h,p->mn);
				const MolMemb*  hnm = UtilAddNewMolMembId(h,p->nm);

				Transporter::carriee c = { R, h, hm, hn, hmn, hnm };
				p->carriees.push_back(c);
			}

			// SW Controller
			LoadSwCtrlSetting(SwPSCtrlMap, p, f->GetElemFirst("SWCtrl"));
		}
	}
	cfg.PSTrCount = cfg.PSTrTbl.size();
	return cfg.PSTrCount;
}

int MembraneSolv::LoadCASetting(const MolMemb* gmn, const CfgElem* ml)
{
	DB2(MESSAGE(string("Load Function Setting of AntiportChannel for ")+gmn->name));
	const char* fname = "AntiportChannel";

	const Molecular* g = gmn->mol;
	const Material*  m = gmn->memb->mat[0];
	const Material*  n = gmn->memb->mat[1];

	for (const CfgElem* f = ml->GetElemFirst(fname);
			f != NULL; f = ml->GetElemNext(f, fname)) {
		//
		TRY(f->isSetID(), gmn->memb->name + "/" + ml->GetName() + " has no ID");

		const int pid = f->GetID();
		const string name_base = string_cast(pid)+string(":CA_");
		const string name0 = name_base + g->name+":" + m->name + ":" + n->name;
		const string name1 = name_base + g->name+":" + n->name + ":" + m->name;
		if(!UtilGetCATranserId(name0)) {
			CATranser* p = new CATranser;
			p->set(cfg.CATrTbl.size(),f,name0,gmn);

			cfg.CATrTbl.push_back(p);
			cfg.CATrNameIdMap.insert(Name2IdMapT::value_type(name0,p->id));
			cfg.CATrNameIdMap.insert(Name2IdMapT::value_type(name1,p->id));

			// CAR
			for (const CfgParam* CAR = f->GetParamFirst("CAR");
				CAR != NULL; CAR = f->GetParamNext(CAR,"CAR")) {
				const char* cmt = CAR->GetComment();
				const Molecular* h = UtilGetMolecularId(cmt);
				TRY(cmt, "CAR has no comment");
				TRY(h, UtilStrFormat("Molecular %s is unknown", cmt));

				SKL_REAL R = SKL_REAL(0.0);
				CAR->GetData(&R)? R: SKL_REAL(0.0);

				const Equation* hm  = UtilGetEquationId(h,p->m);
				const Equation* hn  = UtilGetEquationId(h,p->n);
				const MolMemb*  hmn = UtilAddNewMolMembId(h,p->mn);
				const MolMemb*  hnm = UtilAddNewMolMembId(h,p->nm);

				Transporter::carriee c = { R, h, hm, hn, hmn, hnm };
				p->carriees.push_back(c);
			}

			// SW Controller
			LoadSwCtrlSetting(SwCACtrlMap, p, f->GetElemFirst("SWCtrl"));
		}
	}
	cfg.CATrCount = cfg.CATrTbl.size();
	return cfg.CATrCount;
}

int MembraneSolv::LoadPASetting(const MolMemb* gmn, const CfgElem* ml)
{
	DB2(MESSAGE(string("Load Function Setting of AntiportPump for ") + gmn->name));
	const char* fname = "AntiportPump";

	const Molecular* g = gmn->mol;
	const Material*  m = gmn->memb->mat[0];
	const Material*  n = gmn->memb->mat[1];

	for (const CfgElem* f = ml->GetElemFirst(fname);
			f != NULL; f = ml->GetElemNext(f, fname)) {
		//
		TRY(f->isSetID(), gmn->memb->name + "/" + ml->GetName() + " has no ID");

		const int pid = f->GetID();
		const string name_base = string_cast(pid)+string(":PA_");
		const string name = name_base + g->name+":" + m->name + ":" + n->name;
		if(!UtilGetPATranserId(name)) {
			PATranser* p = new PATranser;
			p->set(cfg.PATrTbl.size(),f,name,gmn);

			cfg.PATrTbl.push_back(p);
			cfg.PATrNameIdMap.insert(Name2IdMapT::value_type(name,p->id));

			// PAR
			for (const CfgParam* PAR = f->GetParamFirst("PAR");
				PAR != NULL; PAR = f->GetParamNext(PAR,"PAR")) {
				const char* cmt = PAR->GetComment();
				const Molecular* h = UtilGetMolecularId(cmt);
				TRY(cmt, "PAR has no comment");
				TRY(h, UtilStrFormat("Molecular %s is unknown", cmt));

				SKL_REAL R = SKL_REAL(0.0);
				PAR->GetData(&R)? R: SKL_REAL(0.0);

				const Equation* hm  = UtilGetEquationId(h,p->m);
				const Equation* hn  = UtilGetEquationId(h,p->n);
				const MolMemb*  hmn = UtilAddNewMolMembId(h,p->mn);
				const MolMemb*  hnm = UtilAddNewMolMembId(h,p->nm);

				Transporter::carriee c = { R, h, hm, hn, hmn, hnm };
				p->carriees.push_back(c);
			}

			// SW Controller
			LoadSwCtrlSetting(SwPACtrlMap, p, f->GetElemFirst("SWCtrl"));
		}
	}
	cfg.PATrCount = cfg.PATrTbl.size();
	return cfg.PATrCount;
}

int MembraneSolv::LoadGpSetting(const MolMemb* gmn, const CfgElem* ml)
{
	DB2(MESSAGE(string("Load Function Setting of GapJunction for ") + gmn->name));
	const char* fname = "GapJunction";

	const Molecular* g = gmn->mol;
	const Material*  m = gmn->memb->mat[0];
	const Material*  n = gmn->memb->mat[1];

	for (const CfgElem* f = ml->GetElemFirst(fname);
			f != NULL; f = ml->GetElemNext(f, fname)) {
		//
		TRY(f->isSetID(), gmn->memb->name + "/" + ml->GetName() + " has no ID");

		const int pid = f->GetID();
		const string name_base = string_cast(pid)+string(":Gp_");
		const string name0 = name_base + g->name+":" + m->name + ":" + n->name;
		const string name1 = name_base + g->name+":" + n->name + ":" + m->name;
		if(!UtilGetGpTranserId(name0)) {
			GpTranser* p = new GpTranser;
			p->set(cfg.GpTrTbl.size(),f,name0,gmn);

			cfg.GpTrTbl.push_back(p);
			cfg.GpTrNameIdMap.insert(Name2IdMapT::value_type(name0,p->id));
			cfg.GpTrNameIdMap.insert(Name2IdMapT::value_type(name1,p->id));

			// SW Controller
			LoadSwCtrlSetting(SwGpCtrlMap, p, f->GetElemFirst("SWCtrl"));
		}
	}
	cfg.GpTrCount = cfg.GpTrTbl.size();
	return cfg.GpTrCount;
}

int MembraneSolv::LoadSwCtrlSetting(SwCtrlMapT& m, const Transporter* t, const CfgElem* e)
{
	if (!e) { return 0; }

	const string assoc = e->GetComment();

	SwController* c = new SwController;
	c->assoc =
		(assoc == "sum")? SwController::SUM:
		(assoc == "mul")? SwController::MUL:
		(assoc == "min")? SwController::MIN:
		(assoc == "max")? SwController::MAX:
		SwController::NONE;
	
	if (c->assoc == SwController::NONE) {
		delete c;
		ERROR(t->name + " :: Associative method is illigal at comment");
	}

	SwFuncParam param;

	// michaelis function
	for (const CfgElem* f = e->GetElemFirst("michaelis");
			f != NULL; f = e->GetElemNext(f,"michaelis")) {
		// search tag
		const CfgParam* Vmax = f->GetParamFirst("Vmax");
		const CfgParam* Km   = f->GetParamFirst("Km");
		
		// pick params
		TRY(Vmax && Vmax->GetData(&param.p.michaelis.Vmax),
			"Vmax tag is illigal");
		TRY(Km   && Km->GetData(&param.p.michaelis.Km),
			"Km tag is illigal");

		// search target
		const CfgParam* Tgt  = f->GetParamFirst("S");
		const char *str = NULL;
		TRY(Tgt && Tgt->GetData(&str), "S tag is illigal.");
		param.gm = UtilGetEquationId(str);
		TRY(param.gm, UtilStrFormat("%s is unknown", str));

		// add function  
		c->funcs.push_back((void*)(SwCtrl::michaelis));
		c->params.push_back(param);
	}                    
                         
	// hill function     
	for (const CfgElem* f = e->GetElemFirst("hill");
			f != NULL; f = e->GetElemNext(f,"hill")) {
		// search tag
		const CfgParam* Kd = f->GetParamFirst("Kd");
		const CfgParam* n  = f->GetParamFirst("n");
		
		// pick params
		TRY(Kd && Kd->GetData(&param.p.hill.Kd), "Kd tag is illigal");
		TRY(n  && n->GetData(&param.p.hill.n),   "n tag is illigal");

		// search target
		const CfgParam* Tgt = f->GetParamFirst("S");
		const char *str = NULL;
		TRY(Tgt && Tgt->GetData(&str), "S tag is illigal.");
		param.gm = UtilGetEquationId(str);
		TRY(param.gm, UtilStrFormat("%s is unknown", str));

		// add function  
		c->funcs.push_back((void*)(SwCtrl::hill));
		c->params.push_back(param);
	}                    
                         
	// linear function   
	for (const CfgElem* f = e->GetElemFirst("linear");
			f != NULL; f = e->GetElemNext(f,"linear")) {
		// search tag
		const CfgParam* a = f->GetParamFirst("a");
		const CfgParam* b = f->GetParamFirst("b");
		
		// pick params
		TRY(a && a->GetData(&param.p.linear.a), "a tag is illigal");
		b ? b->GetData(&param.p.linear.b): param.p.linear.b == SKL_REAL(0.0);

		// search target
		const CfgParam* Tgt = f->GetParamFirst("S");
		const char *str = NULL;
		TRY(Tgt && Tgt->GetData(&str), "S tag is illigal.");
		param.gm = UtilGetEquationId(str);
		TRY(param.gm, UtilStrFormat("%s is unknown", str));

		// add function
		c->funcs.push_back((void*)(SwCtrl::linear));
		c->params.push_back(param);
	}
	
	// expr function
	for (const CfgElem* f = e->GetElemFirst("exp");
			f != NULL; f = e->GetElemNext(f,"exp")) {
		// search tag
		const CfgParam* A = f->GetParamFirst("A");
		const CfgParam* a = f->GetParamFirst("a");
		
		// pick params
		TRY(A && A->GetData(&param.p.exp.A), "A tag is illigal");
		TRY(a && a->GetData(&param.p.exp.a), "a tag is illigal");

		// search target
		const CfgParam* Tgt = f->GetParamFirst("S");
		const char *str = NULL;
		TRY(Tgt && Tgt->GetData(&str), "S tag is illigal.");
		param.gm = UtilGetEquationId(str);
		TRY(param.gm, UtilStrFormat("%s is unknown", str));

		// add function
		c->funcs.push_back((void*)(SwCtrl::exp));
		c->params.push_back(param);
	}
	
	// log function
	for (const CfgElem* f = e->GetElemFirst("log");
			f != NULL; f = e->GetElemNext(f,"log")) {
		// search tag
		const CfgParam* A = f->GetParamFirst("A");
		const CfgParam* a = f->GetParamFirst("a");
		
		// pick params
		TRY(A && A->GetData(&param.p.log.A), "A tag is illigal");
		TRY(a && a->GetData(&param.p.log.a), "a tag is illigal");

		// search target
		const CfgParam* Tgt = f->GetParamFirst("S");
		const char *str = NULL;
		TRY(Tgt && Tgt->GetData(&str), "S tag is illigal.");
		param.gm = UtilGetEquationId(str);
		TRY(param.gm, UtilStrFormat("%s is unknown", str));

		// add function
		c->funcs.push_back((void*)(SwCtrl::log));
		c->params.push_back(param);
	}
	
	// log10 function
	for (const CfgElem* f = e->GetElemFirst("log10");
			f != NULL; f = e->GetElemNext(f,"log10")) {
		// search tag
		const CfgParam* A = f->GetParamFirst("A");
		const CfgParam* a = f->GetParamFirst("a");
		
		// pick params
		TRY(A && A->GetData(&param.p.log10.A), "A tag is illigal");
		TRY(a && a->GetData(&param.p.log10.a), "a tag is illigal");

		// search target
		const CfgParam* Tgt = f->GetParamFirst("S");
		const char *str = NULL;
		TRY(Tgt && Tgt->GetData(&str), "S tag is illigal.");
		param.gm = UtilGetEquationId(str);
		TRY(param.gm, UtilStrFormat("%s is unknown", str));

		// add function
		c->funcs.push_back((void*)(SwCtrl::log10));
		c->params.push_back(param);
	}

	m.insert(SwCtrlMapT::value_type(t,c));

	return m.size();
}

int MembraneSolv::SwControl(const SwCtrlMapT& m, S4D_REAL* SW, int* mcoef)
{
	const int I = cfg.Pdiv.size[0];
	const int J = cfg.Pdiv.size[1];
	const int K = cfg.Pdiv.size[2];
	
	for (SwCtrlMapT::const_iterator i = m.begin(); i != m.end(); i++) {

		const Transporter*  t = i->first;
		const SwController* c = i->second;

		if (!(c->params.size())) { continue; }

		int mflag;
		CheckMembArea(&mflag, t->m, t->n);

		// associative funnction
		SKL_REAL (*assoc)(SKL_REAL,SKL_REAL) =
			(c->assoc == SwController::SUM)? SwCtrl::sum:
			(c->assoc == SwController::MUL)? SwCtrl::mul:
			(c->assoc == SwController::MIN)? SwCtrl::min:
			(c->assoc == SwController::MAX)? SwCtrl::max: NULL;

		for (int k = 0; k < K; k++) {
		for (int j = 0; j < J; j++) {
		for (int i = 0; i < I; i++) {
#define FUNC SKL_REAL (*)(const SKL_REAL S, const SwFuncParam& p) 
			// initial value
			SW->val(i,j,k,t->id)
				= ((FUNC)(c->funcs[0]))(
					P0->val(i,j,k,c->params[0].gm->id),
					c->params[0]);
			// 
			for (int f = 1; f < c->funcs.size(); f++) {
				SW->val(i,j,k,t->id)
					= (*assoc)(
						SW->val(i,j,k,t->id),
						((FUNC)(c->funcs[f]))(
							P0->val(i,j,k,c->params[f].gm->id),
							c->params[f]));
			}
#undef FUNC
			//
			const int nni = static_cast<int>(Nx->val(i,j,k,t->m->id));
			const int nnj = static_cast<int>(Ny->val(i,j,k,t->m->id));
			const int nnk = static_cast<int>(Nz->val(i,j,k,t->m->id));

			const int ni  = nni * mflag + mcoef[0];
			const int nj  = nnj * mflag + mcoef[1];
			const int nk  = nnk * mflag + mcoef[2];

			if (ni || nj || nk) { // is membrane on voxel face
				SW->val(i+ni,j+nj,k+nk,t->id) = SW->val(i,j,k,t->id);
			}
		}}}
	}

	return 0;
}

int MembraneSolv::RegistFields()
{
	// penetration
	for (int ip = 0; ip < cfg.PeneCount; ip++) {
		const penetrater& p = *cfg.PeneTbl[ip];
		const string& name = p.name;
		UtilRegArrayS4DReal("K:"+name,Kpene,ip);
	}

	// channel
	for (int ip = 0; ip < cfg.ChTrCount; ip++) {
		const ChTranser& p = *cfg.ChTrTbl[ip];
		const string& name = p.name;

		UtilRegArrayS4DReal("SW:"+name,   SWcha,   ip);
		UtilRegArrayS4DReal("Den:"+name,  DenCha,  ip);
		UtilRegArrayS4DReal("Fuzzy:"+name,FuzzyCha,ip);
		UtilRegArrayS4DReal("K:"+name,    KCha,    ip);
		UtilRegArrayS4DReal("Ke:"+name,   KeCha,   ip);
	}

	// pump
	for (int ip = 0; ip < cfg.PmTrCount; ip++) {
		const PmTranser& p = *cfg.PmTrTbl[ip];
		const string& name = p.name;

		UtilRegArrayS4DReal("SW:"+name,     SWpum,     ip);
		UtilRegArrayS4DReal("Den:"+name,    DenPum,    ip);
		UtilRegArrayS4DReal("Fuzzy:"+name,  FuzzyPum,  ip);
		UtilRegArrayS4DReal("KmAtp:"+name,  KmPumAtp,  ip);
		UtilRegArrayS4DReal("KmGtp:"+name,  KmPumGtp,  ip);
		UtilRegArrayS4DReal("VmaxAtp:"+name,VmaxPumAtp,ip);
		UtilRegArrayS4DReal("VmaxGtp:"+name,VmaxPumGtp,ip);
	}

	// channel symport
	for (int ip = 0; ip < cfg.CSTrCount; ip++) {
		const CSTranser& p = *cfg.CSTrTbl[ip];
		const string& name = p.name;

		UtilRegArrayS4DReal("SW:"+name,   SWchaSym,   ip);
		UtilRegArrayS4DReal("Den:"+name,  DenChaSym,  ip);
		UtilRegArrayS4DReal("Fuzzy:"+name,FuzzyChaSym,ip);
		UtilRegArrayS4DReal("K:"+name,    KChaSym,    ip);
		UtilRegArrayS4DReal("Ke:"+name,   KeChaSym,   ip);
	}

	// pump symport
	for (int ip = 0; ip < cfg.PSTrCount; ip++) {
		const PSTranser& p = *cfg.PSTrTbl[ip];
		const string& name = p.name;

		UtilRegArrayS4DReal("SW:"+name,     SWpumSym,     ip);
		UtilRegArrayS4DReal("Den:"+name,    DenPumSym,    ip);
		UtilRegArrayS4DReal("Fuzzy:"+name,  FuzzyPumSym,  ip);
		UtilRegArrayS4DReal("KmAtp:"+name,  KmPumSymAtp,  ip);
		UtilRegArrayS4DReal("KmGtp:"+name,  KmPumSymGtp,  ip);
		UtilRegArrayS4DReal("VmaxAtp:"+name,VmaxPumSymAtp,ip);
		UtilRegArrayS4DReal("VmaxGtp:"+name,VmaxPumSymGtp,ip);
	}

	// channel antiport
	for (int ip = 0; ip < cfg.CATrCount; ip++) {
		const CATranser& p = *cfg.CATrTbl[ip];
		const string& name = p.name;

		UtilRegArrayS4DReal("SW:"+name,   SWchaAnti,   ip);
		UtilRegArrayS4DReal("Den:"+name,  DenChaAnti,  ip);
		UtilRegArrayS4DReal("Fuzzy:"+name,FuzzyChaAnti,ip);
		UtilRegArrayS4DReal("K:"+name,    KChaAnti,    ip);
		UtilRegArrayS4DReal("Ke:"+name,   KeChaAnti,   ip);
	}

	// pump antiport
	for (int ip = 0; ip < cfg.PATrCount; ip++) {
		const PATranser& p = *cfg.PATrTbl[ip];
		const string& name = p.name;

		UtilRegArrayS4DReal("SW:"+name,     SWpumAnti,     ip);
		UtilRegArrayS4DReal("Den:"+name,    DenPumAnti,    ip);
		UtilRegArrayS4DReal("Fuzzy:"+name,  FuzzyPumAnti,  ip);
		UtilRegArrayS4DReal("KmAtp:"+name,  KmPumAntiAtp,  ip);
		UtilRegArrayS4DReal("KmGtp:"+name,  KmPumAntiGtp,  ip);
		UtilRegArrayS4DReal("VmaxAtp:"+name,VmaxPumAntiAtp,ip);
		UtilRegArrayS4DReal("VmaxGtp:"+name,VmaxPumAntiGtp,ip);
	}

	// gap junction
	for (int ip = 0; ip < cfg.GpTrCount; ip++) {
		const GpTranser& p = *cfg.GpTrTbl[ip];
		const string& name = p.name;

		UtilRegArrayS4DReal("SW:"+name,   SWgap,   ip);
		UtilRegArrayS4DReal("Den:"+name,  DenGap,  ip);
		UtilRegArrayS4DReal("Fuzzy:"+name,FuzzyGap,ip);
		UtilRegArrayS4DReal("K:"+name,    KGap,    ip);
		UtilRegArrayS4DReal("Ke:"+name,   KeGap,   ip);
	}

	return 1;
}

const Membarea* MembraneSolv::CheckMembArea(int* flag, const Material* mat1, const Material* mat2)
{
	*flag = 0;
	const Membarea* area = UtilGetMembareaId(mat1, mat2);
	if( !area ){
		*flag = 1;
	}

	return area;
}
