//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#ifndef __METABOLISM_SOLVER__
#define __METABOLISM_SOLVER__

#include <map>

#include <Foundation/ecellif/emcsim.h>
#include <Foundation/datatype.h>

class MetaboSolv {
private:
	struct FusAreaT {
		unsigned int matid; // material ID
		SKL_REAL     Vfm;   // total Volume ratio
		SKL_REAL     Vfim;  // total Volume ratio inverse
		SKL_REAL*    fmol;  // total mol num array
		unsigned int cells; // number of Voxels
#ifndef _WITHOUT_ECELL
		emcsim*      sim;   // simulator
#endif // _WITHOUT_ECELL
	};
	typedef pair<string, string> MtbPrbTgtObjT; // pair of mat and FullPN.

	struct MetaboProbeT {
		string fname;
		SKL_REAL x, y, z;
		unsigned int i, j, k;      // local Index 
		unsigned int _i, _j, _k;   // overall Index
		int nID;                   // node ID whitch has probed cell
		vector<MtbPrbTgtObjT> obj; // pair(matname, fullpn)
#ifndef _WITHOUT_ECELL
#ifdef ECELL_3_1
		libecs::DataPointVectorIterator* log_iter; // log output iterator
#else
		std::size_t log_iter; // log output iterator
#endif //ECELL_3_1
#endif // _WITHOUT_ECELL
	};

	typedef vector<MetaboProbeT> MetaboProbeListT;

	const CfgInfo& cfg;
	const int vc;

	MetaboProbeListT MetaboProbeList;
	OutDataMapT      MetaboOutDataMap;

	bool enableMetabo;

	S4D_VOIDP* EmcSim;
	S4D_BOOL*  SimMsk;
	S4D_REAL*  TickMemb;
	S4D_REAL*  fm;

	vector<FusAreaT> FusArea;
#ifndef _WITHOUT_ECELL
	vector<emcsim*>  MtbSim;
#endif // _WITHOUT_ECELL

	const S4D_REAL* Vfm;  // source
	const S4D_REAL* Vfim; // source
	const S4D_REAL* P0;   // source
	const S4D_REAL* Sm;   // 
	const S4D_REAL* Nx;   // 
	const S4D_REAL* Ny;   // 
	const S4D_REAL* Nz;   // 
	const S4D_REAL* SmC;  // 
	const S4D_REAL* SmX;  // 
	const S4D_REAL* SmY;  // 
	const S4D_REAL* SmZ;  // 
	S4D_REAL* E;          // destination E(Phi'^n_(m,l))

	S4D_REAL* EVmdt;     // E(Phi'^n_(m,l)) * Vm * dt
	S4D_BOOL* FusMsk;    // F^n_m

	double clock;
	double comm_clk;
	double wait_clk;  // 2011.05.11

protected:
	int _split   (const size_t I, const size_t J, const size_t K);
	int _refusion(const size_t I, const size_t J, const size_t K);
	int _bind    (const size_t I, const size_t J, const size_t K);
	int _run     (const size_t I, const size_t J, const size_t K);
	int _emitt   (const size_t I, const size_t J, const size_t K);
	int _diff    (const size_t I, const size_t J, const size_t K);
	int _post    (const size_t I, const size_t J, const size_t K);

	void _probe_output();
	void _output_field(const size_t I, const size_t J, const size_t K, const SKL_REAL time, const unsigned int step);
	void _split_tgtstr(const string& tgt, string& matname, string& fullpn);
	void _construct_ecell(const size_t I, const size_t J, const size_t K);
	void _allocate_fields(const size_t I, const size_t J, const size_t K);
	void _init_fusion_mask(const size_t I, const size_t J, const size_t K);
	void _init_fldout(const CfgElem* output, const bool file_initialize);
	void _init_probe(const CfgElem* output, const bool file_initialize);

public:
#ifdef SPHERE175
	//ParaComponent, ParaManager
	SklParaComponent *para_cmp;
	SklParaManager *para_mng;
#endif

	MetaboSolv(const CfgInfo& _cfg);

	int MetaboSolvInitialize(const CfgElem* sim,
			S4D_REAL* res, const S4D_REAL* vfm, const S4D_REAL* vfim,
			const S4D_REAL* s,
			const S4D_REAL* nx, const S4D_REAL* ny, const S4D_REAL* nz,
			const S4D_REAL* p0,
			const S4D_REAL* smc,
			const S4D_REAL* smx, const S4D_REAL* smy, const S4D_REAL* smz);
	int MetaboSolvLoop(const SKL_REAL time, const unsigned int step);
	int MetaboSolvPost();

	int chkCfg();
	double GetClock()   const { return clock;    }
	double GetCommClk() const { return comm_clk; }
	double GetWaitClk() const { return wait_clk; } // 2011.05.11

	bool enable()       const { return enableMetabo; }
};

#endif // __METABOLISM_SOLVER__

