/*
 * Copyright (c) RIKEN, Japan. All right reserved. 2008-2009
 *
 */

#include <SklUtil.h>

#include <iostream>
#include <cmath>
#include <cstring>
#include <map>
#include <cfloat>

#include <Foundation/const.h>
#include <Foundation/util.h>
#include <Foundation/error.h>
#include <Foundation/datatype.h>
#include <Foundation/SklSolverFoundation.h>

#include "PotentialSolv.h"
#include "FortranFuncPotential.h"

#define D_DIFF_BASIC 0
#define D_DIFF_26WAY 1

/////////////////////////////////////////////////////////////////////////////
// Constructor
//
PotentialSolv::PotentialSolv(const CfgInfo& _cfg)
: cfg(_cfg), vc(VC)
{
	clock    = double(0.0);
	comm_clk = double(0.0);
	wait_clk = double(0.0);
	hh_calc = true;
}

/////////////////////////////////////////////////////////////////////////////
// Destractor
//
PotentialSolv::~PotentialSolv()
{

        for( int i=0; i<h_MemPotTbl.size(); i++ ) {
           int napp = h_MemPotTbl[i]->IappList.size();
           for(int j=0; j<napp; j++ ) {

             int nl = h_MemPotTbl[i]->IappList[j]->indexList.size();
             for(int k=0; k<nl; k++) {
                delete h_MemPotTbl[i]->IappList[j]->indexList[k];
             }

             delete h_MemPotTbl[i]->IappList[j];
           }
        
           h_MemPotTbl[i]->IappList.clear(); 

           delete h_MemPotTbl[i];
        }
        h_MemPotTbl.clear(); 

}

/////////////////////////////////////////////////////////////////////////////
// Initialize Main
//
int PotentialSolv::PotentialInitialize(
                  const CfgElem* simElem, const CfgElem* HHRoot,
                  S4D_REAL *P1, const S4D_REAL *P0, 
                  const S4D_REAL *AX, const S4D_REAL *AY, const S4D_REAL *AZ,
                  const S4D_REAL *PHI, const S4D_REAL *PHII,
                  const S4D_REAL *MEMBAREA, const S4D_REAL *MEMBNOMALX,
                  const S4D_REAL *MEMBNOMALY, const S4D_REAL *MEMBNOMALZ,
                  S4D_REAL *DIFFX, S4D_REAL *DIFFY, S4D_REAL *DIFFZ,
                  S4D_REAL *COEFX, S4D_REAL *COEFY, S4D_REAL *COEFZ,
                  const S4D_INT *BCX, const S4D_INT *BCY, const S4D_INT *BCZ,
		  const S4D_REAL *smc,
		  const S4D_REAL *smx, const S4D_REAL *smy, const S4D_REAL *smz)
{
#ifndef ENABLE_MEMBRANEPOTENTIAL
	enableHH = false;
	return 1;
#else

	DBX(MESSAGE("PotentialInitialize"));

#ifdef SPHERE175
	para_cmp = cfg.solver->SklGetParaComponent();
	para_mng = para_cmp->GetParaManager();
#endif

	//pointer set
	h_p1=P1;
	h_p0=P0;
	h_AX=AX;
	h_AY=AY;
	h_AZ=AZ;
	h_phi=PHI;
	h_phii=PHII;
	h_membarea=MEMBAREA;
	h_membnomalx=MEMBNOMALX;
	h_membnomaly=MEMBNOMALY;
	h_membnomalz=MEMBNOMALZ;
	h_diffx=DIFFX;
	h_diffy=DIFFY;
	h_diffz=DIFFZ;
	h_CX=COEFX;
	h_CY=COEFY;
	h_CZ=COEFZ;
	h_BCx = BCX;
	h_BCy = BCY;
	h_BCz = BCZ;
	h_smc = smc;
	h_smx = smx;
	h_smy = smy;
	h_smz = smz;
  //

	unsigned const int nstep = 0;

	const CfgElem*  hh = simElem? simElem->GetElemFirst ("membranePotential"): NULL;
	const CfgParam* exe = hh? hh->GetParamFirst("execution"): NULL;
	const CfgParam* diff_type = hh? hh->GetParamFirst("diff_type"): NULL;

	// set isenable metabolism simulation
	const char* exc = NULL;
	enableHH = (exe && exe->GetData(&exc) && (strcasecmp(exc, "true") == 0));

	if( !enableHH ) { return true; }
        
	// set flag diffusion type
	exc = NULL;
	if(diff_type && diff_type->GetData(&exc) && (strcasecmp(exc, "basic") == 0)){
		diffTypeFlag = D_DIFF_BASIC;
		MESSAGE(string("DiffusionType. basic diffTypeFlag : ") + string_cast(diffTypeFlag));
	}
	else
	{
		diffTypeFlag = D_DIFF_26WAY;
		MESSAGE(string("DiffusionType. 26way diffTypeFlag : ") + string_cast(diffTypeFlag));
	}

	if (!enableHH) {
		MESSAGE("sim_setting membranePotential execution false");
		return 1;
	}

	DBX(MESSAGE(UtilStrFormat("diffTypeFlag : %d",diffTypeFlag)));

	if( !HHRoot )
	{
		MSG_ERROR(UtilStrFormat("membranePotential_setting Elem is undefined in Parameter tag"));
		MSG_ERROR(UtilStrFormat("Membrane Potential Simulation is changed to false"));
		enableHH = false;
		//return 1;
	}

	// get Molecular
	h_mol = NULL;
	for(int i = 0; i < cfg.MolCount; i++) {
		if( cfg.MolTbl[i]->potential ) {
			if( h_mol == NULL ) {
				h_mol = cfg.MolTbl[i];
				DBX(MESSAGE(UtilStrFormat("Membrane Potential Molecular name : %s", h_mol->name.c_str())));
			} else {
				MSG_WARNING(UtilStrFormat("duplicate molecular Elem effective molecular : %s", h_mol->name.c_str()));
			}
		}
	}

	// get voxel size 
	const int imax = cfg.Pdiv.size[0];
	const int jmax = cfg.Pdiv.size[1];
	const int kmax = cfg.Pdiv.size[2];

	// Parse XML
	if( !ParseXML(HHRoot) )
	{
		enableHH = false;
		return -1;
	}

	// LoadPositionSPH
	if( !LoadPositionSPH() ) 
	{
		enableHH = false;
		return -1;
	}

	// zero clear
	for(int i = 0; i < cfg.EqnCount; i++) {
		if (!cfg.PeqnAss[i]) { continue; }

		// Molecular not MembrainePotential -> continue
		int ml = cfg.EqnTbl[i]->mol->id;
		int mt = cfg.EqnTbl[i]->mat->id;
		if( cfg.MolTbl[ml] != h_mol ) { continue; }

		DBX(MESSAGE(UtilStrFormat("P0,CX,CY,CZ, ZERO CLEAR Eqn name %s"
		                         ,cfg.EqnTbl[i]->name.c_str())));

		//p0
		const SKL_REAL *p = UtilGetS4DN(h_p0,i);
		clear4s_(&imax,&jmax,&kmax,&vc,p);

		//CX
		SKL_REAL *CX = UtilGetS4DN(h_CX,i);
		clear4s_(&imax,&jmax,&kmax,&vc,CX);

		//CY
		SKL_REAL *CY = UtilGetS4DN(h_CY,i);
		clear4s_(&imax,&jmax,&kmax,&vc,CY);

		//CZ
		SKL_REAL *CZ = UtilGetS4DN(h_CZ,i);
		clear4s_(&imax,&jmax,&kmax,&vc,CZ);
	}


	// get nID
	// check next node id
	const int& gid = cfg.grpP.gid;

#ifdef SPHERE175
	h_nID[0] = para_mng->IsCommID(-1, 0, 0, gid); // -X
	h_nID[1] = para_mng->IsCommID( 1, 0, 0, gid); // +X
	h_nID[2] = para_mng->IsCommID( 0,-1, 0, gid); // -Y
	h_nID[3] = para_mng->IsCommID( 0, 1, 0, gid); // +Y
	h_nID[4] = para_mng->IsCommID( 0, 0,-1, gid); // -Z
	h_nID[5] = para_mng->IsCommID( 0, 0, 1, gid); // +Z
#else
	h_nID[0] = ParaMngr.GetCommID(-1, 0, 0, gid); // -X
	h_nID[1] = ParaMngr.GetCommID( 1, 0, 0, gid); // +X
	h_nID[2] = ParaMngr.GetCommID( 0,-1, 0, gid); // -Y
	h_nID[3] = ParaMngr.GetCommID( 0, 1, 0, gid); // +Y
	h_nID[4] = ParaMngr.GetCommID( 0, 0,-1, gid); // -Z
	h_nID[5] = ParaMngr.GetCommID( 0, 0, 1, gid); // +Z
#endif

  //
	Allocate();


  // potential loop
	for (int e = 0; e < h_MemPotTbl.size(); e++) {

		const MembranePotential* Poten = h_MemPotTbl[e];
		h_diffcoef = Poten->diffcoef;
		h_V =  Poten->V;
		h_M =  Poten->M;
		h_N =  Poten->N;
		h_H =  Poten->H;

		DBX(MESSAGE(UtilStrFormat("Input XML Membrane Potential Parameter : %s", Poten->name)));
		DBX(MESSAGE(UtilStrFormat("diffcoef: %e V: %e M: %e N: %e H:%e", h_diffcoef,h_V,h_M,h_N,h_H))); 

		const Material* mat1 = Poten->mat1;
		const Material* mat2 = Poten->mat2;
		const Equation* gm = UtilGetEquationId(h_mol,mat1);

		const SKL_REAL  *mbnomalx = UtilGetS4DN(h_membnomalx,mat1->id);
		const SKL_REAL  *mbnomaly = UtilGetS4DN(h_membnomaly,mat1->id);
		const SKL_REAL  *mbnomalz = UtilGetS4DN(h_membnomalz,mat1->id);

		SKL_REAL *CX = UtilGetS4DN(h_CX,gm->id);
		SKL_REAL *CY = UtilGetS4DN(h_CY,gm->id);
		SKL_REAL *CZ = UtilGetS4DN(h_CZ,gm->id);
		SKL_REAL *coefdvdt = UtilGetS4DN(h_coefdvdt,e);

		// make cx, cy, cz
		const Membarea* area = UtilGetMembareaId(mat1, mat2);
		if( area ){
			const SKL_REAL *smc = UtilGetS4DN(h_smc, area->id);
			const SKL_REAL *smx = UtilGetS4DN(h_smx, area->id);
			const SKL_REAL *smy = UtilGetS4DN(h_smy, area->id);
			const SKL_REAL *smz = UtilGetS4DN(h_smz, area->id);

			setcoefreadmembarea_(
				&imax,&jmax,&kmax,&vc,&h_diffcoef,smc,smx,smy,smz,
				mbnomalx,mbnomaly,mbnomalz,CX,CY,CZ,coefdvdt);
		}
		else{
			const SKL_REAL  *mbarea1   = UtilGetS4DN(h_membarea,mat1->id);
			const SKL_REAL  *mbarea2   = UtilGetS4DN(h_membarea,mat2->id);

			setcoef_(&imax,&jmax,&kmax,&vc,&h_diffcoef,mbarea1,mbarea2,mbnomalx
			        ,mbnomaly,mbnomalz,CX,CY,CZ,coefdvdt);
		}

		SKL_REAL *mm = UtilGetS4DN(h_mm,e);
		SKL_REAL *nn = UtilGetS4DN(h_nn,e);
		SKL_REAL *hh = UtilGetS4DN(h_hh,e);

		//const SKL_REAL *val = UtilGetS4DN(h_val,gm->id);
		const SKL_REAL *val = UtilGetS4DN(h_p0,gm->id);

		//set mm,nn,hh,val
		inithh_(&imax,&jmax,&kmax,&vc,&h_M,&h_N,&h_H,&h_V,mm,nn,hh
		       ,coefdvdt,val);

		const SKL_REAL *Ax = UtilGetS4DN(h_AX,mat1->id);
		const SKL_REAL *Ay = UtilGetS4DN(h_AY,mat1->id);
		const SKL_REAL *Az = UtilGetS4DN(h_AZ,mat1->id);

		SKL_REAL *ax = UtilGetS4DN(h_ax,mat1->id);
		SKL_REAL *ay = UtilGetS4DN(h_ay,mat1->id);
		SKL_REAL *az = UtilGetS4DN(h_az,mat1->id);
		//set ax,ay,az
		setopenarea_(ax,ay,az,Ax,Ay,Az,coefdvdt
		            ,&imax,&jmax,&kmax,&vc);

	}


  // equation loop
        for (int i = 0; i < cfg.EqnCount; i++) {
                if (!cfg.PeqnAss[i]) { continue; }

                // Molecular not MembrainePotential -> continue
                int ml = cfg.EqnTbl[i]->mol->id;
                int mt = cfg.EqnTbl[i]->mat->id;
                if( cfg.MolTbl[ml] != h_mol ) { continue; }
                if( !chkPoten(mt) ) { continue; }

                SKL_REAL* _cx = h_cx->GetData();
                SKL_REAL* _cy = h_cy->GetData();
                SKL_REAL* _cz = h_cz->GetData();
                SKL_REAL* _CX = h_CX->GetData();
                SKL_REAL* _CY = h_CY->GetData();
                SKL_REAL* _CZ = h_CZ->GetData();
                int eqns = cfg.EqnCount;


                //setbndcoef
                int eqidx = i+1;

                setbndcoef_hh_(_CX,_CY,_CZ,h_nID,&eqidx, &eqns, &imax, &jmax, &kmax,&vc);

                calccoef_hh_(_cx,_cy,_cz,_CX,_CY,_CZ,h_nID,&eqidx, &eqns, &imax, &jmax, &kmax,&vc);

        }
#endif
	return 1;
}

/////////////////////////////////////////////////////////////////////////////
// Loop Main
//
int PotentialSolv::PotentialLoop(const SKL_REAL dt,  const unsigned int step)
{
// Return value -> -1 : error
//                  0 : forced termination
//                  1 : normal return

        if (!enableHH) { return 1; }

        DBX(MESSAGE("\nPotentialLoop"));

	BEGIN_CLOCK(loop);

  // calc dxi, dyi, dzi
        const SKL_REAL& dx = cfg.pitch[0], dxi = SKL_REAL(1.0)/dx;
        const SKL_REAL& dy = cfg.pitch[1], dyi = SKL_REAL(1.0)/dy;
        const SKL_REAL& dz = cfg.pitch[2], dzi = SKL_REAL(1.0)/dz;

  // get voxel size 
        const int imax = cfg.Pdiv.size[0];
        const int jmax = cfg.Pdiv.size[1];
        const int kmax = cfg.Pdiv.size[2];

  // P0 -> p0 & dvdt zero clear
        for( int i=0; i<cfg.EqnCount; i++ ) {
                if (!cfg.PeqnAss[i]) { continue; }

                // Molecular not MembrainePotential -> continue
                int ml = cfg.EqnTbl[i]->mol->id;
                int mt = cfg.EqnTbl[i]->mat->id;

                if( cfg.MolTbl[ml] != h_mol ) { continue; }

                if( !chkPoten(mt) ) { continue; }
                //DBX(MESSAGE(UtilStrFormat("Eqn name : %s",cfg.EqnTbl[i]->name.c_str())));
                //DBX(MESSAGE(UtilStrFormat("    mat name : %s",cfg.EqnTbl[i]->mat->name.c_str())));
                //DBX(MESSAGE(UtilStrFormat("    mol name : %s",cfg.EqnTbl[i]->mol->name.c_str())));
                //DBX(MESSAGE(UtilStrFormat("Eqn id : %d mol id : %d mat id : %d",i,ml,mt)));

                // copy
                const SKL_REAL *dst = UtilGetS4DN(h_val,i);
                const SKL_REAL *src = UtilGetS4DN(h_p0,i);

                ricscpys3d_(dst, src, &imax, &jmax, &kmax, &vc);

                // dvdt zero clear
                const SKL_REAL *dvdt = UtilGetS4DN(h_dvdt,i);
                clear4s_(&imax,&jmax,&kmax,&vc,dvdt);

        }

  // SetBndCell
        SetBndCell();

  // equation loop
        for (int e = 0; e < cfg.EqnCount; e++) {
                if (!cfg.PeqnAss[e]) { continue; }

                // Molecular not MembrainePotential -> continue
                int ml = cfg.EqnTbl[e]->mol->id;
                int mt = cfg.EqnTbl[e]->mat->id;

                //DBX(MESSAGE(string("    mat name  : ")+string(cfg.EqnTbl[e]->mat->name.c_str())));
                //DBX(MESSAGE(string("    mol name  : ")+string(cfg.EqnTbl[e]->mol->name.c_str())));

                if( cfg.MolTbl[ml] != h_mol ) { continue; }
                if( !chkPoten(mt) ) { continue; }

                const SKL_REAL *val = UtilGetS4DN(h_val,e);
                const SKL_REAL *ax = UtilGetS4DN(h_ax,mt);
                const SKL_REAL *ay = UtilGetS4DN(h_ay,mt);
                const SKL_REAL *az = UtilGetS4DN(h_az,mt);
                SKL_REAL *diffx = UtilGetS4DN(h_diffx,e);
                SKL_REAL *diffy = UtilGetS4DN(h_diffy,e);
                SKL_REAL *diffz = UtilGetS4DN(h_diffz,e);
                const SKL_REAL *cx = UtilGetS4DN(h_cx,e);
                const SKL_REAL *cy = UtilGetS4DN(h_cy,e);
                const SKL_REAL *cz = UtilGetS4DN(h_cz,e);

                DBX(MESSAGE(string("calc diff Eqn name : ")+string(cfg.EqnTbl[e]->name.c_str())));

                // calc diff
                differential_hh_(&diffTypeFlag,val,cx,cy,cz,ax,ay,az,diffx,diffy,diffz,
                              &imax,&jmax,&kmax,&vc,&dxi,&dyi,&dzi);

                DB5(MESSAGE(string("P0 of ")+cfg.EqnTbl[e]->name));
                DB5(MESSAGE(S4D2string(h_val, e)));

                DB5(MESSAGE(string("Area for X of ")+cfg.MatTbl[mt]->name));
                DB5(MESSAGE(S4D2string(h_ax,mt)));
                DB5(MESSAGE(string("Area for Y of ")+cfg.MatTbl[mt]->name));
                DB5(MESSAGE(S4D2string(h_ay,mt)));
                DB5(MESSAGE(string("Area for Z of ")+cfg.MatTbl[mt]->name));
                DB5(MESSAGE(S4D2string(h_az,mt)));

                DB5(MESSAGE(string("VC CoefX of ")+cfg.EqnTbl[e]->name));
                DB5(MESSAGE(S4D2string(h_CX,e)));
                DB5(MESSAGE(string("VC CoefY of ")+cfg.EqnTbl[e]->name));
                DB5(MESSAGE(S4D2string(h_CY,e)));
                DB5(MESSAGE(string("VC CoefZ of ")+cfg.EqnTbl[e]->name));
                DB5(MESSAGE(S4D2string(h_CZ,e)));

                DB5(MESSAGE(string("VC CoefX of ")+cfg.EqnTbl[e]->name));
                DB5(MESSAGE(S4D2string(h_cx,e,true)));
                DB5(MESSAGE(string("VC CoefY of ")+cfg.EqnTbl[e]->name));
                DB5(MESSAGE(S4D2string(h_cy,e,true)));
                DB5(MESSAGE(string("VC CoefZ of ")+cfg.EqnTbl[e]->name));
                DB5(MESSAGE(S4D2string(h_cz,e,true)));

                DB5(MESSAGE(string("Flux for X of ")+cfg.EqnTbl[e]->name));
                DB5(MESSAGE(S4D2string(h_diffx,e,true)));
                DB5(MESSAGE(string("Flux for Y of ")+cfg.EqnTbl[e]->name));
                DB5(MESSAGE(S4D2string(h_diffy,e,true)));
                DB5(MESSAGE(string("Flux for Z of ")+cfg.EqnTbl[e]->name));
                DB5(MESSAGE(S4D2string(h_diffz,e,true)));

        }
 // potential loop
        if( !hh_calc ) {
	  END_CLOCK(clock, loop);
	  return 1;
        }

        for (int i = 0; i < h_MemPotTbl.size(); i++) {

                const MembranePotential* Poten = h_MemPotTbl[i];
                const Material* mat1 = Poten->mat1;
                const Equation* gm = UtilGetEquationId(h_mol,mat1);

                h_Cm = Poten->Cm;
                h_gK = Poten->gK;
                h_gNa = Poten->gNa;
                h_gL = Poten->gL;
                h_vK = Poten->vK;
                h_vNa = Poten->vNa;
                h_vL = Poten->vL;

                //DBX(MESSAGE("Calc Membrane Potential Parameter"));
                DBX(MESSAGE(UtilStrFormat("calc hh Eqn name : %s",
                            gm->name.c_str())));
                //DBX(MESSAGE(UtilStrFormat("Cm : %e",h_Cm)));
                //DBX(MESSAGE(UtilStrFormat("gk : %e gNa : %e gL : %e",
                //            h_gK,h_gNa,h_gL)));
                //DBX(MESSAGE(UtilStrFormat("vk : %e vNa : %e vL : %e",
                //            h_vK,h_vNa,h_vL)));

                // calc hh
                SKL_REAL *mm = UtilGetS4DN(h_mm,i);
                SKL_REAL *nn = UtilGetS4DN(h_nn,i);
                SKL_REAL *hh = UtilGetS4DN(h_hh,i);

                SKL_REAL *coefdvdt = UtilGetS4DN(h_coefdvdt,i);
                SKL_REAL *dvdt =     UtilGetS4DN(h_dvdt,gm->id);
                SKL_REAL *val =      UtilGetS4DN(h_val,gm->id);

                calchh_(&imax, &jmax, &kmax, &vc, &h_Cm, &h_gK, &h_gNa, &h_gL,
                          &h_vK, &h_vNa, &h_vL, &dt, val, mm, nn, hh, dvdt, coefdvdt);

                SKL_REAL time = RicsGetCurrentTime();

                for(int j=0; j<Poten->IappList.size(); j++) {

                   IappTable *iapptbl = Poten->IappList[j];

                   int ii,jj,kk;

                   for(int k=0; k<iapptbl->indexList.size(); k++) {
                       IappIndex *index = iapptbl->indexList[k];
                       if( !UtilCheckIdxPdiv(&ii,&jj,&kk,index->i,index->j,index->k) ) continue;
                       SKL_REAL iapp = iapptbl->GetTimeValueByLinear(time);

                       DBX(MESSAGE(UtilStrFormat("dvdt %e Iapp data I : %d J : %d k : %d iapp : %e",
                                                 h_dvdt->val(ii,jj,kk,gm->id),ii,jj,kk,iapp)));

                       h_dvdt->val(ii,jj,kk,gm->id)+=(iapp*h_coefdvdt->val(ii,jj,kk,i));

                       DBX(MESSAGE(UtilStrFormat("dvdt %e coefdvdt : %e",
                                     h_dvdt->val(ii,jj,kk,gm->id),h_coefdvdt->val(ii,jj,kk,i))));

                   }
                }
                dvdtdvcm_(&imax,&jmax,&kmax,&vc,&h_Cm,dvdt,coefdvdt);
        }

	END_CLOCK(clock, loop);

	return 1;
}

/////////////////////////////////////////////////////////////////////////////
// Post Main
//
int PotentialSolv::PotentialPost()
{
  DBX(MESSAGE("PotentialPost"));
	return 1;
}

/////////////////////////////////////////////////////////////////////////////
// Parse XML config
//
bool PotentialSolv::ParseXML(const CfgElem* hhElem)
{
	if( !hhElem ) return false;

        DBX(MESSAGE("PotentialSolv::ParseXML"));
        int cnt=0;

        //DEBUG FLAG SET
        const CfgParam* dbParam = hhElem->GetParamFirst("hh_calc");
        if( dbParam ) {
          const char* debug = NULL;
          if( dbParam ) {
            if( dbParam->GetData(&debug) && (strcasecmp(debug, "false") == 0)) hh_calc=false;
            //if( !hh_calc ) DBX(MESSAGE("hh_calc false"));
          }
        }

        const CfgElem* mpElem = hhElem->GetElemFirst();
        if( mpElem ) {
          while( mpElem ) {
            const char* name = mpElem->GetName();
            MembranePotential* mempote = new MembranePotential;
            if(!mempote->GetPotentialMat(name)) {continue;}
            mempote->name = name;
            //cnt=cnt+1;
            cnt+=1;
            mempote->id=cnt;
            mempote->ParseXML(mpElem,cfg);

            // add membranePotential parameter table  
            h_MemPotTbl.push_back(mempote);

            mpElem =  hhElem->GetElemNext(mpElem);
          }
        }
        if ( h_MemPotTbl.size() == 0 ) { 
           MSG_ERROR(UtilStrFormat("Membrane Potential Elem is undefined in XML"));
           return false;
        }
	return true;
}

/////////////////////////////////////////////////////////////////////////////
// Allocate Arrays
//
bool PotentialSolv::Allocate()
{

        const size_t& I    = cfg.Pdiv.size[0];
        const size_t& J    = cfg.Pdiv.size[1];
        const size_t& K    = cfg.Pdiv.size[2];
        const size_t& mats = cfg.MatCount;
        const size_t& eqns = cfg.EqnCount;
        const size_t& pots = h_MemPotTbl.size();

        const size_t sz_eqn[4] = { I, J, K, eqns };
        const size_t sz_mat[4] = { I, J, K, mats };
        const size_t sz_pot[4] = { I, J, K, pots };

        const int& gid = cfg.grpP.gid;

#ifdef SPHERE175
        h_val = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray(para_mng, "mem_P0", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
        h_cx = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray(para_mng, "mem_cx", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
        h_cy = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray(para_mng, "mem_cy", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
        h_cz = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray(para_mng, "mem_cz", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
        h_ax = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray(para_mng, "mem_ax", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_mat, vc, gid));
        h_ay = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray(para_mng, "mem_ay", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_mat, vc, gid));
        h_az = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray(para_mng, "mem_az", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_mat, vc, gid));
        h_mm = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray(para_mng, "mem_mm", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_pot, vc, gid));
        h_nn = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray(para_mng, "mem_nn", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_pot, vc, gid));
        h_hh = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray(para_mng, "mem_hh", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_pot, vc, gid));
        h_coefdvdt = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray(para_mng, "mem_coefdvdt", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_pot, vc, gid));
        h_dvdt = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray(para_mng, "mem_dvdt", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
#else
        h_val = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray("mem_P0", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
        h_cx = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray("mem_cx", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
        h_cy = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray("mem_cy", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
        h_cz = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray("mem_cz", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
        h_ax = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray("mem_ax", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_mat, vc, gid));
        h_ay = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray("mem_ay", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_mat, vc, gid));
        h_az = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray("mem_az", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_mat, vc, gid));
        h_mm = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray("mem_mm", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_pot, vc, gid));
        h_nn = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray("mem_nn", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_pot, vc, gid));
        h_hh = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray("mem_hh", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_pot, vc, gid));
        h_coefdvdt = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray("mem_coefdvdt", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_pot, vc, gid));
        h_dvdt = dynamic_cast<S4D_REAL*>(
                                cfg.solver->SklAllocateArray("mem_dvdt", SKL_CLASS_SCALAR4D,
                                                                SKL_ARRAY_DTYPE_REAL, sz_eqn, vc, gid));
#endif
// zero clear
        SKL_REAL *val = h_val->GetData();
        clear4d_(&sz_eqn[0],&sz_eqn[1],&sz_eqn[2],&sz_eqn[3],&vc,val);

        SKL_REAL *cx = h_cx->GetData();
        clear4d_(&sz_eqn[0],&sz_eqn[1],&sz_eqn[2],&sz_eqn[3],&vc,cx);

        SKL_REAL *cy = h_cy->GetData();
        clear4d_(&sz_eqn[0],&sz_eqn[1],&sz_eqn[2],&sz_eqn[3],&vc,cy);

        SKL_REAL *cz = h_cz->GetData();
        clear4d_(&sz_eqn[0],&sz_eqn[1],&sz_eqn[2],&sz_eqn[3],&vc,cz);

        SKL_REAL *ax = h_ax->GetData();
        clear4d_(&sz_mat[0],&sz_mat[1],&sz_mat[2],&sz_mat[3],&vc,ax);

        SKL_REAL *ay = h_ay->GetData();
        clear4d_(&sz_mat[0],&sz_mat[1],&sz_mat[2],&sz_mat[3],&vc,ay);

        SKL_REAL *az = h_az->GetData();
        clear4d_(&sz_mat[0],&sz_mat[1],&sz_mat[2],&sz_mat[3],&vc,az);

        SKL_REAL *mm = h_mm->GetData();
        clear4d_(&sz_pot[0],&sz_pot[1],&sz_pot[2],&sz_pot[3],&vc,mm);

        SKL_REAL *nn = h_nn->GetData();
        clear4d_(&sz_pot[0],&sz_pot[1],&sz_pot[2],&sz_pot[3],&vc,nn);

        SKL_REAL *hh = h_hh->GetData();
        clear4d_(&sz_pot[0],&sz_pot[1],&sz_pot[2],&sz_pot[3],&vc,hh);

        SKL_REAL *coefdvdt = h_coefdvdt->GetData();
        clear4d_(&sz_pot[0],&sz_pot[1],&sz_pot[2],&sz_pot[3],&vc,coefdvdt);

        SKL_REAL *dvdt = h_dvdt->GetData();
        clear4d_(&sz_eqn[0],&sz_eqn[1],&sz_eqn[2],&sz_eqn[3],&vc,dvdt);

        DBX(MESSAGE("PotentialSolv::Allocate"));
	return true;
}

/////////////////////////////////////////////////////////////////////////////
// Check config & print info
//
bool PotentialSolv::chkCfg()
{
//	if (cfg.grpAll.nid != 0) { return false; }
	if (cfg.grpAll.nid == 0)
	{

		MESSAGE("**** Configuration of Potential Solver ****", 0);
		MESSAGE(string("Potential Simulation is ")
		       + string_cast(enableHH, ios::boolalpha));
	}

	if (!enableHH) { return true; }

	// log
	DBX( printLog(cfg.grpAll.nid) );

	return true;
}

bool PotentialSolv::printLog(int nid)
{
	//if( !enableHH ) return true;

	//title
	stringstream wout;
	wout << "Potential Simulation Setting" << endl;
	wout << setw(80) << setfill('=') << "" << setfill(' ') << endl;


	if( nid==0 )
	{
		MESSAGE(wout.str(),nid);
	}
	else
	{
		DBX(MESSAGE(wout.str(),nid));
	}

	return true;
}

int PotentialSolv::SetBndCell()
{
        //DBX(MESSAGE("SetBndCell "));

        const int I = cfg.Pdiv.size[0];
        const int J = cfg.Pdiv.size[1];
        const int K = cfg.Pdiv.size[2];

        for (int e = 0; e < cfg.EqnCount; e++) {
                if (!cfg.PeqnAss[e]) { continue; }

                // Molecular not MembrainePotential -> continue
                int ml = cfg.EqnTbl[e]->mol->id;
                if( cfg.MolTbl[ml] != h_mol ) { continue; }
 
                //DBX(MESSAGE(string("SetBndCell ")+cfg.EqnTbl[e]->name));

                const size_t    m  = cfg.EqnTbl[e]->mat->id;
                const SKL_REAL& dx = cfg.pitch[0];
                const SKL_REAL& dy = cfg.pitch[1];
                const SKL_REAL& dz = cfg.pitch[2];

                // set outter boundary coef.
                // X boundary
                if (h_nID[0] < 0) {
                        for (int k = 0; k < K; k++) {
                        for (int j = 0; j < J; j++) {
                                const int      bc = h_BCx->val(0,j,k,e);
                                const SKL_REAL p  = h_val ->val(0,j,k,e);
                                h_val->val(-1,j,k,e) = CalcBndCell(bc,p,dx,-1);
                        }}
                }
                if (h_nID[1] < 0) {
                        for (int k = 0; k < K; k++) {
                        for (int j = 0; j < J; j++) {
                                const int      bc = h_BCx->val(  1,j,k,e);
                                const SKL_REAL p  = h_val ->val(I-1,j,k,e);
                                h_val->val(I,j,k,e)  = CalcBndCell(bc,p,dx, 1);
                        }}
                }
                // Y boudary
                if (h_nID[2] < 0) {
                        for (int k = 0; k < K; k++) {
                        for (int i = 0; i < I; i++) {
                                const int      bc = h_BCy->val(i,0,k,e);
                                const SKL_REAL p  = h_val ->val(i,0,k,e);
                                h_val->val(i,-1,k,e) = CalcBndCell(bc,p,dy,-1);
                        }}
                }
                if (h_nID[3] < 0) {
                        for (int k = 0; k < K; k++) {
                        for (int i = 0; i < I; i++) {
                                const int      bc = h_BCy->val(i,  1,k,e);
                                const SKL_REAL p  = h_val ->val(i,J-1,k,e);
                                h_val->val(i,J,k,e)  = CalcBndCell(bc,p,dy, 1);
                        }}
                }
                // Z boundary
                if (h_nID[4] < 0) {
                        for (int j = 0; j < J; j++) {
                        for (int i = 0; i < I; i++) {
                                const int      bc = h_BCz->val(i,j,0,e);
                                const SKL_REAL p  = h_val ->val(i,j,0,e);
                                h_val->val(i,j,-1,e) = CalcBndCell(bc,p,dz,-1);
                        }}
                }
                if (h_nID[5] < 0) {
                        for (int j = 0; j < J; j++) {
                        for (int i = 0; i < I; i++) {
                                const int      bc = h_BCz->val(i,j,  1,e);
                                const SKL_REAL p  = h_val ->val(i,j,K-1,e);
                                h_val->val(i,j,K,e)  = CalcBndCell(bc,p,dz, 1);
                        }}
                }

#if 0
                DB5(MESSAGE(S4D2string(p0,e)));
#endif
        }

        return 1;
}
SKL_REAL PotentialSolv::CalcBndCell(
        const int bc, const SKL_REAL c, const SKL_REAL dx, const int face)
{
        const BoundaryCond& cond = cfg.BndCndTbl[bc];
        const SKL_REAL&     val  = cond.value;
        const SKL_REAL      sign = static_cast<SKL_REAL>(face);

        SKL_REAL conc =
                (cond.type == BoundaryCond::DIRICHLET)? SKL_REAL(2.0) * val - c:
                (cond.type == BoundaryCond::NEUMANN)?   c + sign * val * dx:
                                                        0;

        return conc;
}
// check Potential
bool PotentialSolv::chkPoten(const int mt)
{
        // potential loop

        for (int e = 0; e < h_MemPotTbl.size(); e++) {
             const MembranePotential* Poten = h_MemPotTbl[e];
             if( Poten->mat1->id == mt ) return true;
        }
        return false;
}
//exptim
int PotentialSolv::ExptimHH(const SKL_REAL dt)
{
        //if (!enableHH || !hh_calc ) { return 1; }
        if (!enableHH ) { return 1; }
        DBX(MESSAGE(UtilStrFormat("ExptimHH dt : %e",dt)));

  // get voxel size 
        const int imax = cfg.Pdiv.size[0];
        const int jmax = cfg.Pdiv.size[1];
        const int kmax = cfg.Pdiv.size[2];

  // calc dxi, dyi, dzi
        const SKL_REAL& dx = cfg.pitch[0], dxi = SKL_REAL(1.0)/dx;
        const SKL_REAL& dy = cfg.pitch[1], dyi = SKL_REAL(1.0)/dy;
        const SKL_REAL& dz = cfg.pitch[2], dzi = SKL_REAL(1.0)/dz;

  // equation loop
        for (int e = 0; e < cfg.EqnCount; e++) {
                if (!cfg.PeqnAss[e]) { continue; }
                // Molecular not MembrainePotential -> continue
                int ml = cfg.EqnTbl[e]->mol->id;
                int mt = cfg.EqnTbl[e]->mat->id;
                if( cfg.MolTbl[ml] != h_mol ) { continue; }
                if( !chkPoten(mt) ) { continue; }

                SKL_REAL *p1         =   UtilGetS4DN(h_p1,e);
                const SKL_REAL *dvdt =   UtilGetS4DN(h_dvdt,e);

                /*
                const SKL_REAL *val  =   UtilGetS4DN(h_p0,e);
                const SKL_REAL *diffx = UtilGetS4DN(h_diffx,e);
                const SKL_REAL *diffy = UtilGetS4DN(h_diffy,e);
                const SKL_REAL *diffz = UtilGetS4DN(h_diffz,e);
                const SKL_REAL *phii =  UtilGetS4DN(h_phii,mt);

                exptim_diff_(&dt,&imax,&jmax,&kmax,&vc,&dxi,&dyi,&dzi,
                            diffx,diffy,diffz,phii,p1,val);
                */

                if( hh_calc ) exptimhh_(&dt,&imax,&jmax,&kmax,&vc,p1,dvdt); 
        }

        return 1;
}
//iapp position sph load
bool PotentialSolv::LoadPositionSPH()
{
        int i,j,k;

        const vxsize_t* sz = cfg.Pdiv.size;
        const int I = sz[0];
        const int J = sz[1];
        const int K = sz[2];
        S3D_REAL ary(I,J,K,0);

        S4D_REAL* m_vf=NULL;

        int mt=1;

        for (int e = 0; e < h_MemPotTbl.size(); e++) {
           const MembranePotential* Poten = h_MemPotTbl[e];
           const Material* m_mat = Poten->mat1;

           for(int p=0; p<Poten->IappList.size(); p++) {
             //const IappTable* iapptbl = Poten->IappList[p];
             IappTable* iapptbl = Poten->IappList[p];
             if( iapptbl->sphAttr.size() == 0 ) continue;

             for(int s=0; s<iapptbl->sphAttr.size(); s++) {
               const char* attr = iapptbl->sphAttr[s];


               // array Vf
               S4DRN Vf  = UtilGetRegistedArrayS4DReal(UtilStrFormat("Vf_%s",m_mat->name.c_str()));
               if( !(m_vf=Vf.ary) )
               {   
                  return false;
               }  

               if( !UtilLoadSphOrEmmToS3D(&ary,attr,SKL_REAL(0.0),cfg.grpP.gid) )
               {
                  MSG_ERROR(UtilStrFormat("SPH read error. attr=%s",attr));
                  return false;
               }

               for(k=0;k<K;k++){
               for(j=0;j<J;j++){
               for(i=0;i<I;i++){
                  if( int(ceil(m_vf->val(i,j,k,m_mat->id)) * ceil(ary.val(i,j,k))) > 0 )
                     {
                         int idx[3]={i+cfg.Pdiv.head[0],j+cfg.Pdiv.head[1],k+cfg.Pdiv.head[2]};
                         
                         IappIndex* indexList = new IappIndex(idx,attr);
                         iapptbl->indexList.push_back(indexList);

                         DBX(MESSAGE(UtilStrFormat("indexList : %s %d %d %d",indexList->attr,
                             indexList->i,indexList->j,indexList->k)));
                         
                     }
               }}}
               //DBX(MESSAGE(UtilStrFormat("SPH read OK. attr=%s grpP.gid=%d sz=%d %d %d",attr,cfg.grpP.gid,I,J,K)));
               const unsigned int* vfsz = m_vf->GetSize();
               //DBX(MESSAGE(UtilStrFormat("vfsz=%d %d %d",vfsz[0],vfsz[1],vfsz[2])));
             }
           }
        }

        return true;
}
//
TypeEnum PotentialSolv::checkPotenType( const char *name)
{

        string n1,m1,m2;

        splitOutFileName(name,n1,m1,m2);

        if( n1 == "POTENTIAL"  ) return POTENTIAL;

        return UNKNOWN;
}
//
const Element* PotentialSolv::getPotenElem( const char* name )
{
#ifdef ENABLE_MEMBRANEPOTENTIAL
        string n1,mat1,mat2;

        splitOutFileName(name,n1,mat1,mat2);

        for( int e = 0; e < cfg.EqnCount; e++ ) {
                if (!cfg.PeqnAss[e]) { continue; }

                if( strcmp(cfg.EqnTbl[e]->mat->name.c_str(),mat1.c_str() )!= 0 ) { continue; }

                // Molecular not MembrainePotential -> continue
                int ml = cfg.EqnTbl[e]->mol->id;
                if( !cfg.MolTbl[ml]->potential ) { continue; }
                int mt = cfg.EqnTbl[e]->mat->id;

                return cfg.EqnTbl[e];
        }
#endif
        return NULL;
}
//
S3D_REAL* PotentialSolv::getPotenArray( const char *name, S3D_REAL *s3d, S4D_REAL* P0, int eqnId)
{

        const unsigned int I = cfg.Pdiv.size[0];
        const unsigned int J = cfg.Pdiv.size[1];
        const unsigned int K = cfg.Pdiv.size[2];

        for (int e = 0; e < h_MemPotTbl.size(); e++) {

                const MembranePotential* Poten = h_MemPotTbl[e];

                const Material* mat1 = Poten->mat1;
                const Material* mat2 = Poten->mat2;
                const Equation* gm = UtilGetEquationId(h_mol,mat1);

                //DBX(MESSAGE(UtilStrFormat("eqnId : %d  gm->id : %d",eqnId,gm->id))); 
                //DBX(MESSAGE(UtilStrFormat("mat1  : %s  mat2   : %s",mat1->name.c_str(),
                //                                                    mat2->name.c_str()))); 

                if( gm->id != eqnId ) { continue; }

                string t1,m1,m2;
                splitOutFileName( name, t1, m1, m2 );

                //DBX(MESSAGE(UtilStrFormat("m1  : %s  m2   : %s",m1.c_str(),
                //                                                m2.c_str()))); 

                if( mat1->name != m1  || mat2->name != m2 ) { continue; }

                //DBX(MESSAGE(UtilStrFormat(" OK!!  eqnId : %d  gm->id : %d e : %d",eqnId,gm->id,e))); 

                for (int i = 0; i < I; i++) {
                for (int j = 0; j < J; j++) {
                for (int k = 0; k < K; k++) {
                    s3d->val(i,j,k) = P0->val(i,j,k,eqnId)*h_coefdvdt->val(i,j,k,e);
                }}}

                return s3d;

       }

       return NULL;

}
//
void PotentialSolv::splitOutFileName( const char *name, string &n1, 
                                      string &mat_name_1,
                                      string &mat_name_2)
{

       const char colon = ':';
       string nm = string(name);
       string::size_type p1 = nm.find_first_of(colon);
       if( p1 == string::npos ) {
         return ;
       }
       n1 = nm.substr(0,p1);

       string::size_type p2 = nm.find_first_of(colon);
       string n2 = nm.substr(p1+1,p2-p1-1);

       p1 = n2.find_first_of(colon);
       mat_name_1 = n2.substr(0,p1);
       
       p2 = n2.find_first_of(colon);
       mat_name_2 = n2.substr(p1+1,p2-p1-1);

       return ;
}
