/*
 * Copyright (c) RIKEN, Japan. All right reserved. 2008-2009
 *
 */
#include "WarpSolv.h"
#include "FortranFuncWarp.h"

/////////////////////////////////////////////////////////////////////////////
// Constructor
//
WarpBase::WarpBase(const CfgInfo& _cfg, int cnt)
: cfg(_cfg), m_counter(cnt)//, m_paramList(_cfg), m_tableList(_cfg)
{
	para_cmp = cfg.solver->SklGetParaComponent();
	para_mng = para_cmp->GetParaManager();
}

/////////////////////////////////////////////////////////////////////////////
// Destractor
//
WarpBase::~WarpBase()
{
}

/////////////////////////////////////////////////////////////////////////////
// Parse XML config
//
bool WarpBase::ParseXML(const CfgElem* elemRoot)
{
	if( !elemRoot ) return false;

  DB1(MESSAGE("WarpBase::ParseXML"));

  m_elemName  = elemRoot->GetName();
  m_elemID    = elemRoot->GetID();

	// search Param
	const CfgParam* param = elemRoot->GetParamFirst();
	while( param )
	{
		// parse
		m_paramList.Add(S_WARP_PARAM(param));

		// next param
		param = elemRoot->GetParamNext(param);
	}

	// search TableList Elem
	const CfgElem* tblElem = elemRoot->GetElemFirst("TableValue");
	while( tblElem )
	{
		// table list
//		WarpTableValue* table = new WarpTableValue(cfg);
		WarpTableValue* table = new WarpTableValue();
		if( !table->ParseXML(tblElem) )
		{
			delete table;
			return false;
		}

		// add to table map
		m_tableList.Add(table);

		// next Elem
		tblElem = elemRoot->GetElemNext(tblElem,"TableValue");
	}

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// print Log
//
bool WarpBase::printLog(stringstream& wout, int ts, int nid)
{
	//title
	WARP_SET_TS(wout,ts);
	wout << setw(20) << left << string(m_elemName);
	if(m_counter > 0) wout << setw(3) << ":" << m_counter;
	wout << "(id=" << setw(10) << left << m_elemID << ")" << endl;

	ts += WARP_TS;

	// param
	for( int i=0;i<m_paramList.Count();i++ )
	{
		m_paramList[i].printLog(wout,ts,nid);
	}

	// table
	m_tableList.printLog(wout,ts,nid);

	return true;
}

// for param
int WarpBase::GetIntParam(const char* name, int def)
{
	return m_paramList.GetIntByName(name,def);
}

SKL_REAL WarpBase::GetRealParam(const char* name, SKL_REAL def)
{
	return m_paramList.GetRealByName(name,def);
}

string WarpBase::GetStringParam(const char* name, string def)
{
	return m_paramList.GetStringByName(name,def);
}

string WarpBase::GetStringParam(const char* name, WarpList<string>& choice, string def)
{
	string str = m_paramList.GetStringByName(name,def);
	for( int i=0;i<choice.Count();i++ )
	{
		if( !strcasecmp(choice[i].c_str(),str.c_str()) ) return str;
	}

	return def;
}

// for table
SKL_REAL WarpBase::GetTableValue(const char* tableName, SKL_REAL x)
{
	// search table
	WarpTableValue* table = m_tableList.Find(tableName);
	if( !table ) return SKL_REAL(0.0);

	// linear interp
	return table->GetValueByLinear(x);
}

// for check param
E_WARP_VALUE_TYPE WarpBase::CheckValueType()
{
	string vtype = GetStringParam("ValueType");
	if( !strcasecmp( vtype.c_str(), "fixed" ) )
		return E_WARP_FIXED;
	else if( !strcasecmp( vtype.c_str(), "table" ) )
		return E_WARP_TABLE;

	return E_WARP_FIXED;
}

E_WARP_OP WarpBase::CheckOpeType()
{
	E_WARP_OP op = E_WARP_AVE;
	string optype = GetStringParam("OpeType");

	if( !strcasecmp(optype.c_str(),"min") )
		op = E_WARP_MIN;
	else if( !strcasecmp(optype.c_str(),"max") )
		op = E_WARP_MAX;
	else if( !strcasecmp(optype.c_str(),"ave") )
		op = E_WARP_AVE;

	return op;
}

E_WARP_CMP WarpBase::CheckCompType()
{
	E_WARP_CMP cmp = E_WARP_LE;
	string cmptype = GetStringParam("CompType");

	if( !strcasecmp(cmptype.c_str(),"Equal") )
		cmp = E_WARP_EQ;
	else if( !strcasecmp(cmptype.c_str(),"LessThan") )
		cmp = E_WARP_LT;
	else if( !strcasecmp(cmptype.c_str(),"LessEqual") )
		cmp = E_WARP_LE;
	else if( !strcasecmp(cmptype.c_str(),"GreaterThan") )
		cmp = E_WARP_GT;
	else if( !strcasecmp(cmptype.c_str(),"GreaterEqual") )
		cmp = E_WARP_GE;

	return cmp;
}

