//Copyright (C) 2013 The Computational Science Resarch Program (CSRP), RIKEN
//All rights reserved.
//
//This file is part of RIKEN Integrated Cell Simulator for the K-Computer(RICS-K).
//
//RICS-K is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//RICS-K is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with RICS-K.  If not, see <http://www.gnu.org/licenses/>.


#ifndef __WARP_DEFINE_H__
#define __WARP_DEFINE_H__

#include <Skl.h>
#include <SklUtil.h>

#include <iostream>
#include <cmath>
#include <cstring>
#include <map>
#include <cfloat>

#include <Foundation/const.h>
#include <Foundation/util.h>
#include <Foundation/error.h>
#include <Foundation/datatype.h>
#include <Foundation/api.h>
#include <Foundation/ecellif/ecellif.h>
#include <Foundation/ecellif/emcsim.h>
#include <Foundation/ecellif/EmlModel.h>

#ifndef _WITHOUT_ECELL
#include <libecs/libecs.hpp>
#include <libecs/DataPoint.hpp>
#include <libecs/DataPointVector.hpp>
#endif // _WITHOUT_ECELL

#define WARP_TS 2
#define WARP_SET_TS(_WOUT,_TS) { _WOUT << setw(_TS) << setfill(' ') << ""; };

enum E_WARP_TYPE
{
	E_WARP_USER,
	E_WARP_DIRECT,
	E_WARP_MEDIATION,
	E_WARP_CAPILLARY,
};

enum E_WARP_OP
{
	E_WARP_MAX,
	E_WARP_MIN,
	E_WARP_AVE,
};

enum E_WARP_CMP
{
	E_WARP_EQ,  // equal
	E_WARP_LT,  // less than
	E_WARP_LE,  // less equal
	E_WARP_GT,  // greater than
	E_WARP_GE,  // greater equal
};

enum E_WARP_VALUE_TYPE
{
	E_WARP_FIXED,  // fixed value
	E_WARP_TABLE,  // table value
};

enum E_WARP_MOLINFO_TYPE
{
	E_WARP_MOLINFO_UNKNOWN,
	E_WARP_MOLINFO_SOURCE,
	E_WARP_MOLINFO_DISTINATION,
	E_WARP_MOLINFO_SRC_TRIGGER,
	E_WARP_MOLINFO_DST_TRIGGER,
};


enum E_PARAM_TYPE
{
	E_PARAM_UNKNOWN,
	E_PARAM_STRING,
	E_PARAM_INT,
	E_PARAM_REAL,
};

struct S_WARP_TEMPAREA
{
	const Molecular* molecular;
	SKL_REAL mol;

	S_WARP_TEMPAREA()
	{
		molecular = new Molecular;
		mol       = SKL_REAL(0.0);
	}
	S_WARP_TEMPAREA(const Molecular* _molecular, SKL_REAL _mol=SKL_REAL(0.0))
	{
		molecular = _molecular;
		mol       = _mol;
	}

	bool operator==(S_WARP_TEMPAREA value)
	{
		if( molecular != value.molecular ) return false;
		return true;
	}
};

struct S_WARP_PARAM
{
	string       name;
	string       svalue;
	int          ivalue;
	SKL_REAL     rvalue;
	E_PARAM_TYPE type;

	S_WARP_PARAM()
	{
		name   = string("");
		svalue = string("");
		ivalue = 0;
		rvalue = SKL_REAL(0.0);
		type   = E_PARAM_UNKNOWN;
	};

	S_WARP_PARAM( const char* _name, string _value )
	{
		name   = string(_name);
		svalue = _value;
		type   = E_PARAM_STRING;
	};

	S_WARP_PARAM( const char* _name, const char* _value )
	{
		name   = string(_name);
		svalue = string(_value);
		type   = E_PARAM_STRING;
	};

	S_WARP_PARAM( const char* _name, int _value )
	{
		name   = string(_name);
		ivalue = _value;
		type   = E_PARAM_INT;
	};

	S_WARP_PARAM( const char* _name, SKL_REAL _value )
	{
		name   = string(_name);
		rvalue = _value;
		type   = E_PARAM_REAL;
	};

	S_WARP_PARAM( const CfgParam* param )
	{
		name = string(param->GetName());
		if( param->GetDataType() == _STR )
		{
			const char* cval = NULL;
			param->GetData(&cval);
			svalue = string(cval);	
			type   = E_PARAM_STRING;
		}
		else if( param->GetDataType() == _INT )
		{
			param->GetData(&ivalue);
			type = E_PARAM_INT;
		}
		else if( param->GetDataType() == _REAL )
		{
			param->GetData(&rvalue);
			type = E_PARAM_REAL;
		}
	}

	bool operator==(S_WARP_PARAM value)
	{
		if( type != value.type ) return false;
		if( strcasecmp(name.c_str(),value.name.c_str())!=0 ) return false;
		if( type==E_PARAM_INT ){
			if( ivalue != value.ivalue ) return false;
		} else if( type==E_PARAM_REAL ){
			if( rvalue != value.rvalue ) return false;
		} else if( type==E_PARAM_STRING ){
			if( strcasecmp(svalue.c_str(),value.svalue.c_str())!=0 ) return false;
		}
		return true;
	}

	bool operator!=(S_WARP_PARAM value)
	{
		return !(*this==value);
	}

	bool printLog(stringstream& wout, int ts, int nid=-1)
	{
		WARP_SET_TS(wout,ts);
		wout << setw(20) << left << name;
		if( type == E_PARAM_REAL )
			wout << setw(20) << left << UtilStrFormat("%.6e",rvalue) << endl;
		else if( type == E_PARAM_INT )
			wout << setw(20) << left << ivalue << endl;
		else if( type == E_PARAM_STRING )
			wout << setw(20) << left << svalue << endl;

		return true;
	}
};

struct S_REAL_PAIR
{
	SKL_REAL value1;
	SKL_REAL value2;

  S_REAL_PAIR()
  {
		value1 = 0.0;
		value2 = 0.0;
  };

  S_REAL_PAIR( SKL_REAL _value1, SKL_REAL _value2 )
  {
		value1 = _value1;
		value2 = _value2;
  };

	bool operator==(S_REAL_PAIR value)
	{
		if( value1 != value.value1 ) return false;
		if( value2 != value.value2 ) return false;
		return true;
	}

	bool operator!=(S_REAL_PAIR value)
	{
		return !(*this==value);
	}
};

struct S_WARP_INDEX
{
	int         I,J,K;    // local domain index
	int         gI,gJ,gK; // global index
	SKL_REAL    X,Y,Z;
	bool        setCoord;
	const char* sphAttr;

	S_WARP_INDEX()
	{
		I=J=K    = 0;
		gI=gJ=gK = 0;
		X=Y=Z    = SKL_REAL(0.0);
    setCoord = false;
		sphAttr  = NULL;
		
	}
	S_WARP_INDEX( int _I, int _J, int _K )
	{
		I = gI = _I;
		J = gJ = _J;
		K = gK = _K;
		X=Y=Z    = SKL_REAL(0.0);
		setCoord = false;
		sphAttr  = NULL;
	};
	S_WARP_INDEX( int* idx )
	{
		I = gI = idx[0];
		J = gJ = idx[1];
		K = gK = idx[2];
		X=Y=Z    = SKL_REAL(0.0);
		setCoord = false;
		sphAttr  = NULL;
	};
	S_WARP_INDEX( int* idx, SKL_REAL* coord )
	{
		I = gI = idx[0];
		J = gJ = idx[1];
		K = gK = idx[2];
		X = coord[0];
		Y = coord[1];
		Z = coord[2];
		setCoord = true;
		sphAttr  = NULL;
	};
	S_WARP_INDEX( int* idx, const char* _sphAttr )
	{
		I = gI = idx[0];
		J = gJ = idx[1];
		K = gK = idx[2];
		X=Y=Z    = SKL_REAL(0.0);
		setCoord = false;
		sphAttr  = _sphAttr;
	}

	int operator[](int idx)
	{
		if( idx==0 ) return I;
		else if( idx==1 ) return J;
		else if( idx==2 ) return K;
		return 0;
	};

	bool operator==(S_WARP_INDEX idx)
	{
		if( gI != idx.gI ) return false;
		if( gJ != idx.gJ ) return false;
		if( gK != idx.gK ) return false;
		return true;
	}

	bool operator!=(S_WARP_INDEX idx)
	{
		return !(*this==idx);
	}
};

#endif  /*__WARP_DEFINE_H__*/

