/*
 * Copyright (c) RIKEN, Japan. All right reserved. 2008-2009
 *
 */
#include "WarpInfo.h"
#include "FortranFuncWarp.h"

/////////////////////////////////////////////////////////////////////////////
// Calc Direct
//
bool WarpInfo::CalcDirect(WarpTemporaryArea *tempArea)
{
	int i;

	// get elem ID
	int elemID = GetElemID();
	DB1(MESSAGE(UtilStrFormat("WarpInfo::CalcDirect elemID = %d",elemID)));

	// get equation
	string srcEqnName = m_srcMolInfo? m_srcMolInfo->GetEquation()->name: string("NULL");
	string dstEqnName = m_dstMolInfo? m_dstMolInfo->GetEquation()->name: string("NULL");
	DB1(MESSAGE(UtilStrFormat("  Source Equation      = %s", srcEqnName.c_str())));
	DB1(MESSAGE(UtilStrFormat("  Distination Equation = %s", dstEqnName.c_str())));

	///////////////////////////////////////////
	// for warp trigger
	bool checkTrigger = true;
	if( m_srcTrigger )
	{
		checkTrigger = m_srcTrigger->CheckTrigger();
	}
	
	///////////////////////////////////////////
	// for warp source
	if( checkTrigger )
	{
		if( !CalcDirect_Source(tempArea) )
		{
			return false;
		}
	}
	
	///////////////////////////////////////////
	// for warp distination
	if( !CalcDirect_Distination(tempArea) )
	{
		return false;
	}

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// Calc Direct for Warp source
//
bool WarpInfo::CalcDirect_Source(WarpTemporaryArea *tempArea)
{
	if( !m_srcMolInfo ) return false;
	if( !m_dstMolInfo ) return false;

	// time
	SKL_REAL time = RicsGetCurrentTime();
	SKL_REAL dt   = RicsGetDeltaT();

	//type
	E_WARP_MOLINFO_TYPE type = m_srcMolInfo->GetType();
	if( type != E_WARP_MOLINFO_SOURCE ) return false;

	// warp value type
	E_WARP_VALUE_TYPE vtype = m_srcMolInfo->CheckValueType();

	// get warp density
	SKL_REAL warpDensity = SKL_REAL(0.0);
	if( vtype==E_WARP_FIXED )
	{
		warpDensity = m_srcMolInfo->GetRealParam("warpDensity"); // fixed value
		DB1(MESSAGE(UtilStrFormat("    CalcDirect_Source fixed warpDensity=%.6e time=%.6e",warpDensity,time)));
	}
	else if( vtype==E_WARP_TABLE )
	{
		warpDensity = m_srcMolInfo->GetTableValue("warpDensity", time ); // table value
		DB1(MESSAGE(UtilStrFormat("    CalcDirect_Source table warpDensity=%.6e time=%.6e",warpDensity,time)));
	}


	// value define(Density or DensityRatio, default=Density)
	string def = m_srcMolInfo->GetStringParam("ValueDefinition");

	SKL_REAL changed_mol = SKL_REAL(0.0);
	if( !strcasecmp(def.c_str(),"DensityRatio") )
	{
		// for warp density ratio

		// for target operation
		E_WARP_OP op = m_srcMolInfo->CheckOpeType();

		// for Density Ratio
		changed_mol = m_srcMolInfo->CalcSourceTermByDensityRatio(warpDensity,op);
	}
	else
	{
		// for warp density
		changed_mol = m_srcMolInfo->CalcSourceTermByDensity(warpDensity);
	}

	// for plus value
	if( changed_mol<SKL_REAL(0.0) ) changed_mol *= SKL_REAL(-1.0);

	// for temporary area
	const Equation* dstEqn = m_dstMolInfo->GetEquation();
	tempArea->AddMol(dstEqn->mol->name.c_str(),changed_mol);
	DB1(MESSAGE(UtilStrFormat("    CalcDirect_Source mol=%.6e",changed_mol)));

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// Calc Direct for Warp distination
//
bool WarpInfo::CalcDirect_Distination(WarpTemporaryArea *tempArea)
{
	if( !m_dstMolInfo ) return false;

	// time
	SKL_REAL time = RicsGetCurrentTime();
	SKL_REAL dt   = RicsGetDeltaT();

	//type
	E_WARP_MOLINFO_TYPE type = m_dstMolInfo->GetType();
	if( type != E_WARP_MOLINFO_DISTINATION ) return false;

	// equation
	const Equation* eqn = m_dstMolInfo->GetEquation();

	// warp value type
	E_WARP_VALUE_TYPE vtype = m_dstMolInfo->CheckValueType();

	// get warp density
	if( vtype==E_WARP_FIXED )
	{
		// change mol by source warp
		SKL_REAL mol = tempArea->CheckMol(eqn->mol->name.c_str()); // fixed value
		mol = tempArea->GetMol(eqn->mol->name.c_str(),mol); // get all mol
		DB1(MESSAGE(UtilStrFormat("    CalcDirect_Distination fixed mol=%.6e",mol)));

		// change mol per unit time
		SKL_REAL cmol_t = mol / dt;

		// calc source term
		SKL_REAL changed_density = m_dstMolInfo->CalcSourceTermByMol(cmol_t);
	}
	else if( vtype==E_WARP_TABLE )
	{
		// for table value
		SKL_REAL warpDensity = m_dstMolInfo->GetTableValue("warpDensity", time );
		DB1(MESSAGE(UtilStrFormat("    CalcDirect_Distination table warpDensity=%.6e",warpDensity)));

		// calc source term
		SKL_REAL changed_mol = m_dstMolInfo->CalcSourceTermByDensity(warpDensity);

		// for temp area
		SKL_REAL mol = tempArea->GetMol(eqn->mol->name.c_str(),changed_mol);
	}

	return true;
}

