/*
 * Copyright (c) RIKEN, Japan. All right reserved. 2008-2009
 *
 */
#include "WarpInfo.h"
#include "FortranFuncWarp.h"

/////////////////////////////////////////////////////////////////////////////
// Calc Mediation
//
bool WarpInfo::CalcMediation(WarpTemporaryArea *tempArea)
{
	int i;

	// get elem ID
	int elemID = GetElemID();
	DB1(MESSAGE(UtilStrFormat("WarpInfo::CalcMediation elemID = %d",elemID)));

	// get equation
	string srcEqnName = m_srcMolInfo? m_srcMolInfo->GetEquation()->name: string("NULL");
	string dstEqnName = m_dstMolInfo? m_dstMolInfo->GetEquation()->name: string("NULL");
	DB1(MESSAGE(UtilStrFormat("  Source Equation      = %s", srcEqnName.c_str())));
	DB1(MESSAGE(UtilStrFormat("  Distination Equation = %s", dstEqnName.c_str())));

	///////////////////////////////////////////
	// for warp distination  trigger
	bool checkDstTrigger = true;
	if( m_dstTrigger )
	{
		checkDstTrigger = m_dstTrigger->CheckTrigger(tempArea);
	}

	///////////////////////////////////////////
	// for warp distination
	if( checkDstTrigger )
	{
		if( !CalcMediation_Distination(tempArea) )
		{
			return false;
		}
	}

	///////////////////////////////////////////
	// for warp source trigger
	bool checkSrcTrigger = true;
	if( m_srcTrigger )
	{
		checkSrcTrigger = m_srcTrigger->CheckTrigger(tempArea);
	}

	///////////////////////////////////////////
	// for warp source
	if( checkSrcTrigger )
	{
		if( !CalcMediation_Source(tempArea) )
		{
			return false;
		}
	}

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// Calc Mediation for Warp source
//
bool WarpInfo::CalcMediation_Source(WarpTemporaryArea *tempArea)
{
	if( !m_srcMolInfo ) return true;

	// time
	SKL_REAL time = RicsGetCurrentTime();
	SKL_REAL dt   = RicsGetDeltaT();

	//type
	E_WARP_MOLINFO_TYPE type = m_srcMolInfo->GetType();
	if( type != E_WARP_MOLINFO_SOURCE ) return false;

	// warp value type
	E_WARP_VALUE_TYPE vtype = m_srcMolInfo->CheckValueType();

	// get warp density
	SKL_REAL warpDensity = SKL_REAL(0.0);
	if( vtype==E_WARP_FIXED )
	{
		warpDensity = m_srcMolInfo->GetRealParam("warpDensity"); // fixed value
DB1(MESSAGE(UtilStrFormat("    CalcMediation_Source fixed warpDensity=%.6e time=%.6e",warpDensity,time)));
	}
	else if( vtype==E_WARP_TABLE )
	{
		warpDensity = m_srcMolInfo->GetTableValue("warpDensity", time ); // table value
DB1(MESSAGE(UtilStrFormat("    CalcMediation_Source table warpDensity=%.6e time=%.6e",warpDensity,time)));
	}


	// value define(Density or DensityRatio, default=Density)
	WarpList<string> choice;
	choice.Add("Density");
	choice.Add("DensityRatio");
	string def = m_srcMolInfo->GetStringParam("ValueDefinition",choice,"Density");
	DB1(MESSAGE(UtilStrFormat("    ValueDefinition=%s",def.c_str())));

	SKL_REAL changed_mol = SKL_REAL(0.0);
	if( !strcasecmp(def.c_str(),"DensityRatio") )
	{
		// warp by density ratio

		// for target operation
		E_WARP_OP op = m_srcMolInfo->CheckOpeType();

		// for Density Ratio
		changed_mol = m_srcMolInfo->CalcSourceTermByDensityRatio(warpDensity,op);
	}
	else
	{
		// for warp density
		changed_mol = m_srcMolInfo->CalcSourceTermByDensity(warpDensity);
	}

	// for plus value
	if( changed_mol<SKL_REAL(0.0) ) changed_mol *= SKL_REAL(-1.0);

	// for temporary area
	const Equation* srcEqn = m_srcMolInfo->GetEquation();
	const Equation* dstEqn = NULL;
	const char* tempMol = NULL;
	if( !m_dstMolInfo ){
		tempMol = srcEqn->mol->name.c_str();
	}
	else{
		dstEqn = m_dstMolInfo->GetEquation();
		string timing = GetStringParam("MolChangeTiming",string("metabolism"));
		if( !strcasecmp(timing.c_str(),"after" ) || !strcasecmp(timing.c_str(),"metabolism") )
			tempMol = srcEqn->mol->name.c_str();
		else if( !strcasecmp(timing.c_str(),"before") || !strcasecmp(timing.c_str(),"both") )
			tempMol = dstEqn->mol->name.c_str();
		else
			tempMol = srcEqn->mol->name.c_str();
	}

	tempArea->AddMol(tempMol,changed_mol);
DB1(MESSAGE(UtilStrFormat("    CalcMediation_Source changed mol %s %.6e",tempMol,changed_mol)));

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// Calc Mediation for Warp distination
//
bool WarpInfo::CalcMediation_Distination(WarpTemporaryArea *tempArea)
{
	if( !m_dstMolInfo ) return true;

	// time
	SKL_REAL time = RicsGetCurrentTime();
	SKL_REAL dt   = RicsGetDeltaT();

	//type
	E_WARP_MOLINFO_TYPE type = m_dstMolInfo->GetType();
	if( type != E_WARP_MOLINFO_DISTINATION ) return false;

	// equation
	const Equation* eqn = m_dstMolInfo->GetEquation();

	// warp value type
	E_WARP_VALUE_TYPE vtype = m_dstMolInfo->CheckValueType();

	// get warp density
	SKL_REAL warpDensity = SKL_REAL(0.0);
	if( vtype==E_WARP_FIXED )
	{
		warpDensity = m_dstMolInfo->GetRealParam("warpDensity"); // fixed value
	}
	else if( vtype==E_WARP_TABLE )
	{
		warpDensity = m_dstMolInfo->GetTableValue("warpDensity", time ); // table value
	}

	// for temporary area
	const Equation* srcEqn = NULL;
	const Equation* dstEqn = m_dstMolInfo->GetEquation();
	const char* tempMol = NULL;
	if( !m_srcMolInfo ){
		tempMol = dstEqn->mol->name.c_str();
	}
	else{
		srcEqn = m_srcMolInfo->GetEquation();
		string timing = GetStringParam("MolChangeTiming",string("metabolism"));
		if( !strcasecmp(timing.c_str(),"after" ) || !strcasecmp(timing.c_str(),"both") )
			tempMol = srcEqn->mol->name.c_str();
		else if( !strcasecmp(timing.c_str(),"before") || !strcasecmp(timing.c_str(),"metabolism") )
			tempMol = dstEqn->mol->name.c_str();
		else
			tempMol = dstEqn->mol->name.c_str();
	}

	// density to mol
	SKL_REAL vol = m_dstMolInfo->GetAreaVolume();
	//SKL_REAL warpMol = warpDensity*vol;
	SKL_REAL warpMol = warpDensity*vol*dt;
	warpMol = tempArea->GetMol(tempMol,warpMol); // get mol : dt
	if( vol<=SKL_REAL(0.0) )
		warpDensity = SKL_REAL(0.0);
	else
		warpDensity = warpMol/vol;

	if( vtype==E_WARP_FIXED )
DB1(MESSAGE(UtilStrFormat("    CalcMediation_Distination fixed warpDensity=%.6e time=%.6e",warpDensity,time)));
	else if( vtype==E_WARP_TABLE )
DB1(MESSAGE(UtilStrFormat("    CalcMediation_Distination table warpDensity=%.6e time=%.6e",warpDensity,time)));

	// calc source term
	//  change mol per unit time
	warpMol = warpMol / dt;
	SKL_REAL changed_density = m_dstMolInfo->CalcSourceTermByMol(warpMol);

	return true;
}

