/*
 * Copyright (c) RIKEN, Japan. All right reserved. 2008-2009
 *
 */

#ifndef __WARP_LIST_H__
#define __WARP_LIST_H__

#include "WarpDefine.h"

template<class T>
class WarpList
{
protected:

// member variables
	const CfgInfo& cfg;
	vector<T> m_data;

// member functions

public:

// member functions
	WarpList():cfg(*UtilGetCfg()){};
	virtual ~WarpList(){};

	T& operator[](int idx)
	{
		return m_data[idx];
	}

	bool Add(T d)
	{
		m_data.push_back(d);
		return true;
	}

	bool AddUnique(T d)
	{
		bool exist = false;
		for( int i=0;i<m_data.size();i++ )
		{
			if( m_data[i] == d )
			{
				exist = true;
				break;
			}
		}
		if( !exist )
		{
			m_data.push_back(d);
			return true;
		}
		return false;
	}

	size_t Count()
	{
		return m_data.size();
	}

	void Clear()
	{
		m_data.clear();
	}

	void Copy(WarpList& list)
	{
		Clear();
		for( int i=0;i<list.Count();i++ )
		{
			T d = list.m_data[i];
			Add(d);
		}
	}

};

class WarpIndexList : public WarpList<S_WARP_INDEX>
{
protected:
	int *m_index;
	vector<const char*> m_posSph;
	const char* m_memfilter;

public:
	WarpIndexList() : WarpList<S_WARP_INDEX>()
	{
		m_index = NULL;
		m_memfilter = NULL;
	}
	virtual ~WarpIndexList()
	{
		if( m_index )
			delete [] m_index;
		m_index = NULL;
	};

	void Initialize(WarpIndexList& list)
	{
		for( int i=0;i<list.GetSphCount();i++ )
		{
			m_posSph.push_back(list.GetSphAttr(i));
		}
		m_memfilter = list.GetMemFilter();
	}

	int* GetListPtr()
	{
		if( m_index ) return m_index;
		m_index = new int[Count()*3];
		for( int i=0;i<Count();i++ )
		{
			m_index[i*3+0] = m_data[i][0];
			m_index[i*3+1] = m_data[i][1];
			m_index[i*3+2] = m_data[i][2];
		}
		return m_index;
	}

	bool ParseXML(const CfgElem* elemRoot)
	{
		if( !elemRoot ) return false;

		//membrane filter
		const CfgParam *membParam = elemRoot->GetParamFirst("membraneFilter");
		if( membParam )
		{
			membParam->GetData(&m_memfilter);
		}

		//search index Elem
		const CfgElem *idxElem = elemRoot->GetElemFirst("index");
		while( idxElem )
		{
			// get I,J,K value
			int index[3];
			if( !SklUtil::getVecParams(idxElem,"I","J","K",index) )
			{
				MSG_ERROR(UtilStrFormat("specified value in index Elem is invalid"));
				return false;
			}
			AddUnique(S_WARP_INDEX(index));

			// next elem
			idxElem = elemRoot->GetElemNext(idxElem,"index");
		}

		// search coord Elem
		const CfgElem* codElem = elemRoot->GetElemFirst("coord");
		while( codElem )
		{
			// get I,J,K value
			SKL_REAL coord[3];
			if( !SklUtil::getVecParams(codElem,"X","Y","Z",coord) )
			{
				MSG_ERROR(UtilStrFormat("specified value in coord Elem is invalid"));
				return false;
			}
			AddUnique(CoordToIndex(coord));

			// search next elem
			codElem = elemRoot->GetElemNext(codElem,"coord");
		}

		// search SPH Elem
		const CfgElem* sphElem = elemRoot->GetElemFirst("sph");
		while( sphElem )
		{
			const CfgParam* attrParam = sphElem->GetParamFirst("attr");
			while(attrParam)
			{
				const char* attr = NULL;
				if( attrParam->GetData(&attr) )
					m_posSph.push_back(attr);

				attrParam = sphElem->GetParamNext(attrParam,"attr");
			}

			// search next elem
			sphElem = elemRoot->GetElemNext(sphElem,"sph");
		}

		return true;
	}

	S_WARP_INDEX CoordToIndex( SKL_REAL *coord )
	{
		int index[3];
		for( int i=0;i<3;i++ )
		{
			// dist from origin
			SKL_REAL cod_org = coord[i] - cfg.origin[i];
			// to index
			index[i] = int(cod_org/cfg.pitch[i]);
		}
		return S_WARP_INDEX(index,coord);
	}

	const char* GetMemFilter()
	{
		return m_memfilter;
	}

	const Material* GetMemFilterMaterial(int idx)
	{
		if( !m_memfilter ) return NULL;
		if( idx!=1 && idx!=2 ) return NULL;
		string filter = string(m_memfilter);

		const char sep = ':';

		// first token
		string::size_type p1 = filter.find_first_of(sep);
		if( p1 == string::npos )
		{
			if( idx==1 )
				return UtilGetMaterialId(filter);
			else
				return NULL;
		}
		string m1 = filter.substr(0,p1);

		// second token
		string::size_type p2 = filter.find_first_of(sep, p1 + 1);
		string m2 = filter.substr(p1 + 1, p2 - p1 -1);

		if( idx == 1 ) return UtilGetMaterialId(m1);
		return UtilGetMaterialId(m2);
	}

	size_t GetSphCount()
	{
		return m_posSph.size();
	}

	const char* GetSphAttr(int idx)
	{
		return m_posSph[idx];
	}
	
	bool printLog(stringstream& wout, int ts, int nid=-1)
	{
		if( Count()==0 ) return true;

		WARP_SET_TS(wout,ts);
		wout << setw(20) << left << string("IndexList") << Count() << endl;

		ts += WARP_TS;

		WARP_SET_TS(wout,ts);
		wout << setw(20) << left << string("membrane filter");
		if( m_memfilter )
			wout << string(m_memfilter) << endl;
		else
			wout << string("none") << endl;

		for(int i=0;i<Count();i++)
		{
			S_WARP_INDEX idx = m_data[i];
			WARP_SET_TS(wout,ts);
			wout << setw(8) << left << idx.gI;
			wout << setw(8) << left << idx.gJ;
			wout << setw(8) << left << idx.gK;
			if( idx.setCoord )
			{
				wout << setw(20) << left << UtilStrFormat(": coord = %.6e %6e %6e", idx.X,idx.Y,idx.Z);
			}
			else if( idx.sphAttr )
			{
				wout << setw(20) << left << UtilStrFormat(": sph = %s", idx.sphAttr);
			}
			wout << endl;
		}

		return true;
	}
};

class WarpTableValue : public WarpList<S_REAL_PAIR>
{
protected:
	const char* name;
	SKL_REAL    delay;
public:
	WarpTableValue() : WarpList<S_REAL_PAIR>()
	{
		name  = NULL;
		delay = SKL_REAL(0.0);
	};
	virtual ~WarpTableValue(){};

	const char* GetName(){ return name; };

	SKL_REAL GetValueByLinear(SKL_REAL xx)
	{
		SKL_REAL x = xx - delay;

		if( Count() == 0 ) return SKL_REAL(0.0);
		if( x <= m_data[0].value1 ) return m_data[0].value2;
		if( x >= m_data[Count()-1].value1 ) return m_data[Count()-1].value2;

		for( int i=1;i<Count();i++ )
		{
			SKL_REAL x1 = m_data[i-1].value1;
			SKL_REAL x2 = m_data[i  ].value1;
			if( x < x2 )
			{
				SKL_REAL y1 = m_data[i-1].value2;
				SKL_REAL y2 = m_data[i  ].value2;
				if( x1==x2 ) return y2;
				if( y1==y2 ) return y2;
				SKL_REAL a = (y1-y2)/(x1-x2);
				SKL_REAL b = y1 - a * x1;
				return a * x + b;
			}
		}
		return SKL_REAL(0.0);
	}

	bool ParseXML(const CfgElem* tblElem)
	{
		const CfgParam* nameParam = tblElem->GetParamFirst("name");
		if( !nameParam )
		{
			return false;
		}
		if( !nameParam->GetData(&name) )
		{
			return false;
		}

		const CfgParam* delayParam = tblElem->GetParamFirst("delay");
		if( delayParam )
		{
			if( !delayParam->GetData(&delay) ) delay = SKL_REAL(0.0);
		}

		const CfgParam* fnmParam = tblElem->GetParamFirst("fname");
		const char* fname = NULL;
		if( !fnmParam->GetData(&fname) )
		{
			return false;
		}

		ifstream ifs;
		const char sep = ',';
		bool ret = true;

		try 
		{
			ifs.open(fname,ios::in);
			if( ifs.is_open() )
			{
				string line;
				while (getline(ifs, line))
				{
					string::size_type p1 = line.find_first_of(sep);
					if( p1 == string::npos ) continue;
					string::size_type p2 = line.find_first_of(sep, p1 + 1);
					const string f = line.substr(0,p1);
					const string s = line.substr(p1 + 1, p2 - p1 -1);
					const SKL_REAL v1 = static_cast<SKL_REAL>(atof(f.c_str()));
					const SKL_REAL v2 = static_cast<SKL_REAL>(atof(s.c_str()));
					AddUnique(S_REAL_PAIR(v1,v2));
				}
				DB1(MESSAGE(UtilStrFormat("read table value count=%d",int(Count()))));
			}
			else
			{
				ret = false;
			}
		}
		catch(...)
		{
			ret = false;
    }
		ifs.close();

		return ret;
	}

	bool printLog(stringstream& wout, int ts, int nid=-1)
	{
		if( Count()==0 ) return true;

		WARP_SET_TS(wout,ts);
		wout << setw(20) << left << string("TableValue");
		wout << setw(40) << left << string(name) << endl;

		ts += WARP_TS;
		WARP_SET_TS(wout,ts);
		wout << setw(20) << left << string("delay") << UtilStrFormat("%.6e",delay) << endl;
		for(int i=0;i<Count();i++)
		{
			WARP_SET_TS(wout,ts);
			wout << setw(20) << left << UtilStrFormat("%.6e",m_data[i].value1);
			wout << setw(20) << left << UtilStrFormat("%.6e",m_data[i].value2);
			wout << endl;
		}


		return true;
	}
};

class WarpTableList : public WarpList<WarpTableValue*>
{
public:
	WarpTableList() : WarpList<WarpTableValue*>() {};
	virtual ~WarpTableList()
	{
		for(int i=0;i<Count();i++)
		{
			delete m_data[i];
		}
		m_data.clear();
	};

	bool printLog(stringstream& wout, int ts, int nid=-1)
	{
		for(int i=0;i<Count();i++ )
		{
			if( !m_data[i]->printLog(wout,ts,nid) )
			{
				return false;
			}
		}
		return true;
	};

	WarpTableValue* Find(const char* _name)
	{
		for(int i=0;i<Count();i++)
		{
			if( !strcasecmp(_name, m_data[i]->GetName()) )
				return m_data[i];
		}
		return NULL;
	}
};

class WarpParamList : public WarpList<S_WARP_PARAM>
{
public:
	WarpParamList() : WarpList<S_WARP_PARAM>(){};
	virtual ~WarpParamList(){};

	int GetIntByName( const char* name, int def=0 )
	{
		for( int i=0;i<Count();i++ )
		{
			if( m_data[i].type != E_PARAM_INT ) continue;
			if( !strcasecmp( m_data[i].name.c_str(), name ) )
			{
				return m_data[i].ivalue;
			}
		}
		return def;
	}

	SKL_REAL GetRealByName( const char* name, SKL_REAL def=SKL_REAL(0.0) )
	{
		for( int i=0;i<Count();i++ )
		{
			if( m_data[i].type != E_PARAM_REAL ) continue;
			if( !strcasecmp( m_data[i].name.c_str(), name ) )
			{
				return m_data[i].rvalue;
			}
		}
		return def;
	}

	string GetStringByName( const char* name, string def=string("") )
	{
		for( int i=0;i<Count();i++ )
		{
			if( m_data[i].type != E_PARAM_STRING ) continue;
			if( !strcasecmp( m_data[i].name.c_str(), name ) )
			{
				return m_data[i].svalue;
			}
		}
		return def;
	}
};

// temp area class
//
class WarpTemporaryArea : public WarpList<S_WARP_TEMPAREA>
{
protected:
	SKL_REAL *sendbuf;
	SKL_REAL *recvbuf;

#ifndef _WITHOUT_ECELL
	emcsim*     sim;
#endif
	string      emlfile;	// name of path to EML file
	set<string> system;
	SKL_REAL    volume_unit;
	SKL_REAL    time_unit;
	SKL_REAL    volume;	// temp area volume
public:
	WarpTemporaryArea() : WarpList<S_WARP_TEMPAREA>() 
	{
		sendbuf = NULL;
		recvbuf = NULL;

#ifndef _WITHOUT_ECELL
		sim = NULL;
#endif
		emlfile = "";
		volume = 1.0;
	};
	virtual ~WarpTemporaryArea()
	{
		delete [] sendbuf;
		delete [] recvbuf;
	};

	bool ParseXML(const CfgElem* elemRoot)
	{
		if( !elemRoot ) return false;

		DB1(MESSAGE("WarpTemporaryArea::ParseXML"));

		SKL_REAL vol, vu, tu;
		const char* eml = NULL;

		// Temporary area volume
		const CfgParam* TAP = elemRoot->GetParamFirst("TempAreaVolume");
		volume = (TAP && TAP->GetData(&vol) && (vol > 0))? vol: (SKL_REAL)(1.0);

		// name of path to EML file
		const CfgParam* MP = elemRoot->GetParamFirst("metabolism");
		emlfile = (MP && MP->GetData(&eml) && (eml != NULL))? string(eml): string("");

		// volume unit
		const CfgParam* VUP = elemRoot->GetParamFirst("volume_unit");
		volume_unit = (VUP && VUP->GetData(&vu) && (vu > 0))? vu: (SKL_REAL)(1e-3);

		// time unit
		const CfgParam* TUP = elemRoot->GetParamFirst("time_unit");
		time_unit = (TUP && TUP->GetData(&tu) && (tu > 0))? tu: (SKL_REAL)(1e0);

		// system
		const CfgParam* sysParam = elemRoot->GetParamFirst("system");
		if( !sysParam ) { system.insert("/"); }
		for( ;sysParam; sysParam = elemRoot->GetParamNext(sysParam, "system")){
			const char* sysname;
			sysParam->GetData(&sysname);
			system.insert(sysname);
		}

		return true;
	}

	bool Initialize(const CfgElem* elemRoot, E_WARP_TYPE type)
	{
		DB1(MESSAGE("WarpTemporaryArea::Initialize"));

		// Molecular
		for( int i=0;i<cfg.EqnCount;i++ )
		{
			const Molecular* mol= cfg.EqnTbl[i]->mol;
			S_WARP_TEMPAREA molTemp(mol);
			AddUnique(molTemp);
		}

		// Parse XML
		if( !ParseXML(elemRoot) ){
			if((type == E_WARP_DIRECT) || (type == E_WARP_MEDIATION)){
				MSG_ERROR(UtilStrFormat("configuration XML parse error in TempArea Elem"));
				return false;
			}
			else{
				return true;
			}
		}

#ifndef _WITHOUT_ECELL
		if((emlfile.length() != 0) && (type == E_WARP_MEDIATION)){
			// emcsim::loadModel debug message flag
			bool loadModelMsg = false;
			DB1(loadModelMsg = true);

			// calc E-Cell set volume
			const SKL_REAL Ls3 = cfg.Ls * cfg.Ls * cfg.Ls;
			const SKL_REAL VolLs3 = volume * Ls3;
			const SKL_REAL VolLs3pVe = VolLs3 / volume_unit;

			// E-Cell simulator instance
			sim = new emcsim();

			// load eml
			EmlModel eml;
			eml.open(emlfile.c_str());
			sim->loadModel(eml, system, loadModelMsg);
			// sim->testModel(eml);

			// setSize
			sim->setSize(VolLs3pVe);
		}
#endif

		// sendbuf = new SKL_REAL[cfg.EqnCount];
		// recvbuf = new SKL_REAL[cfg.EqnCount];

		// add mol to temp area : debug
		// AddMol("GLC", SKL_REAL(0.0166));
		// AddMol("Pi", SKL_REAL(16.6));
		// AddMol("Hx", SKL_REAL(0.0));

		//

		return true;
	}

	bool RunECell()
	{
#ifndef _WITHOUT_ECELL
		if(!sim){
			return false;
		}

		DB1(MESSAGE("WarpTemporaryArea::RunECell"));

		// Calc KnaNAVolLs3
		// const SKL_REAL Vol = cfg.cvol;
		const SKL_REAL Ls3 = cfg.Ls * cfg.Ls * cfg.Ls;
		const SKL_REAL KnaNA = cfg.Kna * cfg.NA;
		const SKL_REAL KnaNAVolLs3 = KnaNA * volume * Ls3;
		const SKL_REAL iKnaNAVolLs3 = SKL_REAL(1.0) / KnaNAVolLs3;

		// MESSAGE(UtilStrFormat("Debug message : Run E-Cell !!"),0,false);

		DB1(MESSAGE(UtilStrFormat("  Set EntityProperty")));
		// Set Entity Property
		for( int i=0; i<Count(); i++ ){
			const string fullpn = GetFullPn(i);

			// check fullpn
			const bool existpn = sim? sim->isExistEntityProperty(fullpn): false;
			if(!existpn) { continue; }

			const SKL_REAL mass = m_data[i].mol * KnaNAVolLs3;

			// check mass
			TRY(REAL_VALID(mass), m_data[i].molecular->name + "has over flow");

#ifdef ECELL_3_1
			sim->setEntityProperty(fullpn, (mass / volume_unit));
#else
			sim->setEntityProperty(fullpn, libecs::Polymorph(static_cast<libecs::Real>(mass / volume_unit)));
#endif //ECELL_3_1

			// print log
			DB1(MESSAGE(UtilStrFormat("    %s, %.6e",m_data[i].molecular->name.c_str(), m_data[i].mol)));
		}

		// Run E-Cell and Stop E-Cell
		const SKL_REAL dt = cfg.dt * cfg.Ts / time_unit;
		if(sim && sim->hasModel()){
			sim->run(dt);
			sim->stop();
		}

		DB1(MESSAGE(UtilStrFormat("  Get EntityProperty")));
		// Get Entity Property
		for( int i=0; i<Count(); i++ ){
			const string fullpn = GetFullPn(i);

			// check fullpn
			const bool existpn = sim? sim->isExistEntityProperty(fullpn): false;
			if(!existpn) { continue; }

			const SKL_REAL mass = sim->getEntityProperty(fullpn).asReal() * volume_unit;

			m_data[i].mol = (mass * iKnaNAVolLs3);

			// print log
			DB1(MESSAGE(UtilStrFormat("    %s, %.6e",m_data[i].molecular->name.c_str(), m_data[i].mol)));
		}
#endif
		return true;
	}

	bool Allreduce(SklOpType ope=SKL_SUM)
	{
		int i;

		// set send buffer
		for( i=0;i<cfg.EqnCount;i++ )
		{
			if( !cfg.PeqnAss[i] )
			{
				sendbuf[i] = recvbuf[i] = SKL_REAL(0.0);
			}
			else
			{
				sendbuf[i] = recvbuf[i] = m_data[i].mol;
			}
		}

		// allreduce
		SklParaComponent *para_cmp = cfg.solver->SklGetParaComponent();
		if( !para_cmp->Allreduce(sendbuf,recvbuf,(unsigned int)cfg.EqnCount,SKL_ARRAY_DTYPE_REAL,ope,0) )
		{
			return false;
		}

		// set value from recive buffer
		for( i=0;i<cfg.EqnCount;i++ )
		{
			m_data[i].mol = recvbuf[i];
		}

		return true;	
	}

	bool printLog(stringstream& wout, int ts, int nid)
	{
		WARP_SET_TS(wout,ts);
		wout << setw(20) << left << "TemporaryArea" << endl;

		ts += WARP_TS;

		WARP_SET_TS(wout,ts);
		wout << setw(20) << left << "TemporaryAreaVolume";
		wout << setw(20) << left << UtilStrFormat("%.6e",volume) << endl;
		WARP_SET_TS(wout,ts);
		wout << setw(20) << left << "metabolism";
		wout << setw(20) << left << UtilStrFormat("%s",emlfile.c_str()) << endl;
		WARP_SET_TS(wout,ts);
		wout << setw(20) << left << "volume_unit";
		wout << setw(20) << left << UtilStrFormat("%.6e",volume_unit) << endl;
		WARP_SET_TS(wout,ts);
		wout << setw(20) << left << "time_unit";
		wout << setw(20) << left << UtilStrFormat("%.6e",time_unit) << endl;

		set<string>::iterator i = system.begin();
		while( i != system.end() ){
			WARP_SET_TS(wout,ts);
			wout << setw(20) << left << "system";
			wout << setw(20) << left << *i << endl;
			i++;
		}

		return true;
	}

	int GetParamIDByName(const char* name)
	{
		for( int i=0;i<Count();i++ )
		{
			// if( !strcasecmp( m_data[i].name.c_str(), name ) )
			if( !strcasecmp( m_data[i].molecular->name.c_str(), name ) )
			{
				return i;
			}
		}
		return -1;
	}

	void SetVolume(SKL_REAL vol)
	{
		volume = vol;
	}
/*
		void SetVolume(const char* name, SKL_REAL vol)
		{
			int id = GetParamIDByName(name);
			if( id<0 ) return;
			SetVolume(id,vol);
		}
		void SetVolume(int id, SKL_REAL vol)
		{
			m_data[id].volume = vol;
		}
*/
	// get fullid by molecular name
	string GetFullID(const char* name){
		int id = GetParamIDByName(name);
		if( id<0 ) return "";
		return GetFullID(id);
	}
	string GetFullID(int id){
		return (m_data[id].molecular->fullid);
	}

	// get fullpn
	string GetFullPn(const char* name){
		return (GetFullID(name) + ":Value");
	}
	string GetFullPn(int id){
		return (GetFullID(id) + ":Value");
	}

	// check density by name
	//   not changed value
	//   return value = density
	SKL_REAL CheckDensity(const char* name)
	{
		int id = GetParamIDByName(name);
		if( id<0 ) return SKL_REAL(0.0);
		return CheckDensity(id);
	}
	SKL_REAL CheckDensity(int id)
	{
		if( volume <= SKL_REAL(0.0) ) return SKL_REAL(0.0);
		return m_data[id].mol/volume;
	}
		// SKL_REAL CheckDensity(int id)
		// {
		//	if( m_data[id].volume <= SKL_REAL(0.0) ) return SKL_REAL(0.0);
		//	return m_data[id].mol/m_data[id].volume;
		// }

	// add density by name
	//   changed value
	//   return value = density after add 
	SKL_REAL AddDensity(const char* name, SKL_REAL density)
	{
		int id = GetParamIDByName(name);
		if( id<0 ) return SKL_REAL(0.0);
		return AddDensity(id,density);
	}
	SKL_REAL AddDensity(int id, SKL_REAL density)
	{
		if( volume <= SKL_REAL(0.0) ) return SKL_REAL(0.0);
		m_data[id].mol += (density*volume);
		return m_data[id].mol/volume;
	}
		// SKL_REAL AddDensity(int id, SKL_REAL density)
		// {
		//	if( m_data[id].volume <= SKL_REAL(0.0) ) return SKL_REAL(0.0);
		//	m_data[id].mol += (density*m_data[id].volume);
		//	return m_data[id].mol/m_data[id].volume;
		// }

	// get density by name
	//   changed value
	//   return value = gotten density
	SKL_REAL GetDensity(const char* name, SKL_REAL density)
	{
		int id = GetParamIDByName(name);
		if( id<0 ) return SKL_REAL(0.0);
		return GetDensity(id,density);
	}
	SKL_REAL GetDensity(int id, SKL_REAL density)
	{
		if( volume <= SKL_REAL(0.0) ) return SKL_REAL(0.0);
		SKL_REAL _mol = density*volume;
		return GetMol(id,_mol)/volume;
	}
		// SKL_REAL GetDensity(int id, SKL_REAL density)
		// {
		//	if( m_data[id].volume <= SKL_REAL(0.0) ) return SKL_REAL(0.0);
		//	SKL_REAL _mol = density*m_data[id].volume;
		//	return GetMol(id,_mol)/m_data[id].volume;
		// }

	// check mol by name
	//   not changed value
	//   return value = mol num
	SKL_REAL CheckMol(const char* name)
	{
		int id = GetParamIDByName(name);
		if( id<0 ) return SKL_REAL(0.0);
		return CheckMol(id);
	}
	SKL_REAL CheckMol(int id)
	{
		return m_data[id].mol;
	}

	// add mol by name
	//   changed value
	//   return value = mol num after add 
	SKL_REAL AddMol(const char* name, SKL_REAL _mol)
	{
		int id = GetParamIDByName(name);
		if( id<0 ) return SKL_REAL(0.0);
		return AddMol(id,_mol);
	}
	SKL_REAL AddMol(int id, SKL_REAL _mol)
	{
		m_data[id].mol += _mol;
		return m_data[id].mol;
	}

	// get mol by name
	//   changed value
	//   return value = gotten mol num
	SKL_REAL GetMol(const char* name, SKL_REAL _mol)
	{
		int id = GetParamIDByName(name);
		if( id<0 ) return SKL_REAL(0.0);
		return GetMol(id,_mol);
	}
	SKL_REAL GetMol(int id, SKL_REAL _mol)
	{
		if( m_data[id].mol > _mol )
		{
			m_data[id].mol -= _mol;
		}
		else
		{
			_mol = m_data[id].mol;
			m_data[id].mol = SKL_REAL(0.0);
		}

		return _mol;
	}

};

#endif /*__WARP_LIST_H__*/

