/*
 * Copyright (c) RIKEN, Japan. All right reserved. 2008-2009
 *
 */
#include "WarpMolInfo.h"
#include "FortranFuncWarp.h"

/////////////////////////////////////////////////////////////////////////////
// Constructor
//
WarpMolInfo::WarpMolInfo(const CfgInfo& _cfg)
: WarpBase(_cfg)//, m_indexList(_cfg), m_indexListOrg(_cfg)
{
	m_eqn = NULL;
	m_mat = NULL;
	m_mol = NULL;
	m_vf  = NULL;
	m_P0  = NULL;
	m_src = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// Destractor
//
WarpMolInfo::~WarpMolInfo()
{
}

/////////////////////////////////////////////////////////////////////////////
// Parse XML config
//
bool WarpMolInfo::ParseXML(const CfgElem* elemRoot)
{
	if( !elemRoot ) return false;

	// parse param by BaseClass function
	if( !WarpBase::ParseXML(elemRoot) )
	{
		return false;
	}

  DB1(MESSAGE(UtilStrFormat("WarpMolInfo::ParseXML - %s Elem(id=%d)",m_elemName, m_elemID)));

	// search position Elem -> index list
	const CfgElem* posElem = elemRoot->GetElemFirst("position");
	if( posElem )
	{
		if( !m_indexListOrg.ParseXML(posElem) )
		{
			return false;
		}
	}

	// initialize
	if( !EquationInitialize() )
	{
		return false;
	}

	return true;
}

// set material, molecular, equation
bool WarpMolInfo::EquationInitialize()
{
	// equation
	string eqnName = GetStringParam("equation");
	if( eqnName.length()==0 )
	{
		return false;
	}
	m_eqn = UtilGetEquationId(eqnName);
	if( !m_eqn )
	{
		return false;
	}
  DB1(MESSAGE(UtilStrFormat("WarpMolInfo::ParseXML - equation [%s] OK.",m_eqn->name.c_str())));

	// material
	m_mat = m_eqn->mat;
	if( !m_mat )
	{
		return false;
	}
  DB1(MESSAGE(UtilStrFormat("WarpMolInfo::ParseXML - material [%s] OK.",m_mat->name.c_str())));

	// molecular
	m_mol = m_eqn->mol;
	if( !m_mol )
	{
		return false;
	}
  DB1(MESSAGE(UtilStrFormat("WarpMolInfo::ParseXML - molecular [%s] OK.",m_mol->name.c_str())));

	//
	// array Vf
	S4DRN Vf  = UtilGetRegistedArrayS4DReal(UtilStrFormat("Vf_%s",m_mat->name.c_str()));
	if( !(m_vf=Vf.ary) )
	{
		return false;
	}
  DB1(MESSAGE(UtilStrFormat("WarpMolInfo::ParseXML - array Vf OK.")));

	// array P0
	S4DRN P0  = UtilGetRegistedArrayS4DReal(m_eqn->name);
	if( !(m_P0=P0.ary) )
	{
		return false;
	}
  DB1(MESSAGE(UtilStrFormat("WarpMolInfo::ParseXML - array P0 OK.")));

	// array Src
	S4DRN Src = UtilGetRegistedArrayS4DReal(UtilStrFormat("SRC_%s",m_eqn->name.c_str()));
	if( !(m_src=Src.ary) )
	{
		return false;
	}
  DB1(MESSAGE(UtilStrFormat("WarpMolInfo::ParseXML - array Src OK.")));

	//
	// sph to index
	if( !LoadPositionSPH() )
	{
		return false;
	}
  DB1(MESSAGE(UtilStrFormat("WarpMolInfo::ParseXML - LoadPositionSPH() OK.")));

	//
	// update index list to own domain index (for Parallel)
	UpdateDomainIndex();

//debug.s
MESSAGE(string("WarpMolInfo::EquationInitialize OK"));
	GetAreaDensity(E_WARP_MIN);
	GetAreaDensity(E_WARP_MAX);
	GetAreaDensity(E_WARP_AVE);
	GetAreaVolume();
//debug.e

	return true;
}

bool WarpMolInfo::LoadPositionSPH()
{
	int i,j,k;
	if( m_indexListOrg.GetSphCount()==0 ) return true;
	if( !cfg.PeqnAss[m_eqn->id] ) return true;

	const vxsize_t* sz = cfg.Pdiv.size;
	const int I = sz[0];
	const int J = sz[1];
	const int K = sz[2];
	S3D_REAL ary(I,J,K,0);

	for( int i=0;i<m_indexListOrg.GetSphCount();i++ )
	{
		const char* attr = m_indexListOrg.GetSphAttr(i);
		if( !UtilLoadSphOrEmmToS3D(&ary,attr,SKL_REAL(0.0),cfg.grpP.gid) )
		{
			MSG_ERROR(UtilStrFormat("SPH read error. attr=%s",attr));
			return false;
		}
		for(k=0;k<K;k++){
		for(j=0;j<J;j++){
		for(i=0;i<I;i++){
			if( int(ceil(m_vf->val(i,j,k,m_mat->id)) * ceil(ary.val(i,j,k))) > 0 )
			{
				int idx[3]={i+cfg.Pdiv.head[0],j+cfg.Pdiv.head[1],k+cfg.Pdiv.head[2]};
				S_WARP_INDEX index(idx,attr);
				m_indexListOrg.AddUnique(index);
			}
		}}}
  	DB1(MESSAGE(UtilStrFormat("SPH read OK. attr=%s grpP.gid=%d sz=%d %d %d",attr,cfg.grpP.gid,I,J,K)));
		const unsigned int* vfsz = m_vf->GetSize();
  	DB1(MESSAGE(UtilStrFormat("vfsz=%d %d %d",vfsz[0],vfsz[1],vfsz[2])));
	}

	return true;
}

// update index list to own domain index (for Parallel)
bool WarpMolInfo::UpdateDomainIndex()
{
	m_indexList.Initialize(m_indexListOrg);

	if( !m_eqn )
	{
		return false;
	}
	
	if( m_indexListOrg.Count()==0 )
	{
		return true;
	}

	if( !cfg.PeqnAss[m_eqn->id] )
	{
		return true;
	}

	// membrane filter
	S4D_REAL* MembF1 = NULL;
	S4D_REAL* MembF2 = NULL;
	const Material* matF1 = m_indexListOrg.GetMemFilterMaterial(1);
	const Material* matF2 = m_indexListOrg.GetMemFilterMaterial(2);
	if( matF1 )
	{
 		DB1(MESSAGE(UtilStrFormat("membrane filter1=%s",matF1->name.c_str())));
		MembF1 = UtilGetRegistedArrayS4DReal(UtilStrFormat("Sm_%s",matF1->name.c_str())).ary;
	}
	if( matF2 && (matF1!=matF2) )
	{
 		DB1(MESSAGE(UtilStrFormat("membrane filter2=%s",matF2->name.c_str())));
		MembF2 = UtilGetRegistedArrayS4DReal(UtilStrFormat("Sm_%s",matF2->name.c_str())).ary;
	}

	// domain check & membrane filter
	for( int i=0;i<m_indexListOrg.Count();i++ )
	{
		S_WARP_INDEX idx = m_indexListOrg[i];
		int ii,jj,kk;

		// check domain in
		if( !UtilCheckIdxPdiv(&ii,&jj,&kk,idx.gI,idx.gJ,idx.gK) ) continue;

		// check membrane filter
		bool membCheck = true;
		if( MembF1 )
		{
			membCheck = false;

			// check filter 1
			if( int(ceil(MembF1->val(ii,jj,kk,matF1->id))) > 0 )
			{
				if( !MembF2 )
				{
					// filter 1 only
					membCheck = true;
				}
				else
				{
					// check filter 2
					if( int(ceil(MembF2->val(ii,jj,kk,matF2->id))) > 0 )
					{
						membCheck = true;
					}
					else
					{
						// check neighbor cell 
						if( int(ceil(MembF2->val(ii+1,jj,kk,matF2->id))) > 0 ||
						    int(ceil(MembF2->val(ii,jj+1,kk,matF2->id))) > 0 ||
						    int(ceil(MembF2->val(ii,jj,kk+1,matF2->id))) > 0 ||
						    int(ceil(MembF2->val(ii-1,jj,kk,matF2->id))) > 0 ||
						    int(ceil(MembF2->val(ii,jj-1,kk,matF2->id))) > 0 ||
						    int(ceil(MembF2->val(ii,jj,kk-1,matF2->id))) > 0 )
						{
							membCheck = true;
						}
					}
				}
			}
		}

		// check OK or no filter
		if( membCheck )
		{
			idx.I = ii;
			idx.J = jj;
			idx.K = kk;
			m_indexList.Add(idx);
		}
	}

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// print log
//
bool WarpMolInfo::printLog(stringstream& wout, int ts, int nid)
{
	// base
	WarpBase::printLog(wout,ts,nid);

	ts += WARP_TS;

	// index list
	m_indexList.printLog(wout,ts,nid);




	return true;
}

