/*
 * Copyright (c) RIKEN, Japan. All right reserved. 2008-2009
 *
 */
#include "WarpMolInfo.h"
#include "FortranFuncWarp.h"

/////////////////////////////////////////////////////////////////////////////
// for index
//
S_WARP_INDEX WarpMolInfo::GetIndex(int idx)
{
	return m_indexList[idx];
}
size_t WarpMolInfo::GetIndexCount()
{
	return m_indexList.Count();
}


/////////////////////////////////////////////////////////////////////////////
// get mol info type by Elem name
//
E_WARP_MOLINFO_TYPE WarpMolInfo::GetType()
{
	if( !strcasecmp( m_elemName, "SourceInfo" ) )
	{
		return E_WARP_MOLINFO_SOURCE;
	}
	else if( !strcasecmp( m_elemName, "DistinationInfo" ) )
	{
		return E_WARP_MOLINFO_DISTINATION;
	}
	else if( !strcasecmp( m_elemName, "SourceTrigger" ) )
	{
		return E_WARP_MOLINFO_SRC_TRIGGER;
	}
	else if( !strcasecmp( m_elemName, "DistinationTrigger" ) )
	{
		return E_WARP_MOLINFO_DST_TRIGGER;
	}

	return E_WARP_MOLINFO_UNKNOWN;
}

/////////////////////////////////////////////////////////////////////////////
// get density abount index area
//
SKL_REAL WarpMolInfo::GetAreaDensity(E_WARP_OP flag)
{
	if( flag==E_WARP_MAX )
	{
		return GetAreaDensityMax();
	}
	else if( flag==E_WARP_MIN )
	{
		return GetAreaDensityMin();
	}
	else if( flag==E_WARP_AVE )
	{
		return GetAreaDensityAve();
	}

	return SKL_REAL(0.0);
}

/////////////////////////////////////////////////////////////////////////////
// get max density abount index area
//
SKL_REAL WarpMolInfo::GetAreaDensityMax(bool debugWrite)
{
	SKL_REAL maxval = SKL_REAL(0.0);

	if( m_P0 && m_eqn && m_indexList.Count()>0 )
	{
		// first value
		S_WARP_INDEX idx = m_indexList[0];
		maxval = m_P0->val(idx.I,idx.J,idx.K,m_eqn->id);

		// search
		for( int i=1;i<m_indexList.Count();i++ )
		{
			idx = m_indexList[i];
			SKL_REAL val = m_P0->val(idx.I,idx.J,idx.K,m_eqn->id);
			if( val > maxval ) maxval = val;
		}
	}

	// for Parallel
	if( cfg.Parallel )
	{
		SKL_REAL tmp = maxval;
		para_cmp->Allreduce(&tmp,&maxval,1,SKL_ARRAY_DTYPE_REAL,SKL_MAX,cfg.grpAll.gid);
	}
if(debugWrite) DB1(MESSAGE(UtilStrFormat("      WarpMolInfo::GetAreaDensityMax %s %.6e[mol/m3]",m_eqn->name.c_str(),maxval)));

	return maxval;
}

/////////////////////////////////////////////////////////////////////////////
// get min density abount index area
//
SKL_REAL WarpMolInfo::GetAreaDensityMin()
{
	SKL_REAL minval = GetAreaDensityMax(false);

	if( m_P0 && m_eqn && m_indexList.Count()>0 )
	{
		// first value
		S_WARP_INDEX idx = m_indexList[0];
		minval = m_P0->val(idx.I,idx.J,idx.K,m_eqn->id);

		// search
		for( int i=1;i<m_indexList.Count();i++ )
		{
			idx = m_indexList[i];
			SKL_REAL val = m_P0->val(idx.I,idx.J,idx.K,m_eqn->id);
			if( val < minval ) minval = val;
		}
	}

	// for Parallel
	if( cfg.Parallel )
	{
		SKL_REAL tmp = minval;
		para_cmp->Allreduce(&tmp,&minval,1,SKL_ARRAY_DTYPE_REAL,SKL_MIN,cfg.grpAll.gid);
	}

DB1(MESSAGE(UtilStrFormat("      WarpMolInfo::GetAreaDensityMin %s %.6e[mol/m3]",m_eqn->name.c_str(),minval)));
	return minval;
}

/////////////////////////////////////////////////////////////////////////////
// get average density abount index area
//
SKL_REAL WarpMolInfo::GetAreaDensityAve()
{
	int      avecnt = 0;
	SKL_REAL aveval = SKL_REAL(0.0);

	if( m_P0 && m_eqn && m_indexList.Count()>0 )
	{
		// search
		for( int i=0;i<m_indexList.Count();i++ )
		{
			S_WARP_INDEX idx = m_indexList[i];
			SKL_REAL val = m_P0->val(idx.I,idx.J,idx.K,m_eqn->id);
			aveval += val;
			avecnt++;
		}
	}

	// for Parallel
	if( cfg.Parallel )
	{
		SKL_REAL sbuf[2] = {aveval,SKL_REAL(avecnt)};
		SKL_REAL rbuf[2] = {aveval,SKL_REAL(avecnt)};
		para_cmp->Allreduce(sbuf,rbuf,2,SKL_ARRAY_DTYPE_REAL,SKL_SUM,cfg.grpAll.gid);
		aveval = rbuf[0];
		avecnt = int(rbuf[1]);
	}

	if( avecnt>0 )
		aveval /= SKL_REAL(avecnt);
	else
		aveval = SKL_REAL(0.0);
DB1(MESSAGE(UtilStrFormat("      WarpMolInfo::GetAreaDensityAve %s %.6e[mol/m3]",m_eqn->name.c_str(),aveval)));

	return aveval;
}

/////////////////////////////////////////////////////////////////////////////
// get index area volume
//
SKL_REAL WarpMolInfo::GetAreaVolume()
{
	SKL_REAL vol = SKL_REAL(0.0);

	// cell volume
	const SKL_REAL& dx = cfg.pitch[0];
	const SKL_REAL& dy = cfg.pitch[1];
	const SKL_REAL& dz = cfg.pitch[2];
	SKL_REAL vol1 = dx*dy*dz;

	if( m_vf && m_mat && m_indexList.Count()>0 )
	{
		// search
		for( int i=0;i<m_indexList.Count();i++ )
		{
			S_WARP_INDEX idx = m_indexList[i];
			SKL_REAL vf = m_vf->val(idx.I,idx.J,idx.K,m_mat->id);
			vol += (vf*vol1);
		}
	}

	// for Parallel
	if( cfg.Parallel )
	{
		SKL_REAL tmp = vol;
		para_cmp->Allreduce(&tmp,&vol,1,SKL_ARRAY_DTYPE_REAL,SKL_SUM,cfg.grpAll.gid);
	}

	return vol;
}

/////////////////////////////////////////////////////////////////////////////
// get index area mol
//
SKL_REAL WarpMolInfo::GetAreaMol()
{
	SKL_REAL mol = SKL_REAL(0.0);

	// cell volume
	const SKL_REAL& dx = cfg.pitch[0];
	const SKL_REAL& dy = cfg.pitch[1];
	const SKL_REAL& dz = cfg.pitch[2];
	SKL_REAL vol1 = dx*dy*dz;

	if( m_P0 && m_eqn && m_indexList.Count()>0 )
	{
		if( m_vf && m_mat )
		{
			// search
			for( int i=0;i<m_indexList.Count();i++ )
			{
				S_WARP_INDEX idx = m_indexList[i];
				SKL_REAL vf      = m_vf->val(idx.I,idx.J,idx.K,m_mat->id);
				SKL_REAL density = m_P0->val(idx.I,idx.J,idx.K,m_eqn->id);
				mol += (density*vf*vol1);
			}
		}
	}

	// for Parallel
	if( cfg.Parallel )
	{
		SKL_REAL tmp = mol;
		para_cmp->Allreduce(&tmp,&mol,1,SKL_ARRAY_DTYPE_REAL,SKL_SUM,cfg.grpAll.gid);
	}

DB1(MESSAGE(UtilStrFormat("      WarpMolInfo::GetAreaMol %s %.6e[mol]",m_eqn->name.c_str(),mol)));
	return mol;
}

/////////////////////////////////////////////////////////////////////////////
// calc source term by changed density
//
// density : change density per unit time
// return  : changed mol num per DeltaT
SKL_REAL WarpMolInfo::CalcSourceTermByDensity(SKL_REAL density)
{
	// Delta T
	SKL_REAL dt = RicsGetDeltaT();

	// changed mol num
	SKL_REAL mol = SKL_REAL(0.0);

	// cell volume
	const SKL_REAL& dx = cfg.pitch[0];
	const SKL_REAL& dy = cfg.pitch[1];
	const SKL_REAL& dz = cfg.pitch[2];
	SKL_REAL vol1 = dx*dy*dz;

	// for minus value
	if( GetType() == E_WARP_MOLINFO_SOURCE )
		density = abs(density) * SKL_REAL(-1.0);
	else
		density = abs(density);

	// search
	if( m_src && m_eqn && m_indexList.Count()>0 )
	{
		if( m_vf && m_mat )
		{
			for( int i=0;i<m_indexList.Count();i++ )
			{
				S_WARP_INDEX idx = m_indexList[i];

				// calc source term
				m_src->val(idx.I,idx.J,idx.K,m_eqn->id) += density;

				// calc mol num
				SKL_REAL vf = m_vf->val(idx.I,idx.J,idx.K,m_mat->id);
				mol += (density*vf*vol1);
			}
		}
	}

	// for Parallel
	if( cfg.Parallel )
	{
		SKL_REAL tmp = mol;
		para_cmp->Allreduce(&tmp,&mol,1,SKL_ARRAY_DTYPE_REAL,SKL_SUM,cfg.grpAll.gid);
	}

DB1(MESSAGE(UtilStrFormat("      WarpMolInfo::CalcSourceTermByDensity %s %.6e[mol/m3] ->  %.6e[mol]",m_eqn->name.c_str(),density,mol)));

	return mol*dt;
}

/////////////////////////////////////////////////////////////////////////////
// calc source term by changed density ratio
//
// densityratio : warp density ratio per unit time
// op           : operation(min/max/ave) for calc area density
// return       : changed mol num per dt
SKL_REAL WarpMolInfo::CalcSourceTermByDensityRatio(SKL_REAL densityRatio, E_WARP_OP op)
{
	// Delta T
	SKL_REAL dt = RicsGetDeltaT();

	// changed mol num
	SKL_REAL mol = SKL_REAL(0.0);

	// cell volume
	const SKL_REAL& dx = cfg.pitch[0];
	const SKL_REAL& dy = cfg.pitch[1];
	const SKL_REAL& dz = cfg.pitch[2];
	SKL_REAL vol1 = dx*dy*dz;

	// area density
	SKL_REAL op_den = GetAreaDensity(op);

	// change density by ratio;
	SKL_REAL density = op_den*densityRatio;
	if( GetType() == E_WARP_MOLINFO_SOURCE )
		density = abs(density) * SKL_REAL(-1.0);
	else
		density = abs(density);

	// search
	if( m_src && m_eqn && m_indexList.Count()>0 )
	{
		if( m_vf && m_mat )
		{
			for( int i=0;i<m_indexList.Count();i++ )
			{
				S_WARP_INDEX idx = m_indexList[i];

				// calc source term
				m_src->val(idx.I,idx.J,idx.K,m_eqn->id) += density;

				// calc mol num
				SKL_REAL vf = m_vf->val(idx.I,idx.J,idx.K,m_mat->id);
				mol += (density*vf*vol1);
			}
		}
	}

	// for Parallel
	if( cfg.Parallel )
	{
		SKL_REAL tmp = mol;
		para_cmp->Allreduce(&tmp,&mol,1,SKL_ARRAY_DTYPE_REAL,SKL_SUM,cfg.grpAll.gid);
	}

DB1(MESSAGE(UtilStrFormat("      WarpMolInfo::CalcSourceTermByDensityRatio %s %.6e[%] * %.6e[mol/m3] -> %.6e[mol/m3]",m_eqn->name.c_str(),(densityRatio*100),op_den,density)));
DB1(MESSAGE(UtilStrFormat("      WarpMolInfo::CalcSourceTermByDensityRatio %s %.6e[mol/m3] -> %.6e[mol]",m_eqn->name.c_str(),density,mol)));

	return mol*dt;
}

/////////////////////////////////////////////////////////////////////////////
// calc source term by changed mol num
//
// mol    : change mol num per unit time
// return : changed density per dt
SKL_REAL WarpMolInfo::CalcSourceTermByMol(SKL_REAL mol)
{
	// Delta T
	SKL_REAL dt = RicsGetDeltaT();

	// area volume
	SKL_REAL total_vol = GetAreaVolume();
	if( total_vol <= SKL_REAL(0.0) )
	{
		return SKL_REAL(0.0);
	}

	// changed density
	SKL_REAL density = mol / total_vol;
// DB1(MESSAGE(UtilStrFormat("WarpMolInfo::CalcSourceTermByMol %s change_density by mol = %.6e",m_elemName,density)));

	// cell volume
	const SKL_REAL& dx = cfg.pitch[0];
	const SKL_REAL& dy = cfg.pitch[1];
	const SKL_REAL& dz = cfg.pitch[2];
	SKL_REAL vol1 = dx*dy*dz;

	// search
	if( m_src && m_eqn && m_indexList.Count()>0 )
	{
		if( m_vf && m_mat )
		{
			for( int i=0;i<m_indexList.Count();i++ )
			{
				S_WARP_INDEX idx = m_indexList[i];

				// volume ratio
				SKL_REAL vf = m_vf->val(idx.I,idx.J,idx.K,m_mat->id);
				if( vf <= SKL_REAL(0.0) ) continue;

				// source term
				m_src->val(idx.I,idx.J,idx.K,m_eqn->id) += density;
			}
		}
	}

DB1(MESSAGE(UtilStrFormat("      WarpMolInfo::CalcSourceTermByMol %s %.6e[mol] / %.6e[m3] -> %.6e[mol/m3]",m_eqn->name.c_str(),mol,total_vol,density)));

	return density*dt;
}

/////////////////////////////////////////////////////////////////////////////
// check density threshold about index area
//
bool WarpMolInfo::CheckDensityThreshold(SKL_REAL threshold, E_WARP_OP flag1, E_WARP_CMP flag2, WarpTemporaryArea *tempArea)
{
	// check value
	SKL_REAL check_val = SKL_REAL(0.0);

	// check area
	string chkTemp = GetStringParam("CheckTempArea");
	if( !strcasecmp(chkTemp.c_str(),"true") )
	{
		if( !tempArea )
		{
			return false;
		}
		check_val = tempArea->CheckDensity(m_eqn->mol->name.c_str());
	}
	else
	{
		check_val = GetAreaDensity(flag1);
	}

	// check
	bool ret = false;
	if( flag2 == E_WARP_EQ )
	{
		if( check_val == threshold ) ret = true;
	}
	else if( flag2 == E_WARP_LT )
	{
		if( check_val < threshold ) ret = true;
	}
	else if( flag2 == E_WARP_LE )
	{
		if( check_val <= threshold ) ret = true;
	}
	else if( flag2 == E_WARP_GT )
	{
		if( check_val > threshold ) ret = true;
	}
	else if( flag2 == E_WARP_GE )
	{
		if( check_val >= threshold ) ret = true;
	}

DB1(MESSAGE(UtilStrFormat("      WarpMolInfo::CheckDensityThreshold eqn=%s th=%.6e[mol/m3] check=%.6e[mol/m3] %s: %s"
	,m_eqn->name.c_str()
	,threshold
	,check_val
	,GetStringParam("CompType").c_str()
	,string_cast(ret,ios::boolalpha).c_str())));

	return ret;
}

bool WarpMolInfo::CheckTrigger(WarpTemporaryArea *tempArea)
{
	// value type(fixed or table)
	E_WARP_VALUE_TYPE vtype = CheckValueType();

	// area OP type(min or max or ave)
	E_WARP_OP op = CheckOpeType();

	// threshold compare type(Equal or LessThan or LessEqual or GreaterThan or GreaterEqual)
	E_WARP_CMP cmp = CheckCompType();

	// get density threshold
	SKL_REAL threshold = SKL_REAL(0.0);
	if( vtype==E_WARP_FIXED )
	{
		// fixed value
		threshold = GetRealParam("threshold");
	}
	else if( vtype==E_WARP_TABLE )
	{
		// time
		SKL_REAL time = RicsGetCurrentTime();

		// table value
		threshold = GetTableValue("threshold", time );
	}
	else
	{
		return false;
	}

	return CheckDensityThreshold( threshold, op, cmp, tempArea );
}

