/*
 * Copyright (c) RIKEN, Japan. All right reserved. 2008-2009
 *
 */
#include "WarpSolv.h"
#include "FortranFuncWarp.h"

/////////////////////////////////////////////////////////////////////////////
// Constructor
//
WarpSolv::WarpSolv(const CfgInfo& _cfg)
: cfg(_cfg)
{
	clock    = double(0.0);
	comm_clk = double(0.0);
}

/////////////////////////////////////////////////////////////////////////////
// Destractor
//
WarpSolv::~WarpSolv()
{
	int i;

	for( i=0;i<m_warpGrp.size();i++ )
	{
		delete m_warpGrp[i];
	}
	m_warpGrp.clear();
}

/////////////////////////////////////////////////////////////////////////////
// Initialize Main
//
int WarpSolv::WarpInitialize(
        const CfgElem* simElem, const CfgElem* warpRoot
      )
{
  DB1(MESSAGE("WarpInitialize"));

#ifdef SPHERE175
	para_cmp = cfg.solver->SklGetParaComponent();
	para_mng = para_cmp->GetParaManager();
#endif

	const CfgElem*  warp = simElem? simElem->GetElemFirst ("warp"): NULL;
	const CfgParam* exe = warp? warp->GetParamFirst("execution"): NULL;

  // set isenable metabolism simulation
	const char* exc = NULL;
	enableWarp = (exe && exe->GetData(&exc) && (strcmp(exc, "true") == 0));
	if (!enableWarp) {
		return 1;
	}

	if( !warpRoot )
	{
		MSG_ERROR(UtilStrFormat("warp_setting Elem is undefined in Parameter tag"));
		MSG_ERROR(UtilStrFormat("warp Simulation is changed to false"));
		enableWarp = false;
		return 1;
	}

	// Parse XML
	if( !ParseXML(warpRoot) )
	{
		enableWarp = false;
		return -1;
	}

  // init fields
	Allocate();

	return 1;
}

/////////////////////////////////////////////////////////////////////////////
// Loop Main
//
int WarpSolv::WarpLoop()
{
// Return value -> -1 : error
//                  0 : forced termination
//                  1 : normal return

	if (!enableWarp) { return 1; }

  DB1(MESSAGE("WarpLoop"));

	BEGIN_CLOCK(loop);

	for( int i=0;i<m_warpGrp.size();i++ )
	{
		if( !m_warpGrp[i]->Calc() )
		{
			return -1;
		}
	}


	END_CLOCK(clock, loop);

	return 1;
}

/////////////////////////////////////////////////////////////////////////////
// Post Main
//
int WarpSolv::WarpPost()
{
  DB1(MESSAGE("WarpPost"));
	return 1;
}

/////////////////////////////////////////////////////////////////////////////
// Parse XML config
//
bool WarpSolv::ParseXML(const CfgElem* warpElem)
{
	if( !warpElem ) return false;

  DB1(MESSAGE("WarpSolv::ParseXML"));

	// search WarpGroup elem
	int cnt = 0;
	const CfgElem* warpGrpElem = warpElem->GetElemFirst("WarpGroup");
	while( warpGrpElem )
	{
		WarpGroup *grp = new WarpGroup(cfg, ++cnt);
		if( !grp->ParseXML(warpGrpElem) )
		{
			delete grp;
			return false;
		}
		m_warpGrp.push_back(grp);

		warpGrpElem = warpElem->GetElemNext(warpGrpElem, "WarpGroup");
	}

	return true;
}

/////////////////////////////////////////////////////////////////////////////
// Allocate Arrays
//
bool WarpSolv::Allocate()
{
  DB1(MESSAGE("WarpSolv::Allocate"));
	return true;
}

/////////////////////////////////////////////////////////////////////////////
// Check config & print info
//
bool WarpSolv::chkCfg()
{
//	if (cfg.grpAll.nid != 0) { return false; }
	if (cfg.grpAll.nid == 0)
	{

		MESSAGE("**** Configuration of Warp Solver ****", 0);
		MESSAGE(string("Warp Simulation is ")
		       + string_cast(enableWarp, ios::boolalpha));
	}

	if (!enableWarp) { return true; }

	// log
	DB1( printLog(cfg.grpAll.nid) );

	return true;
}

bool WarpSolv::printLog(int nid)
{
	if( !enableWarp ) return true;

	//title
	stringstream wout;
	wout << "Warp Simulation Setting" << endl;
	wout << setw(80) << setfill('=') << "" << setfill(' ') << endl;

	// WarpGroup
	for( int i=0;i<m_warpGrp.size();i++ )
	{
		if( !m_warpGrp[i]->printLog(wout,0,nid) )
		{
			MSG_ERROR(UtilStrFormat("print Warp Setting Log error at WarpGroup::printLog"),nid);
			return false;
		}
	}

	wout << setw(80) << setfill('=') << "" << setfill(' ') << endl;

	if( nid==0 )
	{
		MESSAGE(wout.str(),nid);
	}
	else
	{
		DB1(MESSAGE(wout.str(),nid));
	}

	return true;
}


